/*
 * @(#)Kan.java Kan.java,v 1.2 2002/11/22 23:26:06 kris Exp $
 *
 *
 */

package mokan;

import java.io.File;
import java.io.FileInputStream;
import java.util.Properties;

import gnu.getopt.Getopt;

import mokan.app.Application;
import mokan.comm.KanComm;
import mokan.comm.MachinePort;
import mokan.comm.KanURL;
import mokan.comm.PhysicalNode;
import mokan.debug.Event;
import mokan.nameservice.Globals;

import klogger.Logger;
/**
 * The main class for the <em>Kan</em> system.  Its purpose is to get the
 * system running on the local machine and also connect to the other machines
 * in its current universe.  It connects up to a Kan Virtual Machine at each
 * of the machines in its universe.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author	Jerry James
 * @author	Suk Lee
 * @author	Jing Wang
 * @author	Sean Brydon
 * @author	Sandeep Bhatia
 */
public final class Kan {

  /**
   * The default Kan system properties.  This set of properties is used in
   * case Kan cannot find an external properties file.
   */
  private static final Properties defaultProps = new Properties();

  /**
   * The user-defined Kan system properties.
   */
  public static final Properties systemProps = new Properties(defaultProps);

  // Initialize the property lists
  static {
    // First, set up the default values
    defaultProps.setProperty("apps", "");
    defaultProps.setProperty("controller.port", "6000");
    defaultProps.setProperty("machines", "localhost:5500");
    defaultProps.setProperty("obj.replica.interval", "10");
    defaultProps.setProperty("rmi.server.port", "5005");
    defaultProps.setProperty("rmi.server.retries", "3");
    defaultProps.setProperty("run", "javamokan.Kan");
    defaultProps.setProperty("tcp.port", "5500");
    defaultProps.setProperty("thread.inlining", "false");
    defaultProps.setProperty("thread.pool.adaptive", "true");
    defaultProps.setProperty("thread.pool.interval", "40");
    defaultProps.setProperty("thread.pool.growRatio", "0.2");
    defaultProps.setProperty("thread.pool.shrinkRatio", "0.25");
    defaultProps.setProperty("thread.pool.minimum", "2");
    defaultProps.setProperty("thread.pool.request", "3");

    // Bogus work around
    // http://forum.java.sun.com/thread.jsp?forum=27&thread=231280&tstart=0&trange=15
    System.setProperty("java.util.prefs.syncInterval", "2000000");
    // Now look for the system-wide Kan properties
    final String classPath = System.getProperty("java.class.path");
    final int index = classPath.indexOf("kan.jar");
    boolean notThere = true;
    if (index > 0) {
      final char pathSep = File.pathSeparatorChar;
      final String kanPath = classPath.substring(0, index - 1);
      final int start = kanPath.lastIndexOf(pathSep) + 1;
      final File propFile =
          new File(kanPath.substring(start), "kan.properties");
      try {
        systemProps.load(new FileInputStream(propFile));
        notThere = false;
      } catch (Exception ex) {
        System.err.print("Unable to find ");
        System.err.println(propFile.toString());
      }
    } else {

      final String homeDir = System.getProperty("user.home");

      try {
        final File propFile = new File(homeDir, "kan.properties");
        systemProps.load(new FileInputStream(propFile));
      } catch (Exception ex) {
        // Just ignore this; we'll use the system-wide props & defaults

        System.err.println
            ("Unable to find mokan.properties, using defaults");
      }

    }

    // Finally, look for the user-specific Kan properties
  }

  /**
   *  An interface to the communication system
   */
  public static final KanComm kanComm = new KanComm();

  /**
   * Controlling information for application started from the command line
   */
  public static Application newApp = null;

  /**
   * Whether to print verbose messages to <code>System.out</code>
   */
  public static boolean verbose = false;

  /**
   * The starting point for running the Kan system.  All startups of Kan go
   * through here, whether using the GUI or the command line.
   *
   * @param args the command line arguments to the Kan Virtual Machine
   */
  public static void main(final String[] args) {
    // First read the command line options.
    Getopt options = new Getopt("kan", args, "e:vmj:p:s:");
    int c;
    boolean main = false;
    KanURL joinUrl = null;
    String shapefile = null;

    //The port number is obtained in this order:
    //1. Globals.DEFAULT_PORT, which is 5500
    //2. From system properties
    //3. -p option

    

    int port = Globals.DEFAULT_PORT;
    String portName = systemProps.getProperty("tcp.port");
    if(portName != null) {
        System.out.println ("Checking " + portName);
      port = Integer.parseInt(portName);
    }

    while ( (c = options.getopt()) != -1) {
      switch (c) {
        case 'e': // Set the events of interest
          final String[] eventNames =
              explodeString(options.getOptarg());
          final int[] events = Event.namesToEvents(eventNames);
          Event.setEventMask(events);
          break;

        case 'v':
          verbose = true;
          System.out.println("Verbose messages enabled.");
          break;
        case 'm':
          main = true;
          break;
        case 'j':
          try {
            joinUrl = new KanURL(options.getOptarg());
          } catch (Exception e) {
            System.err.println("Bad kan URL " + options.getOptarg());
          }
          break;
        case 'p':
            port = Integer.parseInt(options.getOptarg());
            if (port == 0) {
              System.err.println("Invalid port number");
              usage();
              return;
            }
            break;
      case 's':
          shapefile = options.getOptarg();
          break;

      default:
          usage();
          return;
      }
    }

    try {
      // Get the argument which should be a URL for a kan application
      //KK for now just use nodename/application-class
      int i = options.getOptind();
      if (i >= args.length) {
        usage();
        return;
      }
      Globals.listenPort = port;
      KanURL appurl = new KanURL(args[i++]);

      // Initialize logging
      System.setProperty ("MAIN", appurl.getAppName());
      Logger.configure ("mokan.properties");

      //If the appurl designates a different port, we can't accomodate it
      //because the appurl is also used as an identifier.

      // Initialize the local network
      kanComm.init();
      if (verbose) {
        System.out.println("Communication established");
      }

      if (shapefile != null) 
          mokan.prefetch.runtime.PrefetchSystem.start(shapefile);
      else {
          System.err.println("***WARNING: no shape specified please use -s option");
      }
      Application app = new Application(appurl);

      // Added by Huahai
      Globals.app = app;

      if (main) {
        final String[] appArgs = new String[args.length - i];
        System.arraycopy(args, i, appArgs, 0, args.length - i);
        app.invoke(appArgs);
        System.out.println("Created " + app);
      } else {
        if (joinUrl == null) {
          joinUrl = new KanURL(appurl, appurl.getAppHost());
        }
        app.join(joinUrl);
      }
    } catch (java.net.ConnectException e) {
      System.err.println("Connection refused or servise not available");
    } catch (java.net.MalformedURLException e) {
      System.err.println(" Application  URL is not valid");
      usage();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private static void usage() {
    System.out.println("Usage: java mokan.Kan [-v] [-e events] AppID [master]");
    System.out.println("  where -v = verbose operation");
    System.out.println("        -e events = list of debugging events to print");
    System.out.println("        -m  = Make this a new application instance");
    System.out.println("        -p port = the port this node will listen to");
    System.out.println("        -s file = shape file");
    System.out.println("    AppID = Kan URL for application");
    System.out.println("    i.e. AppID =  kan://machine/fullclassname");
    System.out.println("    master = node already running application\n");
  }

  /**
   * Fetch node identities from a list of node names, and create an array
   * of corresponding {@linkmokan.comm.MachinePort MachinePorts}
   *
   * @param theString the string of node IDs to parse
   * @return an array of {@linkmokan.comm.MachinePort MachinePorts}
   */
  private static MachinePort[] parseNodeList(final String theString) {
    int elems = 1;
    for (int i = 0; i < theString.length(); i++) {
      if (theString.charAt(i) == ',') {
        elems++;
      }
    }
    final MachinePort[] nodeList = new MachinePort[elems];

    try {
      int start = 0, end = 0;
      for (int i = 0; i < elems - 1; i++) {
        end = theString.indexOf(',', start);
        nodeList[i] = new MachinePort(theString.substring(start, end));
        start = end + 1; // Skip the comma
      }
      nodeList[elems - 1] = new MachinePort(theString.substring(start));
    } catch (Exception e) {
      // We don't understand a node name.  Complain and die.
      e.printStackTrace();
      System.exit(0);
    }
    // The compiler doesn't know that System.exit doesn't return, so we
    // have to put the return statement last.
    return nodeList;
  }

  /**
   * Turn a string containing commas into an array of strings, each
   * corresponding to one comma-separated portion of the original string.
   *
   * @param theString the string to explode
   * @return the exploded array of strings
   */
  private static String[] explodeString(final String theString) {
    int elems = 1;
    for (int i = 0; i < theString.length(); i++) {
      if (theString.charAt(i) == ',') {
        elems++;
      }
    }
    final String[] stringList = new String[elems];

    int start = 0, end = 0;
    for (int i = 0; i < elems - 1; i++) {
      end = theString.indexOf(',', start);
      stringList[i] = theString.substring(start, end);
      start = end + 1; // Skip the comma
    }
    stringList[elems - 1] = theString.substring(start);
    return stringList;
  }
}
