/**
 * @(#)KanSystem.java KanSystem.java,v 1.1.1.1 2002/11/06 22:10:24 kris Exp $
 *
 */

package mokan;

import mokan.comm.LogicalNode;
import mokan.debug.Event;
import mokan.obj.GuardCondition;
import mokan.obj.NoSuchKanObjectException;
import mokan.obj.ObjectID;
import mokan.obj.GlobalObj;

import mokan.thread.AsynchCall;
import mokan.thread.Future;
import mokan.thread.KanThread;
import mokan.thread.ThreadScheduler;

import mokan.transact.Transaction;
import mokan.transact.TransactionID;
import mokan.transact.TransactionException;
import mokan.transact.TransactionUserException;
import mokan.transact.ConcurrencyScheduler;
import mokan.transact.StrictScheduler;
import mokan.transact.OptimisticScheduler;

/**
 * Provides all the methods available to the application programmers,
 * including creation of Kan objects, creation of local and remote threads,
 * beggining and ending of guarded atomic actions, joining child threads,
 * getting results, etc.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:24 $
 * @author	Suk Y. Lee
 * @author	Jing Wang
 * @author	Jerry James
 * @author	Sean Brydon
 * @author	Sandeep Bhatia
 */
public final class KanSystem {

    public final static int IMMOBILE = 0;
    public final static int MOBILE = 1;

    public final static ConcurrencyScheduler scheduler
    //= new OptimisticScheduler();
    = new StrictScheduler();


    /**
     *  Create a new global object
     */
    public static void createObject (ObjectID oid,
                                     GlobalObj obj,
                                     int flag) {
        new mokan.obj.KanObject(oid, obj, scheduler.protocol(flag));
    }


    /**
     *  Create a new global object
     */
    public static void createObject (ObjectID oid,
                                     Class klass,
                                     int   constrNum,
                                     Object[] args) {

        final mokan.obj.ObjectMgr mgr =
            mokan.comm.LogicalNode.getMyNode().getLocalHandler().manager;

        mgr.createRmtObject(oid, klass, constrNum, args);
    }



    /**
     * Create a thread to invoke a method on some <code>KanObject</code>
     *
     * @param objID the {@linkmokan.obj.ObjectID ObjectID} for the object on
     *        which to invoke the method
     * @param fut a {@linkmokan.thread.Future Future} to hold the results of
     *        the method call
     * @param num the method number to invoke
     * @param args an array of objects holding the method arguments
     * @exception TransactionException if the containing transaction (if any)
     *        is aborted
     */
    public static void invokeMethod(final ObjectID objID, final Future fut,
                                    final int num, final Object[] args)
        throws TransactionException {

        scheduler.invokeMethod(objID, fut, num, args);

//         try {
//             final LogicalNode myNode = LogicalNode.getMyNode();
//             final ThreadScheduler myScheduler =
//                 myNode.getLocalHandler().scheduler;
//             myScheduler.callAsyncMethod(objID, fut, num, args);
//         } catch (NoSuchKanObjectException noObjEx) {
//             // TODO: What should we do here?
//             // This shouldn't be possible, given the compiler, but...
//             noObjEx.printStackTrace();
//         }
    }

    // TODO: Implement the following methods
    /*
      public static Object readObjectField(ObjectID objID, String name,
      TransactionID parent) {
      return null;
      }

      public static boolean readBooleanField(ObjectID objID, String name,
      TransactionID parent) {
      return false;
      }

      public static char readCharField(ObjectID objID, String name,
      TransactionID parent) {
      return ' ';
      }

      public static byte readByteField(ObjectID objID, String name,
      TransactionID parent) {
      return 0;
      }

      public static short readShortField(ObjectID objID, String name,
      TransactionID parent) {
      return 0;
      }

      public static int readIntField(ObjectID objID, String name,
      TransactionID parent) {
      return 0;
      }

      public static float readFloatField(ObjectID objID, String name,
      TransactionID parent) {
      return 0.0f;
      }

      public static double readDoubleField(ObjectID objID, String name,
      TransactionID parent) {
      return 0.0;
      }

      public static void writeField(ObjectID objID, String name, Object val,
      TransactionID parent) {
      }
    */

    /**
     * Create a new transaction for the current guarded statement.  The
     * compiler must decide whether the guarded statement joins any
     * {@linkmokan.thread.Future Futures} (<var>blocking</var>), and if so,
     * whether it created any of those {@linkmokan.thread.Future Futures}
     * (<var>nested</var>).
     *
     * @param objID    object on which the guarded action occurs
     * @param gc       Guard Condtion, provided by the user
     * @param write    <code>true</code> if this is a writing transaction
     * @param type whether the transaction should be nonblocking, blocking, or
     *        nested
     * @exception TransactionException if the transaction aborts
     */
    public static void guardBegin(final ObjectID objID,
                                  final GuardCondition gc,
                                  final boolean write, final int type)
        throws TransactionException {

        final AsynchCall call = KanThread.currCall();

        final Transaction newTrans =
            //Transaction.makeTransaction(gc, objID, write, call, type);
            scheduler.makeTransaction(gc, objID, write, call, type);



        call.currTrans = newTrans.transID;
        try {
            scheduler.beginTransaction(newTrans);
        } catch (TransactionException e) {
            Event.DEBUG(Event.TRANSACTION_CREATE, "KanSystem.guardBegin:" + newTrans + " died " + e);
            throw e;
        }
//Commented by Huahai
//        catch (Exception e) {
//            Event.DEBUG(Event.TRANSACTION_CREATE, "KanSystem.guardBegin:" + newTrans + " died " + e);
//            e.printStackTrace();
//        }
        Event.DEBUG(Event.TRANSACTION_CREATE, "KanSystem.guardBegin: ",
                    newTrans);
    }

    /**
     * End a guarded atomic block.  Commit or abort as appropriate, and unlock
     * everything.
     *
     * @exception TransactionException if the transaction was aborted
     */
    public static void guardEnd() throws TransactionException {
        final AsynchCall call = KanThread.currCall();
        final Transaction currTrans = Transaction.lookup(call.currTrans);
        try {
            Event.DEBUG(Event.TRANSACTION, "KanSystem.guardEnd: begin ", currTrans);
            scheduler.endTransaction(currTrans, true);
        } finally {
            call.currTrans = null;
            Event.DEBUG(Event.TRANSACTION, "KanSystem.guardEnd: end ", currTrans);
        }
    }

    /**
     * Abort a transaction by killing it with an exception
     *
     * @param ex the exception with which to kill the transaction
     */
    public static void abort(final Exception ex) {
        try {
            final TransactionID currTransID = KanThread.currCall().currTrans;
            TransactionException tex = null;
            if (ex instanceof TransactionException) {
// 		currTransID.abort((TransactionException)ex);
                tex = (TransactionException)ex;
            } else {
// 		TransactionUserException userEx =
// 		    new TransactionUserException(ex);
// 		currTransID.abort(userEx);
                tex = new TransactionUserException(ex);
            }
            scheduler.abortTransaction(currTransID, tex);
        } catch (NullPointerException nullEx) {
            // Do nothing; there is no current transaction
        }
    }

    /**
     * Wait for all child threads to complete.
     */
    public static void joinAll() {
        KanThread.currCall().joinAll();
    }
}
