/*
 * @(#)$Id
 *
 *
 */

package mokan.app;

import java.io.*;
import mokan.comm.LogicalNode;
import mokan.debug.Event;
import mokan.thread.AsynchCall;
import mokan.thread.ThreadID;
import mokan.thread.ThreadScheduler;

/**
 * A class holding information on system applications
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author	Sandeep Bhatia
 * @author	Sean Brydon
 * @author	Jerry James
 */
public class AppInfo implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 8912459337805103862L;






    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /**
     * The string corresponding to the fully qualified name of the
     * application.  This is a pseudo-final field.
     *
     * @serial a string value holding the fully qualified name of the
     *         application's main class
     */
    private String appName;

    /**
     * An array of {@link java.lang.String String} holding the application
     * arguments.  This is a pseudo-final field.
     *
     * @serial a string array holding the application arguments
     */
    private String[] appArgs;

    /**
     * Whether we should shut down the KVM when the application completes.
     * This is a pseudo-final field.
     *
     * @serial a boolean value that is <code>true</code> if Kan should shut
     *         down when the application exits
     */
    private boolean shutdown;

    /**
     * Create a new <code>AppInfo</code>
     *
     * @param name the name of the class containing the <code>main</code>
     *        method 
     * @param args the arguments to pass to <code>main</code>
     * @param shutdown <code>true</code> if Kan should shut down after the
     *        application exits
     */
    public AppInfo(final String name, final String[] args,
		   final boolean shutdown) {
	appName = name;
	appArgs = args;
	this.shutdown = shutdown;
    }

    /**
     * Start the application represented by this <code>AppInfo</code>
     *
     * @param appEntry the <code>Application</code> to pass to
     *        <code>main</code>
     */
    public void invoke(final Application app) {
	Event.DEBUG(Event.APP_START, "Starting application: ", app);
	final LogicalNode node = LogicalNode.getMyNode();
	final ThreadScheduler myScheduler = node.getLocalHandler().scheduler;
	final Object argObjects[] =
	    { app, (appArgs == null) ? new String[0] : appArgs };
	final AsynchCall rootCall = new AsynchCall(node, appName, argObjects,
						   myScheduler);
	final ThreadID rootThrdID = myScheduler.createLocThread(rootCall);

	/*
	 * Create a thread to watch the AsynchCall so we can detect whether
	 * main returned and do garbage collection later
	 */
	new AppCleaner(rootThrdID, rootCall, app, shutdown);
    }

    /**
     * Produce a string representation of this <code>AppInfo</code>
     *
     * @return a string representation of this <code>AppInfo</code>
     */
    public String toString() {
	return appName;
    }

    /**
     * Save the state of the <code>AppInfo</code> instance to a stream (i.e.,
     * serialize it).
     *
     * @serialData The application name is written out (String), then the
     *             arguments to <code>main</code> (Object[]),followed by an
     *             indication of whether Kan should shut down after running
     *             this application (boolean).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
	out.writeUTF(appName);
	out.writeObject(appArgs);
	out.writeBoolean(shutdown);
    }

    public void readExternal(final ObjectInput in)
	throws ClassNotFoundException, IOException {

	appName = in.readUTF();
	appArgs = (String[])in.readObject();
	shutdown = in.readBoolean();
    }
}
