/*
 * @(#)AppMessage.java AppMessage.java,v 1.2 2002/11/22 23:26:06 kris Exp $
 *
 *
 */

package mokan.comm;

import java.io.*;
import java.util.Collection;
import java.util.Iterator;
import mokan.debug.Event;
import mokan.thread.KanThread;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionID;
import mokan.util.LongHashMap;

import mokan.nameservice.*;

/**
 * A class of {@linkmokan.comm.Message message objects} that are intended for
 * application messages.  These messages are passed between logical nodes for
 * the running application.  The thread-to-logical node binding is set up in
 * the constructor for this class.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author	Sean Brydon
 * @author	Sandeep Bhatia
 * @author	Jerry James
 */
public class AppMessage extends Message implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -6817245558732301116L;






    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /**
     * Message sequence number (sending).
     */
    private static long ssn = 0;

    /**
     * Hash table with send sequence number as key and
     * {@linkmokan.comm.Wakeup Wakeup} as element
     */
    protected static final LongHashMap wakeupTable = new LongHashMap();

    /**
     * Send Sequence Number; the number of messages sent by this node so far
     *
     * @serial the send sequence number for this message
     */
    protected long SSN;

    /**
     * The (logical or physica) source of the message
     *
     * @serial the source node of this message
     */
    protected LogicalNode from;

    /**
     * The (logical or physical) destination of the message
     *
     * @serial the destination node of this message
     */
    public LogicalNode to;

    /**
     * Create a new <code>AppMessage</code>; the deserializing constructor
     *
     * @param needsReset <code>true</code> if the outgoing stream should be
     *        reset before sending this message
     * @param canBlock <code>true</code> if the handling of this message might
     *        block the handling thread
     */
    protected AppMessage(final boolean needsReset, final boolean canBlock) {
        super(needsReset, canBlock);
        //this.from = null;
        this.from = LogicalNode.getMyNode();
    }

    /**
     * Create a message from some {@linkmokan.comm.LogicalNode LogicalNode}
     *
     * @param from the {@linkmokan.comm.LogicalNode LogicalNode} originating
     *        the message, or <code>null</code> for the local
     *        {@linkmokan.comm.LogicalNode LogicalNode}
     * @param needsReset <code>true</code> if the outgoing stream should be
     *        reset before sending this message
     * @param canBlock <code>true</code> if the handling of this message might
     *        block the handling thread
     */
    protected AppMessage(final LogicalNode from, final boolean needsReset,
                         final boolean canBlock) {
        super(needsReset, canBlock);
        this.from = (from != null) ? from : LogicalNode.getMyNode();
    }

    /**
     * Send a point-to-point message
     *
     * @param to the destination {@linkmokan.comm.LogicalNode node} for this
     *        message
     * @return the sequence number for this message
     */
    public final long send(final LogicalNode to) {
        this.to = to;

        // We should synchronize this, but the chances of interruption are too
        // small to justify it
        SSN = ssn++;
        to.send(this);
        return SSN;
    }

    /**
     * Send a point-to-point message that already has a sequence number
     *
     * @param to the destination {@linkmokan.comm.LogicalNode node} for this
     *        message
     * @param sequence the sequence number for this message
     */
    public final void send(final LogicalNode to, final long sequence) {
        this.to = to;
        SSN = sequence;
        to.send(this);
    }

    /**
     * Send a multicast message
     *
     * @param dests the destination {@linkmokan.comm.LogicalNode nodes} for
     *        this message
     * @return the sequence number for this message
     */
    public final long send(final Collection dests) {
        // We should synchronize this, but the chances of interruption are too
        // small to justify it
        SSN = ssn++;
        final Iterator nodes = dests.iterator();
        while (nodes.hasNext()) {
            to = (LogicalNode)nodes.next();
            to.send(this);
        }
        return SSN;
    }

    /**
     * Send a multicast message that already has a sequence number
     *
     * @param dests the destination {@linkmokan.comm.LogicalNode nodes} for
     *        this message
     * @param sequence the sequence number for this message
     */
    public final void send(final Collection dests, final long sequence) {
        SSN = sequence;
        final Iterator nodes = dests.iterator();
        while (nodes.hasNext()) {
            to = (LogicalNode)nodes.next();
            to.send(this);
        }
    }

    /**
     * Send a message and wait for the reply to arrive
     *
     * @param to the node to which the message should be sent
     * @return the message sent in reply
     * @exception TransactionAbortException if the transaction (if any)
     *            sending the message is aborted
     */
    public final AppMessage sendAndWait(final LogicalNode to)
        throws TransactionAbortException {

        // We should synchronize this, but the chances of interruption are too
        // small to justify it
        final long SSN = ssn++;
        final SingleWakeup wake = new SingleWakeup();
        synchronized (wakeupTable) {
            wakeupTable.put(SSN, wake);
        }
        send(to, SSN); //send the message
        try {
            return (AppMessage)wake.await();
        } catch (InterruptedException intEx) {
            final Thread thrd = Thread.currentThread();
            if (thrd instanceof KanThread) {
                final TransactionID trans = KanThread.currCall().currTrans;
                if (trans != null)
                    throw (TransactionAbortException)trans.killed;
            }
            Event.ERROR("SendAndWait interrupted");
            return null;
        } finally {
            synchronized (wakeupTable) {
                wakeupTable.remove(SSN);
            }
        }
    }

    /**
     * Send a message to a set of nodes and wait for all of them to say yes,
     * or for one of them to say no
     *
     * @param to the set of nodes to query
     * @return <code>true</code> if they all say yes; <code>false</code> if at
     *        least one says no
     * @exception TransactionAbortException if the transaction (if any)
     *            sending the message is aborted
     */
    public final boolean sendAndGetConsensus(final Collection to)
        throws TransactionAbortException {

        // We should synchronize this, but the chances of interruption are too
        // small to justify it
        final long SSN = ssn++;
        final GroupWakeup wake = new GroupWakeup(to.size());
        synchronized (wakeupTable) {
            wakeupTable.put(SSN, wake);
        }
        send(to, SSN); //send all messages
        try {
            return wake.await();
        } catch (InterruptedException intEx) {
            final Thread thrd = Thread.currentThread();
            if (thrd instanceof KanThread) {
                final TransactionID trans = KanThread.currCall().currTrans;
                if (trans != null)
                    throw (TransactionAbortException)trans.killed;
            }
            return false;
        } finally {
            wakeupTable.remove(SSN);
        }
    }

    /**
     * Send a message and register myself as a waiter, but wait later
     *
     * @param to the logical node to which the message is to be sent
     * @return a ticket to pass to <code>await</code> when I decide to wait
     *        later
     */
    public final long register(final LogicalNode to) {
        // We should synchronize this, but the chances of interruption are too
        // small to justify it
        final long SSN = ssn++;
        wakeupTable.put(SSN, new SingleWakeup());
        send(to, SSN);
        return SSN;
    }

    /**
     * Send a multicast message and register myself as a waiter, but wait
     * later
     *
     * @param to the logical nodes to which the message is to be sent
     * @return a ticket to pass to <code>await</code> when I decide to wait
     *        later
     */
    public final long register(final Collection to) {
        // We should synchronize this, but the chances of interruption are too
        // small to justify it
        final long SSN = ssn++;
        wakeupTable.put(SSN, new GroupWakeup(to.size()));
        send(to, SSN);
        return SSN;
    }

    /**
     * Wait for a reply to arrive for a previously sent message
     *
     * @param SSN the ticket that was returned from a previous call to
     *        {@link #register(Collection) register}
     * @return the message sent in reply
     * @exception TransactionAbortException if the transaction (if any)
     *            sending the message is aborted
     */
    public static AppMessage await(final long SSN)
        throws TransactionAbortException {

        final SingleWakeup wake = (SingleWakeup)wakeupTable.get(SSN);
        try {
            return (AppMessage)wake.await();
        } catch (InterruptedException intEx) {
            final Thread thrd = Thread.currentThread();
            if (thrd instanceof KanThread) {
                final TransactionID trans = KanThread.currCall().currTrans;
                if (trans != null)
                    throw (TransactionAbortException)trans.killed;
            }
            return null;
        } finally {
            synchronized (wakeupTable) {
                wakeupTable.remove(SSN);
            }
        }
    }

    /**
     * Wait for yes or no replies to arrive for previously sent messages
     *
     * @param SSN the ticket that was returned from a previous call to
     *        {@link #register(Collection) register}
     * @return <code>true</code> if they all said yes; <code>false</code> if
     *        at least one said no
     * @exception TransactionAbortException if the transaction (if any)
     *            sending the message is aborted
     */
    public static boolean getConsensus(long SSN)
        throws TransactionAbortException {

        final GroupWakeup wake = (GroupWakeup)wakeupTable.get(SSN);
        try {
            return wake.await();
        } catch (InterruptedException intEx) {
            final Thread thrd = Thread.currentThread();
            if (thrd instanceof KanThread) {
                final TransactionID trans = KanThread.currCall().currTrans;
                if (trans != null)
                    throw (TransactionAbortException)trans.killed;
            }
            return false;
        } finally {
            synchronized (wakeupTable) {
                wakeupTable.remove(SSN);
            }
        }
    }

    /**
     * Take the receive action for this message.  The default action is merely
     * to set the logical node of the receiving thread.
     */
    public void handle() {
        super.handle();
        LogicalNode.setMyNode(to);
    }

    /**
     * Save the state of the <code>AppMessage</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The send sequence number is written out (long), followed by
     *             the originating node (LogicalNode), and finally the
     *             destination node (LogicalNode) is written out.
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        out.writeLong(SSN);
        out.writeObject(from);
        out.writeObject(to);
    }

    public void readExternal(final ObjectInput in)
        throws IOException, ClassNotFoundException {

        SSN = in.readLong();
        from = (LogicalNode)in.readObject();
        to = (LogicalNode)in.readObject();

        Globals.objectInputMap.put(in, from);
    }

}
