/*
 * @(#)KanComm.java KanComm.java,v 1.1.1.1 2002/11/06 22:10:25 kris Exp $
 *
 *
 */

package mokan.comm;

import java.io.*;
import java.net.*;
import java.util.Iterator;
import java.util.HashMap;
import mokan.app.AppInfo;
import mokan.debug.Event;
import mokan.thread.KanThread;
import mokan.thread.ThreadPool;

import mokan.nameservice.*;
import mokan.obj.ObjectID;

/**
 * The gateway to all Network services.  It provides a means of initialization,
 * access to the local node identity, a list of known neighbors, and access to
 * a means of sending objects between the nodes of a network.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:25 $
 * @author	Jerry James
 * @author	Sean Brydon
 * @author	Sandeep Bhatia
 */
public final class KanComm {

  /**
   *   List of Available Networks to this KVM
   */
  static Network[] networks;

  /**
   *    This is the local Physical Node.
   */
  private static PhysicalNode localPhysicalNode;

  /**
   * All the machines in the universe with this KVM.  We initialize with a
   * load factor of 0.7, which should give good lookup speed and not waste
   * too much space, and a size of 47.  The upshot is that you have 32 slots
   * initially, and we don't expect most applications to be using any more
   * nodes than that.
   */

  /**
   * This map contains (uniqueID, PhysicalNode) pairs.
   */
  public static final HashMap uniqueID2node = new HashMap(47, 0.7f);


  /**
   * This map contains (Network.NodeID, PhysicalNode) pairs.
   */
  public static final HashMap NodeID2node =
      new HashMap(47, 0.7f);

  /**
   *  A private pool for communication listeners
   */
  private static final ThreadPool commpool = new ThreadPool("comm", false);

  /**
   * This represents a controller of some kind.  Currently, only the GUI
   * Commander uses this connection.  However, profiling or debugging tools
   * could conceivably use this mechanism.
   */
  //private static PhysicalNode controller;

  /** Used only for initializing PhysicalNode names */
  private static InheritableThreadLocal receivingFrom = new
      InheritableThreadLocal();

  /**
   * Create initially as part of the first ring of KVMs
   *
   * @param allNodes the physical nodes in the initial set
   * @param server the identity of a server node that controls the Kan
   *        system, or <code>null</code> if there is no controller
   */
  public static void init() throws java.net.UnknownHostException {

    networks = new Network[] {
        new TCPNetwork(Globals.listenPort)};

    localPhysicalNode = PhysicalNode.createLocalPhysicalNode();
    Globals.localPhysicalNode = localPhysicalNode;
    uniqueID2node.put(localPhysicalNode.getUniqueID(), localPhysicalNode);

    //  For each of the Support networks Initilize the network and add this
    //  this class as a listener.  We will listen for network event.
    for (int i = 0; i < networks.length; i++) {
      networks[i].addNetworkListener(new NetworkAdaptor() {
        /**
         *   Event signalled when network has recieved indiciation
         *   of a new node wishing to attach to the current KVM
         */
        public void networkAddNode(NetworkEvent e) {
          // Make sure we have some description of this
          // node.  How Can I determine that several
          // interfaces represent the same physical
          // Node.
          Network.NodeID nid = e.getNodeID();
          PhysicalNode pn = (PhysicalNode) e.getArgument();
          Event.DEBUG(Event.PHYSICAL_NODE, "KanComm.addNode ", nid);

          KanComm.NodeID2node.put(nid, pn);

          // Instantiate a KanThread to handle incomming Messages
          KanThread kt = commpool.getThreadFromPool(true);
          kt.setHandler(pn, localPhysicalNode);

        }

        /**  Event signalled when network has recieved
         *   indiciation of a new node wishing to detach.
         *   This may be failure or a requested
         *   termination.  */
        public void networkRemoveNode(NetworkEvent e) {

          Network.NodeID nid = e.getNodeID();
          Event.DEBUG(Event.PHYSICAL_NODE, "KanComm.removeNode ", nid);

          PhysicalNode pn = (PhysicalNode) NodeID2node.remove(nid);

          if (pn != null) {
            pn.removeNodeID(nid);

            Event.DEBUG(Event.PHYSICAL_NODE, "KanComm.removeNode: ",
                        pn + " -> " + pn.countNetworks());
            if (pn.countNetworks() == 0) {
              Event.DEBUG(Event.PHYSICAL_NODE, "KanComm.removeNode: last interface ",
                          pn);
            }
          }

        }
      });

      Network.NodeID lid = networks[i].init();
      localPhysicalNode.addNodeID(lid);
      NodeID2node.put(lid, localPhysicalNode);
    }
  }

  /**
   *   Turn off all connections and die.
   */
  public static void destroy() {
    Event.DEBUG(Event.NETWORK, "KanComm.destroy");
    // Kill networks and all comm threads
    //commpool.destroy();
  }

  /**
   *   Prepare for a shutdown by turning off  network listening.
   */
  public static void disableListening() {
    Event.DEBUG(Event.NETWORK, "KanComm.disableListening");
    for (int n = 0; n < networks.length; n++) {
      networks[n].disable();
    }
  }


//     /**
//      * Actively make a connection to another KVM and set up the
//      * {@linkmokan.comm.PhysicalNode physical node} and pipes
//      *
//      * @param newNode the {@linkmokan.comm.PhysicalNode physical node} that is
//      *        on the other end of the socket to be created
//      */
//     private static void createPipe(final PhysicalNode newNode) {
// 	// set the outputPipe, and the requestHandler for this node
// 	// NOTE : this method is to be called when we are trying to connect to
// 	// a remote node
// 	final InetAddress connectAddress = newNode.getPhysicalAddr();

// 	for (;;) {
// 	    try {
// 		final Socket socket = new Socket(connectAddress, newNode.port);
// 		// TODO: Make the next 2 lines comes from a config file
// 		socket.setTcpNoDelay(true);
// 		socket.setSoLinger(true, 2);
// 		final ObjectOutputStream oos = new ObjectOutputStream
// 		    (new BufferedOutputStream(socket.getOutputStream()));
// 		oos.flush();

// 		// set the values inside the newNode
// 		newNode.setSendPipe(oos);
// 		Event.DEBUG(Event.SOCKET_OPEN, "Opened socket to ", newNode);
// 		new KanThread(socket, newNode);
// 		return;
// 	    } catch (Exception ioe) {
// 		try {
// 		    // FIXME: Justify/document this literal
// 		    Thread.sleep(1000);
// 		} catch (InterruptedException ie) {
// 		    ie.printStackTrace();
// 		}
// 	    }
// 	}
//     }

//     /**
//      * Passively connect to another KVM.  Set up the
//      * {@linkmokan.comm.PhysicalNode physical node} and pipes.
//      *
//      * @param newSocket the socket just created by the server socket
//      * @exception SocketException if network problems occur
//      */
//     private static void createPipe(final Socket newSocket)
// 	throws SocketException {

// 	//  out who the pipe is from, search for the corresponding node in
// 	// the hashtable, and do the needful
// 	newSocket.setTcpNoDelay(true);
// 	newSocket.setSoLinger(true, 2);
// 	InetAddress myClient = newSocket.getInetAddress();
// 	PhysicalNode currPhysicalNode =
// 	    (PhysicalNode)allPhysicalNodes.get(myClient.hashCode());
// 	Event.DEBUG(Event.SOCKET_ACCEPT, "Accepted socket from ",
// 		    currPhysicalNode);

// 	try {
// 	    // get the necessary fields
// 	    ObjectOutputStream oos = new ObjectOutputStream
// 		(new BufferedOutputStream(newSocket.getOutputStream()));
// 	    oos.flush();

// 	    // set the necessary fields
// 	    currPhysicalNode.setSendPipe(oos);
// 	    new KanThread(newSocket, currPhysicalNode) ;
// 	} catch (IOException ioe) {
// 	    ioe.printStackTrace();
// 	    System.exit(0);
// 	}
//    }

  /**
   * Shut down the entire network of Kan Virtual Machines
   */
//     public static void shutdownAllKVMs() {
// 	final Iterator physNodeList = allPhysicalNodes.valueIterator();
// 	final DieMsg die = new DieMsg();
// 	while (physNodeList.hasNext()) {
// 	    final PhysicalNode node = (PhysicalNode)physNodeList.next();
// // 	    if (node.compareTo(me) != 0)
//  		die.send(node);
// 	}
// 	System.exit(0);
//     }


  /**
   * Retrieve the local physical node
   *
   * @param addr the IP address of the node to lookup
   * @return the <code>PhysicalNode</code> object with that IP address
   */
  public static PhysicalNode getPhysicalNode() {
    return localPhysicalNode;
  }

  /**
   *  Send a message to a PhysicalNode
   *  If the node belongs to several networks KanComm will choose
   *  the route.
   */
  public static void send(PhysicalNode pn, Message m) throws IOException {
    // Choose a network and send with it.

    pn.send(m);

  }

  /**
   *   Recieve a message from a phsyicalNode
   *   TODO:  need a thread to listen to each network interface
   *
   */
  public static Message recv(PhysicalNode pn) throws IOException {
    receivingFrom.set(pn);
    return pn.recv();
  }

  /**
   *   Close all connections to a physicalNode
   *   TODO:  need a thread to listen to each network interface
   */
  public static void close(PhysicalNode node) {

    node.close();

  }

   public static final PhysicalNode connect(String address, int port, String remoteID) {
     TCPNetwork tcp = (TCPNetwork)networks[0];
     return tcp.connect(address, port, remoteID);
   }

  /**
   * Called by an external source (NOT in current ring/NW of KVMs).  This
   * method will notify the other KVMs to start the app by sending
   * {@linkmokan.app.StartAppMsg start application messages} through the
   * normal channels.
   *
   * @param newAppInfo an object describing the application to start
   * @param machines a list of machines on which to start the application
   */
  public static void startApp(final AppInfo newAppInfo,
                              final MachinePort[] machines) {
    // add as a local running application
// 	final PhysicalNode myNode = getPhysicalNode(me);
// 	final Application app = new Application(machines, myNode);

// 	// send msgs to others to start the new application
// 	final StartAppMsg mess = new StartAppMsg(app);
// 	for (int i = 0; i < machines.length; i++) {
// 	    mess.send(getPhysicalNode(machines[i].addr));
// 	}

// 	try {
// 	    newAppInfo.invoke(app);
// 	} catch (Exception e) {
// 	    System.err.println("Cannot invoke " + newAppInfo);
// 	    e.printStackTrace();
// 	}
  }
}
