/*
 * @(#)$Id: KanURL.java,v 1.1.1.1 2002/11/06 22:10:25 kris Exp $
 *
 *
 */


package mokan.comm;

import mokan.Kan;
import java.io.*;
import java.net.MalformedURLException;
import java.util.StringTokenizer;
import mokan.nameservice.Globals;

/**
 * A kan URL is a string representation of a Kan Object.
 * There are several object that a represented by the different
 * forms of url's.
 *
 * <code>PhysicalNode</code> kan://networkaddress
 * <code>Application</code>  kan://networkaddress/applicationid
 * <code>LogicalNode</code>  kan://networkaddress/applicationid/nodename
 * <code>GlobalObject</code> kan://networkaddress/applicationid/nodename/objid
 *
 * A network address should be machine[:port]
 * An applicationid is a fully qualified classname[:instance] (i.e.mokan.example.whiteboard.Whiteboard)
 * A nodename is again a machinename[:port]
 * An objid is a classname:instance
 *
 */
public class KanURL implements Externalizable{

    /**
     * The default port
     */
    //private static final int DEFAULT_PORT;

    private transient String host;
    private transient int port;
    private transient String app;
    private transient int instance;
    private transient String node;
    private transient int    count;

    /** PsuedoFinal
     */
    private String fullurl;
    private int hash ;


    private void parse (String u) throws MalformedURLException {

        int colon;

        app = null;
        node = null;
        count = -1;
        if (u.startsWith ("kan://")) {
            StringTokenizer st = new StringTokenizer (u.substring (6), "/");

            // Physical Node
            String piece = st.nextToken();
            colon = piece.indexOf (':');
            if (colon > 0) {
                host = piece.substring(0, colon);
                port = Integer.parseInt(piece.substring(colon+1));
            } else {
                host = piece;
                port = Globals.DEFAULT_PORT;
            }

            // Application ID
            if (!st.hasMoreTokens()) return;
            piece = st.nextToken();
            colon = piece.indexOf (':');
            if (colon > 0) {
                app = piece.substring(0, colon);
                instance = Integer.valueOf(piece.substring(colon+1)).intValue();
            } else {
                app = piece;
                instance = 0;
            }

            // LogicalNode
            if (!st.hasMoreTokens()) return;
            node = st.nextToken();

            // Object ID
            if (!st.hasMoreTokens()) return;
            count = Integer.valueOf(st.nextToken()).intValue();
        }
    }

    private String canonicalurl () {
        StringBuffer r = new StringBuffer("kan://");
        r.append ( host );
        if (port != Globals.DEFAULT_PORT) {
            r.append (':'); r.append (port);
        }
        if (app != null) {
            r.append ('/'); r.append (app);
            r.append (':'); r.append (instance);
            if (node != null) {
                r .append ('/'); r.append (node);
                if (count >= 0) {
                    r .append ('/'); r.append (count);
                }
            }
        }
        return r.toString();
    }
    private void init (String h, String a, int inst, String _node, int c) {
        host = h;
        port = Globals.DEFAULT_PORT;
        app = a;
        instance = inst;
        node = _node;
        count = c;
    }

    public KanURL () {}
    /**
     *   Create kan URL based on a string
     *   The URL takes the form
     *     kan://host[:port]/class[:instance]/      - Applicaion Identifier
     *     kan://host[:port]/class#instance/host  - Node Identifier Identifier
     *     kan://host[:port]/class#instance/host#instance - Object Identifier
     */
    public KanURL (String u) throws MalformedURLException{
        parse(u);
        fullurl = canonicalurl();
        hash= fullurl.hashCode();
    }
    public KanURL (String h, String a, int inst) {
        init (h,a,inst, null, -1);
        fullurl = canonicalurl();
        hash=fullurl.hashCode();
    }
    public KanURL (String h, String a, int inst, String n) {
        init (h,a,inst, n, -1);
        fullurl = canonicalurl();
        hash=fullurl.hashCode();
    }
    public KanURL (String h, String a, int inst, String n, int c) {
        init (h,a,inst, n, c);
        fullurl = canonicalurl();
        hash=fullurl.hashCode();
    }
    public KanURL (KanURL appURL, String _node) {
        init (appURL.host, appURL.app, appURL.instance, _node, -1);
        fullurl = canonicalurl();
        hash=fullurl.hashCode();
    }
    /**
     *  Return a valid URL representing the Application Identifier
     */
    public KanURL getAppURL () {
        return new KanURL (host, app, instance);
    }
    /**
     *  Return a valid URL representing the Node Identifier
     */
    public KanURL getNodeURL () {
        if (node == null)
            return new KanURL (host, app, instance, host);
        else
            return new KanURL (host, app, instance, node);
    }

    public String getAppHost () {
        return host;
    }
    public int getAppPort () {
        return port;
    }

    public String getAppName () {
        return app;
    }
    public int getAppInst () {
        return instance;
    }

    public String getObjHost () {
        return node;
    }

    public int getObjInst () {
        return count;
    }
    public String toString() {
        return fullurl;
    }

    public int hashCode () {
        return hash;
    }

    public boolean equals (Object o) {
        return (o instanceof KanURL) && fullurl.equals(((KanURL)o).fullurl);
    }
    /**
     * Save the state of the <code>KanURL</code> instance to a stream (i.e.,
     * serialize it).
     *
     * @serialData The superclass data is written out, followed by the send
     *             sequence number of this reply (long).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
	out.writeObject(fullurl);
    }

    public void readExternal(final ObjectInput in)
	throws IOException, ClassNotFoundException {
	fullurl = (String)in.readObject();
        try {
            parse(fullurl);
            hash = fullurl.hashCode();
        } catch (MalformedURLException me) {
            me.printStackTrace();
        }
    }

}
