/*
 * @(#)LogicalNode.java LogicalNode.java,v 1.2 2002/11/22 23:26:06 kris Exp $
 *
 *
 */

package mokan.comm;

import java.io.*;
//import java.net.InetAddress;

import mokan.util.Hash2Map;
import mokan.app.Application;
import mokan.obj.ObjectID;
import mokan.thread.ThreadID;
import mokan.transact.TransactionID;
import mokan.nameservice.*;

import java.util.Collection;

import mokan.debug.Event;

/**
 * A representation of a computing node in the distributed system.  Logical
 * nodes hold objects and threads.  However, logical nodes are only weakly
 * tied to physical machines.  In the event of a failure, a logical node can
 * migrate to a new physical node.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author	Jerry James
 * @author	Sean Brydon
 * @author	Sandeep Bhatia
 */
public final class LogicalNode implements Comparable, Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -6097181915878513550L;

    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
        ObjectStreamClass.NO_FIELDS;

    /**
     * The <code>LogicalNode</code> associated with each thread.
     * Basically this is a hashtable of (key = thread, value = LogicalNode).
     */
    private static final InheritableThreadLocal myNode =
        new InheritableThreadLocal();

    /** Mape from (Application, PhysicalNode) -> LogicalNode */
    private static final Hash2Map logicalNodeMap = new Hash2Map();

    /**
     * A unique number for each node object.  These numbers <em>must</em> be
     * unique.  This is a pseudo-final field.
     *
     * @serial a unique number for this logical node
     */
//KK    private long nodeID;

    /**
     * The hashcode for this <code>LogicalNode</code>.  This is a pseudo-final
     * field.
     *
     * @serial the hashcode for this logical node
     */
    private transient int myHashCode;

    // IMPLEMENTATION NOTE:
    // Originally, we had a single field of type NodeHandlerInterface that
    // held either the LocalLogicalNode or RemoteLogicalNode object, as
    // appropriate.  However, we don't really need the RemoteLogicalNode
    // object, since it was just a placeholder.  So now we have a field that
    // is specifically of type <code>LocalLogicalNode</code> (to avoid all the
    // casting we had to do before).  If it holds <code>null</code>, then the
    // logical node is remote.

    /**
     * The local component of the logical node, if any.  If this field is
     * <code>null</code>, then the logical node lives on a remote node.
     */
    private transient LocalLogicalNode localHandler;

    /**
     *  The URL to represent this LogiccalNode (App,Node)
     */
    private KanURL kanurl;

    /**
     * The physical node currently housing this logical node
     */
    private transient PhysicalNode myPhysNode;

    /**
     * The application of which this logical node is a part
     */
    private transient Application app;

    /**
     * Determine the logical node executing the current thread
     *
     * @return the local logical node
     */
    public static LogicalNode getMyNode() {
        return (LogicalNode) myNode.get();
    }

    /**
     * Register a logical node as the one executing the current thread
     *
     * @param node the logical node executing the current thread
     */
    public static void setMyNode(LogicalNode node) {
        myNode.set(node);
    }

    public static LogicalNode getLogicalNode(Application app, PhysicalNode pn){
        LogicalNode ln = (LogicalNode) logicalNodeMap.get(app, pn);
        return ln;
    }

    public static LogicalNode getLogicalNode(KanURL kurl) {
        Application app = Application.getApplication(kurl);
        String uniqueID = kurl.getObjHost();
        PhysicalNode pn = (PhysicalNode) KanComm.uniqueID2node.get(uniqueID);

        return (LogicalNode) logicalNodeMap.get(app, pn);
    }

    public static void remove(LogicalNode ln) {
        logicalNodeMap.remove(ln.app, ln.myPhysNode);
    }

    /**
     * The no-args constructor, needed for deserialization
     */
    public LogicalNode() {
        app = null; // Have to set this, because it's final
    }

    public LogicalNode(KanURL appURL, PhysicalNode node) throws java.net.
        MalformedURLException {
        app = Application.getApplication(appURL);
        myPhysNode = node;
        if (myPhysNode == KanComm.getPhysicalNode()) {
            localHandler = new LocalLogicalNode(this);
        }
        kanurl = new KanURL(appURL, node.toString());

        //app.addNode (this);
        setMyNode(this);

        logicalNodeMap.put(app, myPhysNode, this);

        Event.DEBUG(Event.LOG_NODE_CREATE, "LogicalNode.LogicalNode ", this);
    }

    /**
     * Create a new <code>LogicalNode</code> for the main application
     *
     * @param applic the application of which this logical node is a part
     * @param physNode the physical node currently holding this logical node
     * @param nodeID a unique number representing this logical node
     * @param local <code>true</code> if this logical node exists on the local
     *        physical node
     */
    public LogicalNode(final Application applic, final PhysicalNode physNode) {
        app = applic;
        myPhysNode = physNode;
        localHandler = null;
        if (physNode == KanComm.getPhysicalNode()) {
            localHandler = new LocalLogicalNode(this);
        }
        kanurl = new KanURL(applic.getURL(), physNode.getUniqueID());
        //app.addNode (this);
        setMyNode(this);
        logicalNodeMap.put(app, myPhysNode, this);

        Event.DEBUG(Event.LOG_NODE_CREATE, "LogicalNode.LogicalNode ", this);
    }

    /**
     * Prepare for the removal of this logical node from the system
     */
    public void destroy() {
        Event.DEBUG(Event.LOGICAL_NODE, "LogicalNode.destroy:");

        if (!isLocal()) {
            ObjectID.removeObjectRefs(this);
            TransactionID.abortTransactions(this, new mokan.app.LeaveException());
        }

        KanComm.close(myPhysNode);
    }

    /**
     *  Stop this not from doing anymore processing. Part of the
     *  process during removal.
     */
    public void stop() {
        //Event.assert (localHandler != null, "LogicalNode.stop: Not Local");
        Event.DEBUG(Event.LOGICAL_NODE, "LogicalNode.stop:");

        localHandler.allowObjectCreations(false);
        // This will request that objects be dropped
        ObjectID.removeObjectRefs(this);
        TransactionID.abortTransactions(null, new mokan.app.LeaveException());
    }

    /**
     * Retrieve the local handler for this node
     *
     * @return the local handler for this node, or <code>null</code> if it is
     *         a remote node
     */
    public LocalLogicalNode getLocalHandler() {
        return localHandler;
    }

    /**
     * retrienve the physical node
     */
    public PhysicalNode getPhysicalNode() {
        return myPhysNode;
    }

    /**
     * Send a message to this node.  If the node is local, no network activity
     * occurs.
     *
     * @param message the message to send
     */
    public void send(final Message message) {
        try {
            if (localHandler == null) {

                // Added by Huahai
                if (message instanceof AppMessage) {
                    Globals.neighbors.add(this);
                }

                KanComm.send(myPhysNode, message);
            } else {
                // This is a local logical node.  We do not want to send the
                // message over the network, but instead just process it locally.
                message.handle();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    /**
     * Get the application of which this logical node is a part
     *
     * @return the application of which this logical node is a part
     */
    public Application getApplication() {
        return app;
    }

    /**
     * Get the index of this node in the application node list
     *
     * @return the index of this node in the application node list
     */
//     public int getIndex() {
// 	//return (int)(nodeID & 0xff);
//         return 0;
//     }

    public KanURL getURL() {
        return kanurl;
    }

    /**
     * Determine whether this node is local
     *
     * @return <code>true</code> if this node is local; <code>false</code> if
     *         it is remote
     */
    public boolean isLocal() {
        return localHandler != null;
    }

    /**
     * Serialize this logical node for the first time
     *
     * @param out the object stream on which to serialize
     * @exception IOException if anything goes wrong during serialization
     */
    public void writeInitially(final ObjectOutput out) throws IOException {
        //out.writeObject(myPhysNode.getPhysicalAddr());
    }

    /**
     * Save the state of the <code>LogicalNode</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The unique identifier for this logical node is written out
     *             (long).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        out.writeObject(kanurl);
    }

    public void readExternal(final ObjectInput in) throws ClassNotFoundException,
        IOException {
        kanurl = (KanURL) in.readObject();
    }

    public Object readResolve() throws ObjectStreamException {

        Object node = getLogicalNode(kanurl);
        Event.DEBUG(Event.LOGICAL_NODE, "readresolve " + kanurl + '=' + node);

        if (node == null) {
            Application app = Application.getApplication(kanurl);
            String uniqueID = kanurl.getObjHost();
            PhysicalNode pn = (PhysicalNode) KanComm.uniqueID2node.get(uniqueID);
            if(pn == null) {
                pn = new PhysicalNode(uniqueID);
                KanComm.uniqueID2node.put(uniqueID, pn);
            }
            if(app == null || pn == null) {
                throw new RuntimeException("LogicalNode.readResolve: invalid kanurl " + kanurl);
            }
            node = new LogicalNode(app, pn);
        }
        return node;
    }

    // This is okay, because we only compare LogicalNodes of the same
    // application, and they have the same upper 48 bits
    public int compareTo(final LogicalNode node) {
        return myPhysNode.compareTo(node.myPhysNode);
    }

    public int compareTo(final Object obj) throws ClassCastException {
        return compareTo( (LogicalNode) obj);
    }

//     public int compareTo(final long nodeID) {
// 	return (int)(this.nodeID - nodeID);
//     }

    public boolean equals(final LogicalNode node) {
        return myPhysNode.equals(node.myPhysNode);
    }

//     public boolean equals(final long nodeID) {
// 	return this.nodeID == nodeID;
//     }

    public boolean equals(final Object obj) {
        return (obj instanceof LogicalNode) && equals( (LogicalNode) obj);
    }

    public int hashCode() {
        return myHashCode;
    }

    public String toString() {
        return myPhysNode.toString();
    }
}
