/*
 * @(#)$Id: PhysicalNode.java,v 1.2 2002/11/22 23:26:06 kris Exp $
 *
 *
 */
package mokan.comm;

import java.io.*;
import java.net.*;
import java.util.*;
import mokan.debug.Event;

import mokan.nameservice.Globals;

/**
 * The address and pipes to a physical node on the network
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author	Sean Brydon
 * @author	Sandeep Bhatia
 * @author	Jerry James
 */
public final class PhysicalNode implements Comparable, Externalizable {

  /* This stuff is for checking message sizes
       private static final int INIT_SIZE = 4096;
       private transient final ByteArrayOutputStream byteOut =
    new ByteArrayOutputStream(INIT_SIZE);
       private transient final ObjectOutputStream objOut;
   */

  /**
   * The <var>serialVersionUID</var> for this class.  If you change the
   * signature of this class (field names or types, method names, parameter
   * names or types, adding or removing fields or methods, etc.), you
   * <strong>must</strong> change this value.  The recommended way is to use
   * the <code>serialver</code> tool.
   */
   static final long serialVersionUID = 7635499198051741382L;

  /**
   * The list of persistent fields for this class.  This is empty, because
   * that is the fastest option for <code>Externalizable</code> classes when
   * we are not worried about versioning.
   */
  private static final ObjectStreamField[] serialPersistentFields =
      ObjectStreamClass.NO_FIELDS;

  /**
   * A unique timestamp for physical nodes, for distinguishing different
   * incarnations of a node in between crashes
   */
  //private static long timeStamp = -1L;

  /**
   * Each PhysicalNode has a unique ID which is a global reference to the
   * PhysicalNode. A unique ID is generated ONLY when a local PhysicalNode
   * is created.
   * Format:
   *  "address:port,timestamp"
   */
  private transient String uniqueID;

  /**
   * An array of Network.NodeID's.  This PhysicalNode participates
   *  In the following networks and is accessible by them.
   */
  private transient ArrayList nids;
  /**
   *   This physical node has a preferred comminucation path
   */
  private transient Network.NodeID preferredNid;

  private boolean connected = false;

  /**
   * The no-args constructor, needed for deserialization
   */
  public PhysicalNode() {
    super();

    nids = new ArrayList();
    //machineName = new ArrayList();
    uniqueID = "tmp";
    preferredNid = null;
  }

  /**
   * Create a non-local, not-connected physical node.
   * Note:
   * A PhysicalNode can NOT exist without a uniqueID;
   * A uniqueID is generated ONLY when creating a local PhysicalNode;
   * A PhysicalNode may be or may be not CONNECTED;
   * A PhysicalNode is created when a connection is established, or when
   * reading a PhysicalNode object from a stream. In the latter case, the
   * PhysicalNode may be not connected.
   *
   * @author Huahai He
   * @param uniqueID
   */
  public PhysicalNode(String uniqueID) {
    nids = new ArrayList();
    this.uniqueID = uniqueID;
    preferredNid = null;
    connected = false;
    Event.DEBUG(Event.PHYS_NODE_CREATE, "Created physical node ", this);
  }


  /**
   * Create a new <code>PhysicalNode</code> to represent the node with a
   * particular IP address and port number for listening, using a particular
   * output stream for sending objects
   *
   *
   * @param mach the machine to represent
   * @param output the output stream to use for communication with this
   *        physical node
   * @param time the timestamp for this physical node
   */
  public PhysicalNode(final Network.NodeID nid, String uniqueID) {
    nids = new ArrayList();
    nids.add(nid);
    preferredNid = nid;
    this.uniqueID = uniqueID;
    connected = true;
    Event.DEBUG(Event.PHYS_NODE_CREATE, "Created physical node ", this);
  }

  /**
   * Create the local physical node.
   * @return
   */
  static public PhysicalNode createLocalPhysicalNode() throws java.net.
      UnknownHostException {
    String uniqueID = InetAddress.getLocalHost().getHostName() + ':' +
        Globals.listenPort + ',' + System.currentTimeMillis();
    return new PhysicalNode(uniqueID);
  }

  /**
   * Set the connected status of the physical node
   * @param _connected
   */
  public void setConnected(boolean _connected) {
    connected = _connected;
  }
  public void close() {
    try {
      preferredNid.network.close(preferredNid);
    } catch (IOException io) {
      Event.DEBUG(Event.PHYSICAL_NODE, "PhysicalNode.close exception", io);
    }
  }

  public void setPreferredNetworkID(Network.NodeID nid) {
    preferredNid = nid;
  }

  public String getUniqueID() {
    return uniqueID;
  }

  /**
   * Get the DNS-resolvable name of the node represented by this physical
   * node
   * @depracated
   * @return the name of the node
   */
  public String getMachineName() {
// 	return machineNames.get(0);
//        return preferredNid.getName();
    return uniqueID.substring(0, uniqueID.indexOf(':'));
  }

  /**
   *  Receive a message from this Physical Node.
   *  Deprecated -- use KanComm.recv
   */
  public Message recv() throws IOException {
    if(!connected) {
      String address = uniqueID.substring(0, uniqueID.indexOf(':'));
      int port = Integer.parseInt(uniqueID.substring(uniqueID.indexOf(':')+1, uniqueID.indexOf(',')));
      KanComm.connect(address, port, uniqueID);
    }
    return preferredNid.network.recv(preferredNid);
  }

  /**
   * Send a message to this physical node
   *  Deprecated -- use KanComm.send
   *
   * @param message the message to send
   */
  public void send(final Message message) throws IOException {
    //KK TODO PhysicalAddr.send (choose network type).
    if (this == Globals.localPhysicalNode) {
      message.handle();
    } else {
      if (!connected) {
        String address = uniqueID.substring(0, uniqueID.indexOf(':'));
        int port = Integer.parseInt(uniqueID.substring(uniqueID.indexOf(':')+1, uniqueID.indexOf(',')));
        KanComm.connect(address, port, uniqueID);
      }
      preferredNid.network.send(preferredNid, message);
    }
  }

  /** PhysicalNode keep track of which networks they belong to
   */
  public void addNodeID(Network.NodeID id) {
    //KK Shouldn't always set preferredNid.
    nids.add(id);
    preferredNid = id;

    Event.DEBUG(Event.PHYSICAL_NODE, "PhysicalNode.addNodeID: " + id + "=" + this);
  }

  /** PhysicalNode keep track of which networks they belong to
   */
  public void removeNodeID(Network.NodeID id) {
    Event.DEBUG(Event.PHYSICAL_NODE, "PhysicalNode.removeNodeID: ", id);
    //machineNames.remove(id.getName());
    int index = nids.indexOf(id);
    if (index < 0) {
      return;
    }
    nids.remove(id);

    if (nids.size() == 0) {
      connected = false;
    }
  }

  public int countNetworks() {
    return nids.size();
  }

  public boolean containsNodeID(Network.NodeID id) {
    return nids.indexOf(id) >= 0;
  }

  /**
   * Save the state of the <code>PhysicalNode</code> instance to a stream
   * (i.e., serialize it).
   *
   * @serialData The IP address of the node is written out (InetAddress).
   */
  public void writeExternal(final ObjectOutput out) throws IOException {
    out.writeObject(uniqueID);
  }

  public void readExternal(final ObjectInput in) throws ClassNotFoundException,
      IOException {
    uniqueID = (String) in.readObject();
  }

  public Object readResolve() throws ObjectStreamException {
    PhysicalNode pn = (PhysicalNode) KanComm.uniqueID2node.get(uniqueID);
    if(pn != null) {
      return pn;
    } else {
      KanComm.uniqueID2node.put(uniqueID, this);
      return this;
    }
  }

  public int compareTo(final PhysicalNode node) throws ClassCastException {
    return uniqueID.compareTo(node.uniqueID);
  }

//     public int compareTo(final InetAddress addr) throws ClassCastException {
// 	return compareIP(physAddr.getAddress(), addr.getAddress());
//     }

  public int compareTo(final Object obj) throws ClassCastException {
    return compareTo( (PhysicalNode) obj);
  }

  /**
   * Impose a total ordering on IP addresses.  Compare a.b.c.d and w.x.y.z
   * by comparing a and w; if the are equal, then compoare b and x; if they
   * are equal ...  This is equivalent to converting each IP address to its
   * corresponding 32-bit number and comparing the result as integers
   *
   * @param ip1 the first IP address to compare
   * @param ip2 the second IP address to compare
   */
//     private static int compareIP(final byte[] ip1, final byte[] ip2) {
// 	for (int i = 0; i < ip1.length; i++) {
// 	    if (ip1[i] != ip2[i])
// 		return (int)ip1[i] - (int)ip2[i];
// 	}
// 	return 0;
//     }

  /** Physical nodes are equal when they have a Nid in common */
  public boolean equals(final PhysicalNode node) {
    //KK2002-02-17 return machineName.equals(node.machineName);

    return uniqueID.equals(node.uniqueID);
  }

//     public boolean equals(final InetAddress nodePhysicalAddr) {
//         return physAddr.equals(nodePhysicalAddr);
//     }

  public boolean equals(final Object obj) {
    return (obj instanceof PhysicalNode) && equals( (PhysicalNode) obj);
  }

//     public int hashCode() {
// 	return uniqueID.hashCode();
//     }

  public String toString() {
    return uniqueID;
  }

}
