/*
 * @(#)$Id: SysMessage.java,v 1.2 2002/11/22 23:26:06 kris Exp $
 *
 *
 */

package mokan.comm;

import java.io.*;
import java.util.Collection;
import java.util.Iterator;
import mokan.util.LongHashMap;

/**
 * A class of {@linkmokan.comm.Message message objects} that are intended for
 * system messages.  These messages are passed between physical nodes.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author	Jerry James
 */
public abstract class SysMessage extends Message implements Externalizable {

  /**
   * The <var>serialVersionUID</var> for this class.  If you change the
   * signature of this class (field names or types, method names, parameter
   * names or types, adding or removing fields or methods, etc.), you
   * <strong>must</strong> change this value.  The recommended way is to use
   * the <code>serialver</code> tool.
   */
  static final long serialVersionUID = 6686601412282981482L;

  /**
   * The list of persistent fields for this class.  This is empty, because
   * that is the fastest option for <code>Externalizable</code> classes when
   * we are not worried about versioning.
   */
  private static final ObjectStreamField[] serialPersistentFields =
      ObjectStreamClass.NO_FIELDS;

  /**
   * Message sequence number (sending).
   */
  private static long ssn = 0;

  /**
   * Hash table with send sequence number as key and
   * {@linkmokan.comm.Wakeup Wakeup} as element
   */
  protected static final LongHashMap wakeupTable = new LongHashMap();

  /**
   * Send Sequence Number; the number of messages sent by this node so far
   *
   * @serial the send sequence number for this message
   */
  protected long SSN;

  /**
   * The (logical or physical) source of the message
   *
   * @serial the source node of this message
   */
  protected PhysicalNode from;

  /**
   * The (logical or physical) destination of the message
   *
   * @serial the destination node of this message
   */
  protected PhysicalNode to;

  /**
   * Create a new <code>SysMessage</code>
   *
   * @param needsReset <code>true</code> if the outgoing stream should be
   *        reset before sending this message
   * @param canBlock <code>true</code> if the handling of this message might
   *        block the handling thread
   */
  protected SysMessage(final boolean needsReset, final boolean canBlock) {
    super(needsReset, canBlock);
  }

  /**
   * Send a point-to-point message
   *
   * @param to the destination {@linkmokan.comm.PhysicalNode node} for this
   *        message
   * @return the sequence number for this message
   */
  public final long send(final PhysicalNode to) throws IOException {
    this.to = to;
    // We should synchronize this, but the chances of interruption are too
    // small to justify it
    SSN = ssn++;
    to.send(this);
    return SSN;
  }

  /**
   * Send a point-to-point message that already has a sequence number
   *
   * @param to the destination {@linkmokan.comm.PhysicalNode node} for this
   *        message
   * @param sequence the sequence number for this message
   */
  public final void send(final PhysicalNode to, final long sequence) throws
      IOException {
    this.to = to;
    SSN = sequence;
    to.send(this);
  }

  /**
   * Send a multicast message
   *
   * @param dests the destination <code>Node</code>s for this message
   * @return the sequence number for this message
   */
  public final long send(final Collection dests) throws IOException {
    // We should synchronize this, but the chances of interruption are too
    // small to justify it
    SSN = ssn++;
    final Iterator nodes = dests.iterator();
    while (nodes.hasNext()) {
      to = (PhysicalNode) nodes.next();
      to.send(this);
    }
    return SSN;
  }

  /**
   * Send a multicast message that already has a sequence number
   *
   * @param dests the destination <code>Node</code>s for this message
   * @param sequence the sequence number for this message
   */
  public final void send(final Collection dests, final long sequence) throws
      IOException {
    SSN = sequence;
    final Iterator nodes = dests.iterator();
    while (nodes.hasNext()) {
      to = (PhysicalNode) nodes.next();
      to.send(this);
    }
  }

  /**************** Active message stuff ***************/

  /**
   * Send a message and wait for the reply to arrive
   *
   * @param to the node to which the message should be sent
   * @return the message sent in reply
   */
  public final SysMessage sendAndWait(final PhysicalNode to) throws IOException {
    // We should synchronize this, but the chances of interruption are too
    // small to justify it
    final long SSN = ssn++;
    final SingleWakeup wake = new SingleWakeup();
    synchronized (wakeupTable) {
      wakeupTable.put(SSN, wake);
    }
    send(to, SSN); //send the message
    try {
      return (SysMessage) wake.await();
    } catch (InterruptedException intEx) {
      // This should be impossible
      // FIXME: Assert that this doesn't happen
      return null;
    }
    finally {
      synchronized (wakeupTable) {
        wakeupTable.remove(SSN);
      }
    }
  }

  /**
   * Send a message to a set of nodes and wait for all of them to say yes,
   * or for one of them to say no
   *
   * @param to the set of nodes to query
   * @return <code>true</code> if they all say yes; <code>false</code> if at
   *        least one says no
   */
  public final boolean sendAndGetConsensus(Collection to) throws IOException {
    // We should synchronize this, but the chances of interruption are too
    // small to justify it
    final long SSN = ssn++;
    final GroupWakeup wake = new GroupWakeup(to.size());
    synchronized (wakeupTable) {
      wakeupTable.put(SSN, wake);
    }
    send(to, SSN); //send all messages
    try {
      return wake.await();
    } catch (InterruptedException intEx) {
      // This should be impossible
      // FIXME: Assert that this doesn't happen
      return false;
    }
    finally {
      wakeupTable.remove(SSN);
    }
  }

  /**
   * Send a message and register myself as a waiter, but wait later
   *
   * @param to the node to which the message is to be sent
   * @return a ticket to pass to <code>await</code> when I decide to wait
   *        later
   */
  public final long register(final PhysicalNode to) throws IOException {
    // We should synchronize this, but the chances of interruption are too
    // small to justify it
    final long SSN = ssn++;
    wakeupTable.put(SSN, new SingleWakeup());
    send(to, SSN);
    return SSN;
  }

  /**
   * Send a multicast message and register myself as a waiter, but wait
   * later
   *
   * @param to the nodes to which the message is to be sent
   * @return a ticket to pass to <code>await</code> when I decide to wait
   *        later
   */
  public final long register(final Collection to) throws IOException {
    // We should synchronize this, but the chances of interruption are too
    // small to justify it
    final long SSN = ssn++;
    wakeupTable.put(SSN, new GroupWakeup(to.size()));
    send(to, SSN);
    return SSN;
  }

  /**
   * Wait for a reply to arrive for a previously sent message
   *
   * @param SSN the ticket that was returned from a previous call to
   *        <code>register</code>
   * @return the message sent in reply
   */
  public final SysMessage await(final long SSN) {
    final SingleWakeup wake = (SingleWakeup) wakeupTable.get(SSN);
    try {
      return (SysMessage) wake.await();
    } catch (InterruptedException intEx) {
      // This should be impossible
      // FIXME: Assert that this doesn't happen
      return null;
    }
    finally {
      synchronized (wakeupTable) {
        wakeupTable.remove(SSN);
      }
    }
  }

  /**
   * Wait for yes or no replies to arrive for previously sent messages
   *
   * @param SSN the ticket that was returned from a previous call to
   *        <code>register</code>
   * @return <code>true</code> if they all said yes; <code>false</code> if
   *        at least one said no
   */
  public final boolean getConsensus(final long SSN) {
    final GroupWakeup wake = (GroupWakeup) wakeupTable.get(SSN);
    try {
      return wake.await();
    } catch (InterruptedException intEx) {
      // This should be impossible
      // FIXME: Assert that this doesn't happen
      return false;
    }
    finally {
      synchronized (wakeupTable) {
        wakeupTable.remove(SSN);
      }
    }
  }

  /**
   * Save the state of the <code>SysMessage</code> instance to a stream
   * (i.e., serialize it).
   *
   * @serialData The send sequence number is written out (long), then the
   *             originating logical node (LogicalNode), followed by the
   *             destination logical node (LogicalNode).
   */
  public void writeExternal(final ObjectOutput out) throws IOException {
    out.writeLong(SSN);
    out.writeObject(from);
    out.writeObject(to);
  }

  public void readExternal(final ObjectInput in) throws IOException,
      ClassNotFoundException {
    SSN = in.readLong();
    from = (PhysicalNode) in.readObject();
    to = (PhysicalNode) in.readObject();
  }
}
