/*
 * @(#)CompileContext.java CompileContext.java,v 1.1.1.1 2002/11/06 22:10:25 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler;

import java.util.LinkedList;
import mokan.compiler.stmt.StatementAST;

/**
 * Information about the compilation context so that individual AST nodes can
 * set up needed values
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:25 $
 * @author Jerry James
 */
public final class CompileContext {

    /**
     * Whether the current construct is a field (top-level construct in a
     * class), or not.  This is used to distinguish local variables from
     * static and instance variables.
     */
    public boolean isField = true;

    /**
     * The next statement in sequence after the current one
     */
    public StatementAST nextStmt;

    /**
     * The current modifiers
     */
    public ModifierAST mods;

    /**
     * The current variable or array base type
     */
    public mokan.compiler.expr.TypeAST type;

    /**
     * The current sign of a number being read, <code>true</code> for negative
     * or <code>false</code> for positive.  This is necessary to avoid a
     * {@link java.lang.NumberFormatException NumberFormatException} when
     * reading in the most negative number.
     */
    public boolean negative;

    /**
     * The current targets for <code>break</code> statements; i.e., the
     * closest enclosing <code>for</code>, <code>while</code>,
     * <code>do-while</code>, or <code>switch</code> statements.
     */
    private final LinkedList breakTarget = new LinkedList();

    /**
     * The current targets for <code>continue</code> statements; i.e., the
     * closest enclosing <code>for</code>, <code>while</code>, or
     * <code>do-while</code> statements.
     */
    private final LinkedList continueTarget = new LinkedList();

    /**
     * Push a new break target onto the stack
     *
     * @param target the new break target
     */
    public void pushBreak(final KanAST target) {
	breakTarget.addFirst(target);
    }

    /**
     * Pop a break target off of the stack
     */
    public void popBreak() {
	breakTarget.removeFirst();
    }

    /**
     * Get the current break target
     *
     * @return the current break target
     */
    public StatementAST breakTarget() {
	return (StatementAST)breakTarget.getFirst();
    }

    /**
     * Push a new continue (and break) target onto the stack
     *
     * @param target the new continue target
     */
    public void pushContinue(final KanAST target) {
	breakTarget.addFirst(target);
	continueTarget.addFirst(target);
    }

    /**
     * Pop a continue (and break) target off of the stack
     */
    public void popContinue() {
	breakTarget.removeFirst();
	continueTarget.removeFirst();
    }

    /**
     * Get the current continue target
     *
     * @return the current continue target
     */
    public StatementAST continueTarget() {
	return (StatementAST)continueTarget.getFirst();
    }
}
/* arch-tag: 3f787d7b-279d-4be5-92c0-4a00c65544ce */