/*
 * @(#)FileAST.java FileAST.java,v 1.1.1.1 2002/11/06 22:10:25 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler;

import antlr.CommonHiddenStreamToken;
import java.io.*;
import java.util.HashMap;

/**
 * An AST node that represents the contents of a file
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:25 $
 * @author Jerry James
 */
public final class FileAST extends KanAST implements KanTokenTypes {

    /**
     * The file currently being evaluated
     */
    static FileAST currFile;

    /**
     * The file that was parsed to create this AST
     */
    final File theFile;

    /**
     * The package for this file, or <code>null</code> for the default package
     */
    String pkg;

    /**
     * The import statements for this file
     */
    String[] imports;

    /**
     * The top-level class and interface definitions for this file
     */
    TypeAST[] types;

    /**
     * A mapping from class names to types, in the context of the import list
     * for this file
     */
    private final HashMap map = new HashMap();

    /**
     * Create a new file AST node
     *
     * @param file the file parsed to create this AST
     */
    public FileAST(final File file) {
	super();
	initialize(FILE, "");
	theFile = file;

	// Seed the map with the primitive data types
	map.put("boolean", Type.booleanType);
	map.put("byte",	   Type.byteType   );
	map.put("char",	   Type.charType   );
	map.put("double",  Type.doubleType );
	map.put("float",   Type.floatType  );
	map.put("int",	   Type.intType    );
	map.put("long",	   Type.longType   );
	map.put("short",   Type.shortType  );
	map.put("void",	   Type.voidType   );
    }

    /**
     * Set the initial hidden token for the file.  This ensures that leading
     * comments and whitespace are printed during a tree traversal.
     *
     * @param hiddenTok the initial hidden token
     */
    void setInitialHiddenToken(final CommonHiddenStreamToken hiddenTok) {
	hiddenAfter = hiddenTok;
    }

    /**
     * Find a type with the given name, using the import list for this file
     *
     * @param name the name of the type to look up
     * @return a type object representing this type
     * @exception ClassNotFoundException if the class cannot be found
     */
    public Type getType(final String name) throws ClassNotFoundException {
	// Case 1: we have looked up this name before
	Type type = (Type)map.get(name);
	if (type != null) {
	    return type;
	}

	// Case 2: it is an array; look up the base type
	if (name.endsWith("[]")) {
	    try {
		final int index = name.indexOf('[');
		final Type baseType = getType(name.substring(0, index));
		if (baseType == null) {
		    System.err.print("**** Failed to resolve ");
		    System.err.println(name.substring(0, index));
		}
		type =
		    Type.forName(baseType.getName() + name.substring(index));
		map.put(name, type);
		return type;
	    } catch (ClassNotFoundException ex0) {
	    }
	}

	// Case 3: it is a fully qualified name, or a partially qualified
	// inner class
	final int index = name.lastIndexOf('.');
	if (index != -1) {
	    // First see if it is a fully qualified name
	    try {
		type = Type.forName(name);
		map.put(name, type);
		return type;
	    } catch (ClassNotFoundException ex1) {
		// Now see if it is a partially qualified inner class
		try {
		    final Type t = getType(name.substring(0, index));
		    type = t.getInner('$' + name.substring(index + 1));
		    if (type != null) {
			map.put(name, type);
			return type;
		    }
		} catch (ClassNotFoundException ex2) {
		}
	    }
	}

	// Case 4: it is an unqualified inner class in this file
	final String dollarName = '$' + name;
	for (int i = 0; i < types.length; i++) {
	    final TypeAST[] inner = types[i].inner;
	    for (int j = 0; j < inner.length; j++) {
		if (inner[j].name.endsWith(dollarName)) {
		    type = inner[j].retrieveType();
		    map.put(name, type);
		    return type;
		}
	    }
	}

	// Case 5: it is in the same package
	final String dotName = '.' + name;
	try {
	    type = Type.forName(pkg + dotName);
	    map.put(name, type);
	    return type;
	} catch (ClassNotFoundException ex3) {
	}

	// Case 6: it is in the java.lang package
	try {
	    type = Type.forName("java.lang." + name);
	    map.put(name, type);
	    return type;
	} catch (ClassNotFoundException ex4) {
	}

	// Case 7: it is an imported class
	for (int i = 0; i < imports.length; i++) {
	    String fullName;
	    final int lastIndex = imports[i].length() - 1;
	    if (imports[i].charAt(lastIndex) == '*') {
		fullName = imports[i].substring(0, lastIndex) + name;
	    } else if (imports[i].endsWith(dotName)) {
		fullName = imports[i];
	    } else {
		continue;
	    }

	    try {
		type = Type.forName(fullName);
		map.put(name, type);
		return type;
	    } catch (ClassNotFoundException ex5) {
	    }
	}

	// Case 8: it is a class in the default package
	try {
	    type = Type.forName(name);
	    map.put(name, type);
	    return type;
	} catch (ClassNotFoundException ex6) {
	}

	// Case 9: it is an unqualified inner class in a superclass of some
	// class in this file
	for (int i = 0; i < types.length; i++) {
	    type = types[i].retrieveType().getInner(dollarName);
	    if (type != null) {
		map.put(name, type);
		return type;
	    }
	}

	// Case 10: it is an unqualified inner class in an imported class
	for (int i = 0; i < imports.length; i++) {
	    final int lastIndex = imports[i].length() - 1;
	    if (imports[i].charAt(lastIndex) != '*') {
		try {
		    final Type t = Type.forName(imports[i]);
		    type = t.getInner(dollarName);
		    if (type != null) {
			map.put(name, type);
			return type;
		    }
		} catch (ClassNotFoundException ex7) {
		}
	    }
	}

	// Nope, couldn't find it
	throw new ClassNotFoundException(name);
    }

    public void parseComplete() {
	for (int i = 0; i < types.length; i++) {
	    types[i].parseComplete();
	}
    }

    // Kan extension
    /**
     * Make sure a future type exists
     *
     * @param meth the method for which this class of futures is being created
     * @return the future type
     */
    public Type registerFuture(final Method meth) {
	// See if it exists already
	final String name = meth.futureName();
	try {
	    return getType(name);
	} catch (ClassNotFoundException classEx) {
	    // It does not exist already
	}

	// No such type.  See if the source file exists already.
	final File futFile = new File(theFile.getParentFile(), name + ".java");
	if (!futFile.exists()) {
	    // Nope.  Create it.
	    try {
		final PrintWriter futOut = new PrintWriter
		    (new BufferedWriter(new FileWriter(futFile)));
		futOut.print("// **** THIS IS A GENERATED FILE ****\n// **** DO NOT EDIT IT!!!!!!!!!! ****\n// It was created as part of the Kan compilation process\n\npackage ");
		futOut.print(pkg);
		futOut.print(";\n\n/**\n * A class of futures with a computed return type and computed exceptions\n */\npublic final class ");
		futOut.print(name);
		futOut.print(" extends mokan.thread.Future {\n\n    /**\n     * Wait for the asynchronous method to complete and\n     * ");
		final Type retType = meth.getReturnType();
		final boolean isVoid = retType == Type.voidType;
		if (!isVoid)
		    futOut.print("return its results or ");
		futOut.print("throw any resulting exceptions\n     */\n    public ");
// 		final String retName = retType.isGlobal()
// 		    ? retType.getName() + ".KanID"
// 		    : retType.getName();
		final String retName = retType.getModifiedName();
		futOut.print(retName);
		futOut.print(" join()");
		boolean needsException = true;
		final Type[] exceptions = meth.getExceptionTypes();
		final boolean hasExceptions = exceptions.length > 0;
		final StringBuffer buf = new StringBuffer();
		if (hasExceptions) {
		    final Type jlangException = Type.forClass(Exception.class);
		    buf.append(" throws ");
		    buf.append(exceptions[0].getName());
		    if (exceptions[0] == jlangException)
			needsException = false;
		    for (int i = 1; i < exceptions.length; i++) {
			buf.append(", ");
			buf.append(exceptions[i].getName());
			if (exceptions[0] == jlangException)
			    needsException = false;
		    }
		}
		final String exceptionList = buf.toString();
		futOut.print(exceptionList);
		futOut.print(" {\n        super.topJoin();\n        ");
		if (!isVoid)
		    futOut.print("return ");
		futOut.print("result();\n    }\n\n    /**\n     * ");
		futOut.print(isVoid
			     ? "T"
			     : "Return the results of the asynchronous method call or\n     * t");
		futOut.print("hrow any resulting exceptions\n     */\n    public ");
		futOut.print(retName);
		futOut.print(" result()");
		futOut.print(exceptionList);
		futOut.print(" {\n        ");
		if (hasExceptions || needsException)
		    futOut.print("try {\n            ");
		if (!isVoid) {
		    futOut.print("return (");
		    futOut.print(retName);
		    futOut.print(')');
		}
		futOut.print("super.");
		futOut.print(isVoid ? "topR" : "r");
		futOut.print("esult");
		futOut.print(getResultType(retType));
		futOut.println("();");
		if (hasExceptions || needsException)
		    futOut.print("        }");
		for (int i = 0; i < exceptions.length; i++) {
		    futOut.print(" catch (");
		    futOut.print(exceptions[i].getName());
		    futOut.print(" ex");
		    futOut.print(i);
		    futOut.print(") {\n            throw ex");
		    futOut.print(i);
		    futOut.println(';');
		    futOut.println("        }");
		}
		if (needsException) {
		    futOut.println(" catch (Exception ex) {\n            throw (RuntimeException)ex;\n        }");
		}
		futOut.println("    }\n}");
		futOut.close();
	    } catch (IOException ioEx) {
		ioEx.printStackTrace();
	    }
	}
	try {
	    // The file exists now, so parse it into a FileAST
	    final FileAST file = Type.parseFile(futFile);

	    // There is only one type in the file.  Return it.
	    return file.types[0].retrieveType();
	} catch (IOException ex) {
	    // FIXME: What should we do here?
	    return null;
	}
    }

    // Kan extension
    /**
     * Get the extension to the result and join method names in order to
     * return the correct type
     *
     * @param type the return type of the method
     * @return a string representing the method name extension for the result
     * and join method names for this type
     */
    private static String getResultType(final Type type) {
	if (type == Type.booleanType) {
	    return "Boolean";
	} else if (type == Type.byteType) {
	    return "Byte";
	} else if (type == Type.charType) {
	    return "Character";
	} else if (type == Type.doubleType) {
	    return "Double";
	} else if (type == Type.floatType) {
	    return "Float";
	} else if (type == Type.intType) {
	    return "Integer";
	} else if (type == Type.longType) {
	    return "Long";
	} else if (type == Type.shortType) {
	    return "Short";
	} else if (type == Type.voidType) {
	    return "";
	} else {
	    return "Object";
	}
    }

    public String toString() {
	return theFile.toString();
    }
}

// Local Variables: //
// tab-width: 8 //
// End: //


/* arch-tag: b11ab201-33ca-41f1-87fb-2ba72d210d99 */