/*
 * @(#)KanAST.java KanAST.java,v 1.1.1.1 2002/11/06 22:10:25 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler;

import antlr.CommonASTWithHiddenTokens;
import antlr.CommonHiddenStreamToken;
import antlr.Token;
import java.io.PrintWriter;

/**
 * An AST node that is a superclass for all Kan AST types.
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:25 $
 * @author Jerry James
 */
public class KanAST extends CommonASTWithHiddenTokens {

    /**
     * An empty array of types, for use in parameter and exception lists when
     * there are none; avoids creating new arrays when it is unnecessary
     * without resorting to lots of <code>null</code> checks
     */
    protected static final Type[] noTypes = new Type[0];

    /**
     * The current symbol table under construction
     */
    protected static SymbolTable currSymTable;

    /**
     * The current compilation context
     */
    protected static CompileContext context;

    /**
     * The symbols in the context of this AST node
     */
    public final SymbolTable symTable;

    /**
     * The top-level node for this file
     */
    public final FileAST topLevel;

    /**
     * The type (class or interface) containing this AST node
     */
    public final TypeAST typeAST;

    /**
     * A line number for element and error reporting
     */
    public int lineNumber;

    /**
     * Create a new Kan AST
     */
    public KanAST() {
	super();
	symTable = currSymTable;
	topLevel = FileAST.currFile;
	typeAST = TypeAST.currType;
    }

    /**
     * Create a new Kan AST with an existing symbol table
     *
     * @param table the symbol table to use
     */
    public KanAST(final SymbolTable table) {
	super();
	symTable = currSymTable = table;
	topLevel = FileAST.currFile;
	typeAST = TypeAST.currType;
    }

    /**
     * Create a new Kan AST from a token
     *
     * @param token the token represented by this AST node
     */
    public KanAST(final Token token) {
	super(token);
	symTable = currSymTable;
	topLevel = FileAST.currFile;
	typeAST = TypeAST.currType;
    lineNumber = token.getLine();
    }

    /**
     * Create a new Kan AST from a token, with an existing symbol table
     *
     * @param token the token represented by this AST node
     * @param table the symbol table to use
     */
    public KanAST(final Token token, final SymbolTable table) {
	super(token);
	symTable = currSymTable = table;
	topLevel = FileAST.currFile;
	typeAST = TypeAST.currType;
    }

    /**
     * Print a representation of this AST node, and its following hidden
     * tokens
     *
     * @param output the output stream on which to print
     */
    public final void print(final PrintWriter output) {
	output.print(getText());
	for (CommonHiddenStreamToken t = getHiddenAfter(); t != null;
	     t = t.getHiddenAfter())
	    output.print(t.getText());
    }

    // Kan extension
    /**
     * Print a representation of this AST node, a suffix, and its following
     * hidden tokens
     *
     * @param output the output stream on which to print
     * @param suffix the extra string to print after this node
     */
    public final void print(final PrintWriter output, final String suffix) {
	output.print(getText());
	if (suffix != null)
	    output.print(suffix);
	for (CommonHiddenStreamToken t = getHiddenAfter(); t != null;
	     t = t.getHiddenAfter())
	    output.print(t.getText());
    }

    /**
     * Print any hidden tokens after this AST node
     *
     * @param output the output stream on which to print
     */
    public final void printHiddenAfter(final PrintWriter output) {
	for (CommonHiddenStreamToken t = getHiddenAfter(); t != null;
	     t = t.getHiddenAfter())
	    output.print(t.getText());
    }

    /**
     * Print any hidden tokens before this AST node
     *
     * @param output the output stream on which to print
     */
    public final void printHiddenBefore(final PrintWriter output) {
	CommonHiddenStreamToken tok = getHiddenBefore();
	CommonHiddenStreamToken last = null;
	while (tok != null) {
	    last = tok;
	    tok = tok.getHiddenBefore();
	}
	for (CommonHiddenStreamToken t = last; t != null;
	     t = t.getHiddenAfter())
	    output.print(t.getText());
    }

    /**
     * Compute any derived attributes that must be evaluated <em>after</em>
     * the initial parse.  The default action is to just tell your children
     * that the parse is complete.
     */
    public void parseComplete() {
	for (KanAST a = (KanAST)getFirstChild(); a != null;
	     a = (KanAST)a.getNextSibling()) {
	    a.parseComplete();
	}
    }
}
/* arch-tag: 6a60c2a4-1495-4b80-9a3f-53eed23e6354 */