/*
 * @(#)Kanc.java Kanc.java,v 1.2 2002/11/22 23:26:06 kris Exp $
 *
 * This file is part of the Kan compiler.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler;

import java.io.*;
import java.util.ArrayList;

import antlr.RecognitionException;
import antlr.TokenStreamHiddenTokenFilter;
import antlr.collections.AST;

import mokan.compiler.expr.FutureFixup;

/**
 * Transform Kan programs to plain Java.  This is accomplished by invoking an
 * <a href="http://www.antlr.org/">ANTLR</a> parser to build an Abstract
 * Syntax Tree for each input file on the command line, then invoking an
 * <a href="http://www.antlr.org/">ANTLR</a> tree walker on each AST which
 * emits the Java output.
 *
 * @version $Revision: 1.2 $, $Date: 2002/11/22 23:26:06 $
 * @author Jerry James
 * @author Kris Kvilekval
 */
public final class Kanc {

    /**
     * The amount by which to grow the <code>FileAST</code> array each time it
     * runs out of room
     */
    private static final int GROW_SIZE = 32;

    /**
     * The number of {@linkmokan.compiler.FileAST FileAST} objects generated so
     * far
     */
    private static int numFiles = 0;

    /**
     * The <code>FileAST</code> objects created during this compilation
     */
    private static FileAST[] fileAST = new FileAST[GROW_SIZE];

    /**
     * The main method of the Kan-to-Java transformer.  It calls the parser on
     * each class given on the command line (or all classes in a directory
     * given on the command line), then emits Java code for each one.
     *
     * @param args the command line parameters
     * @exception Exception if <em>anything</em> goes wrong during the compile
     */
    public static void main(final String[] args) throws Exception {
        if (args.length == 0) {
            System.err.println("Usage: java Kanc <directory or file name>");
            return;
        }

        for (int i = 0; i < args.length; i++) {
            if (!(args[i].charAt(0) == '-')) {	// Ignore options
                doFile(new File(args[i]));
            }
        }


        // Swizzle all futures
        FutureFixup.fixupFutures();

        // Now emit code for each class
        for (int i = 0; i < numFiles; i++) {
            emitCode(fileAST[i]);
        }

        // Finally, invoke the Java compiler for all generated .java files
        compileCode(args);
    }

    /**
     * Decide what action to take based on the type of file we are looking at
     *
     * @param f a file or directory name given on the command line
     * @exception FileNotFoundException if the file or directory does not exist
     */
    public static void doFile(final File f) throws IOException {
        // If this is a directory, walk each file/dir in that directory
        if (f.isDirectory()) {
            final String[] files = f.list();
            for (int i = 0; i < files.length; i++)
                doFile(new File(f, files[i]));
        } else {
            // Otherwise, parse the file, if it exists
            if (!f.exists())
                return;
            final String name = f.getCanonicalPath();

            // Even if we don't use the AST now, this makes the classes &
            // interfaces in f be registered with Type
            final FileAST file = Type.parseFile(f);
            if (name.endsWith(".kan")) {
                if (numFiles == fileAST.length) {
                    final FileAST[] temp = new FileAST[numFiles + GROW_SIZE];
                    System.arraycopy(fileAST, 0, temp, 0, numFiles);
                    fileAST = temp;
                }
                fileAST[numFiles++] = file;
            }
        }
    }

    /**
     * Emit the Java code for one Kan source file
     *
     * @param file a <code>FileAST</code> object describing one source class
     * @exception IOException if any file writing errors occur
     */
    public static void emitCode(final FileAST file)
        throws IOException, RecognitionException {

        final KanTranslator translator = new KanTranslator();
        translator.setASTNodeClass("mokan.compiler.KanAST");
        final String fullName = file.theFile.getPath();
        final String rootName = fullName.substring(0, fullName.length() - 4);
        final PrintWriter out = new PrintWriter
            (new BufferedWriter(new FileWriter(rootName + ".java")));
        try {
            translator.compilationUnit(file, out);
        } catch (Exception ex) {
            ex.printStackTrace();
            System.err.println("Dumping type state:");
            Type.dumpTypes();
        }
        out.close();
    }

    /**
     * Invoke the Java compiler to compile all of the generated files
     *
     * @param args the command line arguments given to the Kan compiler
     * @exception IOException if the exec call throws an exception
     */
    public static void compileCode(final String[] args) throws IOException {

        // IMPLEMENTATION NOTES:
        // 1. We don't need to list any of the Future types, because those
        //    that are actually used are referred to in one of the KanID
        //    classes.  We'll let the Java compiler be smart instead of being
        //    smart ourselves.
        // 2. Because of 1, we don't try to interpret command line options in
        //    any way.  All the Kan compiler does is read in the *.kan files
        //    mentioned on that command line, and generate Java files from
        //    them.  Hence, the Java compiler command line is the Kan compiler
        //    command line with Foo.java replacing Foo.kan.

        if (false) {  // KGK libc bug for 1.3.1 remove when using 1.4.1

            ArrayList cmdArray = new ArrayList();
            // FIXME: Make this a settable option (so other compilers can be used)
            cmdArray.add("javac");
            for (int i = 0; i < args.length; i++) {
                final String arg = args[i];
                cmdArray.add(arg.endsWith(".kan")
                             ? arg.substring(0, arg.length() - 3) + "java"
                             : arg);
            }

            final String[] cmds = new String[cmdArray.size()];
            cmdArray.toArray(cmds);
            System.out.println(cmdArray);
            try {
                final Process pjava = Runtime.getRuntime().exec(cmds);
                final BufferedReader err = new BufferedReader
                    (new InputStreamReader(pjava.getErrorStream()));
                String s = err.readLine();
                while (s != null) {
                    System.err.println(s);
                    s = err.readLine();
                }
                pjava.waitFor();	// REVIEW: Is this necessary?
            } catch (InterruptedException intEx) {
                System.err.println("This can't happen.");
                intEx.printStackTrace();
            } catch (IOException ioEx) {
                System.err.print("Compile aborted: ");
                System.err.print(ioEx);
            }
        } //KGK
    }
}
/* arch-tag: a2503c51-aad1-4875-890a-56b123a6a902 */