/*
 * @(#)AssignAST.java AssignAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.expr;

import antlr.Token;
import mokan.compiler.KanTokenTypes;
import mokan.compiler.Type;
import mokan.compiler.VarList;
import mokan.transact.Transaction;

/**
 * An AST node that represents an assignment
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public final class AssignAST extends ExpressionAST implements KanTokenTypes {

    /**
     * The left-hand side of the assignment
     */
    private ExpressionAST lhs;

    /**
     * The right-hand side of the assignment
     */
    private ExpressionAST rhs;

    /**
     * Create a new assignment AST
     *
     * @param token the token represented by this AST node
     */
    public AssignAST(final Token token) {
        super(token);
    }

    public void parseComplete() {
        lhs = (ExpressionAST)getFirstChild();
        rhs = (ExpressionAST)lhs.getNextSibling();
        lhs.parseComplete();
        rhs.parseComplete();

        // Kan extension: Is this assigning to a Future or an array of Futures?
        ExpressionAST change;
        int levels = 0;
        int type = lhs.getType();
        for (change = lhs; type == ARRAY_DECLARATOR || type == INDEX_OP;
             change = (ExpressionAST)change.getFirstChild(),
                 type = change.getType()){
            levels++;
        }
        while (type == DOT) {
            change = (ExpressionAST)change.getFirstChild().getNextSibling();
            type = change.getType();
            levels++;
        }
        final VarAST var = symTable.getVar(((IdentifierAST)change).getName());
        if (var != null) {
            final String typeName = var.getTypeName().getName();
            if (typeName.equals("Future") || typeName.startsWith("Future[")){
                FutureFixup.register(var, this);
            }
        }
    }

    protected Type computeType() {
        // The type of an assignment is the type of its left-hand side
        final Type type = lhs.retrieveType();
        if (type == Type.stringType && getType() == PLUS_ASSIGN)
            setType(CONCAT_ASSIGN);
        return type;
    }

    protected Type[] computeExceptions() {
        // The lhs of an assignment can't throw any checked exceptions
        return rhs.getExceptionTypes();
    }

    protected Object computeValue() {
        return rhs.getValue();
    }

    public VarList getVarList() {
        final VarList leftList = lhs.getVarList();
        final VarList rightList = rhs.getVarList();
        return new VarList(leftList, rightList, true);
    }

    /**
     * Get the left-hand-side of this assignment expression
     *
     * @return the lhs of the expression
     */
    public ExpressionAST getLeft() {
        return lhs;
    }

    /**
     * Get the right-hand-side of this assignment expression
     *
     * @return the rhs of the expression
     */
    public ExpressionAST getRight() {
        return rhs;
    }

    // Kan extension
    public int transType() {
        final int lType = lhs.transType();
        if (lType == Transaction.NESTED)
            return lType;
        final int rType = rhs.transType();
        return (lType > rType) ? lType : rType;
    }

    // Kan extension
    /**
     * Change the type of any arrays of Futures created by this expression
     */
    void swizzleFuture() {
        if (rhs instanceof NewAST) {
            ((NewAST)rhs).swizzleFuture(lhs.retrieveType());
        }
    }

    // Kan extension
    /**
     * Change the type of the variable assigned to by this expression, given
     * the future type identified by the rhs
     *
     * @param decl the variable declaration to change
     */
    public void swizzleFuture(final VarAST decl) {
        // Get the future type
        Type futType =
            topLevel.registerFuture(((MethodCallAST)rhs).getMethod());

        // Now change the type of the lhs
        ExpressionAST change = lhs;
        int levels = 0, typecode;
        for (typecode = change.getType();
             typecode == ARRAY_DECLARATOR || typecode == INDEX_OP;
             typecode = change.getType()) {
            change.type = futType;
            futType = futType.getArrayType();
            levels++;
            change = (ExpressionAST)change.getFirstChild();
        }
        change.type = futType;

        // Then change the type of the variable
        decl.changeType(futType, levels);
    }
}
/* arch-tag: f22b4a6c-cb53-4863-ba9b-f09e9db579c6 */