/*
 * @(#)$Id: BooleanAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.expr;

import antlr.Token;
import mokan.compiler.KanTokenTypes;
import mokan.compiler.Type;
import mokan.compiler.VarList;
import mokan.transact.Transaction;

/**
 * An AST node that represents a boolean expression
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public final class BooleanAST extends ExpressionAST implements KanTokenTypes {

    /**
     * The left expression
     */
    private ExpressionAST left;

    /**
     * The right expression
     */
    private ExpressionAST right;

    /**
     * Create a new boolean expression AST
     *
     * @param token the token represented by this AST node
     */
    public BooleanAST(final Token token) {
	super(token);
    }

    public void parseComplete() {
	left = (ExpressionAST)getFirstChild();
	right = (ExpressionAST)left.getNextSibling();
	left.parseComplete();
	if (right != null)
	    right.parseComplete();
    }

    protected Type computeType() {
	return Type.booleanType;
    }

    protected Type[] computeExceptions() {
	return (right == null)
	    ? left.getExceptionTypes()	// Logical not (!) operator
	    : Type.mergeTypeLists(left.getExceptionTypes(),
				  right.getExceptionTypes());
    }

    protected Object computeValue() {
	final Object leftObj = left.getValue();
	if (leftObj == nonconstant)
	    return nonconstant;

	// Watch out for short-circuiting!
	final int operator = getType();
	if ((operator == LOR && ((Boolean)leftObj).booleanValue()) ||
	    (operator == LAND && !((Boolean)leftObj).booleanValue()))
	    return leftObj;
	// Don't do the same for rightObj if leftObj is (or contains) a method
	// call, since we can't predict things like program exit or
	// exceptions.  Without method calls, it is safe to do the same for
	// rightObj.

	if (operator == LNOT)
	    return new Boolean(!((Boolean)leftObj).booleanValue());

	if (right == null)
	    return leftObj;

	final Object rightObj = right.getValue();
	if (rightObj == nonconstant)
	    return nonconstant;

	// Both constant!  We have a winner!
	switch (operator) {
	case LOR:
	    return new Boolean(((Boolean)leftObj).booleanValue() ||
			       ((Boolean)rightObj).booleanValue());
	case LAND:
	    return new Boolean(((Boolean)leftObj).booleanValue() &&
			       ((Boolean)rightObj).booleanValue());
	case NOT_EQUAL:
	    return new Boolean(!leftObj.equals(rightObj));
	case EQUAL:
	    return new Boolean(leftObj.equals(rightObj));
	case LITERAL_instanceof:
	    return new Boolean
		(right.retrieveType() == Type.forClass(leftObj.getClass()));
	default:
	    break;
	}

	final Type compareType = Type.arithType(left.retrieveType(),
						right.retrieveType());
	if (compareType == Type.doubleType) {
	    final double leftDouble = ((Number)leftObj).doubleValue();
	    final double rightDouble = ((Number)rightObj).doubleValue();
	    switch (operator) {
	    case LT:
		return new Boolean(leftDouble < rightDouble);
	    case GT:
		return new Boolean(leftDouble > rightDouble);
	    case LE:
		return new Boolean(leftDouble <= rightDouble);
	    case GE:
		return new Boolean(leftDouble >= rightDouble);
	    }
	} else if (compareType == Type.floatType) {
	    final float leftFloat = ((Number)leftObj).floatValue();
	    final float rightFloat = ((Number)rightObj).floatValue();
	    switch (operator) {
	    case LT:
		return new Boolean(leftFloat < rightFloat);
	    case GT:
		return new Boolean(leftFloat > rightFloat);
	    case LE:
		return new Boolean(leftFloat <= rightFloat);
	    case GE:
		return new Boolean(leftFloat >= rightFloat);
	    }
	} else if (compareType == Type.longType) {
	    final long leftLong = ((Number)leftObj).longValue();
	    final long rightLong = ((Number)rightObj).longValue();
	    switch (operator) {
	    case LT:
		return new Boolean(leftLong < rightLong);
	    case GT:
		return new Boolean(leftLong > rightLong);
	    case LE:
		return new Boolean(leftLong <= rightLong);
	    case GE:
		return new Boolean(leftLong >= rightLong);
	    }
	} else /* if (compareType == Type.intType) */ {
	    final long leftLong = ((Number)leftObj).longValue();
	    final long rightLong = ((Number)rightObj).longValue();
	    switch (operator) {
	    case LT:
		return new Boolean(leftLong < rightLong);
	    case GT:
		return new Boolean(leftLong > rightLong);
	    case LE:
		return new Boolean(leftLong <= rightLong);
	    case GE:
		return new Boolean(leftLong >= rightLong);
	    }
	}
	return nonconstant;
    }

    public VarList getVarList() {
	return (right == null)
	    ? left.getVarList()
	    : new VarList(left.getVarList(), right.getVarList());
    }

    /**
     * Get the left-hand-side of this boolean expression
     *
     * @return the lhs of the expression
     */
    public ExpressionAST getLeft() {
	return left;
    }

    /**
     * Get the right-hand-side of this boolean expression
     *
     * @return the rhs of the expression
     */
    public ExpressionAST getRight() {
	return right;
    }

    public int transType() {
	final int lType = left.transType();
	if (lType == Transaction.NESTED || right == null)
	    return lType;
	final int rType = right.transType();
	return (lType > rType) ? lType : rType;
    }
}
/* arch-tag: 12c052c8-42cc-4883-9b4a-b59948e2f2c2 */