/*
 * @(#)ExpressionAST.java ExpressionAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.expr;

import antlr.Token;
import mokan.compiler.HasExceptions;
import mokan.compiler.KanAST;
import mokan.compiler.Type;
import mokan.compiler.VarList;

/**
 * An AST node that represents an expression
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public abstract class ExpressionAST extends KanAST implements HasExceptions {

    /**
     * A dummy object used to determine when the value has been computed
     */
    private static final Object noVal = new Object();

    /**
     * A dummy object that indicates that an expression is not constant
     */
    public static final Object nonconstant = new Object();

    /**
     * The type of this expression
     */
    protected Type type;

    /**
     * The checked exceptions this expression might throw
     */
    protected Type[] exceptions;

    /**
     * The constant value of this expression (possibly wrapped), or
     * <code>null</code> if the expression is not constant
     */
    protected Object value = noVal;

    /**
     * Create a new expression AST
     */
    protected ExpressionAST() {
	super();
    }

    /**
     * Create a new expression AST
     *
     * @param token the token represented by this AST node
     */
    protected ExpressionAST(final Token token) {
	super(token);
    }

    /**
     * Retrieve the type of this expression
     *
     * @return the type of this expression
     */
    public final Type retrieveType() {
	if (type == null)
	    type = computeType();
	return type;
    }

    /**
     * Compute the type of this expression
     *
     * @return the type of this expression
     */
    protected abstract Type computeType();

    /**
     * Get the checked exception types that might be thrown by this expression
     *
     * @return an array of types representing the exceptions that this
     * expression might throw
     */
    public final Type[] getExceptionTypes() {
	if (exceptions == null)
	    exceptions = computeExceptions();
	return exceptions;
    }

    /**
     * Compute the checked exception types that might be thrown by this
     * expression
     *
     * @return an array of types representing the exceptions that this
     * expression might throw
     */
    protected abstract Type[] computeExceptions();

    /**
     * Get the constant value of this expression, if it has one, or
     * <code>null</code> if it does not
     *
     * @return the constant value (if any) of this expression, as a wrapped
     * primitive type or an object type
     */
    public final Object getValue() {
	if (value == noVal)
	    value = computeValue();
	return value;
    }

    /**
     * Compute the constant value of this expression, if any
     *
     * @return the constant value (if any) of this expression, as a wrapped
     * primitive type or an object type
     */
    protected abstract Object computeValue();

    /**
     * Get the list of variables read and written by this expression
     *
     * @return the list of variables read and written by this expression
     */
    public abstract VarList getVarList();

    /**
     * Determine the transaction type of this expression, either nonblocking,
     * blocking, or nested
     *
     * @return the transaction type
     * @see mokan.transact.Transaction
     */
    public abstract int transType();
}
/* arch-tag: 24c01ce0-57d7-4cca-811c-30b1e4d6e700 */