/*
 * @(#)$Id: FutureFixup.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 *
 */
package mokan.compiler.expr;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import mokan.compiler.KanTokenTypes;
import mokan.compiler.Type;

/**
 * A collection of variable declarations of type Future (or array of Future)
 * that needs to be fixed up when it is safe to do so.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author	Jerry James
 */
public final class FutureFixup implements KanTokenTypes {

    /**
     * The collection of <code>FutureFixup</code>s
     */
    private static final ArrayList list = new ArrayList();

    /**
     * The variable declaration for a Future type (or array of Future types)
     */
    private VarAST decl;

    /**
     * Any assignments to this variable (or elements of this variable, if it
     * is an array)
     */
    private final HashSet assigns = new HashSet();

    /**
     * Create a new <code>FutureFixup</code> to register a variable
     * declaration that needs to have its Future type fixed up when it is safe
     * to do so
     *
     * @param varDecl the variable declaration that needs to be fixed up
     */
    FutureFixup(final VarAST varDecl) {
	decl = varDecl;
	list.add(find(varDecl.getName()), this);
    }

    /**
     * Fixup the variable declaration with the correct future type
     */
    void fixup() {
	// First, get the type of individual Futures
	Iterator iter = assigns.iterator();
	while (iter.hasNext()) {
	    final AssignAST assign = (AssignAST)iter.next();
	    if (assign.getRight().getType() == ASYNC_METHOD_CALL) {
		assign.swizzleFuture(decl);
		try {
		    iter.remove();
		} catch (UnsupportedOperationException unsuppEx) {
		    // This is just an optimization, so don't worry about it
		}
	    }
	}

	// Now, fixup any array creations
	iter = assigns.iterator();
	while (iter.hasNext()) {
	    final AssignAST assign = (AssignAST)iter.next();
	    if (assign.getRight().getType() != ASYNC_METHOD_CALL &&
		decl.retrieveType() == Type.futType) {
		assign.swizzleFuture();
	    }
	}

	// Don't do this more than once
	assigns.clear();

	// Finally, check the initializer.  Do this LAST, in case the
	// initializer creates an array.  We need to know the real type of the
	// variable to do that
	if (decl.getInitializer() != null) {
	    decl.swizzleFuture();
	}
    }

    /**
     * Register a variable declaration with an initializer, so that its type
     * can be used to modify the initializer appropriately
     *
     * @param var the variable declaration associated with this assignment
     */
    static void register(final VarAST var) {
	final FutureFixup fix = find(var);
	if (fix == null) {
	    new FutureFixup(var);	// This does the registration
	}
    }

    /**
     * Register an assignment statement with a fixup, so that its type can be
     * used to determine the future type later
     *
     * @param var the variable declaration associated with this assignment
     * @param assign the assignment statement to register
     */
    static void register(final VarAST var, final AssignAST assign) {
	FutureFixup fix = find(var);
	if (fix == null) {
	    fix = new FutureFixup(var);
	}
	fix.assigns.add(assign);
    }

    /**
     * Fixup all assignment statements and new expressions with Future type
     */
    public static void fixupFutures() {
	final Iterator iter = list.iterator();
	while (iter.hasNext()) {
	    final FutureFixup fix = (FutureFixup)iter.next();
	    fix.fixup();
	}
	list.clear();
    }

    /**
     * Do a binary search on the list to find the position where the named
     * variable is or ought to be
     *
     * @param name the variable name to search for
     * @return the index where this variable name goes in sorted order
     */
    static int find(final String name) {
	final int size = list.size();
	int low = 0;
	int high = size - 1;
	while (low <= high) {
	    final int mid = (low + high) / 2;
	    final FutureFixup fix = (FutureFixup)list.get(mid);
	    final int compare = name.compareTo(fix.decl.getName());
	    if (compare < 0) {
		high = mid - 1;
	    } else if (compare > 0) {
		low = mid + 1;
	    } else {
		low = mid;
		high = mid - 1;
	    }
	}
	while (low > 0 && name.compareTo(((FutureFixup)list.get(low - 1))
					 .decl.getName()) == 0) {
	    low--;
	}
	return low;
    }

    /**
     * Do a binary search on the list to find the fixup for a variable
     * declaration
     *
     * @param var the variable to search for
     * @return the fixup for this variable, or <code>null</code> if there is
     * not one
     */
    static FutureFixup find(final VarAST var) {
	final String name = var.getName();
	final int size = list.size();
	int low = 0;
	int high = size - 1;
	while (low <= high) {
	    final int mid = (low + high) / 2;
	    final FutureFixup fix = (FutureFixup)list.get(mid);
	    final int compare = name.compareTo(fix.decl.getName());
	    if (compare < 0) {
		high = mid - 1;
	    } else if (compare > 0) {
		low = mid + 1;
	    } else {
		low = mid;
		high = mid - 1;
	    }
	}

	FutureFixup fix;
	if (low < size) {
	    fix = (FutureFixup)list.get(low);
	    if (fix.decl == var) {
		return fix;
	    }
	}

	int oldLow = low--;
	while (low > 0 &&
	       name.compareTo(((FutureFixup)list.get(low)).decl.getName())==0){
	    fix = (FutureFixup)list.get(low--);
	    if (fix.decl == var) {
		return fix;
	    }
	}

	low = oldLow + 1;
	while (low < size &&
	       name.compareTo(((FutureFixup)list.get(low)).decl.getName())==0){
	    fix = (FutureFixup)list.get(low++);
	    if (fix.decl == var) {
		return fix;
	    }
	}

	return null;
    }
}
/* arch-tag: 377cda6d-c65d-4fad-9071-218b4bbc032a */