/*
 * @(#)$Id: VarAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.expr;

import antlr.Token;
import antlr.collections.AST;
import mokan.compiler.*;
import mokan.transact.Transaction;

/**
 * An AST node that represents a variable definition or formal parameter
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public final class VarAST extends IdentifierAST implements KanTokenTypes {

    /**
     * The modifiers for this variable
     */
    private ModifierAST mods;

    /**
     * The declared type of this variable
     */
    private TypeAST declType;

    /**
     * Any trailing array brackets on the declaration
     */
    private KanAST brackets;

    /**
     * The initializer (if any) for this variable
     */
    private InitializerAST initializer;

    /**
     * <code>true</code> if this is a field (static or instance variable);
     * <code>false</code> if this is a local variable (including parameters)
     */
    private boolean field;

    /**
     * Create a new variable AST
     *
     * @param token the token represented by this AST node
     */
    public VarAST(final Token token) {
	super(token);

	// Register this variable with the symbol table
	symTable.addVar(this);
    }

    public void parseComplete() {
	mods = context.mods;
	declType = context.type;
	AST a = getNextSibling();
	if (a != null && a.getType() == ARRAY_DECLARATOR) {
	    brackets = (KanAST)a;
	    do {
		a = a.getNextSibling().getNextSibling();
	    } while (a != null && a.getType() == ARRAY_DECLARATOR);
	}
	field = context.isField;
	if (a != null && a instanceof InitializerAST) {
	    initializer = (InitializerAST)a;
	    initializer.parseComplete();

	    // Kan extension: Is this a global type?
	    final String typeName = declType.getName();
	    if (typeName.equals("Future") || typeName.startsWith("Future[")) {
		FutureFixup.register(this);
	    }
	}
    }

    protected Type computeType() {
	final Type baseType = declType.retrieveType();
	if (brackets == null) {
	    return baseType;
	}
	final StringBuffer buf = new StringBuffer(baseType.getName());
	for (AST b = brackets; b != null && b.getType() == ARRAY_DECLARATOR;
	     b = b.getNextSibling().getNextSibling())
	    buf.append("[]");
	try {
	    return Type.forName(buf.toString());
	} catch (ClassNotFoundException classEx) {
	    return null;
	}
    }

    protected Type[] computeExceptions() {
	return (initializer == null)
	    ? noTypes
	    : initializer.getExceptionTypes();
    }

    protected Object computeValue() {
	return (mods.isFinal() || initializer == null)
	    ? nonconstant
	    : initializer.getValue();
    }

    public VarList getVarList() {
	final VarList thisVar = new VarList(this);
	return (initializer == null)
	    ? thisVar
	    : new VarList(thisVar, initializer.getVarList());
    }

    /**
     * Get the modifiers for this variable
     *
     * @return the modifiers
     */
    public ModifierAST getModifiers() {
	return mods;
    }

    /**
     * Get the name of the declared type for this variable
     *
     * @return the type name
     */
    public TypeAST getTypeName() {
	return declType;
    }

    /**
     * Get the brackets modifying the type name, if those brackets follow the
     * variable name
     *
     * @return brackets modifying the type name, or <code>null</code> if there
     * are none
     */
    public KanAST getBrackets() {
	return brackets;
    }

    /**
     * Get the initializer for this variable
     *
     * @return the initializer for this variable, or <code>null</code> if
     * there is not one
     */
    public InitializerAST getInitializer() {
	return initializer;
    }

    /**
     * Determine whether this variable is a field (static or instance
     * variable) or a local variable or parameter
     *
     * @return <code>true</code> if this variable is a field;
     * <code>false</code> otherwise
     */
    public boolean isField() {
	return field;
    }

    public int transType() {
	return (initializer == null)
	    ? Transaction.NONBLOCKING
	    : initializer.transType();
    }

    // Kan extension
    /**
     * Change the type of this variable from Future to a more specific type,
     * based on its initializer
     */
    void swizzleFuture() {
	final ExpressionAST theInit = initializer.getRight();
	if (theInit instanceof MethodCallAST) {
	    final MethodCallAST rhs = (MethodCallAST)theInit;
	    Type futType = topLevel.registerFuture(rhs.getMethod());

	    // Now change the type of this variable
	    int levels = 0;
	    for (Type varType = retrieveType(); varType.isArray();
		 varType = varType.getComponentType()) {
		futType = futType.getArrayType();
		levels++;
	    }
	    changeType(futType, levels);
	} else if (theInit instanceof NewAST) {
	    // In this case, the type of *this* variable declaration has
	    // already been swizzled by an assignment expression ... or else
	    ((NewAST)theInit).swizzleFuture(retrieveType());
	}
    }

    // Kan extension
    /**
     * Change the type of this variable to a Future type
     *
     * @param futType the Future type to change to
     * @param levels the number of levels of array on <var>futType</var>
     */
    void changeType(Type futType, int levels) {
	String futName = futType.getName();
	declType.setName(futName);
	declType.type = futType;

	// If it has a "new" initializer, set its type
	if (initializer != null && initializer.getRight() instanceof NewAST) {
	    ((NewAST)initializer.getRight()).swizzleFuture(futType);
	}

	IdentifierAST ident = (IdentifierAST)declType.getFirstChild();
	while (levels-- > 0) {
	    ident.setName(futName);
	    ident.type = futType;
	    ident = (IdentifierAST)ident.getFirstChild();
	    futType = futType.getComponentType();
	    futName = futName.substring(0, futName.length() - 2);
	}
	ident.setName(futName);
	ident.setText(futName);
	ident.type = futType;
    }

    public String toString() {
	final StringBuffer buf = new StringBuffer();
	if (declType != null) {
	    buf.append(declType.toString());
	    buf.append(' ');
	}
	buf.append(super.toString());
	for (AST b = brackets; b != null && b.getType() == ARRAY_DECLARATOR;
	     b = b.getNextSibling().getNextSibling())
	    buf.append("[]");
	return buf.toString();
    }
}
/* arch-tag: c34528b3-9d19-49a9-91ad-8d0109f8ce02 */