header {
/*
 * @(#)kan.trans.g mokan.trans.g,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 * Copyright (C) 2000 Jeremiah W. James
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler;

import antlr.CommonHiddenStreamToken;
import java.io.PrintWriter;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import mokan.compiler.expr.*;
import mokan.compiler.stmt.*;
import mokan.transact.Transaction;
}

/**
 * The Kan code emitter.  It is based on a Java 1.2 tree walker that prints
 * out the parsed tree, exactly as it was read in.  That tree walker, in turn,
 * is based on, but differs significantly from, the public domain ANTLR tree
 * walker identified as follows:
 *
 * Java 1.2 AST Recognizer Grammar
 *
 * Author:
 *     Terence Parr    parrt@jguru.com
 *
 * Version tracking now done with following ID:
 *
 * $Id: mokan.trans.g,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * This grammar is in the PUBLIC DOMAIN
 *
 * BUGS
 */
class KanTranslator extends TreeParser;

options {
	ASTLabelType = "KanAST";		// Change the default AST node type
	k = 3;							// Three tokens of lookahead
	importVocab = Kan;				// Use the "Kan" vocabulary
	codeGenMakeSwitchThreshold = 2;	// Some optimizations
	codeGenBitsetTestThreshold = 3;
}

{
	/**
	 * A list of the constructors in the class
	 */
	private final ArrayList kanConstructors = new ArrayList();

	/**
	 * The guard expressions for which methods must be created
	 */
	private GuardExpr guardList = null;

	/**
	 * Where to send the output of the tree walker
	 */
	private PrintWriter out;

	/**
	 * The current type being compiled
	 */
	private SourceType type;

	/**
	 * The number of constructors processed so far
	 */
	private int constrNum;

	/**
	 * The number of methods processed so far
	 */
	private int methodNum;

	/**
	 * Get the name of the wrapper class for a primitive type
	 *
	 * @param type the type to wrap
	 * @return the name of the wrapper for that tyep
	 */
	private static final String getWrapper(final KanAST type) {
		switch(type.getType()) {
		case LITERAL_boolean:
			return "Boolean";
		case LITERAL_byte:
			return "Byte";
		case LITERAL_char:
			return "Character";
		case LITERAL_double:
			return "Double";
		case LITERAL_float:
			return "Float";
		case LITERAL_int:
			return "Integer";
		case LITERAL_long:
			return "Long";
		case LITERAL_short:
			return "Short";
		default:
			return "";
		}
	}

	/**
	 * Get the name of the method to call to unwrap a wrapped primitive type
	 *
	 * @param type the type to unwrap
	 * @return the name of the method to call to unwrap a <var>type</var>
	 */
	private static final String getUnwrapper(final KanAST type) {
		switch(type.getType()) {
		case LITERAL_boolean:
			return "booleanValue";
		case LITERAL_byte:
			return "byteValue";
		case LITERAL_char:
			return "charValue";
		case LITERAL_double:
			return "doubleValue";
		case LITERAL_float:
			return "floatValue";
		case LITERAL_int:
			return "intValue";
		case LITERAL_long:
			return "longValue";
		case LITERAL_short:
			return "shortValue";
		default:
			return "";
		}
	}

	/**
	 * Print some text representing a typed Java construct, and also put it in
	 * a wrapper if it is a primitive type
	 *
	 * @param type the type of the Java construct
	 * @param text the text to print
	 */
	private final void printMaybeWrapped(final KanAST type, final String text)
	{
		final String wrapper = getWrapper(type);
		out.print((wrapper.length() == 0)
				  ? text
				  : "new " + wrapper + '(' + text + ')');
	}

	/**
	 * Print the synchronous and asynchronous stubs for a method
	 *
	 * @param meth the method for which the stubs should be printed
	 * @param num a unique number for this method
	 * @param typeAST the TypeAST representing the current type being compiled
	 */
	private final void printMethodStub(final Method meth, final int num,
			final TypeAST typeAST) {
		out.print("\n    ");
		// Print the modifiers, except we don't want synchronized, native, or
		// abstract on the stub
		final String mods = Modifier.toString(meth.getModifiers() &
			~(Modifier.SYNCHRONIZED | Modifier.NATIVE | Modifier.ABSTRACT));
		out.print(mods);
		out.print(' ');
		out.print(meth.getReturnType().getModifiedName());
		out.print(' ');
		out.print(meth.getName());
		out.print('(');
		final Type[] params = meth.getParameterTypes();
		if (params.length > 0) {
			out.print(params[0].getModifiedName());
			out.print(" p0");
			for (int i = 1; i < params.length; i++) {
				out.print(", ");
				out.print(params[i].getModifiedName());
				out.print(" p");
				out.print(i);
			}
		}
		out.print(") ");
		final Type[] except = meth.getExceptionTypes();
		if (except.length > 0) {
			out.print("throws ");
			out.print(except[0].getModifiedName());
			for (int i = 1; i < except.length; i++) {
				out.print(", ");
				out.print(except[i].getModifiedName());
			}
			out.print(' ');
		}
		out.print("{\n        ");
		final Type theClass = meth.getDeclaringClass();
		final String className = theClass.getName();
		out.print(className);
		out.print(" target = (");
		out.print(className);
		out.print(")callDirect();\n        if (target == null) {\n            ");
		final boolean isVoid = meth.getReturnType() == Type.voidType;
		if (!isVoid) {
			out.print("return ");
		}
		out.print(meth.getName());
		out.print("KanAsync(");
		if (params.length > 0) {
			out.print("p0");
			for (int i = 1; i < params.length; i++) {
				out.print(", p");
				out.print(i);
			}
		}
		out.print(").join();\n        } else {\n");
                out.print("            startCall();\n            ");
		if (!isVoid) {
			out.print("final ");
			out.print(meth.getReturnType().getModifiedName());
			out.print(" retval = ");
		}
		out.print("target.");
		out.print(meth.getName());
		out.print('(');
		if (params.length > 0) {
			out.print("p0");
			for (int i = 1; i < params.length; i++) {
				out.print(", p");
				out.print(i);
			}
		}
		out.println(");\n            endCall();");
		if (!isVoid)
			out.println("            return retval;");
		out.print("        }\n    }\n\n    ");
		out.print(mods);
		out.print(' ');
		final String futName = meth.futureName();
		out.print(futName);
		out.print(' ');
		out.print(meth.getName());
		out.print("KanAsync(");
		if (params.length > 0) {
			out.print(params[0].getModifiedName());
			out.print(" p0");
			for (int i = 1; i < params.length; i++) {
				out.print(", ");
				out.print(params[i].getModifiedName());
				out.print(" p");
				out.print(i);
			}
		}
		out.print(") {\n        final ");
		out.print(futName);
		out.print(" fut = new ");
		out.print(futName);
		out.print("();\n        mokan.KanSystem.invokeMethod(this, fut, ");
		out.print(num /*KK + theClass.minMethodNum() */);
		out.print(", ");
		if (params.length == 0) {
			/* KK out.print("new Object[0]"); */
                out.print ("null");
		} else {
			out.print("new Object[] {");
			out.print(params[0].wrapText("p0"));
			for (int i = 1; i < params.length; i++) {
				out.print(", ");
				out.print(params[i].wrapText("p" + i));
			}
			out.print('}');
		}
		out.println(");\n        return fut;\n    }");

		// Register the future
		typeAST.topLevel.registerFuture(meth);
	}

	/**
	 * Print out the fast calling information for a method
	 *
	 * @param meth the method for which fast calling information is to be
	 * printed
	 * @param num a unique number for <var>meth</var>
	 */
	private final void printFastCall(final Method meth, final int num) {
		final Type methType = meth.getReturnType();
		final String methName = meth.getName();
		final boolean isVoid = methType == Type.voidType;
		out.print("        case ");
		out.print(num);
		out.print(":\n            ");
		if (!isVoid) {
			out.print("return ");
		}
		final StringBuffer buf = new StringBuffer(methName);
		buf.append('(');
		final Type[] params = meth.getParameterTypes();
		for (int i = 0; i < params.length; i++) {
			if (i > 0) buf.append(", ");
			final Type paramType = params[i];
			if (paramType.isPrimitive()) {
				buf.append("((");
				buf.append(paramType.getWrapper());
				buf.append(")args[");
				buf.append(i);
				buf.append("])");
				buf.append(paramType.getUnwrapper());
			} else {
				buf.append('(');
				buf.append(paramType.getModifiedName());
				buf.append(")args[");
				buf.append(i);
				buf.append(']');
			}
		}
		buf.append(')');
		out.print(methType.wrapText(buf.toString()));
		out.println(';');
		if (isVoid)
			out.println("            return null;");
	}

	/**
	 * Add extra methods and declarations onto the end of a global class as
	 * needed to support the Kan interface
	 *
	 * @param typeAST the TypeAST representing the current type being compiled
	 * @exception RecognitionException if walking guard expressions causes a
	 * tree walker error
	 */
	private final void addKanInfo(final TypeAST typeAST)
		throws RecognitionException {

		// Print the method calling info
		Type parent;
		boolean globalParent;
		try {
			parent = type.getSuperclass();
			globalParent = parent.isGlobal();
		} catch (ClassNotFoundException classEx) {
			parent = null;
			globalParent = false;
		}
		final int minMethNum = type.minMethodNum();
		out.print("\n    // Kan additions follow this line\n\n    public final static int kanMinMethod = ");
		out.print(minMethNum);
		out.print(";\n\n    public Object kanCallMethod(final int methodNum, final Object[] args)\n        throws Exception {\n\n        ");
		if (globalParent)
			out.print("if (methodNum < kanMinMethod)\n            return super.kanCallMethod(methodNum, args);\n        ");
		out.println("switch(methodNum) {");

		int methNum = minMethNum;
		final KanAST[] members = typeAST.getMembers();
		for (int i = 0; i < members.length; i++) {
			if (members[i].getType() == METHOD_DEF) {
				final MethAST d = (MethAST)members[i];
				final int m = d.getModifiers();
				if (!Modifier.isPrivate(m) && !Modifier.isStatic(m)) {
					printFastCall(d, methNum++);
				}
			}
		}

		// If the parent is not global, we also need to print fast call
		// information for its methods
		if (!globalParent) {
			final Method[] myMeths = type.getMethods();
			for (int i = 0; i < myMeths.length; i++) {
				final Method meth = myMeths[i];
				final int mod = meth.getModifiers();
				final Type declType = meth.getDeclaringClass();
				if (declType != type &&	!Modifier.isStatic(mod) &&
					meth.isStubAccessible(type) &&
					!meth.getName().equals("equals") &&
					!meth.getName().equals("hashCode") &&
					!meth.getName().equals("finalize") &&
					!meth.getName().equals("toString") &&
					(declType != Type.objectType || !Modifier.isFinal(mod))) {
					printFastCall(meth, -1-i);
				}
			}
		}
		out.println("        default:\n            return null;\n        }\n    }\n");

		// Print the global object controllers
		final String name = type.getName();
		if (!globalParent) {
			out.print("    public ");
			out.print(name);
			out.println(".KanID kanObjID;\n\n    public mokan.obj.ObjectID kanGetObjID() {\n        return kanObjID;\n    }\n");
			if (kanConstructors.size() == 0 &&
				!Modifier.isAbstract(type.getModifiers())) {
				// We need a Kan-ized default constructor
				out.print("    public ");
				final int index = name.lastIndexOf('.');
				out.print((index < 0) ? name : name.substring(index + 1));
				out.print("(final ");
				out.print(name);
				out.println(".KanID kanObjID) {\n        this.kanObjID = kanObjID;\n    }\n");
			}
			out.println("    public mokan.obj.GlobalObj kanClone() {\n        try {\n            return (mokan.obj.GlobalObj)clone();\n        } catch (Exception cloneEx) {\n            return null;\n        }\n    }\n");
		}
		out.println("    public void kanSetObject(mokan.obj.GlobalObj obj) {");
		if (globalParent)
			out.println("        super.kanSetObject(obj);");
		boolean first = true;
		for (int i = 0; i < members.length; i++) {
			if (members[i] instanceof DeclarationAST) {
				final DeclarationAST decl = (DeclarationAST)members[i];
				// FIXME: Some final variables are okay.  For example, final
				// arrays can be handled with a System.arraycopy.  Should we
				// have an interface, e.g., Settable, that defines a method
				// for copying state from one to another?
				if (!decl.getModifiers().isFinal()) {
					if (first) {
						out.println("        final " + name + " castObj = (" +
							name + ")obj;");
						first = false;
					}
					final VarAST[] vars = decl.getVariables();
					for (int j = 0; j < vars.length; j++) {
						out.println("        " + vars[j].getName() +
							" = castObj." + vars[j].getName() + ';');
					}
				}
			}
		}

		// Print the guards
		final int minGuardNum = type.minGuardNum();
		out.print("    }\n\n    // Guard code follows this line\n\n    public final static int kanMinGuard = ");
		out.print(minGuardNum);
		out.println(';');
		for (GuardExpr g = guardList; g != null; g = g.next) {
			out.print("\n    public final boolean kanGuard");
			out.print(g.number + minGuardNum);
			out.print('(');
			final ParameterAST[] params = g.params;
			if (params.length > 0) {
				out.print(params[0].toString());
				for (int i = 1; i < params.length; i++) {
					out.print(", ");
					out.print(params[i].toString());
				}
			}
			out.print(") {\n        return ");
			expression(g.expr);
			out.println(";\n    }");
		}
		out.print("\n    public boolean kanCallGuard(final int guardNum, final Object[] args) {\n        ");
		if (guardList == null) {
			// No guards
			out.print("return ");
			out.println(globalParent
				? "super.kanCallGuard(guardNum, args);"
				: "true;");
		} else {
			out.print("switch(guardNum) {\n");
			for (GuardExpr g = guardList; g != null; g = g.next) {
				final int guardNum = g.number + minGuardNum;
				out.print("        case ");
				out.print(guardNum);
				out.print(":\n            return kanGuard");
				out.print(guardNum);
				out.print('(');
				final ParameterAST[] params = g.params;
				if (params.length > 0) {
					parameterCasted(params[0], 0);
					for (int i = 1; i < params.length; i++) {
						out.print(", ");
						parameterCasted(params[i], i);
					}
				}
				out.println(");");
			}
			out.print("        default:\n            ");
			if (!globalParent) {
				out.print("System.err.println(\"Bad guard number: \" + guardNum + \", \" + this);\n            ");
			}
			out.print("return ");
			out.print(globalParent
				? "super.kanCallGuard(guardNum, args)"
				: "false");
			out.println(";\n        }");
		}
		out.println("    }\n");

		// Emit the stub object definition
		typeDefinitionStub(typeAST, parent, globalParent);
	}
}

compilationUnit[PrintWriter output]
	:	{ out = output; }
		#(FILE { #FILE.print(output); }
			(packageDefinition)?
			(importDefinition)*
			(typeDefinition)*
		)
		{ output.flush(); }
	; 

packageDefinition
	:	#( p:"package" {#p.print(out);} identifier[null]
			SEMI {#SEMI.print(out);} )
	;

importDefinition
	:	#( i:"import" {#i.print(out);} identifierStar SEMI {#SEMI.print(out);})
	;

typeDefinition { SourceType oldType = type; }
	:	#(cl:"class" { type = (SourceType)((TypeAST)#cl).retrieveType(); }
			modifiers id1:IDENT {#cl.print(out); #id1.print(out);}
			(extendsClause)? implementsClause objBlock) { type = oldType; }
	|	#(in:"interface" { type = (SourceType)((TypeAST)#in).retrieveType(); }
			modifiers id2:IDENT {#in.print(out); #id2.print(out);}
			(extendsClause)? interfaceBlock ) { type = oldType; }
	|	SEMI {#SEMI.print(out);}
	;

protected
typeDefinitionStub[Type parent, boolean globalParent]
	:	#("class" modifiers IDENT ("extends")? ("implements")?
			{
				out.print("static class KanID extends ");
				if (globalParent) {
					out.print(parent.getName());
					out.print(".KanID");
				} else {
					out.print("mokan.obj.ObjectID");
				}
				out.print(' ');
				methodNum = constrNum = 0;
			}
			objBlockStub)
	;

typeSpec
	:	#(TYPE typeSpecArray)
	;

typeSpecArray
	:	#( ARRAY_DECLARATOR (typeSpecArray)? RBRACK
			{ #ARRAY_DECLARATOR.print(out); #RBRACK.print(out); } )
	|	type
	;

type { String append = null; }
	:	{
			Type t = ((mokan.compiler.expr.ExpressionAST)#type).retrieveType();
			if (t != null && t.isGlobal()) {
				append = ".KanID";
			}
		}
		identifier[append]
	|	builtInType
	;

builtInType
    :   v:"void"	{#v.print(out);}
    |   b:"boolean"	{#b.print(out);}
    |   y:"byte"	{#y.print(out);}
    |   c:"char"	{#c.print(out);}
    |   s:"short"	{#s.print(out);}
    |   i:"int"		{#i.print(out);}
    |   f:"float"	{#f.print(out);}
    |   l:"long"	{#l.print(out);}
    |   d:"double"	{#d.print(out);}
    ;

modifiers
	:	#( MODIFIERS (modifier)* )
	;

modifier
    :   u:"public"			{#u.print(out);}
    |   p:"private"			{#p.print(out);}
    |   r:"protected"		{#r.print(out);}
    |   s:"static"			{#s.print(out);}
    |   f:"final"			{#f.print(out);}
    |   y:"synchronized"	{#y.print(out);}
    |   v:"volatile"		{#v.print(out);}
    |   t:"transient"		{#t.print(out);}
    |   n:"native"			{#n.print(out);}
    |   a:"abstract"		{#a.print(out);}
	|	i:"strictfp"		{#i.print(out);}
	// New Kan modifiers -- don't print them!
	|	"global"
	|	"immobile"
    ;

extendsClause
	:	#(e:"extends" {#e.print(out);} identifier[null]
			(COMMA {#COMMA.print(out);} identifier[null])* )
	;

implementsClause
	:	#(i:"implements" {#i.print(out);} identifier[null]
			(COMMA {#COMMA.print(out);} identifier[null])*)
		{
			if (type.isGlobal()) {
				final Type clone = Type.forClass(Cloneable.class);
				if (!type.implementsInterface(clone)) {
					out.print(", Cloneable");
				}
				final Type global = Type.forClass(mokan.obj.GlobalObj.class);
				if (!type.implementsInterface(global)) {
					out.print(", mokan.obj.GlobalObj");
				}
			}
		}
	|	/* Nothing */
		{
			if (type.isGlobal()) {
				final Type clone = Type.forClass(Cloneable.class);
				final Type global = Type.forClass(mokan.obj.GlobalObj.class);
				final boolean isCloneable = type.implementsInterface(clone);
				final boolean isGlobal = type.implementsInterface(global);
				if (!isCloneable || !isGlobal) {
					out.print(" implements");
					if (!isCloneable) {
						out.print(" Cloneable");
						if (!isGlobal) {
							out.print(',');
						}
					}
					if (!isGlobal) {
						out.print(" mokan.obj.GlobalObj");
					}
				}
			}
		}
	;

interfaceBlock
	:	#(	OBJBLOCK {#OBJBLOCK.print(out);}
			(	methodDecl
			|	variableDefs
			|	typeDefinition
			)*
			RCURLY {#RCURLY.print(out);}
		)
	;
	
objBlock
	:	#(	OBJBLOCK {#OBJBLOCK.print(out);}
			(	ctorDef
			|	methodDef
			|	variableDefs
			|	typeDefinition
			|	#(s:"static" {#s.print(out);} slist)
			|	#(INSTANCE_INIT slist)
			)*
			RCURLY
			// If this is the class block for a global class, add some stuff
			{
				if (#RCURLY.typeAST.retrieveType().isGlobal())
					addKanInfo(#RCURLY.typeAST);
				#RCURLY.print(out);
			}
		)
	;
	
protected
objBlockStub
{
	int methNum = 0;
	final String name = type.getName();
	final int mods = type.getModifiers();
}
	:	#(	OBJBLOCK
			{
				#OBJBLOCK.print(out);

				// Print out the special-purpose constructors
				out.print("public KanID() {}\n\n");
                                //out.print("public KanID(final mokan.comm.LogicalNode node) {\n        super(node);\n    }\n\n    ");

				// If there are no constructors, we need a default one
				final int consSize = kanConstructors.size();
				if (!Modifier.isAbstract(mods) && consSize == 0) {
					out.print("public KanID(final mokan.comm.LogicalNode node) {\n"
                                                  +"        super(node);\n"
                                                  +"        if (node == null || node.isLocal()) {\n"
                                                  +"            mokan.KanSystem.createObject(this, new ");
					out.print(name);
					out.print("(this), mokan.KanSystem.");
					out.print(ModifierAST.isImmobile(mods)
						? "IMMOBILE" : "MOBILE");
					out.print(");\n        } else {\n               mokan.KanSystem.createObject (this, ");
					out.print(name);
					out.print(".class, 0, new Object[0]);\n        }\n    }\n\n    ");
				}
			}
			(	ctorDefStub
			|	def:METHOD_DEF
				{
					final MethAST d = (MethAST)#def;
					final int m = d.getModifiers();
					if (!Modifier.isPrivate(m) && !Modifier.isStatic(m)) {
						printMethodStub(d, d.getDeclaringClass().minMethodNum() 
                                                                   + methNum++, #def.typeAST);
					}
				}
			|	VARIABLE_DEFS
			|	"class"
			|	"interface"
			|	SEMI
			|	"static"
			|	INSTANCE_INIT
			)*
			RCURLY
			{
				Type parent;
				boolean globalParent;
				try {
					parent = type.getSuperclass();
					globalParent = parent.isGlobal();
				} catch (ClassNotFoundException classEx) {
					parent = null;
					globalParent = false;
				}
				if (!globalParent) {
					// Emit stubs for the parent's methods.  Can't just get
					// the parent's methods directly, because *this* class may
					// have overridden some of them.  So instead we get all
					// the methods available to this class, and print stubs
					// for the ones not declared in this class.
					final Method[] myMeths = type.getMethods();
					for (int i = 0; i < myMeths.length; i++) {
						final Method meth = myMeths[i];
						final int mod = meth.getModifiers();
						final Type declType = meth.getDeclaringClass();
						if (declType != type && !Modifier.isStatic(mod) &&
							meth.isStubAccessible(type) &&
							!meth.getName().equals("equals") &&
							!meth.getName().equals("hashCode") &&
							!meth.getName().equals("finalize") &&
							!meth.getName().equals("toString") &&
							(declType != Type.objectType ||
								!Modifier.isFinal(mod))) {
							printMethodStub(meth, -1-i, #RCURLY.typeAST);
						}
					}
				}

				// Print the remote object creator
				if (!Modifier.isAbstract(mods)) {
					final ConstrAST[] constrs =
						new ConstrAST[kanConstructors.size()];
					kanConstructors.toArray(constrs);
					out.print("\n    public final static ");
					//out.print(name);
					out.print("KanID kanCreate");
					//out.print(name.replace('.', '$'));
					out.println("(int constrNum, final mokan.comm.LogicalNode node, final Object[] args) {\n        switch (constrNum) {");
					for (int i = 0; i < constrs.length; i++) {
						out.print("        case ");
						out.print(i);
						out.print(":\n            return new ");
						//out.print(name);
						out.print("KanID(node");
						final Type[] params = constrs[i].getParameterTypes();
						for (int j = 0; j < params.length; j++) {
							out.print(", ");
							final Type paramType = params[j];
							if (paramType.isPrimitive()) {
								out.print("((");
								out.print(paramType.getWrapper());
								out.print(")args[");
								out.print(j);
								out.print("])");
								out.print(paramType.getUnwrapper());
							} else {
								out.print('(');
								out.print(paramType.getModifiedName());
								out.print(")args[");
								out.print(j);
								out.print(']');
							}
						}
						out.println(");");
					}
					if (constrs.length == 0) {
						out.print("        default:\n            return new ");
						//out.print(name);
						out.println("KanID(node);");
					} else {
						out.println("        default:\n            return null;");
					}
					out.print("        }\n    }");
				}

				#RCURLY.print(out);
			}
		)
	;

ctorDef
	:	#(c:CTOR_DEF modifiers ctorHead ctorSlist)
		{ 
          if (!type.isInner())
               kanConstructors.add(#c);
       		//  System.err.println (kanConstructors.size() + " "+ #c);
		 }
	;

protected
ctorDefStub { final int mods = type.getModifiers(); }
	:       #(CTOR_DEF
			{
				// FIXME: Ugly hack that may break in future versions of ANTLR
				if (Modifier.isAbstract(mods)) {
					_retTree = #ctorDefStub.getNextSibling();
					return;
				}
			}
			modifiers head:ctorHeadStub SLIST)
		{
			#SLIST.print(out);
			final String baseName = ((KanAST)#head).typeAST.name;
			out.print("super(node);\n        if (node.isLocal()) {\n            final ");
			out.print(baseName);
			out.print(" obj = new ");
			out.print(baseName);
			out.print("(this");
			final KanAST params =
				(KanAST)#head.getNextSibling().getNextSibling();
			if (params.getFirstChild() != null)
				out.print(", ");
			parameterNameList(params);
			out.print(");\n            mokan.KanSystem.createObject(this, obj, mokan.KanSystem.");
			out.print(ModifierAST.isImmobile(mods) ? "IMMOBILE" : "MOBILE");
			out.print(");\n        } else {\n           mokan.KanSystem.createObject(this, ");
			out.print(#head.getText());
			out.print(".class, ");
			out.print(constrNum++);
			out.print(", new Object[] {");
			parameterWrappedList(params);
			out.print("});\n        }\n    }\n\n    ");
		}
	;

// Kan extension: Add a parameter to each constructor for the object ID
ctorHead
	:	IDENT LPAREN {#IDENT.print(out); #LPAREN.print(out);}
		#( p:PARAMETERS
			{
				if (type.isGlobal()) {
					out.print("final KanID kanID");
					if (p.getFirstChild() != null) {
						out.print(", ");
					}
				}
			}
			(parameterDef (COMMA {#COMMA.print(out);} parameterDef)* )? )
		RPAREN {#RPAREN.print(out);} declaratorBrackets (throwsClause)?
	;

protected
ctorHeadStub
	:	IDENT LPAREN {out.print("KanID"); #LPAREN.print(out);}
		#( p:PARAMETERS
			{
				out.print("final mokan.comm.LogicalNode node");
				if (#p.getFirstChild() != null) {
					out.print(", ");
				}
			}
			(parameterDef (COMMA {#COMMA.print(out);} parameterDef)* )? )
		RPAREN {#RPAREN.print(out);} declaratorBrackets (throwsClause)?
	;

// Kan extension: Add a setting of the object ID parameter on the last line
ctorSlist
	:	#( SLIST {#SLIST.print(out);} (stat)* RCURLY )
		{
			if (type.isGlobal()) {
				out.print("kanObjID = kanID; ");
			}
			#RCURLY.print(out);
		}
	;

methodDecl
	:	#(METHOD_DEF modifiers typeSpec methodHead SEMI {#SEMI.print(out);})
	;

methodDef
	:	#(m:METHOD_DEF mod:modifiers typeSpec methodHead
			(slist | SEMI {#SEMI.print(out);} ))
	; exception
      catch  [RuntimeException ex] {
           System.err.println ("internal error:" + ex + " in method " + #m + " of " + #m.topLevel);
           throw ex;
      }

variableDefs
	:	#(VARIABLE_DEFS modifiers typeSpec variableDef
			(COMMA {#COMMA.print(out);} variableDef)* SEMI {#SEMI.print(out);}
		)
	;

variableDef
	:	#(VARIABLE_DEF variableDeclarator (varInitializer)? )
	;

variableDeclarator
	:	IDENT {#IDENT.print(out);} declaratorBrackets
	;

declaratorBrackets
	:	(l:ARRAY_DECLARATOR r:RBRACK {#l.print(out); #r.print(out);} )*
	;

varInitializer
	:	#(ASSIGN {#ASSIGN.print(out);} initializer)
	;

parameterDeclarationList
	:	#(PARAMETERS
			(parameterDef (COMMA {#COMMA.print(out);} parameterDef)* )? )
	;

parameterNameList
	:	#(PARAMETERS
			(parameterName (COMMA {#COMMA.print(out);} parameterName)* )? )
	;

parameterWrappedList
	:	#(PARAMETERS (parameterWrapped (COMMA {#COMMA.print(out);}
					parameterWrapped)* )? )
	;

parameterDef
	:	#(PARAMETER_DEF modifiers typeSpec IDENT {#IDENT.print(out);}
			declaratorBrackets)
	;

parameterName
	:	#(PARAMETER_DEF MODIFIERS TYPE IDENT {#IDENT.print(out);})
	;

parameterWrapped
	:	#(PARAMETER_DEF MODIFIERS t:TYPE id:IDENT)
		{ printMaybeWrapped((KanAST)#t.getFirstChild(), #id.getText()); }
	;

parameterCasted[int num]
	:	#(PARAMETER_DEF MODIFIERS t:TYPE id:IDENT)
		{
			final Type paramType = ((VarAST)#id).retrieveType();
			if (paramType.isPrimitive()) {
				final KanAST typeName = (KanAST)#t.getFirstChild();
				out.print("((");
				out.print(getWrapper(typeName));
				out.print(")args[");
				out.print(num);
				out.print("]).");
				out.print(getUnwrapper(typeName));
				out.print("()");
			} else {
				out.print('(');
				out.print(paramType.getModifiedName());
				out.print(")args[");
				out.print(num);
				out.print(']');
			}
		}
	;

initializer
	:	expression
	|	arrayInitializer
	;

arrayInitializer
	:	#(ARRAY_INIT {#ARRAY_INIT.print(out);}
			(initializer (commaInitializer)? )?
			RCURLY {#RCURLY.print(out);} )
	;

commaInitializer
	:	#(COMMA {#COMMA.print(out);} (initializer (commaInitializer)? )? )
	;

methodHead
	:	IDENT LPAREN {#IDENT.print(out); #LPAREN.print(out);}
		parameterDeclarationList RPAREN {#RPAREN.print(out);}
		declaratorBrackets (throwsClause)?
	;

throwsClause
	:	#( t:"throws" {#t.print(out);} identifier[null]
			(COMMA {#COMMA.print(out);} identifier[null])* )
	;

identifier[String append]
	:	IDENT {#IDENT.print(out, append);}
	|	#( DOT identifier[null]
			IDENT {#DOT.print(out); #IDENT.print(out, append);} )
	;

identifierStar
	:	IDENT {#IDENT.print(out);}
	|	#(DOT identifier[null] {#DOT.print(out);}
		  (STAR {#STAR.print(out);} | IDENT {#IDENT.print(out);} ) )
	;

slist
	:	#( SLIST {#SLIST.print(out);} (stat)* RCURLY {#RCURLY.print(out);} )
	;

stat:	variableDefs
	|	#(TYPE_STAT typeDefinition)
	|	#(EXPRESSION_STAT expression sm1:SEMI) {sm1.print(out);}
	|	#(LABELED_STAT COLON {#LABELED_STAT.print(out); #COLON.print(out);}
			stat)
	|	#(i:"if" l1:LPAREN {i.print(out); #l1.print(out);} expression r1:RPAREN
			{#r1.print(out);} stat (e:"else" {#e.print(out);} stat)? )
	|	#(f:"for" l2:LPAREN {f.print(out); #l2.print(out);}
			#(FOR_INIT (variableDefs | elist sm2:SEMI {#sm2.print(out);}) )
			#(FOR_CONDITION (expression)? sm3:SEMI {#sm3.print(out);} )
			#(FOR_ITERATOR elist)
			r2:RPAREN {#r2.print(out);} stat
		)
	|	#(w:"while" l3:LPAREN {#w.print(out); #l3.print(out);} expression
		  r3:RPAREN {#r3.print(out);} stat)
	|	#(d:"do" {#d.print(out);} stat wh:"while" l4:LPAREN
			{#wh.print(out); #l4.print(out);} expression r4:RPAREN sm4:SEMI
			{#r4.print(out); #sm4.print(out);} )
	|	#(b:"break" {#b.print(out);} (id1:IDENT {#id1.print(out);})?
			sm5:SEMI {#sm5.print(out);} )
	|	#(c:"continue" {#c.print(out);} (id2:IDENT {#id2.print(out);})?
			sm6:SEMI {#sm6.print(out);} )
	|	#(r:"return" {#r.print(out);} (expression)?
			sm7:SEMI {#sm7.print(out);} )
	|	#(sw:"switch" l5:LPAREN {#sw.print(out); #l5.print(out);} expression
			r5:RPAREN LCURLY {#r5.print(out); #LCURLY.print(out);}
			(caseGroup)* RCURLY {#RCURLY.print(out);} )
	|	#(t:"throw" {#t.print(out);} expression sm8:SEMI {#sm8.print(out);} )
	|	#(sy:"synchronized" l6:LPAREN {#sy.print(out); #l6.print(out);}
			expression r6:RPAREN {#r6.print(out);} stat)
	|	tryBlock
	|	slist // nested SLIST
	|	EMPTY_STAT {#EMPTY_STAT.print(out);}
	// Kan extension: a guarded statement
	|	#(g:"guard" GUARD_MODIFIERS l7:LPAREN exp:noPrintExpression r7:RPAREN
			{
				final GuardAST guard = (GuardAST)#g;
				final boolean trivial = guard.isTrivial();
				if (!trivial) {
					guardList = new GuardExpr(astFactory, guardList,
									(mokan.compiler.expr.ExpressionAST)#exp,
									type.guardCount++);
				}
				out.print(" for ");
				#l7.print(out);
				out.print(";;");
				#r7.print(out);
				out.print("{ try { mokan.KanSystem.guardBegin(kanObjID, ");
				if (trivial) {
					out.print("null");
				} else {
					out.print("new mokan.obj.GuardCondition(");
					out.print(guardList.number);
					out.print(" + kanMinGuard, new Object[] {");
					final ParameterAST[] vars = guardList.params;
					if (vars.length > 0) {
						printMaybeWrapped
							((KanAST)vars[0].getTypeName().getFirstChild(),
							 vars[0].getParamName().getName());
						for (int j = 1; j < vars.length; j++) {
							out.print(", ");
							printMaybeWrapped
								((KanAST)vars[j].getTypeName().getFirstChild(),
								 vars[j].getParamName().getName());
						}
					}
					out.print("})");
				}
				out.print(", ");

				out.print(guard.isWriteTransaction());
				out.print(", mokan.transact.Transaction.");
				switch (guard.transType()) {
				case Transaction.NONBLOCKING:
					out.print("NONBLOCKING");
					break;
				case Transaction.BLOCKING:
					out.print("BLOCKING");
					break;
				default:
					out.print("NESTED");
					break;
				}
				out.print(");");
			}
			s:stat)
		{
			// FIXME: Gross hack for now, flow analysis later
			StatementAST st = (StatementAST)#s;
			if (st instanceof CompoundAST) {
				final StatementAST[] list = ((CompoundAST)st).getList();
				st = list[list.length - 1];
			}
			if (!(st instanceof BreakAST || st instanceof ContinueAST ||
					st instanceof ThrowAST || st instanceof ReturnAST)) {
				out.print("break; ");
			}
			out.print("} catch(mokan.transact.TransactionAbortException ex) {} finally { mokan.KanSystem.guardEnd(); } }");
		}
	;

caseGroup
	:	#(CASE_GROUP (( c:"case" {#c.print(out);} expression
					  | d:"default" {#d.print(out);})
					 COLON {#COLON.print(out);} )+ (stat)*)
	;

tryBlock
	:	#(t:"try" {#t.print(out);} slist (handler)*
			(f:"finally" {#f.print(out);} slist)? )
	;

handler
	:	#(c:"catch" LPAREN {#c.print(out); #LPAREN.print(out);} parameterDef
		  RPAREN {#RPAREN.print(out);} slist )
	;

elist
	:	#( ELIST (expression (COMMA {#COMMA.print(out);} expression)* )? )
	;

// Don't print these; they'll be printed later.  What I REALLY want is a way
// to say "Match any one token and just shut up about what you matched."
noPrintExpression
	:	QUESTION | ASSIGN | PLUS_ASSIGN | CONCAT_ASSIGN | MINUS_ASSIGN
	|	STAR_ASSIGN | DIV_ASSIGN | MOD_ASSIGN | SR_ASSIGN | BSR_ASSIGN
	|	SL_ASSIGN | BAND_ASSIGN | BXOR_ASSIGN | BOR_ASSIGN | LOR | LAND | BOR
	|	BXOR | BAND | NOT_EQUAL | EQUAL | LT | GT | LE | GE | SL | SR | BSR
	|	PLUS | CONCATENATION | MINUS | DIV | MOD | STAR | INC | DEC
	|	POST_INC | POST_DEC | BNOT | LNOT | "instanceof" | UNARY_MINUS
	|	UNARY_PLUS | IDENT | DOT | INDEX_OP | METHOD_CALL | ASYNC_METHOD_CALL
	|	CONSTRUCTOR_CALL | TYPECAST | PAREN_EXPR | "new" | NUM_INT
	|	CHAR_LITERAL | STRING_LITERAL | NUM_FLOAT | "super" | "true" | "false"
	|	"this" | "kanThis" | "null" | TYPE
	;

expression
	:	#(QUESTION expression {#QUESTION.print(out);} expression
			COLON {#COLON.print(out);} expression)
	|	#(ASSIGN expression {#ASSIGN.print(out);} expression)
	|	#(PLUS_ASSIGN expression {#PLUS_ASSIGN.print(out);} expression)
	|	#(CONCAT_ASSIGN expression {#CONCAT_ASSIGN.print(out);} expression)
	|	#(MINUS_ASSIGN expression {#MINUS_ASSIGN.print(out);} expression)
	|	#(STAR_ASSIGN expression {#STAR_ASSIGN.print(out);} expression)
	|	#(DIV_ASSIGN expression {#DIV_ASSIGN.print(out);} expression)
	|	#(MOD_ASSIGN expression {#MOD_ASSIGN.print(out);} expression)
	|	#(SR_ASSIGN expression {#SR_ASSIGN.print(out);} expression)
	|	#(BSR_ASSIGN expression {#BSR_ASSIGN.print(out);} expression)
	|	#(SL_ASSIGN expression {#SL_ASSIGN.print(out);} expression)
	|	#(BAND_ASSIGN expression {#BAND_ASSIGN.print(out);} expression)
	|	#(BXOR_ASSIGN expression {#BXOR_ASSIGN.print(out);} expression)
	|	#(BOR_ASSIGN expression {#BOR_ASSIGN.print(out);} expression)
	|	#(LOR expression {#LOR.print(out);} expression)
	|	#(LAND expression {#LAND.print(out);} expression)
	|	#(BOR expression {#BOR.print(out);} expression)
	|	#(BXOR expression {#BXOR.print(out);} expression)
	|	#(BAND expression {#BAND.print(out);} expression)
	|	#(NOT_EQUAL expression {#NOT_EQUAL.print(out);} expression)
	|	#(EQUAL expression {#EQUAL.print(out);} expression)
	|	#(LT expression {#LT.print(out);} expression)
	|	#(GT expression {#GT.print(out);} expression)
	|	#(LE expression {#LE.print(out);} expression)
	|	#(GE expression {#GE.print(out);} expression)
	|	#(SL expression {#SL.print(out);} expression)
	|	#(SR expression {#SR.print(out);} expression)
	|	#(BSR expression {#BSR.print(out);} expression)
	|	#(PLUS expression {#PLUS.print(out);} expression)
	|	#(CONCATENATION expression {#CONCATENATION.print(out);} expression)
	|	#(MINUS expression {#MINUS.print(out);} expression)
	|	#(DIV expression {#DIV.print(out);} expression)
	|	#(MOD expression {#MOD.print(out);} expression)
	|	#(STAR expression {#STAR.print(out);} expression)
	|	#(INC {#INC.print(out);} expression)
	|	#(DEC {#DEC.print(out);} expression)
	|	#(POST_INC expression {#POST_INC.print(out);})
	|	#(POST_DEC expression {#POST_DEC.print(out);})
	|	#(BNOT {#BNOT.print(out);} expression)
	|	#(LNOT {#LNOT.print(out);} expression)
	|	#(i:"instanceof" expression {#i.print(out);} expression)
	|	#(UNARY_MINUS {#UNARY_MINUS.print(out);} expression)
	|	#(UNARY_PLUS {#UNARY_PLUS.print(out);} expression)
	|	primaryExpression
	;

primaryExpression
    :   IDENT { #IDENT.print(out); }
    |   #(	d:DOT
			(	expression {#d.print(out);}
				(	IDENT {#IDENT.print(out);}
				|	arrayIndex
				|	th:"this" {#th.print(out);}
				|	kth:"kanThis"
					{
						out.print("kanObjID");
						#kth.printHiddenAfter(out);
					}
				|	c:"class"     {#c.print(out);}
				|	#( ne:"new" IDENT {#ne.print(out); #IDENT.print(out);}
						lp:LPAREN {#lp.print(out);} elist
						rp:RPAREN {#rp.print(out);} )
				|	su:"super" {#su.print(out);}
				)
			|	#(ARRAY_DECLARATOR type RBRACK)
				{#ARRAY_DECLARATOR.print(out); #RBRACK.print(out);}
				(cl:"class" {#d.print(out); #cl.print(out);})?
			|	builtInType (cl2:"class" {#d.print(out); #cl2.print(out);})?
			)
		)
	|	arrayIndex
	|	#(call1:METHOD_CALL primaryExpression {#call1.print(out);} elist
			r2:RPAREN {#r2.print(out);})
	|	#(call2:CONSTRUCTOR_CALL primaryExpression
			{
				#call2.print(out);
				// Kan extension: print an extra first parameter for
				// constructor ("this" or "super") calls, but only if the
				// corresponding class is global
				final String name =
					((MethodCallAST)#call2).getMethodName().getName();
				try {
					if ((name.equals("this") && type.isGlobal()) ||
						(name.equals("super") &&
							type.getSuperclass().isGlobal())) {
						out.print("kanID, ");
					}
				} catch (ClassNotFoundException classEx) {
					// Do nothing; we don't know if the parent is global
				}
			}
			elist r3:RPAREN {#r3.print(out);})
	|	#(t:TYPECAST {#t.print(out);} typeSpec r4:RPAREN {#r4.print(out);}
			expression)
	|   #(p:PAREN_EXPR {#p.print(out);} expression
			r5:RPAREN {#r5.print(out);} )
	|   newExpression
	|   constant
    |   s:"super"	   {#s.print(out);}
    |   tr:"true"	   {#tr.print(out);}
    |   f:"false"	   {#f.print(out);}
    |   thi:"this"	   {#thi.print(out); 

/*
			Type t = ((mokan.compiler.expr.ExpressionAST)#type).retrieveType();
			if (t != null && t.isGlobal()) {
                out.print(".kanObjID");
			}
*/


    }
    |   kthi:"kanThis" {out.print("kanObjID"); #kthi.printHiddenAfter(out);}
    |   nu:"null"	   {#nu.print(out);}
	|	typeSpec // type name used with instanceof
	// Kan extension: asynchronous method call
	|	#(call3:ASYNC_METHOD_CALL
			{
				if (((MethodCallAST)#call3).getObject() == null) {
					out.print("kanObjID.");
				}
			}
			primaryExpression { out.print("KanAsync"); #call3.print(out); }
			elist r6:RPAREN {#r6.print(out);} "async")
	;

arrayIndex
	:	#(INDEX_OP primaryExpression {#INDEX_OP.print(out);} expression
		  RBRACK {#RBRACK.print(out);} )
	;

constant
    :   NUM_INT {#NUM_INT.print(out);}
    |   CHAR_LITERAL {#CHAR_LITERAL.print(out);}
    |   STRING_LITERAL {#STRING_LITERAL.print(out);}
    |   NUM_FLOAT {#NUM_FLOAT.print(out);}
    ;

newExpression { SourceType oldType = type; }
	:	#(	n:"new" {#n.print(out);}
			(LPAREN e:noPrintExpression RPAREN { out.print(' '); })? t:type
			(	newArrayDeclarator (arrayInitializer)?
			|	LPAREN
				{
					#LPAREN.print(out);
					final mokan.compiler.expr.ExpressionAST expr =
						(mokan.compiler.expr.ExpressionAST)#t;

                   if (expr == null || expr.retrieveType() == null ) {
                        throw new SemanticException("Unknown type:"+expr);
                    }

					if (expr.retrieveType().isGlobal()) {
						if (#e != null) {
							expression(#e);
						} else {
							out.print("null");
						}
						if (#LPAREN.getNextSibling().getFirstChild() != null) {
							out.print(',');
						}
					}
				}
				elist RPAREN {#RPAREN.print(out);}
				// DON'T PRINT THAT "class"!
				(#(cl:"class"
					{ type = (SourceType)((TypeAST)#cl).retrieveType(); }
					objBlock) { type = oldType; } )?
			)
		)
	; exception 
      catch [SemanticException ex] {
           reportError (ex.toString());
      }

newArrayDeclarator
	:	#( ARRAY_DECLARATOR (newArrayDeclarator)?
			{#ARRAY_DECLARATOR.print(out);} (expression)?
			RBRACK {#RBRACK.print(out);} )
	;


/* arch-tag: f212c4a4-569b-43a2-b035-4622408a6f07 */