/*
 * @(#)$Id: DeclarationAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.stmt;

import antlr.collections.AST;
import java.util.ArrayList;
import mokan.compiler.KanTokenTypes;
import mokan.compiler.ModifierAST;
import mokan.compiler.Type;
import mokan.compiler.VarList;
import mokan.compiler.expr.VarAST;
import mokan.compiler.expr.TypeAST;
import mokan.transact.Transaction;

/**
 * An AST node that represents a variable declaration statement
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public final class DeclarationAST extends StatementAST implements KanTokenTypes
{

    /**
     * The modifiers for this declaration
     */
    private final ModifierAST modifiers;

    /**
     * The type for this declaration
     */
    private final TypeAST typeSpec;

    /**
     * The variables declared in this declaration
     */
    private final VarAST[] variables;

    /**
     * Create a new declaration statement AST
     *
     * @param mods the modifiers for this declaration
     * @param type the type for this declaration
     * @param vars the varaibles declared in this declaration
     */
    public DeclarationAST(final ModifierAST mods, final TypeAST type,
			  AST vars) {
	initialize(VARIABLE_DEFS, "VARIABLE_DEFS");
	modifiers = mods;
	typeSpec = type;

	// Collect the variable declarations
	final ArrayList list = new ArrayList();
	while (vars != null) {
	    list.add(vars.getFirstChild());
	    vars = vars.getNextSibling();	// Skip the comma
	    if (vars == null)
		break;
	    vars = vars.getNextSibling();
	}

	// Create the variables array
	variables = new VarAST[list.size()];
	list.toArray(variables);
    }

    public void parseComplete() {
	super.parseComplete();
	context.mods = modifiers;
	context.type = typeSpec;
	for (int i = 0; i < variables.length; i++) {
	    variables[i].parseComplete();
	}
    }

    protected Type[] computeExceptions() {
	Type[] exceptions = noTypes;
	for (int i = 0; i < variables.length; i++)
	    exceptions = Type.mergeTypeLists(exceptions,
					     variables[i].getExceptionTypes());
	return exceptions;
    }

    protected StatementAST[] computeControl() {
	return new StatementAST[] { next };
    }

    public VarList getVarList() {
	return new VarList(variables);
    }

    /**
     * Get the modifiers for this declaration
     *
     * @return the modifiers
     */
    public ModifierAST getModifiers() {
	return modifiers;
    }

    /**
     * Get the type name for this declaration
     *
     * @return the type name
     */
    public TypeAST getTypeName() {
	return typeSpec;
    }

    /**
     * Get the variables declared in this declaration
     *
     * @return an array of variables
     */
    public VarAST[] getVariables() {
	return variables;
    }

    public int transType() {
	int type = Transaction.NONBLOCKING;
	for (int i = 0; i < variables.length; i++) {
	    int vType = variables[i].transType();
	    if (vType > type)
		type = vType;
	    if (type == Transaction.NESTED)	// Might as well stop now
		break;
	}
	return type;
    }
}
/* arch-tag: 4fe689f0-0a85-4c59-9521-8c2c80756dee */