/*
 * @(#)ForAST.java ForAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.stmt;

import antlr.Token;
import antlr.collections.AST;
import mokan.compiler.HasExceptions;
import mokan.compiler.KanTokenTypes;
import mokan.compiler.Type;
import mokan.compiler.VarList;
import mokan.compiler.expr.ListAST;
import mokan.compiler.stmt.DeclarationAST;
import mokan.transact.Transaction;

/**
 * An AST node that represents a for statement
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public final class ForAST extends StatementAST implements KanTokenTypes {

    /**
     * The initializer, either a
     * {@link mokan.compiler.stmt.DeclarationAST DeclarationAST} or a
     * {@link mokan.compiler.expr.ListAST ListAST}
     */
    private HasExceptions init;

    /**
     * The condition test
     */
    private mokan.compiler.expr.ExpressionAST cond;

    /**
     * The increment
     */
    private mokan.compiler.expr.ExpressionAST incr;

    /**
     * The statement
     */
    private StatementAST stmt;

    /**
     * Create a new for statement AST
     *
     * @param token the token represented by this AST node
     */
    public ForAST(final Token token) {
        super(token);
    }

    public void parseComplete() {
        super.parseComplete();
        final AST initNode = getFirstChild().getNextSibling();
        init = (HasExceptions)initNode.getFirstChild();
        final AST condNode = initNode.getNextSibling();
        final AST theCond = condNode.getFirstChild();
        cond = (theCond.getType() == SEMI)
            ? null
            : (mokan.compiler.expr.ExpressionAST)theCond;
        final AST incrNode = condNode.getNextSibling();
        incr = (mokan.compiler.expr.ExpressionAST)incrNode.getFirstChild();
        stmt = (StatementAST)incrNode.getNextSibling().getNextSibling();

        // Set up the break/continue context
        context.pushContinue(this);
        if (init != null && init instanceof ListAST) {
            ((ListAST)init).parseComplete();
        } else {
            ((DeclarationAST)init).parseComplete();
        }
        if (cond != null)
            cond.parseComplete();
        if (incr != null)
            incr.parseComplete();
        stmt.parseComplete();
        context.popContinue();
    }

    protected Type[] computeExceptions() {
        Type[] retVal = (cond == null)
            ? init.getExceptionTypes()
            : Type.mergeTypeLists(init.getExceptionTypes(),
                                  cond.getExceptionTypes());
        retVal = Type.mergeTypeLists(retVal, incr.getExceptionTypes());
        return Type.mergeTypeLists(retVal, stmt.getExceptionTypes());
    }

    protected StatementAST[] computeControl() {
        return stmt.nextControlPoints();
    }

    public VarList getVarList() {
        final VarList initList = (init instanceof ListAST)
            ? ((ListAST)init).getVarList()
            : ((DeclarationAST)init).getVarList();
        return new VarList(new VarList(initList, cond.getVarList()),
                           new VarList(incr.getVarList(), stmt.getVarList()));
    }

    /**
     * Get the initialization clause of the for loop
     *
     * @return the initialization clause, which is either a
     * {@link jparse.stmt.DeclarationAST DeclarationAST} or a
     * {@link jparse.expr.ListAST ListAST}
     */
    public HasExceptions getInit() {
        return init;
    }

    /**
     * Get the condition clause of the for loop
     *
     * @return the condition clause
     */
    public mokan.compiler.expr.ExpressionAST getCondition() {
        return cond;
    }

    /**
     * Get the increment clause of the for loop
     *
     * @return the increment clause
     */
    public mokan.compiler.expr.ExpressionAST getIncrement() {
        return incr;
    }

    /**
     * Get the body of the for loop
     *
     * @return the body
     */
    public StatementAST getBody() {
        return stmt;
    }

    public int transType() {
        return (((init instanceof ListAST) &&
                 ((ListAST)init).transType() == Transaction.NESTED) ||
                ((DeclarationAST)init).transType() == Transaction.NESTED ||
                cond.transType() == Transaction.NESTED ||
                incr.transType() == Transaction.NESTED ||
                stmt.transType() == Transaction.NESTED)
            ? Transaction.NESTED
            : Transaction.BLOCKING;
    }
}
/* arch-tag: 30f8c6f0-6246-4c9e-8bd2-70ab0015b9c1 */