/*
 * @(#)IfElseAST.java IfElseAST.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 * JParse: a freely available Java parser
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jerry James
 * Email: james@eecs.ukans.edu, james@ittc.ukans.edu, jamesj@acm.org
 * Address: EECS Department
 *          415 Snow Hall
 *          University of Kansas
 *          Lawrence, KS  66045
 */
package mokan.compiler.stmt;

import antlr.Token;
import antlr.collections.AST;
import java.util.HashSet;
import mokan.compiler.Type;
import mokan.compiler.VarList;
import mokan.transact.Transaction;

/**
 * An AST node that represents an if-else statement
 *
 * @version $Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author Jerry James
 */
public final class IfElseAST extends StatementAST {

    /**
     * The conditional
     */
    private mokan.compiler.expr.ExpressionAST condition;

    /**
     * The "then" statement
     */
    private StatementAST thenStmt;

    /**
     * The "else" statement
     */
    private StatementAST elseStmt;

    /**
     * Create a new if-else statement AST
     *
     * @param token the token represented by this AST node
     */
    public IfElseAST(final Token token) {
	super(token);
    }

    public void parseComplete() {
	super.parseComplete();
	condition = (mokan.compiler.expr.ExpressionAST)getFirstChild()
	    .getNextSibling();
	condition.parseComplete();
	thenStmt = (StatementAST)condition.getNextSibling().getNextSibling();
	thenStmt.parseComplete();
	final AST elseLiteral = thenStmt.getNextSibling();
	if (elseLiteral != null) {
	    elseStmt = (StatementAST)elseLiteral.getNextSibling();
	    elseStmt.parseComplete();
	}
    }

    protected Type[] computeExceptions() {
	final Type[] body = (elseStmt == null)
	    ? thenStmt.getExceptionTypes()
	    : Type.mergeTypeLists(thenStmt.getExceptionTypes(),
				  elseStmt.getExceptionTypes());
	return Type.mergeTypeLists(condition.getExceptionTypes(), body);
    }

    protected StatementAST[] computeControl() {
	// Optimize for the common case
	if (elseStmt == null)
	    return thenStmt.nextControlPoints();

	final HashSet control = new HashSet();
	StatementAST[] points = thenStmt.nextControlPoints();
	for (int i = 0; i < points.length; i++) {
	    control.add(points[i]);
	}
	points = elseStmt.nextControlPoints();
	for (int i = 0; i < points.length; i++) {
	    control.add(points[i]);
	}

	points = new StatementAST[control.size()];
	return (StatementAST[])control.toArray(points);
    }

    public VarList getVarList() {
	return (elseStmt == null)
	    ? new VarList(condition.getVarList(), thenStmt.getVarList())
	    : new VarList(condition.getVarList(), thenStmt.getVarList(),
			  elseStmt.getVarList());
    }

    /**
     * Get the condition of the <code>if</code> statement
     *
     * @return the condition
     */
    public mokan.compiler.expr.ExpressionAST getCondition() {
	return condition;
    }

    /**
     * Get the <code>then</code> part of the <code>if</code> statement
     *
     * @return the then part
     */
    public StatementAST getThen() {
	return thenStmt;
    }

    /**
     * Get the <code>else</code> part of the <code>if</code> statement
     *
     * @return the else part
     */
    public StatementAST getElse() {
	return elseStmt;
    }

    public int transType() {
	int type = condition.transType();
	if (type == Transaction.NESTED)
	    return type;
	int tType = thenStmt.transType();
	if (tType > type) {
	    type = tType;
	    if (type == Transaction.NESTED) {
		return type;
	    }
	}
	if (elseStmt != null) {
	    int eType = elseStmt.transType();
	    if (eType > type) {
		type = eType;
	    }
	}
	return type;
    }
}
/* arch-tag: c3148c45-13f9-49fa-be22-5027d968937b */