/*
 * @(#)ProcessEventList.java ProcessEventList.java,v 1.1.1.1 2002/11/06 22:10:26 kris Exp $
 *
 *
 */

package mokan.debug;

import java.io.*;
import java.util.ArrayList;

/**
 * Process the Kan system event list, to create the source files for
 * {@link mokan.debug.Event mokan.debug.Event} and {@link mokan.ui.DebugList
 * mokan.ui.DebugList}.  The event list is expected to be in the same directory,
 * in a file named "event.list".  Those wishing to add more event types should
 * edit event.list, then run this program, rather than editing Event.java and
 * DebugList.java directly.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:26 $
 * @author	Jerry James
 */
public final class ProcessEventList {

    // ***********************************************************************
    // The non-static portion of the class.
    // This part contains information on a single user-selectable event
    // ***********************************************************************

    /**
     * The name of this event
     */
    private final String name;

    /**
     * The subevents of this event
     */
    private final ArrayList children = new ArrayList();
    // TODO: Turn this into a real array (more complicated file read method)

    /**
     * Create a ProcessEventList object to hold one event name and references
     * to its children
     *
     * @param name the name of the event represented by this object
     */
    private ProcessEventList(final String name) {
        this.name = name;
    }

    /**
     * Write out this event and all of its descendants as a constant
     *
     * @param num the number for this event
     * @param output the {@link java.io.PrintWriter PrintWriter} on which to
     *        write
     * @return the number of the next event to write
     */
    private int printEvent(int num, final PrintWriter output) {
        output.print("    public static final int ");
        output.print(name);
        output.print(" = ");
        output.print(num++);
        output.println(";");
        final int size = children.size();
        for (int i = 0; i < size; i++) {
            num = ((ProcessEventList)children.get(i)).printEvent(num, output);
        }
        return num;
    }

    /**
     * Write out this event and all of its descendants as a string
     *
     * @param pos the character position at which to start printing
     * @param comma <code>true</code> if we need to print a comma first
     * @param output the {@link java.io.PrintWriter PrintWriter} on which to
     *        write
     * @return the character position at which we left the cursor
     */
    private int printString(int pos, final boolean comma,
                            final PrintWriter output) {
        if (comma) {
            output.print(',');
            pos++;
        }

        // The length is one for a separating space, 2 for the quotes, and the
        // length of the string itself
        final int length = 3 + name.length();

        // Assume a line of 80 characters.  We also have to leave one for a
        // comma after this event name, for the next event.  Finally, (X)Emacs
        // wraps if you actually *reach* 80, so stay in 79 characters since we
        // are all big (X)Emacs fans here.
        if (pos + length > 78) {
            output.println();
            output.print("         ");
            pos = 10;
        }
        output.print(" \"");
        output.print(name);
        output.print('"');
        pos += length;
        final int size = children.size();
        for (int i = 0; i < size; i++) {
            pos = ((ProcessEventList)children.get(i)).printString(pos, true,
                                                                  output);
        }
        return pos;
    }

    /**
     * Write out the event propagator.  This is code that checks whether a
     * high-level event is enabled and, if so, enables all of its children as
     * well.
     *
     * @param output the {@link java.io.PrintWriter PrintWriter} on which to
     *        write
     */
    private void propagate(final PrintWriter output) {
        final int size = children.size();
        if (size > 0) {
            output.print("        if (print[");
            output.print(name);
            output.println("])");
            output.print("            ");
            for (int i = 0; i < size; i++) {
                final ProcessEventList child =
                    (ProcessEventList)children.get(i);
                output.print("print[");
                output.print(child.name);
                output.print("] = ");
            }
            output.println("true;");
            for (int i = 0; i < size; i++) {
                ((ProcessEventList)children.get(i)).propagate(output);
            }
        }
    }

    /**
     * Print out the code that will construct GUI elements, allowing the user
     * to choose debug events to print.
     *
     * @param output the {@link java.io.PrintWriter PrintWriter} on which to
     *        write
     */
    private void makeGUI(final PrintWriter output) {
        final int size = children.size();

        if (size > 0) {
            for (int i = 0; i < size; i++) {
                ((ProcessEventList)children.get(i)).makeGUI(output);
            }
            output.print("        final DebugEntry[] ");
            final String arrayName = name.toLowerCase();
            output.print(arrayName);
            output.print(" = new DebugEntry[");
            output.print(size);
            output.println("];");
            for (int i = 0; i < size; i++) {
                output.print("        ");
                output.print(arrayName);
                output.print('[');
                output.print(i);
                output.print("] = new DebugEntry(Event.");
                final ProcessEventList child =
                    (ProcessEventList)children.get(i);
                child.makeArray(output);
                output.println(");");
            }
        }
    }

    /**
     * Print out an array entry for this event
     *
     * @param output the {@link java.io.PrintWriter PrintWriter} on which to
     *        write
     */
    private void makeArray(final PrintWriter output) {
        output.print(name);
        output.print(", \"");
        output.print(name);
        output.print("\", ");
        if (children.size() > 0) {
            output.print(name.toLowerCase());
        } else {
            output.print("null");
        }
    }


    // ***********************************************************************
    // The static portion of the class.
    // This part contains the driving methods (including main).
    // ***********************************************************************

    /**
     * The top-level event
     */
    private static ProcessEventList topEvent;

    /**
     * The main entry point for the ProcessEventList program.  This program
     * reads in the event list and generates Event.java and DebugList.java.
     *
     * @param args the command line arguments (ignored)
     * @exception IOException if any files cannot be read or written
     */
    public static void main(final String[] args) throws IOException {
        writeEventClass(readEventList());
        //writeDebugListClass();  KK removed as no longer used
    }

    /**
     * Read in the event.list file and create a tree of ProcessEventList
     * objects to represent its contents.
     *
     * @return the number of events in event.list
     */
    private static int readEventList() throws IOException {
        final FileReader fileInput = new FileReader("event.list");
        final BufferedReader input = new BufferedReader(fileInput);

        // Read in content
        int numEvents = 0;
        final ArrayList levels = new ArrayList();  // Stack of current events
        String line;

        for (;;) {
            line = input.readLine();
            if (line == null) break;
            // Skip any blank lines
            if (line.length() == 0)
                continue;
            
            // Skip comment lines
            if (line.length() > 1 &&
                line.charAt(0) == '/' && line.charAt(1) == '/')
                continue;

            final int level = getLevel(line);
            for (int i = levels.size() - 1; i >= level; i--)
                levels.remove(i);
            final ProcessEventList event = new ProcessEventList(line.trim());
            levels.add(event);
            if (level > 0)
                ((ProcessEventList)levels.get(level - 1)).children.add(event);
            numEvents++;
        }

        input.close();
        topEvent = (ProcessEventList)levels.get(0);
        return numEvents;
    }

    /**
     * Determine the level of this event by counting the number of spaces at
     * the beginning of the string.  The level is half the number of spaces.
     *
     * @param s the string containing the event name
     * @return the level of this event name
     */
    private static int getLevel(final String s) {
        for (int i = 0; i < s.length(); i++)
            if (s.charAt(i) != ' ')
                return i / 2;
        return 0;
    }

    /**
     * Write out the Event.java file in the current package
     *
     * @param numEvents the number of events to represent
     * @exception IOException if the file cannot be opened for writing
     */
    private static void writeEventClass(final int numEvents)
        throws IOException {

        final FileWriter fileOutput = new FileWriter("Event.java");
        final BufferedWriter bufOutput = new BufferedWriter(fileOutput);
        final PrintWriter output = new PrintWriter(bufOutput);

        output.println("/**************************************************************************");
        output.println(" * THIS IS A GENERATED FILE!  DO NOT EDIT IT!");
        output.println(" * Instead, edit event.list and regenerate this file with ProcessEventList.");
        output.println(" *************************************************************************/");
        output.println();
        output.println("package mokan.debug;");
        output.println();
        output.println("/**");
        output.println(" * A list of event types of interest in the Kan system, together with a hook");
        output.println(" * for deciding when to take action on an event.");
        output.println(" */");
        output.println("public final class Event {");
        output.println("    public static boolean threads = true;");
        output.println();
        output.println("    // The event list");

        // Do a depth-first walk of the event tree to print the types
        topEvent.printEvent(0, output);

        output.println();
        output.println("    /**");
        output.println("     * The list of event types to report to the user.  A true element means to");
        output.println("     * print that event");
        output.println("     */");
        output.print("    private static final boolean[] print = new boolean[");
        output.print(numEvents);
        output.println("];");
        output.println();
        output.println("    /**");
        output.println("     * The list of all event names");
        output.println("     */");
        output.println("    private static final String[] eventNames =");
        output.print("        {");

        // Do a depth-first walk of the event tree to print the names.
        // The literal 10 refers to the (1-based) character position at which
        // to start printing ... i.e., one more than the length of the string
        // in the last output.print() above
        topEvent.printString(10, false, output);

        output.println();
        output.println("        };");
        output.println();
        output.println("    /**");
        output.println("     * Transform a list of event names into an event list suitable for");
        output.println("     * {@link #setEventMask(int[]) setEventMask}");
        output.println("     *");
        output.println("     * @param names an array of strings containing event names");
        output.println("     * @return an event list corresponding to the names");
        output.println("     */");
        output.println("    public static int[] namesToEvents(final String[] names) {");
        output.println("        final int[] mask = new int[names.length];");
        output.println("        for (int i = 0; i < names.length; i++) {");
        output.println("            mask[i] = -1;");
        output.println();
        output.println("            // FIXME: Linear search ... yuck!");
        output.println("            for (int j = 0; j < eventNames.length; j++) {");
        output.println("                if (names[i].equalsIgnoreCase(eventNames[j])) {");
        output.println("                    mask[i] = j;");
        output.println("                    break;");
        output.println("                }");
        output.println("            }");
        output.println("        }");
        output.println("        return mask;");
        output.println("    }");
        output.println();
        output.println("    /**");
        output.println("     * Set the events that will be reported to the user.");
        output.println("     *");
        output.println("     * @param mask an array of event types (see list above) to report to the");
        output.println("     *        user");
        output.println("     */");
        output.println("    public static void setEventMask(final int[] mask) {");
        output.println("        // Clear out the old mask");
        output.println("        for (int i = 0; i < print.length; i++)");
        output.println("            print[i] = false;");
        output.println();
        output.println("        // Set the events in mask");
        output.println("        if (mask == null) return;");
        output.println("        for (int i = 0; i < mask.length; i++)");
        output.println("            if (mask[i] >= 0 && mask[i] < print.length)");
        output.println("                print[mask[i]] = true;");
        output.println();
        output.println("        // Propagate parent events down to the children");

        // Do a depth-first walk of the tree to print the propagator
        topEvent.propagate(output);

        output.println("    }");

        output.println();
        output.println("    /**");
        output.println("     * Print an error message");
        output.println("     *");
        output.println("     * @param type the type of event for this debugging message");
        output.println("     * @param message the message to print for this event");
        output.println("     */");
        output.println("    public static void ERROR(final String message) {");
        output.println("        if (threads) System.err.print(Thread.currentThread().toString()+':');");
        output.println("        System.err.println(message);");
        output.println("    }");
        output.println();

        output.println();
        output.println("    /**");
        output.println("     * Print a debugging message, if its type has been enabled for printing");
        output.println("     *");
        output.println("     * @param type the type of event for this debugging message");
        output.println("     * @param message the message to print for this event");
        output.println("     */");
        output.println("    public static void DEBUG(final int type, final String message) {");
        output.println("        if (print[type]){");
        output.println("            if (threads) System.err.print(Thread.currentThread().toString()+':');");
        output.println("            System.err.println(message);");
        output.println("    }}");
        output.println();
        output.println("    /**");
        output.println("     * Print a debugging message, if its type has been enabled for printing.");
        output.println("     * The purpose of this method is to avoid calling <code>toString()</code>");
        output.println("     * on the object parameter unless the event type is selected.");
        output.println("     *");
        output.println("     * @param type the type of event for this debugging message");
        output.println("     * @param message the message to print for this event");
        output.println("     * @param obj an object whose representation should also be printed");
        output.println("     */");
        output.println("    public static void DEBUG(final int type, final String message,");
        output.println("        final Object obj) {");
        output.println();
        output.println("        if (print[type]) {");
        output.println("            if (threads) System.err.print(Thread.currentThread().toString()+':');");
        output.println("            System.err.println(message + obj);");
        output.println("        }");
        output.println("    }");
        output.println();
        output.println("    /**");
        output.println("     * Print a complex debugging message, if its type has been enabled for");
        output.println("     * printing");
        output.println("     *");
        output.println("     * @param type the type of event for this debugging message");
        output.println("     * @param objs objects to print for this event");
        output.println("     */");
        output.println("    public static void DEBUG(final int type, final Object[] objs) {");
        output.println("        if (print[type]) {");
        output.println("            StringBuffer buf = new StringBuffer();");
        output.println("            for (int i = 0; i < objs.length; i++) {");
        output.println("                buf.append(objs[i].toString());");
        output.println("                buf.append(' ');");
        output.println("            }");
        output.println("            if (threads) System.err.print(Thread.currentThread().toString()+':');");
        output.println("            System.err.println(buf.toString());");
        output.println("        }");
        output.println("    }");
        output.println("}");

        output.close();
    }

    /**
     * Write out the DebugList.java file in the mokan.ui package
     *
     * @exception IOException if the file cannot be opened for writing
     */
    private static void writeDebugListClass() throws IOException {

        final FileWriter fileOutput = new FileWriter("../ui/DebugList.java");
        final BufferedWriter bufOutput = new BufferedWriter(fileOutput);
        final PrintWriter output = new PrintWriter(bufOutput);

        output.println("/**************************************************************************");
        output.println(" * THIS IS A GENERATED FILE!  DO NOT EDIT IT!");
        output.println(" * Instead, edit kan/debug/event.list and regenerate this file");
        output.println(" * with mokan.debug.ProcessEventList.");
        output.println(" *************************************************************************/");
        output.println();
        output.println("package mokan.ui;");
        output.println();
        output.println("import javax.swing.JTree;");
        output.println("import javax.swing.tree.*;");
        output.println("import mokan.debug.Event;");
        output.println();
        output.println("/**");
        output.println(" * A graphical view of the debug events a user can choose to view");
        output.println(" */");
        output.println("final class DebugList extends JTree {");
        output.println();
        output.println("    /**");
        output.println("     * Create a new <code>DebugList</code> for displaying a choice of debug");
        output.println("     * events to the user");
        output.println("     */");
        output.println("    DebugList() {");
        output.println("        super(createNodes());");
        output.println("        getSelectionModel().setSelectionMode");
        output.println("            (TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);");
        output.println("    }");
        output.println();
        output.println("    /**");
        output.println("     * Retrieve a list of the events chosen by the user");
        output.println("     *");
        output.println("     * @return an array containing all of the user-selected events");
        output.println("     */");
        output.println("    int[] getList() {");
        output.println("        final TreePath[] paths = getSelectionPaths();");
        output.println("        if (paths == null)");
        output.println("            return new int[0];");
        output.println("        final int[] list = new int[paths.length];");
        output.println("        for (int i = 0; i < paths.length; i++) {");
        output.println("            final DebugEntry entry =");
        output.println("                (DebugEntry)paths[i].getLastPathComponent();");
        output.println("            list[i] = entry.getValue();");
        output.println("        }");
        output.println("        return list;");
        output.println("    }");
        output.println();
        output.println("    /**");
        output.println("     * Create the tree of <code>DebugEntry</code> objects from which the user");
        output.println("     * can choose");
        output.println("     *");
        output.println("     * @return the root of the created tree");
        output.println("     */");
        output.println("    private static DebugEntry createNodes() {");
	
        // Do a depth-first walk of the tree to print the GUI creators
        topEvent.makeGUI(output);

        output.print("        return new DebugEntry(Event.");
        output.print(topEvent.name);
        output.print(", \"");
        output.print(topEvent.name);
        output.print("\", ");
        output.print(topEvent.name.toLowerCase());
        output.println(");");
        output.println("    }");
        output.println("}");

        output.close();
    }
}

/* arch-tag: 430a7f62-2440-4ed7-9791-34d694649019 */
