package mokan.nameservice;

import java.util.*;
import mokan.comm.*;
import mokan.obj.*;
import java.lang.reflect.*;
import mokan.app.*;
import mokan.transact.*;
import mokan.*;
import mokan.thread.*;

/**
 * This class implements an arrow directory protocol. For each object, there is
 * a forwarding address on each node. The forwarding addresses, or arrows,
 * are linked one after another and finally point to the valid home node of
 * the object. Thus they form a (minimum) spanning tree.
 * When the home node moves, the arrows change respectively.
 * If a node is about to leave the network, either expected or unexpected,
     * the arrows should be modified such that the spanning tree is not disconnected.
 *
 * A forwarding address at a node is generated when an ObjectID is received at
 * the first time.
 */

public class ArrowNameService {

  /**
       * The local reference owner map. The entry is (ObjectID, LogicalNode<owner>).
   * Whenever a node N0 receives an ObjectID from a node N1, N0 has N1 as the
   * owner of the ObjectID.
   */
  private HashMap ownerMap = new HashMap();

  /**
   * The local cached object home map. The entry is <ObjectID, LogicalNode>.
   */
  private HashMap homeMap = new HashMap();

  /**
   *
   */
  public ArrowNameService() {
  }

  /**
   * Put the owner of an ObjectID. May overwrite an existing entry.
   * @param oid The object ID.
   * @param owner The node that the ObjectID comes from.
   */
  public void putOwner(ObjectID oid, LogicalNode owner) {
    System.out.println("putOwner: " + oid + "(" + oid.getClass().getName() +
                       "), " + owner.getPhysicalNode().getUniqueID());
    ownerMap.put(oid, owner);
  }

  /**
   * Add an ownership entry. May NOT overwrite an existing entry.
   * @param oid The object ID.
   * @param owner The node that the ObjectID comes from.
   */
  public void addOwner(ObjectID oid, LogicalNode owner) {
    if (!ownerMap.containsKey(oid)) {
      //System.out.println("addOwner: " + oid + "(" + oid.getClass().getName() +
      //                   "), " + owner.getPhysicalNode().getUniqueID());
      ownerMap.put(oid, owner);
      //System.out.println("ownerMap: " + ownerMap);
    }
  }

  /**
   * Remove an owner
   * @param oid
   * @return
   */
  public LogicalNode removeOwner(ObjectID oid) {
    return (LogicalNode) ownerMap.remove(oid);
  }

  /**
   * Get the owner of an ObjectID
   * @param oid
   * @return
   */
  public LogicalNode getOwner(ObjectID oid) {
    LogicalNode owner = (LogicalNode) ownerMap.get(oid);
    return owner;
  }

  /**
   * Put the home node of an ObjectID. May overwrite an existing entry.
   * @param oid The object ID.
   * @param home The home of the object.
   */
  public void putHome(ObjectID oid, LogicalNode home) {
    homeMap.put(oid, home);
  }

  /**
   * Look up the home node of an object from local. The home node may be out
   * of date.
   * @param oid The object ID.
   * @return The home node of the object.
   */
  public LogicalNode lookupHome(ObjectID oid) {
    LogicalNode home;
    if (homeMap.containsKey(oid)) {
      home = (LogicalNode) homeMap.get(oid);
    }
    // The original home node
    else {
      home = oid.getNode();
    }
    return home;
  }

  /**
   * Discover the home node of an object through the network when the local
   * cached home node is out of date.
   * @param oid The object ID.
   * @return The home node of the object.
   */
  public LogicalNode discoverHome(ObjectID oid) {
    // send the query message
    LogicalNode localNode = LogicalNode.getMyNode();
    QueryHomeMsg msg = new QueryHomeMsg(oid, localNode);
    LogicalNode owner = (LogicalNode) ownerMap.get(oid);
    if (owner == null) {
      throw new NameServiceException("Unkown owner of " + oid);
    }

    //System.out.println("Discover home: the owner of " + oid + "(" +
    //                   oid.getClass().getName() + ")" + " is (" +
    //                   owner.getPhysicalNode().getUniqueID() + ")");
    // wait for the reply from the home node
    QueryHomeReply reply = (QueryHomeReply) msg.sendAndWait(owner);
    LogicalNode home = reply.getHome();
    putHome(oid, home);
    //System.out.println("Discover home: the home of " + oid + "(" +
    //                   oid.getClass().getName() + ")" + " is (" +
    //                   home.getPhysicalNode().getUniqueID() + ")");
    return home;
  }

  /**
   * Actions taken When a node is going to leave.
   * This method will:
   * 1. Transfer objects homed at this node to the new home.
   * 2. Restructure owner maps.
   *
   * @param newHome The new home node
   */
  public void onLeave(LogicalNode newHome) {

    // Transfer objects homed at this node to the new home
    LogicalNode local = LogicalNode.getMyNode();
    Iterator it = ownerMap.keySet().iterator();
    while (it.hasNext()) {
      ObjectID oid = (ObjectID) it.next();
      LogicalNode owner = (LogicalNode) ownerMap.get(oid);

      // Test if the ObjectID is homed at this logical node
      if (local.equals(owner)) {
        if (!oid.isLocalValid()) {
          continue;
        }
        KanObject kObj = oid.getLocKanObject();
        Protocol protocol = kObj.getProtocol();
        // Check if the object is running on a home protocol
        if (protocol instanceof GeneralProtocol
            || protocol instanceof MigratoryHomeProtocol
            || protocol instanceof ReplicaHomeProtocol
            || protocol instanceof OptimisticHomeProtocol) {

          // Transfer the object
          transferObject(oid, newHome);

          // Update the owner map
          ownerMap.put(oid, newHome);

        }
      }
    }

    // Restructure the ownership of the directory protocol.
    // Send the owner map to neighbors and let them update their
    // owner maps respectively.
    Iterator it1 = ownerMap.keySet().iterator();
    while (it1.hasNext()) {
      ObjectID oid = (ObjectID) it1.next();
      LogicalNode owner = (LogicalNode) ownerMap.get(oid);
    }

    OwnerMapMsg msg = new OwnerMapMsg(local, ownerMap);
    msg.send(Globals.neighbors);

  }

  /**
   * Update the local owner map upon receiving an OwnerMapMsg from a node.
   * @param sender
   * @param ownerMap1
   */
  void updateOwnerMap(LogicalNode sender, HashMap ownerMap1) {

    Iterator it1 = ownerMap1.keySet().iterator();
    while (it1.hasNext()) {
      ObjectID oid = (ObjectID) it1.next();
      LogicalNode owner = (LogicalNode) ownerMap1.get(oid);
    }
    // If an ObjectID's owner is equal to the sender, then update it to
    // the owner of the sender's
    Iterator it = ownerMap.keySet().iterator();
    while (it.hasNext()) {
      ObjectID oid = (ObjectID) it.next();
      LogicalNode owner = (LogicalNode) ownerMap.get(oid);
      if (sender.equals(owner)) {
        LogicalNode newOwner = (LogicalNode) ownerMap1.get(oid);
        ownerMap.put(oid, newOwner);
        System.out.println("The owner of " + oid.getClass().getName() +
                           " changed from (" +
                           owner.getPhysicalNode().getUniqueID() + ") to (" +
                           newOwner.getPhysicalNode().getUniqueID() + ")");
      }
    }
  }

  /**
   * Transfer an object to the new home node.
   * @param oid
   * @param newHome
   */
  private void transferObject(ObjectID oid, LogicalNode newHome) {

    // Get the object information
    LogicalNode local = LogicalNode.getMyNode();
    GlobalObj objCopy = oid.getRealObj();

    KanObject kObj = oid.getLocKanObject();
    Protocol protocol = kObj.getProtocol();

    // Lock the object locally
    TransactionID tranID = new TransactionID(local);
    LockRequest lockReq = new LockRequest(tranID, local, null, oid, false, false);
    protocol.lockLocally(lockReq);

    // Invalidate replicas
    InvalidateMsg invalidateMsg = new InvalidateMsg(local, oid);
    HashSet replicas = kObj.getProtocol().getCachedNodes();
    invalidateMsg.send(replicas);

    // Transfer the object to the new home
    TransferObjMsg transferMsg = new TransferObjMsg(local, oid,
        objCopy, Globals.app.getNodes(), protocol.getType());
    transferMsg.send(newHome);

    // Abort nonpreemptible waiting lock requests
    protocol.abortWaiters();

  }

}