package mokan.nameservice;

import mokan.comm.*;
import mokan.debug.*;
import java.io.*;

import mokan.obj.*;

/**
 * This message is used to invalidate object replicas after transferring an
 * object to a new home node.
 */

final class InvalidateMsg extends AppMessage implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -7173592977979302587L;

    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
        ObjectStreamClass.NO_FIELDS;

    /**
     * The <code>ObjectID</code> for the copied object.  This is a
     * pseudo-final field.
     *
     * @serial the <code>ObjectID</code> for the copied object
     */
    private ObjectID objID;

    public InvalidateMsg() {
        super(true, false);
    }

    /**
     * Create a new <code>InvalidateMsg</code>
     *
     * @param from the originator of this message
     * @param objID the <code>ObjectID</code> of the object to replicate
     * @param objCopy a copy of the object
     * @param version the version number of this object
     * @param rqst the request that prompted this object move
     * @param homeify <code>true</code> if the object should switch to general
     *        type on arrival
     * @param replica the node holding a replica, if <var>homeify</var> is
     *        true
     */
    public InvalidateMsg(final LogicalNode from, final ObjectID objID) {
        super(from, true, false);
        this.objID = objID;
        Event.DEBUG(Event.MSG_CREATE, "Created ", this);
    }

    /**
     * Set this node to the new home node and create a new kan object with
     * the general protocol.
     */
    public void handle() {
        super.handle();
        objID.getLocKanObject();
        objID.setRealObject(null);
    }

    /**
     * Convert the <code>InvalidateMsg</code> to a string
     *
     * @return a string representation of this <code>InvalidateMsg</code>
     */
    public String toString() {
        return "InvalidateMsg(" + objID + "): ";
    }

    /**
     * Save the state of the <code>InvalidateMsg</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The superclass data is written out, then the ID of the
     *             object to move (ObjectID), then the object itself
     *             (GlobalObj), then the version number of the object (long),
     *             then the request that prompted the move (AppMessage),
     *             followed by an indication of whether this object should
     *             switch to the general type on arrival (boolean), and if
     *             true, followed by a logical node holding a replica.
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeObject(objID);
    }

    public void readExternal(final ObjectInput in)
        throws IOException, ClassNotFoundException {

        super.readExternal(in);
        objID = (ObjectID)in.readObject();

    }
}
