package mokan.nameservice;

import mokan.comm.*;
import mokan.debug.*;
import java.io.*;
import java.util.*;

import mokan.obj.*;

/**
 * This message is used for transferring an object to a new home node.
 */

final class TransferObjMsg extends AppMessage implements Externalizable {

  /**
   * The <var>serialVersionUID</var> for this class.  If you change the
   * signature of this class (field names or types, method names, parameter
   * names or types, adding or removing fields or methods, etc.), you
   * <strong>must</strong> change this value.  The recommended way is to use
   * the <code>serialver</code> tool.
   */
  static final long serialVersionUID = -6593232420054470735L;

  /**
   * The <code>ObjectID</code> for the copied object.  This is a
   * pseudo-final field.
   *
   * @serial the <code>ObjectID</code> for the copied object
   */
  private ObjectID objID;

  /**
   * The object itself.  This is a pseudo-final field.
   *
   * @serial the object itself
   */
  private GlobalObj objCopy;

  private HashSet appNodes;

  private int protocolType;

  /**
   * The no-args constructor, needed for deserialization
   */
  public TransferObjMsg() {
    super(true, false);
  }

  /**
   * Create a new <code>TransferObjMsg</code>
   *
   * @param from the originator of this message
   * @param objID the <code>ObjectID</code> of the object to replicate
   * @param objCopy a copy of the object
   * @param version the version number of this object
   * @param rqst the request that prompted this object move
   * @param homeify <code>true</code> if the object should switch to general
   *        type on arrival
   * @param replica the node holding a replica, if <var>homeify</var> is
   *        true
   */
  public TransferObjMsg(final LogicalNode from, final ObjectID objID,
                        final GlobalObj objCopy, HashSet appNodes,
                        int _protocolType) {
    super(from, true, false);
    this.objID = objID;
    this.objCopy = objCopy;
    this.appNodes = appNodes;
    protocolType = _protocolType;

    Event.DEBUG(Event.MSG_CREATE, "Created ", this);
  }

  /**
   * Set this node to the new home node and create a new kan object with
   * the general protocol.
   */
  public void handle() {
    super.handle();
    System.out.println("Received TransferObjMsg with object " +
                       objID.getClass().getName());
    LogicalNode local = LogicalNode.getMyNode();
    Globals.nameService.putHome(objID, local);
    Globals.nameService.putOwner(objID, local);
    objID.setRealObject(null);
    KanObject kanObj;
    if (protocolType == Protocol.OPTIMISTIC_HOME) {
      kanObj = new KanObject(objID, objCopy, Protocol.OPTIMISTIC_HOME,
                             local.getLocalHandler().manager,
                             0L);
    } else {
      kanObj = new KanObject(objID, objCopy, Protocol.GENERAL,
                             local.getLocalHandler().manager, 0L);
    }

    //new KanObject(objID, objCopy, Protocol.GENERAL);

    Globals.app.getNodes().addAll(appNodes);
    Globals.app.getNodes().remove(from);
  }

  /**
   * Convert the <code>TransferObjMsg</code> to a string
   *
   * @return a string representation of this <code>TransferObjMsg</code>
   */
  public String toString() {
    return "TransferObjMsg(" + objID + ")";
  }

  /**
   * Save the state of the <code>TransferObjMsg</code> instance to a stream
   * (i.e., serialize it).
   *
   * @serialData The superclass data is written out, then the ID of the
   *             object to move (ObjectID), then the object itself
   *             (GlobalObj), then the version number of the object (long),
   *             then the request that prompted the move (AppMessage),
   *             followed by an indication of whether this object should
   *             switch to the general type on arrival (boolean), and if
   *             true, followed by a logical node holding a replica.
   */
  public void writeExternal(final ObjectOutput out) throws IOException {
    super.writeExternal(out);
    out.writeObject(objID);
    out.writeObject(objCopy);
    out.writeObject(appNodes);
    out.writeInt(protocolType);
  }

  /**
   * Read the state of this message from external.
   * @param in
   * @throws IOException
   * @throws java.lang.ClassNotFoundException
   */
  public void readExternal(final ObjectInput in) throws IOException,
      ClassNotFoundException {
    super.readExternal(in);
    objID = (ObjectID) in.readObject();
    objCopy = (GlobalObj) in.readObject();
    appNodes = (HashSet) in.readObject();
    protocolType = in.readInt();
  }
}
