/*
 * @(#)$Id: CreateObjReply.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.obj;

import java.io.*;
import mokan.comm.AppReply;
import mokan.debug.Event;

import mokan.nameservice.*;

/**
 * Reply to a <code>CreateObjMsg</code> with the
 * {@linkmokan.obj.ObjectID ObjectID} of the newly created object, or an
 * exception if something went wrong.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Sean Brydon
 * @author	Jerry James
 */
final class CreateObjReply extends AppReply implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -5958214599928196745L;

    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /**
     * The <code>ObjectID</code> of the created object.  This is a
     * pseudo-final field.
     *
     * @serial the <code>ObjectID</code> of the created object
     */
    ObjectID objID;

    /**
     * The exception thrown while trying to create the object.  This is a
     * pseudo-final field.
     *
     * @serial the exception thrown while trying to create the object
     */
    Exception ex;

    /**
     * The no-args constructor, needed for deserialization
     */
    public CreateObjReply() {
	super(false, false);
    }

    /**
     * Constructs a new <code>CreateObjReply</code> from the specified
     * {@linkmokan.obj.CreateObjMsg CreateObjMsg} to report the creation of the
     * object identified by <var>objID</var>
     *
     * @param msg the message to which this is a reply
     * @param objID the <code>ObjectID</code> of the new object
     */
    CreateObjReply(final CreateObjMsg msg, final ObjectID objID) {
	super(msg, false, false);
	this.objID = objID;
	ex = null;
	Event.DEBUG(Event.MSG_CREATE, "Created ", this);
    }

    /**
     * Constructs a new <code>CreateObjReply</code> from the specified
     * {@linkmokan.obj.CreateObjMsg CreateObjMsg} to report an exception when
     * trying to create an object
     *
     * @param msg the message to which this is a reply
     * @param ex the exception thrown while trying to create the object
     */
    CreateObjReply(final CreateObjMsg msg, final Exception ex) {
	super(msg, false, false);
	objID = null;
	this.ex = ex;
	Event.DEBUG(Event.MSG_CREATE, "Created ", this);
    }

    /**
     * Convert the <code>CreateObjReply</code> to a string
     */
    public String toString() {
	return "CreateObjReply(" +
	    (objID != null ? objID.toString() : ex.toString()) + ')';
    }

    /**
     * Save the state of the <code>CreateObjReply</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The superclass data is written out, then the ID of the
     *             newly created object (ObjectID), followed by an exception
     *             thrown by the constructor (Exception).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
	super.writeExternal(out);
	// REVIEW: Possible optimization
	// Exactly one of objID and ex is non-null
	out.writeObject(objID);
	out.writeObject(ex);
    }

    public void readExternal(final ObjectInput in)
	throws IOException, ClassNotFoundException {

	super.readExternal(in);
	objID = (ObjectID)in.readObject();
	ex = (Exception)in.readObject();
    }
}
