/*
 * @(#)GeneralProtocol.java GeneralProtocol.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.obj;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import mokan.app.Application;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.debug.Event;

import mokan.transact.LockRequest;
import mokan.transact.IsolationRequest;
import mokan.transact.LockMsg;
import mokan.transact.UnlockMsg;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;
import mokan.util.IntIntMap;

import mokan.nameservice.*;

/**
 * A protocol for managing general (replicated) objects on the home node
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 */
final public class GeneralProtocol extends Protocol {

    /**
     * The set of nodes caching this <code>KanObject</code>
     */
    private final HashSet cachedNodes = new HashSet();

    /**
     * Count the number of writes from each node.  The indexes are those
     * assigned to {@linkmokan.comm.LogicalNode LogicalNodes}.
     */
    private final IntIntMap writes;

    /**
     * Count the number of reads from each node.  The indexes are those
     * assigned to {@linkmokan.comm.LogicalNode LogicalNodes}.
     */
    private final IntIntMap reads;

    /**
     * The number of times this object has been accessed (read or written)
     * during the current replication strategy interval
     */
    private int accessCount = 0;

    /**
     * Create a new <code>GeneralProtocol</code> object
     *
     * @param kObj the {@linkmokan.obj.KanObject KanObject} to manage
     * @param nodes the number of nodes in this application
     */
    GeneralProtocol(final KanObject kObj, final int nodes) {
        super(kObj);
        writes = new IntIntMap();
        reads = new IntIntMap();
    }

    /**
     * Convert to a general (replicating) protocol from some other protocol
     *
     * @param prot the old protocol object
     * @param nodes the number of nodes in this application
     */
    GeneralProtocol(final Protocol prot, final int nodes) {
        this(prot, nodes, null);
    }

    /**
     * Convert to a general (replicating) protocol from some other protocol
     *
     * @param prot the old protocol object
     * @param nodes the number of nodes in this application
     * @param replica an initial replica
     */
    GeneralProtocol(final Protocol prot, final int nodes,
                    final LogicalNode replica) {
        super(prot);
        writes = new IntIntMap();
        reads = new IntIntMap();
        if (replica != null)
            cachedNodes.add(replica);
    }

    /**
     * Count an access to this object, for purposes of determining the
     * replication scheme
     *
     * @param local the local logical node
     * @param request the lock request for the current node
     * @param accessor the logical node initiating the access
     * @param write <code>true</code> if the access is a write
     */
    private void countAccess(final LogicalNode local,
                             final LockRequest request,
                             final boolean write) {
        LogicalNode accessor = request.locker.getAccessor();
        int accessorIndex = accessor.hashCode();
        if (write) {
            // Count reads by accessing node.  Technically, this should be
            // synchronized, but we can stand a small amount of inaccuracy
            writes.put (accessorIndex, writes.get(accessorIndex)+1);
            Event.DEBUG(Event.OBJECT_WRITE, "Write by " + accessor + " of ",
                        kanObj);
        } else {
            // Count reads by accessing node.  Technically, this should be
            // synchronized, but we can stand a small amount of inaccuracy
            reads.put (accessorIndex, reads.get(accessorIndex)+1);
            Event.DEBUG(Event.OBJECT_READ, "Read by " + accessor + " of ",
                        kanObj);
        }

        // Reexamine the replication scheme if enough accesses have taken
        // place.  The increment of <var>accessCount</var> technically should
        // be synchronized, but it won't kill us if an interleaving happens.
        if (++accessCount == REPL_INTERVAL) {
            final Application app = local.getApplication(); // The application
            final Collection nodes = app.getNodes();	// All the nodes
            //final int localIndex = local.hashCode();

            // Should anybody else get a replica?  First, count the total
            // number of writes
            int writeCount = 0;
            for (Iterator i = writes.entryIterator(); i.hasNext(); ) {
                writeCount += ((IntIntMap.Entry)i.next()).value;
            }
            // Now compute the replication formula
            for (Iterator i = nodes.iterator(); i.hasNext(); ) {
                accessor = (LogicalNode)i.next();
                int hash = accessor.hashCode();

                Event.DEBUG(Event.OBJECT_REPLICATE, "GP.countAccess " + accessor + " local:" + local);

                // Can only cache if not home node, not caching already, and
                // sufficient reads from that node
                if (accessor != local && !cachedNodes.contains(accessor) &&
                    3 * reads.get(hash) > writeCount - writes.get(hash)) {

                    Event.DEBUG(Event.OBJECT_REPLICATE, "Sending replica of "
                                + kanObj + " to " + accessor);
                    new ReplicaMsg(kanObj.objID, kanObj.obj,
                                   kanObj.version,
                                   Protocol.REPLICA).send(accessor);
                    cachedNodes.add(accessor);
                    // TODO: Make it so that the unlock won't have to go
                    // unlock all those new replicas!
                }
            }

            // Should we switch to a migratory protocol?
            // First, can't be more than 1 replica
            if (cachedNodes.size() <= 1) {
                // Second, no more than one node accessed during the period
                LogicalNode owner = null;
                int count = 0;

                for (Iterator i = nodes.iterator(); i.hasNext(); ) {
                    accessor = (LogicalNode)i.next();
                    int hash = accessor.hashCode();
                    if (writes.get(hash) > 0 || reads.get(hash) > 0) {
                        //found = (found == -1) ? hash : -2;
                        owner = accessor;
                        count++;
                        System.out.println ("writes="+writes.get(hash)+
                                            " reads="+reads.get(hash));
                    }
                }
                //KK  What's all this???

                System.out.println ("count="+count+
                                    " owner="+owner+
                                    " callers="+kanObj.callers);
                synchronized (kanObj) {
                    if ((count == 0 ||
                         owner == kanObj.myMgr.parentNode)
                        && cachedNodes.size() == 0) {
                        Event.DEBUG(Event.OBJECT_TYPE_CHANGE,
                                    "Change from general to migratory: ",
                                    kanObj);
                        kanObj.protocol = new MigratoryHomeProtocol(this);
                    } else if (count == 1 && kanObj.callers == 0) {
                        Event.DEBUG(Event.OBJECT_TYPE_CHANGE,
                                    "Change from general to migratory: ",
                                    kanObj);
                        //final LogicalNode owner = nodes[found];
                        new MoveObjMsg(local, kanObj.objID, kanObj.obj,
                                       kanObj.version, null,
                                       false, null).send(owner);
                        kanObj.uninstall();
                        kanObj.protocol =
                            new MigratoryHomeProtocol(this, owner);
                    }
                }
            }

            // Reset the counters
            writes.clear();
            reads.clear();
            accessCount = 0;
            Event.DEBUG(Event.OBJECT_ACCESS,
                        "Access counters reset to zero for ", kanObj);
        }
    }

    /**
     * Return a copy of the <code>GlobalObj</code> managed by this
     * <code>Protocol</code>, suitable for sending over the network while the
     * original object is still in use.
     *
     * @param target the target of the copy
     * @return a consistent copy of the <code>GlobalObj</code> managed by this
     *         <code>KanObject</code>
     */
    protected GlobalObjCopy copy(final LogicalNode target) {
        try {
            IsolationRequest lr = lock.lockForCopy();

            final GlobalObjCopy copy =
                new GlobalObjCopy(kanObj.getObject().kanClone(),
                                  kanObj.version,
                                  Protocol.REPLICA);
            lock.unlockForCopy(lr);
            cachedNodes.add(target);
            return copy;
        } catch (Exception intEx) {
            // This can't happen
            intEx.printStackTrace();
            return null;
        }
    }

    /**
     * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
     * holds the parameters of the lock.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for specifying when a lock must be
     * acquired globally.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    protected void lock(final LockRequest request)
        throws TransactionAbortException {

        Event.DEBUG(Event.LOCK, "GeneralProtocol.lock: "+request);

        lock.lock(request);
        // Nonpreemptible write locks on home also lock all replicas
        if (!request.isReadOnly() && !request.isPreemptible() && !cachedNodes.isEmpty()) {
            Event.DEBUG(Event.LOCK,"requesting  replica locks");

            LogicalNode home = kanObj.myMgr.parentNode;
            request.requester = home;
            LockMsg msg = new LockMsg(home, request);

            if (!msg.sendAndGetConsensus(cachedNodes)) {
                throw new TransactionAbortException
                    ("Couldn't lock all replicas");
            }

        }
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that originally locked the object.
     * This is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install)
        throws TransactionException {

        unlock(lock.lookup(transID), install, null);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.  This
     * is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install)
        throws TransactionException {

        unlock(request, install, null);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>TransactionID</code> object identifies the transaction that
     * originally locked the object.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install,
                          final GlobalObj gObj) throws TransactionException {
        unlock(lock.lookup(transID), install, gObj);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.  This is essentially just a wrapper around the
     * <code>Lock</code> object associated with this <code>KanObject</code>,
     * with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install,
                          final GlobalObj gObj) throws TransactionException {
        final LogicalNode from = kanObj.myMgr.parentNode;
        if (request.isReadOnly())
            countAccess(from, request, false);

        if (request.isReadOnly() || request.isPreemptible()) {
            lock.unlock(request, install, gObj);
        } else {
            // Count nonpreemptible write accesses on the home node
            countAccess(from, request, true);

            // Nonpreemptible write locks must unlock all replicas
            boolean reply = true;
            if (!cachedNodes.isEmpty()) {
                GlobalObj sendObj = (gObj != null)
                    ? gObj : (install ? kanObj.obj : null);
                UnlockMsg msg = new UnlockMsg(from, request.locker, install,
                                              kanObj.objID, sendObj);
                reply = msg.sendAndGetConsensus(cachedNodes);
                Event.DEBUG(Event.UNLOCK, "Updated and unlocked all replicas: "
                            + reply);
            }
            lock.unlock(request, install, gObj);
            if (!reply)
                throw new TransactionAbortException("Abort on replica");
            // NOTE: We weren't preemptible, so this is BAD
        }
    }


    /**
     * Inform one node of the disappearnce of a replica from another.
     *
     * @param site the site of the dropped replica
     */
    public void dropReplica(final LogicalNode site) {
        Event.DEBUG(Event.GENERAL_OBJECT, "GeneralProtocol.dropReplica", site);
        cachedNodes.remove(site);
    }

    /**
     *  Drop the local copy of the object.
     */
    protected void dropLocal () {
        //cachedNodes.remove(site);
    }


    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    protected boolean executeHere(final LogicalNode from, final LogicalNode to,
                                  final AppMessage msg) {
        return true;
    }

    /**
     * Take any actions that have been pending while waiting for all threads
     * to stop using the object
     */
    protected void noThreads() { }

    public HashSet getCachedNodes() {
      return cachedNodes;
    }

    public void abortLockRequests() {

    }

    //Added by Huahai
    public int getType() {
      return GENERAL;
    }
}
