/*
 * @(#)ImmobileProtocol.java ImmobileProtocol.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.obj;

import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.transact.LockRequest;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;

/**
 * A protocol for managing immobile objects
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 */
final class ImmobileProtocol extends Protocol {

    /**
     * Create a new <code>ImmobileProtocol</code> object
     *
     * @param kObj the <code>KanObject</code> to manage
     */
    ImmobileProtocol(final KanObject kObj) {
	super(kObj);
    }

    /**
     * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
     * holds the parameters of the lock.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for specifying when a lock must be
     * acquired globally.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    protected void lock(final LockRequest request)
	throws TransactionAbortException {

	lock.lock(request);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that originally locked the object.
     * This is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install)
	throws TransactionException {

	lock.unlock(lock.lookup(transID), install, null);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.  This
     * is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install)
	throws TransactionException {

	lock.unlock(request, install, null);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>TransactionID</code> object identifies the transaction that
     * originally locked the object.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install,
			  final GlobalObj gObj)	throws TransactionException {
	lock.unlock(lock.lookup(transID), install, gObj);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.  This is essentially just a wrapper around the
     * <code>Lock</code> object associated with this <code>KanObject</code>,
     * with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install,
			  final GlobalObj gObj)	throws TransactionException {

	lock.unlock(request, install, gObj);
    }

    /**
     * Inform one node of the disappearance of a replica from another.  This
     * is a no-op for replicas.
     *
     * @param site the site of the dropped replica
     */
    protected void dropReplica(LogicalNode site) { }
    protected  void dropLocal() { }

    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    protected boolean executeHere(final LogicalNode from, final LogicalNode to,
				  final AppMessage msg) {
	return true;
    }

    /**
     * Take any actions that have been pending while waiting for all threads
     * to stop using the object
     */
    protected void noThreads() { }

    //Added by Huahai
    public int getType() {
      return IMMOBILE;
    }

}
