/*
 * @(#)KanObject.java KanObject.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.obj;

import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.debug.Event;
import mokan.thread.KanThread;


import mokan.transact.LockRequest;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;

/**
 * A representation of a distributed object in the Kan system.  There is one
 * <code>KanObject</code> per object copy (whether primary copy or replica) in
 * the system.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Suk Y. Lee
 * @author	Jerry James
 */
public final class KanObject {

    /**
     * The Object Manager for this object
     *
     * @seemokan.obj.ObjectMgr
     */
    final ObjectMgr myMgr;


    /**
     * The <code>ObjectID</code> for this global object
     */
    final ObjectID objID;

    /**
     * The global object itself.  This can be an obsolete version.
     */
    final GlobalObj obj;

    /**
     * The version number of the object.  This is incremented by one at each
     * write.
     */
    public long version;

    /**
     * <code>true</code> if this copy of the object is valid (has not been
     * invalidated by the home/owner node)
     */
    public boolean valid = true;

    /**
     * The consistency protocol for this object
     *
     * @seemokan.obj.Protocol
     */
    Protocol protocol;

    /**
     * The number of threads running on this object
     */
    int callers;

    /**
     * Construct a new <code>KanObject</code> with this as the home node
     *
     * @param objID the <code>ObjectID</code> identifying this object
     * @param gObj the state of the global object
     * @param type the type of consistency protocol to use
     */
    public KanObject(final ObjectID objID, final GlobalObj gObj,
                     final int type) {
        this(objID, gObj, type, objID.getManager(), 0L);
    }

    /**
     * Reconstruct a <code>KanObject</code> on a remote node
     *
     * @param objID the <code>ObjectID</code> identifying this object
     * @param gObj the state of the global object
     * @param type the type of consistency protocol to use
     * @param manager the object manager for this <code>KanObject</code>
     * @param version the version of the object represented by <var>gObj</var>
     */
    public KanObject(final ObjectID objID, final GlobalObj gObj, final int type,
              final ObjectMgr manager, final long version) {
        myMgr = manager;
        this.objID = objID;
        obj = gObj;
        this.version = version;
        switch (type) {
        case Protocol.IMMOBILE:
            protocol = new ImmobileProtocol(this);
            break;
        case Protocol.GENERAL:
            final int numNodes = objID.getNode().getApplication().numNodes();
            protocol = new GeneralProtocol(this, numNodes);
            break;
        case Protocol.REPLICA:
            final int numbNods = objID.getNode().getApplication().numNodes();
            protocol = new ReplicaProtocol(this, numbNods);
            break;
        case Protocol.MIGRATORY:
            protocol = new MigratoryProtocol(this);
            break;
        case Protocol.MIGRATORY_HOME:
            protocol = new MigratoryHomeProtocol(this);
            break;
        case Protocol.OPTIMISTIC:
            protocol = new OptimisticProtocol(this);
            break;
        case Protocol.OPTIMISTIC_HOME:
            protocol = new OptimisticHomeProtocol(this);
            break;

        default:
          throw new RuntimeException("KanObject.KanObject: Unknown protocol type: " + type);
        }

        objID.setRealObject(this);
        Event.DEBUG(Event.OBJECT, "Creating ", this + " : " + gObj + " : " + protocol);
    }

    /**
     * Install a copy of the object on the local node
     *
     * @param gObj the object to install
     * @param version the version number of the object
     */
//     void install(final GlobalObj gObj, final long version) {
// 	protocol.install(gObj, version);
// 	objID.setRealObject(this);
//     }

    /**
     * Install a copy of the object on the local node and possibly change the
     * controlling protocol
     *
     * @param gObj the object to install
     * @param version the version number of the object
     * @param homeify <code>true</code> if the object should switch to general
     *        type on arrival
     * @param replica the node holding a replica, if <var>homeify</var> is
     *        true
     * KK: Called only from MoveObjMsg
     * KK: Is there a way to generalize this?
     */
    public void install(final GlobalObj gObj, final long version,
                 final boolean homeify, final LogicalNode replica) {
        protocol.install(gObj, version);
        objID.setRealObject(this);
        if (homeify && !(protocol instanceof GeneralProtocol)) {
            final int numNodes = objID.getNode().getApplication().numNodes();
            protocol = new GeneralProtocol(protocol, numNodes, replica);
        }
    }

    /**
     * Uninstall the object from the local node
     */
    void uninstall() {
        valid = false;
        objID.setRealObject(null);
    }
    /**
     * Lighweight install the object from the local node.  Used for
     * updating replicas
     */
    void install(GlobalObj obj, long version) {
        //objID.setRealObject(obj);
        this.obj.kanSetObject( obj );
        this.version = version;
        valid = true;
    }

    void invalidate() {
        protocol.invalidateReplica();
    }


    /**
     * Test whether we can make a replica/copy  of this object.
     */
    public final boolean isMobile () {
        return !(protocol instanceof ImmobileProtocol);
    }

    /**
     * Fetch the current state of this object.  Note that the object might be
     * in an inconsistent state.  This reference can only be used safely to
     * invoke methods on the object.
     *
     * @return the global object managed by this <code>KanObject</code>
     */
    public GlobalObj getObject() {
        return obj;
    }

    /**
     * Fetch the Object ID of this object.  Note that the object might be
     * in an inconsistent state.
     *
     * @return the object ID of this <code>KanObject</code>
     */
    public ObjectID getID () {
        return objID;
    }

    /**
     * Return a copy of the <code>GlobalObj</code> managed by this
     * <code>KanObject</code>, suitable for sending over the network while the
     * original object is still in use.
     *
     * @param target the target of the copy
     * @return a consistent copy of the <code>GlobalObj</code> managed by this
     *         <code>KanObject</code>
     */
    GlobalObjCopy copy(LogicalNode target) {
        if (isMobile()) {
            Event.DEBUG(Event.OBJECT_COPY, "Making a copy of ", this);
            return protocol.copy(target);
        }
        Event.DEBUG(Event.OBJECT_COPY, "Refused to copy[immobile] ", this);
        return null;
    }

    /**
     * Return a copy of the <code>GlobalObj</code> for restoration in case of
     * transaction abort.  To guarantee consistent results, this method should
     * only be called by a lock owner.
     *
     * @return a copy of the <code>GlobalObj</code>
     */
    public GlobalObj copyGlobalObj() {
        return obj.kanClone();
    }

    /**
     * Lock this Kan object.  The <code>LockRequest</code> object holds the
     * parameters of the lock.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    public void lock(final LockRequest request)
        throws TransactionAbortException {

        protocol.lock(request);
    }

    /**
     * Acquire a lock, ensuring that a guard condition is satisfied first.
     * The lock can be acquired in read/write and preemptible/nonpreemptible
     * modes.  This version just wraps the parameters into a
     * <code>LockRequest</code> and uses the other entry point.
     *
     * @param guardCond the condition that must be satisfied when the lock is
     *        granted
     * @param read <code>true</code> for a read lock; <code>false</code> for a
     *        write lock
     * @param preemptible <code>true</code> for a preemptible lock;
     *        <code>false</code> for a nonpreemptible lock
     * @return a <code>LockRequest</code> object representing this request
     * @exception TransactionAbortException if the transaction is aborted
     *            while waiting for the lock
     */
    public LockRequest lock(final GuardCondition guardCond,
                            final boolean read, final boolean preemptible)
        throws TransactionAbortException {

        final TransactionID creator = KanThread.currCall().currTrans;
        final LockRequest lr =
            new LockRequest(creator, creator.getNode(), guardCond, objID,
                            read, preemptible);
        lock(lr);
        return lr;
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    public void unlock(final LockRequest request, final boolean install)
        throws TransactionException {

        protocol.unlock(request, install, null);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that first locked the object.
     *
     * @param locker the <code>TransactionID</code> identifying the
     *        transaction that has the object locked
     * @param install whether to install a new version
     * @param obj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    public void unlock(final TransactionID locker, final boolean install,
                       final GlobalObj gObj)
        throws TransactionException {

        protocol.unlock(locker, install, gObj);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param obj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    public void unlock(final LockRequest request, final boolean install,
                       final GlobalObj gObj) throws TransactionException {
        protocol.unlock(request, install, gObj);
    }

    /**
     * Upgrade a lock from preemptible to nonpreemptible.  This requires going
     * through the same algorithm as the original lock, only we are careful to
     * avoid having the <code>LockRequest</code> conflict with itself.
     *
     * @param lockReq a previously granted preemptible lock request
     * @return the upgraded <code>LockRequest</code>
     * @exception TransactionAbortException if the locker was preempted before
     *            making the upgrade
     */
    public LockRequest upgrade(final LockRequest lockReq)
        throws TransactionAbortException {

        return protocol.upgrade(lockReq);
    }

    /**
     * Test whether this <code>KanObject</code> is locked.  This method is
     * useless unless the caller has synchronized on the lock.
     *
     * @return <code>true</code> if there are no current or waiting lockers;
     *         <code>false</code> otherwise
     */
    boolean idle() {
        return protocol.idle();
    }

    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    public boolean executeHere(final LogicalNode from, final LogicalNode to,
                               final AppMessage msg) {
        return protocol.executeHere(from, to, msg);
    }

    /**
     * Record the start of a method call on this object
     *
     * @return <code>true</code> if the method call can proceed;
     * <code>false</code> if the object cannot be used
     */
    synchronized boolean startCall() {
        callers++;
        //System.out.println ("startCall="+callers + " "+this );
        if (valid) {
            return true;
        }
        return false;
    }

    /**
     * Record the end of a method call on this object
     */
    synchronized void endCall() {
        if (--callers == 0)
            protocol.noThreads();
        //System.out.println ("endCall="+callers + " " + this);
    }

    /**
     *  Drop this object from this node.
     */
    public void dropReplica (LogicalNode node) {
        protocol.dropReplica(node);
    }


    /**
     * Convert the <code>KanObject</code> to a string
     */
    public String toString() {
        return "KObj["+ objID + ',' +  version + ']';
    }

    // Added by Huahai
    public Protocol getProtocol() {
      return protocol;
    }
  }
