/*
 * @(#)MigratoryHomeProtocol.java MigratoryHomeProtocol.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.obj;

import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.comm.YesNoReply;

import mokan.transact.LockRequest;
import mokan.transact.LockMsg;
import mokan.transact.UnlockMsg;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;

/**
 * A protocol for managing migratory objects on the home node
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 */
final public class MigratoryHomeProtocol extends MigratoryProtocol {

    /**
     * The current (approximate) location of the object
     */
    private LogicalNode location;

    /**
     * Create a new <code>MigratoryHomeProtocol</code> object
     *
     * @param kObj the <code>KanObject</code> to manage
     */
    MigratoryHomeProtocol(final KanObject kObj) {
        this(kObj, kObj.objID.getNode());
    }

    /**
     * Create a new <code>MigratoryHomeProtocol</code> object
     *
     * @param kObj the <code>KanObject</code> to manage
     * @param node the current location of the object
     */
    MigratoryHomeProtocol(final KanObject kObj, final LogicalNode node) {
        super(kObj);
        location = node;
    }

    /**
     * Convert to a migratory protocol from some other protocol
     *
     * @param prot the old protocol object
     */
    MigratoryHomeProtocol(final Protocol prot) {
      //Modified by Huahai
      //this(prot, prot.kanObj.objID.getNode());
      this(prot, LogicalNode.getMyNode());
    }

    /**
     * Convert to a migratory protocol from some other protocol
     *
     * @param prot the old protocol object
     * @param node the current location of the object
     */
    MigratoryHomeProtocol(final Protocol prot, final LogicalNode node) {
        super(prot);
        location = node;
    }

    /**
     * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
     * holds the parameters of the lock.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for specifying when a lock must be
     * acquired globally.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    protected void lock(final LockRequest request)
        throws TransactionAbortException {

        if (location.isLocal()) {
            super.lock(request);
        } else {
            //Modified by Huahai
            //final LogicalNode local = kanObj.myMgr.parentNode;
            final LogicalNode local = LogicalNode.getMyNode();

            final LockMsg msg = new LockMsg(local, request);
            final YesNoReply reply = (YesNoReply)msg.sendAndWait(location);
            if (!reply.yes) {
                throw new TransactionAbortException
                    ("Couldn't get lock from home node");
            }
        }
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that originally locked the object.
     * This is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install)
        throws TransactionException {

        if (location.isLocal())
            super.unlock(transID, install);
        else
            unlock(transID, install, null);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.  This
     * is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install)
        throws TransactionException {

        if (location.isLocal())
            super.unlock(request, install);
        else
            unlock(request.locker, install, null);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>TransactionID</code> object identifies the transaction that
     * originally locked the object.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install,
                          final GlobalObj gObj) throws TransactionException {
        if (location.isLocal()) {
            super.unlock(transID, install, gObj);
        } else {
            final LogicalNode local = kanObj.myMgr.parentNode;
            final GlobalObj sendObj = (gObj != null) ? gObj
                : (install ? kanObj.obj : null);
            final UnlockMsg msg = new UnlockMsg(local, transID, install,
                                                kanObj.objID, sendObj);
            final YesNoReply reply = (YesNoReply)msg.sendAndWait(location);
            if (!reply.yes)
                throw new TransactionAbortException("Abort on home");
        }
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.  This is essentially just a wrapper around the
     * <code>Lock</code> object associated with this <code>KanObject</code>,
     * with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install,
                          final GlobalObj gObj) throws TransactionException {
        if (location.isLocal())
            super.unlock(request, install, gObj);
        else
            unlock(request.locker, install, gObj);
    }

    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    protected boolean executeHere(final LogicalNode from, final LogicalNode to,
                                  final AppMessage msg) {
        if (location.isLocal())
            return super.executeHere(from, to, msg);
        msg.send(location);
        return false;
    }

    /**
     * Notify the protocol of a new location for the object
     *
     * @param node the new location of the object
     */
    protected void newLocation(final LogicalNode node) {
        location = node;
    }

    /**
     * Take any actions that have been pending while waiting for all threads
     * to stop using the object
     */
    protected void noThreads() { }

    //Added by Huahai
    public int getType() {
      return MIGRATORY_HOME;
    }

}
