/*
 * @(#)MoveObjMsg.java MoveObjMsg.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 */

package mokan.obj;

import java.io.*;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.debug.Event;

import mokan.nameservice.*;

/**
 * Move a migratory object to a new node
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 */
final public class MoveObjMsg extends AppMessage implements Externalizable {

  /**
   * The <var>serialVersionUID</var> for this class.  If you change the
   * signature of this class (field names or types, method names, parameter
   * names or types, adding or removing fields or methods, etc.), you
   * <strong>must</strong> change this value.  The recommended way is to use
   * the <code>serialver</code> tool.
   */
  static final long serialVersionUID = 6778103292859153254L;

  /**
   * The list of persistent fields for this class.  This is empty, because
   * that is the fastest option for <code>Externalizable</code> classes when
   * we are not worried about versioning.
   */
  private static final ObjectStreamField[] serialPersistentFields =
      ObjectStreamClass.NO_FIELDS;

  /**
   * The <code>ObjectID</code> for the copied object.  This is a
   * pseudo-final field.
   *
   * @serial the <code>ObjectID</code> for the copied object
   */
  private ObjectID objID;

  /**
   * The object itself.  This is a pseudo-final field.
   *
   * @serial the object itself
   */
  private GlobalObj objCopy;

  /**
   * The version number of this object.  This is a pseudo-final field.
   *
   * @serial the version number of this object
   */
  private long version;

  /**
   * The request that prompted this move.  This is a pseudo-final field.
   *
   * @serial the request that prompted this move
   */
  private AppMessage rqst;

  /**
   * Whether this object should switch to general type on arrival.  This is
   * a pseudo-final field.
   *
   * @serial whether this object should switch to general type on arrival
   */
  private boolean homeify;

  /**
   * A logical node holding a replica of the object, in case it should
   * switch to general on arrival.  This is a pseudo-final field.
   *
   * @serial a logical node holding a replica of the object, only if
   * <var>homeify</var> is true
   */
  private LogicalNode replica;

  /**
   * The no-args constructor, needed for deserialization
   */
  public MoveObjMsg() {
    super(true, false);
  }

  /**
   * Create a new <code>MoveObjMsg</code>
   *
   * @param from the originator of this message
   * @param objID the <code>ObjectID</code> of the object to replicate
   * @param objCopy a copy of the object
   * @param version the version number of this object
   * @param rqst the request that prompted this object move
   * @param homeify <code>true</code> if the object should switch to general
   *        type on arrival
   * @param replica the node holding a replica, if <var>homeify</var> is
   *        true
   */
  public MoveObjMsg(final LogicalNode from, final ObjectID objID,
                    final GlobalObj objCopy, final long version,
                    final AppMessage rqst, final boolean homeify,
                    final LogicalNode replica) {
    super(from, true, false);
    this.objID = objID;
    this.objCopy = objCopy;
    this.version = version;
    this.rqst = rqst;
    this.homeify = homeify;
    this.replica = replica;
    Event.DEBUG(Event.MSG_CREATE, "Created ", this);
  }

  /**
   * Handle this message by creating moving the transported object into the
   * local node and setting up a protocol handler for it
   */
  public void handle() {

    super.handle();
    final KanObject kanObj = objID.getLocKanObject();
    kanObj.install(objCopy, version, homeify, replica);
    if (rqst != null) {
      rqst.handle();
    }

    //Added by Huahai
    //Should we change the ownership to this node???
    //Set the ownership
    //Globals.nameService.addOwner(objID, from);
  }

  /**
   * Convert the <code>MoveObjMsg</code> to a string
   *
   * @return a string representation of this <code>MoveObjMsg</code>
   */
  public String toString() {
    return "MoveObjMsg(" + objID + "): " + version;
  }

  /**
   * Save the state of the <code>MoveObjMsg</code> instance to a stream
   * (i.e., serialize it).
   *
   * @serialData The superclass data is written out, then the ID of the
   *             object to move (ObjectID), then the object itself
   *             (GlobalObj), then the version number of the object (long),
   *             then the request that prompted the move (AppMessage),
   *             followed by an indication of whether this object should
   *             switch to the general type on arrival (boolean), and if
   *             true, followed by a logical node holding a replica.
   */
  public void writeExternal(final ObjectOutput out) throws IOException {
    super.writeExternal(out);
    out.writeObject(objID);
    out.writeObject(objCopy);
    out.writeLong(version);
    out.writeObject(rqst);
    out.writeBoolean(homeify);
    if (homeify) {
      out.writeObject(replica);
    }
  }

  public void readExternal(final ObjectInput in) throws IOException,
      ClassNotFoundException {

    super.readExternal(in);
    objID = (ObjectID) in.readObject();

    objCopy = (GlobalObj) in.readObject();
    version = in.readLong();
    rqst = (AppMessage) in.readObject();
    homeify = in.readBoolean();
    if (homeify) {
      replica = (LogicalNode) in.readObject();
    }
  }
}
