/*
 * @(#)ObjectID.java ObjectID.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.obj;

import java.io.*;
import java.util.Iterator;
import mokan.comm.LogicalNode;
import mokan.thread.AsynchCall;
import mokan.thread.KanThread;
import mokan.transact.TransactionAbortException;
import mokan.util.ID;
import mokan.util.CanonicalizingIDSet;

import mokan.debug.Event;
import mokan.nameservice.*;

/**
 * A unique identifier for a distributed object.  This class is subclassed by
 * stubs for user objects.  Those stubs are generated by the Kan compiler.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Suk Y. Lee
 * @author	Jerry James
 */
public class ObjectID extends ID implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
     static final long serialVersionUID = -1782826666185622557L;
    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /*
     * A container for canonicalizing <code>ObjectID</code>s
     */
    private static final CanonicalizingIDSet allObj =
	new CanonicalizingIDSet();

    /**
     * A unique sequence number generator for each node.
     */
    protected static long objCounter = 0;

    /**
     * A reference to the object, when it is local.  This is used to shortcut
     * the tortuous method calling path and go directly to the proper method
     * when the call is (1) synchronous, and (2) not in a transaction.
     */
    private transient GlobalObj realObj;

    /**
     * A reference to the Kan object, when it is local.  This is used to
     * shortcut the tortuous method calling path and go more directly to the
     * proper method when the call is (1) asynchronous, and (2) not in a
     * transaction.
     */
    private transient KanObject realKobj;

    /**
     * Whenther we can be a replica
     */
    private transient boolean allowReplicas = true;

    /**
     * The no-args constructor, needed for deserialization
     */
    public ObjectID() {
        super();
    }

    /**
     * Create an <class>ObjectID</class> for a particular
     * <code>LogicalNode</code>
     *
     * @param _node the node for this object
     */
    public ObjectID(final LogicalNode _node) {
        super(_node, _node.isLocal() ? objCounter++ : -1);
        if (count != -1)
            allObj.add(this);

        // Added by Huahai
        Globals.nameService.addOwner(this, LogicalNode.getMyNode());
    }

    /**
     * Copy constructor for <class>ObjectID</class>s
     *
     * @param objID the <class>ObjectID</class> to copy
     */
    protected ObjectID(ObjectID objID) {
        super(objID);
    }

    /**
     * Remove this <code>ObjectID</code> from the set of canonical object IDs
     */
    final void remove() {
        allObj.remove(this);
    }

    /**
     * Update a partial <code>ObjectID</code> with the version sent from a
     * remote node
     *
     * @param objID the <code>ObjectID</code> from the remote node
     */
    final void update(final ObjectID objID) {
        allObj.remove(this);  // Have to do this to get the hashing right
        allObj.remove(objID);
        count = objID.count;
        allObj.add(this);
    }

    /**
     * If the object is local, fetch a copy
     *
     * @return a copy of the object, if one exists locally
     */
    public final KanObject getLocKanObject() {
        return realKobj;
    }

    /**
     *   Return the actual java object if one is available.
     */
    public final GlobalObj getRealObj() {
        return this.realObj ;
    }


    /**
     * Fetch a copy of an object from its home node
     * KK:  This code is used only from a error condition in methodInfo
     * KK:  while attempting to execute a method on the local node when
     * KK:  the object is not found.
     * KK:  Probably can be removed safely or made more generic.
     *
     * @return a <code>KanObject</code> representing the fetched object
     */
    public final KanObject getObjectCopy() {
        if (allowReplicas) {
            CopyObjMsg msg = new CopyObjMsg(this);
            try {
                Event.DEBUG(Event.OBJECT_COPY,"Copy Requested of " + this);
                //Modified by Huahai
                //final CopyObjReply reply = (CopyObjReply)msg.sendAndWait(node);
                LogicalNode home = Globals.nameService.discoverHome(this);
                final CopyObjReply reply = (CopyObjReply)msg.sendAndWait(home);

                Event.DEBUG(Event.OBJECT_COPY,"Copy Reply " + reply);
                if (reply.kObj == null) {
                    Event.DEBUG(Event.OBJECT_COPY, "Copy Attempt refused for ",
                                this);
                    allowReplicas = false;
                }
                return reply.kObj;
            } catch (TransactionAbortException abortEx) {
            }
        }
        return null;
    }

    /**
     * Pointer swizzling.  Tell this identifier that the object it identifies
     * is local, and give it a reference.
     *
     * @param realObj a reference to the object that this identifier identifies
     */
    final public void setRealObject(final KanObject realObj) {
        this.realKobj = realObj;
        this.realObj = (realObj == null) ? null : realObj.obj;

    }

    /**
     * Get the object manager for this object identifier
     *
     * @return the object manager for this object identifier
     */
    final ObjectMgr getManager() {
        return node.getLocalHandler().manager;
    }

    /**
     * Check whether the direct method calling optimization can be used.  This
     * assumes that it is only called when a synchronous method call is about
     * to be made.  The compiler must enforce that restriction, or chaos will
     * result.
     *
     * @return a reference to the object if it can be called directly;
     *        <code>null</code> otherwise
     */
    public final GlobalObj callDirect() {
        // This method name has been brought to you by the fine folks at AT&T
        try {
            final AsynchCall call = KanThread.currCall();
            return ((call == null || call.currTrans == null) &&
                    realKobj != null && realKobj.valid) ? realObj : null;
        } catch (ClassCastException castEx) {
            // REVIEW: Not a KanThread, so no current transaction ... right?
            return (realKobj != null && realKobj.valid) ? realObj : null;
        }
    }

    /**
     * Is our local copy valid
     */
    public final boolean isLocalValid () {
        return realKobj!=null && realKobj.valid;
    }

    /**
     * Record the start of a method call on this object
     *
     * @return <code>true</code> if the method call can proceed;
     * <code>false</code> if the object cannot be used
     */
    public final boolean startCall() {
        return realKobj != null && realKobj.startCall();
    }

    /**
     * Record the end of a method call on this object
     */
    public final void endCall() {
        if (realKobj != null)
            realKobj.endCall();
    }

    /**
     *   Remove reference to a
     */
    public static void removeObjectRefs(LogicalNode node) {

        Iterator ids = allObj.iterator();
        while (ids.hasNext()) {
            ObjectID oid = (ObjectID)ids.next();

            if (oid.realKobj != null) {
                oid.realKobj.dropReplica (node);
                //oid.remove();
            }
            if (oid.getNode() == node) {
                System.err.println ("trying to remove " + oid);
                //oid.remove();
            }
        }
    }


    public void invalidate() {
        if (realKobj!=null) {
            realKobj.invalidate();
        }
    }

    //Added by Huahai
    public void readExternal(final ObjectInput in)
        throws IOException, ClassNotFoundException {
      super.readExternal(in);
      LogicalNode from = (LogicalNode) Globals.objectInputMap.get(in);
      if(from == null) {
        throw new NameServiceException("Invalid from node");
      }
      Globals.nameService.addOwner(this, from);
    }

    public Object readResolve() throws ObjectStreamException {
        ObjectID tempID = (ObjectID) allObj.canonicalize(this);
        if(tempID != this) {
          LogicalNode owner = Globals.nameService.removeOwner(this);
          Globals.nameService.addOwner(tempID, owner);
        }

        return tempID;
    }

    public String toString() {
        return "Obj" + super.toString() + this.getClass().getName();
    }

    //Added by Huahai
    public LogicalNode getNode() {
      Event.DEBUG(Event.OBJECT_ACCESS, "Warning: ObjectID.getNode() does not return the home node.");
      return super.getNode();
    }

}
