/**
 * OptimisticHomeProtocol
 * Copyright kkvilekval@acm.org.   See COPYRIGHT for details.
 */
package mokan.obj;


import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import mokan.util.IntIntMap;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;

import mokan.obj.Protocol;


import mokan.transact.LockRequest;
import mokan.transact.IsolationRequest;
import mokan.transact.LockMsg;
import mokan.transact.UnlockMsg;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;

import mokan.debug.Event;

/**
 *  The <code>OptimisticHomeProtocol</code> description.
 *  @version $Revision$, $Date$
 *  @author Kristian Kvilekval (kkvilekval@acm.org)
 */
public class OptimisticHomeProtocol extends OptimisticProtocol{
    /* Private/Public Variables */

    /**
     * The set of nodes caching this <code>KanObject</code>
     */
    private final HashSet cachedNodes = new HashSet();

    /**
     * Create a new <code>OptimisticProtocol</code> object
     *
     * @param kObj the <code>KanObject</code> to manage
     */
    OptimisticHomeProtocol(final KanObject kObj) {
        super(kObj);
    }

    /**
     * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
     * holds the parameters of the lock.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for specifying when a lock must be
     * acquired globally.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    protected void lock(final LockRequest request)
        throws TransactionAbortException {

        Event.DEBUG(Event.LOCK, "OptimisticHome.lock:"  +request);

        lock.lock(request);

    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that originally locked the object.
     * This is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install)
        throws TransactionException {

        lock.unlock(lock.lookup(transID), install, null);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.  This
     * is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install)
        throws TransactionException {

        lock.unlock(request, install, null);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>TransactionID</code> object identifies the transaction that
     * originally locked the object.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install,
                          final GlobalObj gObj)	throws TransactionException {
        lock.unlock(lock.lookup(transID), install, gObj);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.  This is essentially just a wrapper around the
     * <code>Lock</code> object associated with this <code>KanObject</code>,
     * with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install,
                          final GlobalObj gObj)	throws TransactionException {


        final LogicalNode local = kanObj.myMgr.parentNode;
        final LogicalNode from = request.requester;

        lock.unlock(request, install, gObj);

        if (!request.isPreemptible()) {
            // Update the home node.. may be incompatible
            // We could eagerly send abort messages at this point
            // for forward validition.  Current we use backward.

            // Forward validation model

            // Send drop message (with new version #) to all cached
            // nodes. On recieving drop mesage: replica would abort
            // local transactions using that replica.


        }
    }



    /**
     * Return a copy of the <code>GlobalObj</code> managed by this
     * <code>Protocol</code>, suitable for sending over the network while the
     * original object is still in use.
     *
     * @param target the target of the copy
     * @return a consistent copy of the <code>GlobalObj</code> managed by this
     *         <code>KanObject</code>
     */
    protected GlobalObjCopy copy(final LogicalNode target) {
//         try {
//             IsolationRequest lr = lock.lockForCopy();

            final GlobalObjCopy copy =
                new GlobalObjCopy(kanObj.getObject().kanClone(),
                                  kanObj.version,
                                  Protocol.OPTIMISTIC);
//             lock.unlockForCopy(lr);
            cachedNodes.add(target);
            return copy;
//         } catch (InterruptedException intEx) {
//             // This can't happen
//             intEx.printStackTrace();
//             return null;
//         }
    }

    /**
     * Upgrade a lock from optimistic to strict.  This requires going
     * through the same algorithm as the original lock, only we are careful to
     * avoid having the <code>LockRequest</code> conflict with itself.
     *
     * @param lockReq a previously granted preemptible lock request
     * @return the upgraded <code>LockRequest</code>
     * @exception TransactionAbortException if the locker was preempted before
     *            making the upgrade
     */
     protected LockRequest upgrade(final LockRequest lockReq)
         throws TransactionAbortException {

        Event.DEBUG(Event.UPGRADE, "OptimisticHomeProtocol.upgrade lock for ",
                    lockReq);
         return lock.upgrade(lockReq);
     }



    /**
     *  Inform the protocol that replica has been sent to another site
     *  This is currently used only by <code>OptimisticHomeProtocol</code>
     */
    protected void addReplica (LogicalNode site) {
        cachedNodes.add(site);
    }

    /**
     * Inform one node of the disappearance of a replica from another.  This
     * is a no-op for replicas.
     *
     * @param site the site of the dropped replica
     */
    protected void dropReplica(LogicalNode site) {
        Event.DEBUG(Event.GENERAL_OBJECT, "OptimisticHome.dropReplica @", site);
        cachedNodes.remove(site);

    }
    protected  void dropLocal() { }

    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    protected boolean executeHere(final LogicalNode from, final LogicalNode to,
                                  final AppMessage msg) {
        return true;
    }

    /**
     * Take any actions that have been pending while waiting for all threads
     * to stop using the object
     */
    protected void noThreads() { }

    public HashSet getCachedNodes() {
      return cachedNodes;
    }

    //Added by Huahai
    public int getType() {
      return OPTIMISTIC_HOME;
    }
}
