/**
 * OptimisticProtocol
 * Copyright kkvilekval@acm.org.   See COPYRIGHT for details.
 */
package mokan.obj;

import mokan.comm.AppMessage;
import mokan.comm.YesNoReply;
import mokan.comm.LogicalNode;

import mokan.transact.LockRequest;
import mokan.transact.LockMsg;
import mokan.transact.UnlockMsg;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;
import mokan.nameservice.*;

import mokan.debug.Event;

/**
 *  The <code>OptimisticProtocol</code> description.
 *  @version $Revision$, $Date$
 *  @author Kristian Kvilekval (kkvilekval@acm.org)
 */
class OptimisticProtocol extends Protocol {

  boolean pendingDrop = false;

  /**
   * Create a new <code>OptimisticProtocol</code> object
   *
   * @param kObj the <code>KanObject</code> to manage
   */
  OptimisticProtocol(final KanObject kObj) {
    super(kObj);
  }

  /**
   * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
   * holds the parameters of the lock.  This is essentially just a wrapper
   * around the <code>Lock</code> object associated with this
   * <code>KanObject</code>, with code for specifying when a lock must be
   * acquired globally.
   *
   * @param request the <code>LockRequest</code> giving the parameters for
   *        this request
   * @exception TransactionAbortException if the lock cannot be granted
   * @seemokan.obj.LockRequest
   */
  protected void lock(final LockRequest request) throws
      TransactionAbortException {

    Event.DEBUG(Event.LOCK, "OptimisticProtocol.lock" + request);

    lock.lock(request);
  }

  /**
   * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
   * object identifies the transaction that originally locked the object.
   * This is essentially just a wrapper around the <code>Lock</code> object
   * associated with this <code>KanObject</code>, with code for handling
   * object types
   *
   * @param transID the <code>TransactionID</code> identifying this request
   * @param install whether to install a new version
   * @exception TransactionException if the lock was stolen
   */
  protected void unlock(final TransactionID transID, final boolean install) throws
      TransactionException {

    lock.unlock(lock.lookup(transID), install, null);
  }

  /**
   * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
   * object holds the parameters used when first locking the object.  This
   * is essentially just a wrapper around the <code>Lock</code> object
   * associated with this <code>KanObject</code>, with code for handling
   * object types.
   *
   * @param request the <code>LockRequest</code> giving the parameters for
   *        this request
   * @param install whether to install a new version
   * @exception TransactionException if the lock was stolen
   * @seemokan.obj.LockRequest
   */
  protected void unlock(final LockRequest request, final boolean install) throws
      TransactionException {

    lock.unlock(request, install, null);
  }

  /**
   * Unlock and update this <code>KanObject</code>.  The
   * <code>TransactionID</code> object identifies the transaction that
   * originally locked the object.  This is essentially just a wrapper
   * around the <code>Lock</code> object associated with this
   * <code>KanObject</code>, with code for handling object types.
   *
   * @param transID the <code>TransactionID</code> identifying this request
   * @param install whether to install a new version
   * @param gObj the new version of the object to install
   * @exception TransactionAbortException if the lock was stolen
   */
  protected void unlock(final TransactionID transID, final boolean install,
                        final GlobalObj gObj) throws TransactionException {
    lock.unlock(lock.lookup(transID), install, gObj);
  }

  /**
   * Unlock and update this <code>KanObject</code>.  The
   * <code>LockRequest</code> object holds the parameters used when first
   * locking the object.  This is essentially just a wrapper around the
   * <code>Lock</code> object associated with this <code>KanObject</code>,
   * with code for handling object types.
   *
   * @param request the <code>LockRequest</code> giving the parameters for
   *        this request
   * @param install whether to install a new version
   * @param gObj the new version of the object to install
   * @exception TransactionAbortException if the lock was stolen
   * @seemokan.obj.LockRequest
   */
  protected void unlock(final LockRequest request, final boolean install,
                        final GlobalObj gObj) throws TransactionException {

    final LogicalNode local = kanObj.myMgr.parentNode;
    final LogicalNode from = request.requester;

    if (!request.isPreemptible()) {
      // Update the home node.. may be incompatible

      final GlobalObj sendObj = (gObj != null)
          ? gObj : (install ? kanObj.obj : null);
      final UnlockMsg msg =
          new UnlockMsg(local, request.locker, install,
                        kanObj.objID, sendObj);

      //Modified by Huahai
      //LogicalNode home = kanObj.getID().getNode();
      LogicalNode home = Globals.nameService.lookupHome(kanObj.getID());

      final YesNoReply reply = (YesNoReply) msg.sendAndWait(home);

      if (!reply.yes) {
        kanObj.uninstall();
        request.locker.kill(
            new TransactionAbortException("Abort on home"));
      }
    }
    lock.unlock(request, install, gObj);
  }

  /**
   * Upgrade a lock from optimistic to strict.  This requires going
   * through the same algorithm as the original lock, only we are careful to
   * avoid having the <code>LockRequest</code> conflict with itself.
   *
   * @param lockReq a previously granted preemptible lock request
   * @return the upgraded <code>LockRequest</code>
   * @exception TransactionAbortException if the locker was preempted before
   *            making the upgrade
   */
  protected LockRequest upgrade(LockRequest lockReq) throws
      TransactionAbortException {

    Event.DEBUG(Event.UPGRADE, "OptimisticProtocol.upgrade lock for ",
                lockReq);

    // Upgrade our local copy
    lockReq = lock.upgrade(lockReq);

    if (!lockReq.isPreemptible()) {
      //Modified by Huahai
      //LogicalNode home = kanObj.objID.getNode();
      LogicalNode home = Globals.nameService.discoverHome(kanObj.objID);

      final LogicalNode local = kanObj.myMgr.parentNode;
      //final LogicalNode from = request.requester;
      //  If non preemptible then we must ask home node

      Event.DEBUG(Event.LOCK, "requesting home lock");
      // Nonpreemptible write locks on replicas ask home for lock
      final LockMsg msg = new LockMsg(local, lockReq);

      final YesNoReply reply = (YesNoReply) msg.sendAndWait(home);

      Event.DEBUG(Event.LOCK, "received home lock = " + reply.yes);
      if (!reply.yes) {
        Event.DEBUG(Event.UPGRADE,
                    "OptimisticProtocol.upgrade lock for " + lockReq + "failed");

        invalidateReplica();
        lockReq.locker.kill(
            new TransactionAbortException("Upgrade failed"));
      }
    }
    return lockReq;
  }

  /**
   * Inform one node of the disappearance of a replica from another.  This
   * is a no-op for replicas.
   *
   * @param site the site of the dropped replica
   */
  protected void dropReplica(LogicalNode site) {

  }

  protected void dropLocal() {}

  /**
   *  Invalidate the local copy either aborting any pending
   *  transaction or restarting them with a new copy
   */
  void invalidateReplica() {
    Event.DEBUG(Event.OPTIMISTIC_OBJECT, "OptimisticProtocol.invalidate");

    // Backward validation:
    //   Do nothing as operation will abort when upgrading

    // Forward validation:
    //   Home sends out drop message to replicas:
    //   We should abort any local transaction using
    //   an out-of-date version.
    kanObj.uninstall();
    lock.nudge();
  }

  /**
   * Determine whether a request for a method execution should be filled
   * locally, the request should be forwarded elsewhere, or the object
   * should be shipped to the invoker.  In other words, should I stay or
   * should I go?
   *
   * @param from the remote logical node
   * @param to the local logical node
   * @param msg the request to forward or fill locally
   * @return <code>true</code> if the request should be filled locally;
   *         <code>false</code> if the request was forwarded or the object
   *         was shipped
   */
  protected boolean executeHere(final LogicalNode from, final LogicalNode to,
                                final AppMessage msg) {
    return true;
  }

  /**
   * Take any actions that have been pending while waiting for all threads
   * to stop using the object
   */
  protected void noThreads() {
    if (pendingDrop) {
      Event.DEBUG(Event.OBJECT_DEREPLICATE, "Removing ", kanObj);
      //Modified by Huahai
      //LogicalNode home = kanObj.objID.getNode();
      LogicalNode home = Globals.nameService.discoverHome(kanObj.objID);

      new ReplicaDropMsg(kanObj.objID).send(home);

      kanObj.uninstall();
    }
  }

  //Added by Huahai
  public int getType() {
    return OPTIMISTIC;
  }

}
