/*
 * @(#)Protocol.java Protocol.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.obj;

import mokan.Kan;
import mokan.KanSystem;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;



import mokan.transact.IsolationPrimitive;
import mokan.transact.LockRequest;
import mokan.transact.StrictLock;

import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;

import mokan.thread.KanThread;

import java.util.*;

/**
 * An abstract parent to all classes implementing object type protocols
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 */
public abstract class Protocol {

    // Object types

    /**
     * A code for a general (replicating) object
     */
    public static final int IMMOBILE       = 0;

    /**
     * A code for a general (replicating) object
     */
    public static final int GENERAL        = 1;

    /**
     * A code for a replica of an object
     */
    public static final int REPLICA        = 2;

    /**
     * A code for a migratory object not on its home node
     */
    public static final int MIGRATORY      = 3;

    /**
     * A code for a migratory object on its home node
     */
    public static final int MIGRATORY_HOME = 4;

    /**
     * A code for an optimistic object not on its home node
     */
    public static final int OPTIMISTIC      = 5;

    /**
     * A code for a migratory object on its home node
     */
    public static final int OPTIMISTIC_HOME = 6;

    //Added by Huahai
    public static final int REPLICA_HOME = 7;

    /**
     * The number of accesses after which a typing scheme is reexamined
     */
    protected static final int REPL_INTERVAL;

    static {
        final String intervalName =
            Kan.systemProps.getProperty("obj.replica.interval");
        int interval = 10;
        try {
            interval = Integer.parseInt(intervalName);
        } catch (NumberFormatException numEx) {
        }
        REPL_INTERVAL = interval;
    }

    /**
     * The lock associated with this object
     */
    protected final IsolationPrimitive lock;

    /**
     * The <code>KanObject</code> this protocol object is managing
     */
    protected final KanObject kanObj;

    /**
     * Create a new <code>Protocol</code> object to manage an object of some
     * particular type
     *
     * @param kObj the <code>KanObject</code> to manage
     */
    protected Protocol(final KanObject kObj) {
        //lock = new StrictLock(kObj);
        lock = KanSystem.scheduler.makePrimitive(kObj,0);
        kanObj = kObj;
    }

    /**
     * Create a new <code>Protocol</code> object to manage an object that was
     * previously of a different type
     *
     * @param prot the <code>Protocol</code> that used to manage the object
     */
    protected Protocol(final Protocol prot) {
        lock = prot.lock;
        kanObj = prot.kanObj;
    }

    /**
     * Test whether this <code>KanObject</code> is locked.  This method is
     * useless unless the caller has synchronized on the lock.
     *
     * @return <code>true</code> if there are no current or waiting lockers;
     *         <code>false</code> otherwise
     */
    protected final boolean idle() {
        return lock.idle();
    }

    /**
     * Install a copy of the object on the local node
     *
     * @param gObj the object to install
     * @param version the version number of the object
     */
    protected final void install(final GlobalObj gObj, final long version) {
        try {
            final TransactionID creator = KanThread.currCall().currTrans;

            System.out.println ("creator="+creator+" kanObj="+kanObj);
            final LockRequest lockReq =
                new LockRequest(creator, creator.getNode(), null, kanObj.objID,
                                false, false);
            lock.lock(lockReq);
            //lockReq.gobj = gObj;
            lockReq.version = version;
            lock.unlock(lockReq, true, gObj, version);
        } catch (TransactionException transEx) {
            // This can't happen
            // REVIEW: Can it?
        }
    }

    /**
     * Return a copy of the <code>GlobalObj</code> managed by this
     * <code>Protocol</code>, suitable for sending over the network while the
     * original object is still in use.
     *
     * @param target the target of the copy
     * @return a consistent copy of the <code>GlobalObj</code> managed by this
     *         <code>KanObject</code>
     */
    protected GlobalObjCopy copy(final LogicalNode target) {
//         try {
//             return lock.copy();
//         } catch (InterruptedException intEx) {
//             // This can't happen
//             intEx.printStackTrace();
//             return null;
//         }
        // Each protocol should define a valid copying procedure.
        //KK throw some sort of exception (badprotocol)
        return null;
    }

    /**
     * Upgrade a lock from preemptible to nonpreemptible.  This requires going
     * through the same algorithm as the original lock, only we are careful to
     * avoid having the <code>LockRequest</code> conflict with itself.
     *
     * @param lockReq a previously granted preemptible lock request
     * @return the upgraded <code>LockRequest</code>
     * @exception TransactionAbortException if the locker was preempted before
     *            making the upgrade
     */
    protected  LockRequest upgrade(final LockRequest lockReq)
        throws TransactionAbortException {

        return lock.upgrade(lockReq);
    }

    /**
     * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
     * holds the parameters of the lock.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    protected abstract void lock(LockRequest request)
        throws TransactionAbortException;

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that originally locked the object.
     * This is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     */
    protected abstract void unlock(TransactionID transID, boolean install)
        throws TransactionException;

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.  This
     * is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected abstract void unlock(LockRequest request, boolean install)
        throws TransactionException;

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>TransactionID</code> object identifies the transaction that
     * originally locked the object.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     */
    protected abstract void unlock(TransactionID transID, boolean install,
                                   GlobalObj gObj) throws TransactionException;

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.  This is essentially just a wrapper around the
     * <code>Lock</code> object associated with this <code>KanObject</code>,
     * with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected abstract void unlock(LockRequest request, boolean install,
                                   GlobalObj gObj) throws TransactionException;

    /**
     * Inform one node of the disappearnce of a replica from another.  This is
     * a no-op for all protocols except <code>GeneralProtocol</code>.
     *
     * @param site the site of the dropped replica
     */
    //KK Please change name to replicaDropped
    protected abstract void dropReplica(LogicalNode site);

    /**
     *  Inform the protocol that replica has been sent to another site
     *  This is currently used only by <code>OptimisticHomeProtocol</code>
     */
    protected void addReplica (LogicalNode site) {}

    /**
     *  Drop the local copy of the object.
     */
    protected abstract void dropLocal ();

    /**
     *  Invalidate the local copy either aborting any pending
     *  transaction or restarting them with a new copy
     */
    void invalidateReplica () {}

    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    protected abstract boolean executeHere(LogicalNode from, LogicalNode to,
                                           AppMessage msg);

    /**
     * Take any actions that have been pending while waiting for all threads
     * to stop using the object
     */
    protected abstract void noThreads();

    //Added by Huahai
    public HashSet getCachedNodes() {
      return null;
    }

    //Added by Huahai
    /**
     * Abort all lock requests.
     */
    public void abortWaiters() {
      lock.abortWaiters();
    }

    /**
     * Lock the object locally.
     * @param lockReq
     */
    public void lockLocally(LockRequest lockReq) {
      lock.lock(lockReq);
    }

    //Added by Huahai
    public int getType() {
      return -1;
    }
}
