/*
 * @(#)ReplicaProtocol.java ReplicaProtocol.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.obj;

import java.util.Collection;
import java.util.Iterator;

import mokan.util.IntIntMap;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.comm.YesNoReply;
import mokan.debug.Event;
import mokan.transact.LockRequest;
import mokan.transact.LockMsg;
import mokan.transact.UnlockMsg;
import mokan.transact.TransactionAbortException;
import mokan.transact.TransactionException;
import mokan.transact.TransactionID;

import mokan.nameservice.*;

/**
 * A protocol for managing general (replicated) objects on a replica node
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 */
final class ReplicaProtocol extends Protocol {

    /**
     * Count the number of writes from each node.  The indexes are those
     * assigned to <code>LogicalNode</code>s.
     *
     * @seemokan.comm.LogicalNode
     */
    private final IntIntMap writes;

    /**
     * Count the number of reads from each node.  The indexes are those
     * assigned to <code>LogicalNode</code>s.
     *
     * @seemokan.comm.LogicalNode
     */
    private final IntIntMap reads;

    /**
     * The number of times this object has been accessed (read or written)
     * during the current replication strategy interval
     */
    private int accessCount = 0;

    /**
     * <code>true</code> if this replica will be dropped once all threads
     * finish; <code>false</code> otherwise
     */
    private boolean pendingDrop = false;

    /**
     * Create a new <code>ReplicaProtocol</code> object
     *
     * @param kObj the <code>KanObject</code> to manage
     * @param nodes the number of nodes in this application
     */
    ReplicaProtocol(final KanObject kObj, final int nodes) {
        super(kObj);
        writes = new IntIntMap();
        reads = new IntIntMap();
    }

    /**
     * Convert to a replica protocol from some other protocol
     *
     * @param prot the old protocol object
     * @param nodes the number of nodes in this application
     */
    ReplicaProtocol(final Protocol prot, final int nodes) {
        super(prot);
        writes = new IntIntMap();
        reads = new IntIntMap();
    }

    /**
     * Count an access to this object, for purposes of determining the
     * replication scheme
     *
     * @param local the local logical node
     * @param accessor the logical node initiating the access
     * @param write <code>true</code> if the access is a write
     */
    private void countAccess(final LogicalNode local,
                             final LogicalNode accessor,
                             final boolean write) {
        final int accessorIndex = accessor.hashCode();
        if (write) {
            // Count reads by accessing node.  Technically, this should be
            // synchronized, but we can stand a small amount of inaccuracy
            writes.put(accessorIndex, writes.get(accessorIndex)+1);
        } else {
            // Count reads by accessing node.  Technically, this should be
            // synchronized, but we can stand a small amount of inaccuracy
            reads.put(accessorIndex, reads.get(accessorIndex)+1);
        }

        // Reexamine the replication scheme if enough accesses have taken
        // place.  The increment of <var>accessCount</var> technically should
        // be synchronized, but it won't kill us if an interleaving happens.
        if (++accessCount == REPL_INTERVAL) {
            final LogicalNode home = kanObj.objID.getNode();  // Home node
            //final Application app = local.getApplication(); // The application
            //final Collection nodes = app.getNodes();	// All the nodes
            final int localIndex = local.hashCode();

            // Should anybody else get a replica?  First, count the total
            // number of writes
            int writeCount = 0;
            for (Iterator i = writes.entryIterator(); i.hasNext(); ) {
                writeCount += ((IntIntMap.Entry)i.next()).value;
            }
            // I am a replica node.  Should I keep my replica?
            int readCount = reads.get(localIndex);
            if (3 * readCount <= writeCount - writes.get(localIndex)) {
                // Dereplicate!
                pendingDrop = true;	// Wait for current caller to exit
            }

            // Reset the counters
            writes.clear();
            reads.clear();
            accessCount = 0;
        }
    }

    /**
     * Lock this <code>KanObject</code>.  The <code>LockRequest</code> object
     * holds the parameters of the lock.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for specifying when a lock must be
     * acquired globally.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @exception TransactionAbortException if the lock cannot be granted
     * @seemokan.obj.LockRequest
     */
    protected void lock(final LockRequest request)
        throws TransactionAbortException {

        final LogicalNode local = kanObj.myMgr.parentNode;
        final LogicalNode from = request.requester;

        //Event.DEBUG(Event.LOCK, "ReplicaProtocol.lock: "+request +" from " + from + " local " + local);

        if (from != local) {
            // Somebody else started this; just lock it
            //Event.DEBUG(Event.LOCK, "requesting local lock from" + from);
            lock.lock(request);
        } else {
            // I'm the initiator; I have to behave
            if (request.isReadOnly() || request.isPreemptible()) {
                //Event.DEBUG(Event.LOCK, "requesting local lock");
                lock.lock(request);
            } else {
                //Event.DEBUG(Event.LOCK, "requesting home lock");
                // Nonpreemptible write locks on replicas ask home for lock

                // Modified by Huahai
                //LogicalNode home = kanObj.objID.getNode();
                //final LockMsg msg = new LockMsg(local, request);
                //final YesNoReply reply = (YesNoReply) msg.sendAndWait(home);

                // Discover home node every time, not atomic though
                LogicalNode home = Globals.nameService.discoverHome(kanObj.objID);
                LockMsg msg = new LockMsg(local, request);
                YesNoReply reply = (YesNoReply)msg.sendAndWait(home);

                if (!reply.yes) {
                    throw new TransactionAbortException
                        ("Couldn't get lock from home node");
                }
            }
        }
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>TransactionID</code>
     * object identifies the transaction that originally locked the object.
     * This is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install)
        throws TransactionException {

        unlock(lock.lookup(transID), install, null);
    }

    /**
     * Unlock this <code>KanObject</code>.  The <code>LockRequest</code>
     * object holds the parameters used when first locking the object.  This
     * is essentially just a wrapper around the <code>Lock</code> object
     * associated with this <code>KanObject</code>, with code for handling
     * object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @exception TransactionException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install)
        throws TransactionException {

        unlock(request, install, null);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>TransactionID</code> object identifies the transaction that
     * originally locked the object.  This is essentially just a wrapper
     * around the <code>Lock</code> object associated with this
     * <code>KanObject</code>, with code for handling object types.
     *
     * @param transID the <code>TransactionID</code> identifying this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     */
    protected void unlock(final TransactionID transID, final boolean install,
                          final GlobalObj gObj)	throws TransactionException {
        unlock(lock.lookup(transID), install, gObj);
    }

    /**
     * Unlock and update this <code>KanObject</code>.  The
     * <code>LockRequest</code> object holds the parameters used when first
     * locking the object.  This is essentially just a wrapper around the
     * <code>Lock</code> object associated with this <code>KanObject</code>,
     * with code for handling object types.
     *
     * @param request the <code>LockRequest</code> giving the parameters for
     *        this request
     * @param install whether to install a new version
     * @param gObj the new version of the object to install
     * @exception TransactionAbortException if the lock was stolen
     * @seemokan.obj.LockRequest
     */
    protected void unlock(final LockRequest request, final boolean install,
                          final GlobalObj gObj)	throws TransactionException {
        // This will happen if this is a new replica and the requester held a
        // lock before the replica was made.  Just pretend it unlocked
        // successfully.
        if (request == null)
            return;

        final LogicalNode local = kanObj.myMgr.parentNode;
        final LogicalNode from = request.requester;
        if (from != local) {
            // Somebody else started this; just install and unlock
            countAccess(local, request.locker.getAccessor(), true);
            lock.unlock(request, install, gObj);
        } else {
            // I started it; I have to behave
            if (request.isReadOnly())
                countAccess(local, request.locker.getAccessor(), false);

            if (request.isReadOnly() || request.isPreemptible()) {
                lock.unlock(request, install, gObj);
            } else {
                // Nonpreemptible write locks on replicas ask home to unlock
                final GlobalObj sendObj = (gObj != null)
                    ? gObj : (install ? kanObj.obj : null);
                final UnlockMsg msg =
                    new UnlockMsg(local, request.locker, install,
                                  kanObj.objID, sendObj);

                // Modified by Huahai
                //LogicalNode home = kanObj.objID.getNode();
                LogicalNode home = Globals.nameService.discoverHome(kanObj.objID);

                YesNoReply reply = (YesNoReply) msg.sendAndWait(home);

                if (!reply.yes)
                    throw new TransactionAbortException("Abort on home");
            }
        }
    }

    /**
     * Inform one node of the disappearance of a replica from another.  This
     * is a no-op for replicas.
     *
     * @param site the site of the dropped replica
     */
    protected void dropReplica(LogicalNode site) {
        Event.DEBUG(Event.REPLICA_OBJECT, "ReplicaProtocol.dropReplica", site);
        pendingDrop = true;
    }

    /**
     *  Drop the local copy of the object.
     */
    protected  void dropLocal () {
        pendingDrop = true;
    }

    /**
     * Determine whether a request for a method execution should be filled
     * locally, the request should be forwarded elsewhere, or the object
     * should be shipped to the invoker.  In other words, should I stay or
     * should I go?
     *
     * @param from the remote logical node
     * @param to the local logical node
     * @param msg the request to forward or fill locally
     * @return <code>true</code> if the request should be filled locally;
     *         <code>false</code> if the request was forwarded or the object
     *         was shipped
     */
    protected boolean executeHere(final LogicalNode from, final LogicalNode to,
                                  final AppMessage msg) {
        return kanObj.valid;
    }

    /**
     * Take any actions that have been pending while waiting for all threads
     * to stop using the object
     */
    protected void noThreads() {
        if (pendingDrop) {
            Event.DEBUG(Event.OBJECT_DEREPLICATE, "Removing ", kanObj);
            LogicalNode home = kanObj.objID.getNode();
            new ReplicaDropMsg(kanObj.objID).send(home);

            kanObj.uninstall();
        }
    }

    //Added by Huahai
    public int getType() {
      return REPLICA;
    }
}
