/**
 * AbstractClassAnalysis
 * tag: kkvilekal@acm.org 2002/08/20 17:31:56
 * Copyright kkvilekval@acm.org.   See COPYRIGHT for details.
 */
package mokan.prefetch.analysis.classanalysis;

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.LinkedList;

import mokan.prefetch.util.WorkSet;
import mokan.prefetch.util.RTClass;
import mokan.prefetch.util.RTMethod;

import agl2.Graphs;
import agl2.Bigraph;
import agl2.BigraphBuilder; 
import agl2.DirectedPath;

import klogger.Logger;
/**
 *  The <code>AbstractClassAnalysis</code> provides a based
 *  implementation for class analysis.
 *  @version $Revision: 1.2 $, $Date: 2003/01/24 18:23:17 $
 *  @author Kristian Kvilekval (kkvilekval@acm.org)
 */
public abstract class AbstractClassAnalysis 
                      implements ClassAnalysis, SpecialMethods{
    /* Private/Public Variables */
    final static Logger log = 
      Logger.getLogger("prefetch.classanalysis");

    ClassSet          classes;
    Bigraph           classgraph;
    Bigraph           callgraph;
    WorkSet           classworkq;
    WorkSet           classlinkq;
    WorkSet           callworkq;
    BigraphBuilder    classgb;
    BigraphBuilder    callgb;
    HashSet           initialized_classes;
    boolean           pumping = false;
    protected HashMap method2callsites;
    protected RTMethod currentMethod;

    /** Given a message target and a handle, find all methods targets
     *  that might recieve this message.
     */
    abstract RTMethod[] getVirtualTargets(Callsite cs, 
                                          String handle, 
                                          String target);

    /** Visit the method.  This must initialize the set of the
     *  callsite array of method.
     */
    abstract boolean visitMethod (RTMethod m);

    /** Visit a class.. This is called when the class is loaded */
    boolean visitClass (RTClass cl) { return true; }
    
    ///////////////////////////////////////////
    // Class graph

    public final  RTClass loadClass (String clname){
        RTClass cl = classes.findClass (clname);
        if (cl == null) {
            log.debug ("loadClass " + clname);
            cl = classes.loadClass (clname);
            verifyMethods (cl);
            classlinkq.add (cl);
        }
        return cl;
    }
    public void addClassGraph (RTClass clz){
        log.debug ("ClassAnalysis","addClassGraph", "linking " +clz);
        classgb.addVertex (clz);
        String[] interfaces = clz.getInterfaceNames();
        for (int ii=0; ii < interfaces.length; ii++) {
            RTClass inter = loadClass (interfaces[ii]);
            log.debug("ClassAnalysis","classgraph","  +inter "+interfaces[ii]);
            classgb.addEdge (inter, clz, null);
        }
        // Add the super classes 
        String supername = clz.getSuperclassName();
        if (supername != null) {
            RTClass supercl = loadClass (supername);
            log.debug ("ClassAnalysis","addClassGraph", "  +super "+supername);
            classgb.addEdge (supercl, clz, null);
            clz.setSuperClass (supercl);
        }
        visitClass (clz);
    }
    public void verifyMethods (RTClass cl) {
        //log.debug ("ClassAnalysis","verify", cl.toString());
        String[] dependents = classes.getDependents(cl);
        for (int di=0; di < dependents.length; di++)
            classworkq.addLast (dependents[di]);
    }
    // loadAllClasses 
    // 
    public void loadAllClasses () {
        log.debug("ClassAnalysis", "loadAllClasses", "start");
        ArrayList classnames = classes.listClasses ();
        log.debug("ClassAnalysis", "loadAllClasses", "initially loaded="+classnames);

        classworkq.addAll (classnames);
        pumpClassQueue();
        log.debug("ClassAnalysis", "loadAllClasses", "end");
    }
    protected void pumpClassQueue () {
        while (!classworkq.isEmpty () || !classlinkq.isEmpty()){
            while  (!classworkq.isEmpty())
                loadClass ((String)classworkq.removeFirst());
            while  (!classlinkq.isEmpty())
                addClassGraph ((RTClass)classlinkq.removeFirst());
        }
    }


    ///////////////////////////////////////////
    // Call graph

    /** Find the unique base method for the callsite by looking
     *  up through the classgrpah.  The virtual methods are found
     *  by looking down the classgraph.  This routine 
     *  finds the highest overriden method for a particular method
     *  if it exists 
     */
    public RTMethod findMethod (RTClass base, String target) {
        RTMethod found = null;

        while (found == null && base != null) {
            if (base.isSystemClass()) break ;
            found = base.getMethod (target);
            if (found != null && found.isAbstract()){
                found = null;
                break;
            }

            base = base.getSuperClass();
            if (base == classes.getObjectClass() ) break; // Sanity break.
        }
        return found;
    }


    private RTMethod getStaticMethod(Callsite cs,String handle,String target){
	RTClass clz = classes.findClass (handle);
        if (clz== null) {
            log.warn ("classgraph broken");
            return null;
        }

        if (clz.isSystemClass()) {
            cs.setSystemCall (true);
            return null;
        }

	RTMethod m =  clz.getMethod (target);
        if (m.isNative()) {
            cs.setSystemCall(true);
            return null;
        }

        return m;
    }


    // mokan hack to force analysis to look around the 
    // the mokan method invocation based on reflection
    // Change all  asynchronous calls to the synchronous
    // style where eventually the standard method is called.
    //  
    private String demangleMokan(Callsite cs, String handle, String target) {

        int async = target.indexOf("KanAsync(");
        if (async == -1) 
            return target;

        // Find return list
        int retval = target.indexOf(')', async+9);
        // Unmangle future return value
        if (retval == -1) {
            log.severe("missing future");
            return target;
        }
        StringBuffer sb = new StringBuffer(target.length());
        // method name
        sb.append(target.substring(0, async));
        // method args
        sb.append(target.substring(async+8,retval+1));

        // Remove L<package>Future<type>;
        int future = target.indexOf("Future", retval)+6;
        int semi = target.indexOf(';', future);
        // Replace any array specifiers
        while (target.charAt(future) == 'A') {
            future += 1;
            sb.append('[');
        }
        // check for an object type (insert L<type>;)
        if (semi - future > 1) {
            sb.append('L');
            sb.append(target.substring(future).replace('$','/'));
        } else {
            sb.append(target.substring(future,semi));
        }
        String v = sb.toString();
        log.debug("Mokan demangle on " + cs +':'+handle);
        log.debug("Mokan demangle " + target +"->" + v);
        return v;
        

    }


    /** Check for special handling of a method call. 
     *  This can occur when Class.newInstance  or Thread.start 
     *  is called and the stadard flow analysis will lose 
     *  track of what is actually called.
     */
    RTMethod[] visitSpecialCall (Callsite cs, String handle, String target) {
        //  HACK threadstart eventually calls run.
        //   Switch a call to start run method on a runnable object */
        if (target.equals (THREADSTART))
            target = THREADRUN;
        else if (target.equals(NEWINSTANCE)){
            target = CLASSINIT;
            handle = "java.lang.Object";
            log.warn("called Class.newInstance in "+cs.caller.getFullname());
        } else {
            //  MOKAN
            //  Mokan code passes through most calls through a reflection
            //  interface.   This code by passes the reflection to the
            //  final call destination.
            target = demangleMokan(cs, handle, target);
        }
        return getVirtualTargets (cs, handle, target);
    }

    protected RTMethod[] getTargets (Callsite cs) {
        if (cs.callees == null) {
            String target = cs.target;
            String handle = cs.handleclass;

            log.debug ("getTargets: " 
                       + (cs.isVirtual()? 
                          (cs.isInterface()? "interface ":"class ") : 
                          "static ") + handle +'.'+ target);

            
            RTClass base = classes.findClass (handle);
            if (base == null) {
                log.severe ("GetTargets: call to unloaded class:" + handle);
                //addClass(handle);
                //base = classes.getClass (handle);
            }
            if (classgb.hasVertex (base) == false) {
                log.severe ("getTargets: classgraph broken  "+handle);
                log.severe ("getTargets: called from  "+ cs.caller);
            }
            if (base.isSystemClass ()) cs.setSystemCall(true);

            if (cs.isVirtual ()) {
                RTMethod[] callees=visitSpecialCall(cs, handle, target);
                if (callees == null) 
                    callees = getVirtualTargets (cs, handle, target);
                cs.callees = callees;
            } else {
                RTMethod callee = getStaticMethod (cs, handle,target); 
                if (callee != null)
                    cs.callees =  new RTMethod[] { callee };
                else 
                    cs.callees = new RTMethod[] { };
            }
            log.debug ("getTargets: found " + cs.callees.length +':'+ Arrays.asList (cs.callees));
        }
        return cs.callees;
    }


    /** Initialize all classes mentioned in this class
     *  of this method.  This is equivalnet to loading
     *  the class on it's first call 
     */
    private void addClassInitializers (RTMethod caller, RTClass cl) {
        if (cl.isSystemClass () ) return;

        if (!initialized_classes.contains (cl)){
            initialized_classes.add (cl);

            String[] depends = classes.getDependents (cl);
            for (int di=0; di < depends.length; di++) {
                RTClass dc = classes.findClass (depends[di]);
                if (dc == null) {
                    log.severe ("ClassAnalysis", "addClassInitializer",
                                "missing " + depends[di]);
                    continue;
                }
                addClassInitializers (caller, dc);
            }
            RTMethod clinit = cl.getMethod (STATICCLASSINIT);
            if (clinit != null)
                addCallGraph (caller, clinit);
        }
    }

    private void analyzeMethod (RTMethod method) {

        if (method.isSystemMethod ()) return;
        if (callgb.hasVertex(method)) return;
        
        callgb.addVertex (method);

        RTMethod lastMethod = currentMethod;
        currentMethod = method;

        log.debug ("ClassAnalsis","analyzeMethod", method.toString());

        visitMethod(method);
        Callsite[] sites = (Callsite[])method2callsites.get (method);

        // For all the callsites in this method
        for (int si = 0; si < sites.length; si++) {
            Callsite callsite = sites[si];
            RTMethod[] callees = getTargets (callsite);
            for (int ci = 0; ci < callees.length; ci++){
                addCallGraph (method, callees[ci]);
            }

            if (callees.length < 1 && !callsite.isSystemCall ()) {
                log.severe ("callsite:" + si + " in " + 
                            method + " has no targets:"+ callsite );
            }
        }
        currentMethod = lastMethod;
    }
    private void addCallGraph (RTMethod caller, RTMethod callee) {
        log.debug ("callgraph " + caller+ " -> " + callee);
        if (callee.isSystemMethod ()) return;
        if (!callgb.hasVertex (callee)) {
            // Add any class initializers before the method
            // is actually called
            RTClass calleeclass = callee.getDefiningClass();
            addClassInitializers (caller, calleeclass);

            analyzeMethod (callee);
        }
        callgb.addEdge (caller, callee, null);
    }

    ///////////////////////////////////////////////////////////
    // Main entry points to classanalysis

    /** Reinitialize all structure needed to build the graphs */
    void initializeGraphs (String classpath, String apppath) {
        classworkq = new WorkSet();
        classlinkq = new WorkSet();
        classgb = new BigraphBuilder ();
        callgb = new BigraphBuilder ();
        initialized_classes = new HashSet();
        classes = new ClassSet (classpath, apppath);
    }

    /** Build the classgraph and the callgraph.
     *  @param classpath A single top level directory representing the
     *  class path entry for this application.
     *  @param apppath The actual top level directory of classes in
     *  the application.  Used to limit the amount of classes that are
     *  instantiated during RTA.
     *  @param mainclassname The full path to the mainclass.
     */
    protected void buildGraphs (String classpath, 
                                String apppath,
                                String mainclassname){
        log.entering ("ClassAnalysis", "buildGraphs");

        initializeGraphs (classpath, apppath);

        currentMethod = new RTMethod ("<premain>");
        method2callsites = new HashMap();

        loadClass("java.lang.Object");
        loadClass("java.lang.Class");
        RTClass mainclass = loadClass(mainclassname);
        loadAllClasses();

        RTMethod mainmethod;
        mainmethod = mainclass.getMethod(MOKANMAIN);
        
        if (mainmethod == null) 
            mainmethod = mainclass.getMethod (MAIN);

        log.debug ("buildGraphs -> " + mainmethod);
        addCallGraph (currentMethod, mainmethod);

        classgb.setRoot (classes.findClass ("java.lang.Object"));
        classgraph = classgb.getBigraph();

        callgb.setRoot (mainmethod);
        callgraph = callgb.getBigraph();

        log.info ("main calls ->"+ 
                  callgraph.getAdjacentVertices (mainmethod).size());

        Callsite[] mains= (Callsite[])method2callsites.get(mainmethod);
        for (int mi=0; mi<mains.length; mi++){
            Callsite callsite = mains[mi];
            log.info ("   callsite:" +mi + " system:" + 
                      RTClass.isSystemClass (callsite.handleclass));
            for (int ti=0; ti < callsite.callees.length; ti++){
                log.info ("      "+ callsite.callees[ti]);
            }
        }
        log.exiting ("ClassAnalysis", "buildGraphs");
    }

    //////////////////////////////////////////////////
    //  Help routines for using the graph structures

    public RTClass getClass (String name) {
        return classes.findClass (name);
    }
    public Collection getLoadedClasses (){
        return classes.values ();
    }
    public boolean isAncestorClass(RTClass ancestor, RTClass descendant){
        return new DirectedPath(classgraph).exists (ancestor, descendant);
    }

    /** Return the methods called by m at instr*/
    public RTMethod[] getTargetSet(RTMethod m, int instr) {
        Callsite[] callsites = (Callsite[])method2callsites.get (m);

        int l = 0;
        int u = callsites.length-1;
        int i;

        while (l<=u) {
            i = (u + l)/2;
            if (callsites[i].instr == instr) 
                return callsites[i].callees;
            if (callsites[i].instr > instr) 
                u = i-1;
            else 
                l = i+1;
        }
        return null;
    }
    public Bigraph getStaticClassGraph () {
        return classgraph;
    }
    public Bigraph getStaticCallGraph (){
        return callgraph;
    }
}
