/*
 * 
 *
 * tag: kkvilekval@acm.org Wed Nov 14 10:01:02 2001 (ShapeAnalysis)
 *
 */


package mokan.prefetch.analysis.shapeanalysis;

import java.io.Serializable;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Arrays;

import agl2.DirectedGraph;
import agl2.Bigraph;
import agl2.alg.TopologicalSort;
import agl2.alg.StrongComponents;

//import mokan.prefetch.analysis.Precache;
//import pretime.classgraph.ClassGraph;
//import pretime.callgraph.CallGraph;
import mokan.prefetch.util.RTMethod;
import mokan.prefetch.util.RTClass;
import klogger.Logger;

import mokan.prefetch.analysis.classanalysis.ClassAnalysis;

/**
 *  The <code>AccessAnalysis</code>  perfroms an access analyss 
 *  on a method.  An access analysis determines the the set
 *  of variables accessed by the method and it's submethods.
 *  For example:   
 *    C.method (C self, A a, B b) {
 *       self.f  = a.f        { self, access (r, a.f, 1) } 
 *       B.method (self.f);   { all var used by method on self.f }
 *       local c = a.f.x;     { a.f }
 *       return c.y           { a.f.y }
 *   Note that all accesses are based in terms of the 
 *   original parameters.
 *    
 *  @version $Revision: 1.1.1.1 $, $Date: 2002/12/19 23:48:36 $
 *  @author Kristian Kvilekval
 */
public class ShapeAnalysis implements Analysis, Serializable {

    private final static Logger log = Logger.getLogger ("prefetch.analysis.ShapeAnalysis");

    /** This class is a forward block visitor.
     *  It visits each block and creates 
     */

    /* Private/Public Variables */
//    private ClassGraph classgraph;
//    private CallGraph  callgraph;
    private MethodTable methods;
    private ClassAnalysis classanalysis;

    private transient ShapePool  currentPool;
    private transient Set          currentSCC;
    private transient BlockState currentBlock;
    private transient StackVisitor stackVisitor;
    private transient AccessFlowVisitor flowVisitor;

    private int       component_count, method_count;

    private static void dumpBlockGraph (
                            int level,
                            RTMethod m, 
                            BlockGraph bg, 
                            List toporder){
        log.log(level, RTMethod.dumpListing (m));
        log.log(level, bg.toString());
        
        StringBuffer bf = new StringBuffer();
        Iterator ti = toporder.iterator();
        while (ti.hasNext()) {
            bf.append ((BlockNode)ti.next());
        }
        log.log (level, "Blocks in top order" + bf);
    }

    public void visitBlocks (RTMethod m) {

        BlockGraph bg = new BlockGraph (m);
        List toporder = bg.getTopologicalOrder();
        LinkedList worklist = new LinkedList (toporder);
        
        if (log.isEnabled(Logger.DEBUG)) {
            dumpBlockGraph (Logger.DEBUG, m, bg, toporder);
        }

        flowVisitor.clearState();

        try {
            // We visit each block at least once in topological order
            // Internally we vill visit it again when it modifies the
            // state.
            while (!worklist.isEmpty()) {
                BlockNode b = (BlockNode)worklist.removeFirst();
                flowVisitor.visitBlock (b, worklist);
            }
        } catch (RuntimeException e) {
            dumpBlockGraph (Logger.SEVERE, m, bg, toporder);
        }
    }


    public void visitMethod (RTMethod m) {
        log.entering ("ShapeAnalysis", "visitMethod");

        // Don't bother with native, abstract, and system methods
        if (m.isNative ()) return ;
        if (m.isAbstract()) return ;
        if (m.isSystemMethod()) return;

        log.trace ("ShapeAnalysis", "visitMethod", "Analyzing " + m);

        stackVisitor.initMethod (m);
        visitBlocks (m);
        stackVisitor.finalMethod (m);

        method_count++;

        if (log.isEnabled (Logger.DEBUG)) {
            MethodState ms = methods.getState (m);
            log.trace ("ShapeAnalysis", "visitMethod" , "final->"+ms);
            ms.logShapes(log, Logger.DEBUG);
        }
        log.exiting ("ShapeAnalysis", "visitMethod");
    }

    /** Analyze a strongly connected component 
     *  The set argument must be of type Set<RTMethod>
     */
     public void visitComponent (Set c) {
         // Grab each method analyze it and throw the result
        // in a common bag.
        log.trace ("ShapeAnalysis","visitComponent", "visiting SCC "+ c);
        currentSCC = c; 

        currentPool = new ShapePool();

        //Iterator mi = callgraph.getMethods (c);
        Iterator mi = c.iterator();
        while (mi.hasNext()) {
            stackVisitor.prepareMethod ((RTMethod)mi.next(), currentPool);
        }            

        mi = c.iterator();
        while (mi.hasNext()) {
            visitMethod ((RTMethod)mi.next());
        }            

        component_count++;
    }

    public void visitProgram() {
        log.entering ("ShapeAnalysis", "visitProgram");
        long begin = System.currentTimeMillis();

        //REM Iterator topiter = callgraph.getTopologicalSCCIterator();


        log.info ("Call Graph:" + classanalysis.getStaticCallGraph().getVertices().size()+" vertices" );
        
        DirectedGraph components =
            StrongComponents.getComponents (classanalysis.getStaticCallGraph(),
                                            null);

        log.info ("Component Graph:" + components.getVertices().size()+" vertices" );
        Iterator topiter = 
            TopologicalSort.nodeIterator (components, null);
                                     

        while (topiter.hasNext()) {
            visitComponent ((Set)topiter.next());
        }
        //log.message("AccessAnalysis","visitProgram","time="+(System.currentTimeMillis() - begin));


        log.info ("ShapeAnanlysis","visitProgram", "analyzed " 
                  + component_count + " components " 
                  + method_count + " methods");
        log.exiting ("ShapeAnalysis", "visitProgram");
    }

    /**
     *  The default constructor
     */
    public ShapeAnalysis (ClassAnalysis ca, String main) {
        //REM classgraph=hier;
        //REM callgraph=calls;
        classanalysis = ca;

        //KKNEW
        methods = new MethodTable();
        stackVisitor = new StackVisitor (methods, ca);
        flowVisitor = new AccessFlowVisitor (stackVisitor);
        //KKNEW
    }

    

/*
    public CallGraph getCallGraph () {
        return callgraph;
    }
    public ClassGraph getClassGraph () {
        return classgraph;
    }
*/
    /** Return the MethodTable of the analyzed program */
    public MethodTable getMethods () {
        return methods;
    }

};
