/*
 * tag: kkvilekval@acm.org Thu Nov 15 17:07:05 2001 (RTClass)
 */


package mokan.prefetch.util;

import java.io.Serializable;
import java.io.File;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

import de.fub.bytecode.Repository;
import de.fub.bytecode.classfile.JavaClass;
import de.fub.bytecode.classfile.Method;
import de.fub.bytecode.classfile.ConstantPool;
import de.fub.bytecode.generic.ConstantPoolGen;
import de.fub.bytecode.classfile.ClassParser;


/**
 *  The <code>RTClass</code> is used to 
 *  represent a runtime class.  Also is covers most 
 *  JavaClass functions.
 *  @version $Revision: 1.2 $, $Date: 2003/01/24 18:23:17 $
 *  @author Kristian Kvilekval
 */
public class RTClass implements Comparable, Serializable {

    public final static int NONE = 0;
    public final static int JAVAX = 1;
    public final static int JAVA = 2;

    /** Class prefix exlcusion  .. only classes used in tests*/
    private final static String[] nonePrefix = {
        "mokan.prefetch.",
        "mokan.util.",
        "mokan.obj.",
        "mokan.transact.",
        "mokan.thread.",
        "mokan.nameservice.",
        "mokan.app.",
        "mokan.comm.",
        "mokan.debug.",
        "mokan.KanSystem",
        "mokan.Kan",
        "klogger.",
        "spec.harness.",
        "spec.io."
    };

    /** Most java, sun packages are eliminated */
    private final static String[] unsafePrefix = { 
        "java.awt.", 
        "java.applet.",
        "java.beans.",
        "java.io.",
        "java.im.",
        "java.lang.",
        "java.net.",
        "java.rmi.",
        "java.security.",
        "java.sql.",
        "java.text.",
        "java.util.jar.",
        "java.util.zip.",
        "javax.", 
        "sun." ,
        "com.sun.",
        "mokan.prefetch.",
        "mokan.util.",
        "mokan.obj.",
        "mokan.transact.",
        "mokan.thread.",
        "mokan.nameservice.",
        "mokan.app.",
        "mokan.comm.",
        "mokan.debug.",
        "mokan.KanSystem",
        "mokan.Kan",
        "klogger.",
        "agl.",
        "agl2.",
        "de.",
        "gnu.",
        "spec.harness.",
        "spec.io."
    };
    

    /** All java, sun prefixes are eliminated */
    private final static String[] allSystemPrefix = { 
        "java.", 
        "javax.", 
        "sun." ,
        "com.",
        "org.",
        "mokan.prefetch.",
        "mokan.util.",
        "mokan.obj.",
        "mokan.transact.",
        "mokan.thread.",
        "mokan.nameservice.",
        "mokan.app.",
        "mokan.comm.",
        "mokan.debug.",
        "mokan.KanSystem",
        "mokan.Kan",
        "klogger.",
        "agl.",
        "agl2.",
        "de.",
        "gnu.",
        "spec.harness.",
        "spec.io."
    };



    private static String[]  systemPrefixes = allSystemPrefix;

    public static void setSystemClassLevel (int level) {
        switch (level) {
        case NONE: systemPrefixes =nonePrefix; break;
        case JAVAX: systemPrefixes = unsafePrefix; break;
        default: systemPrefixes = allSystemPrefix; break;
        }

        System.out.println ("Setting system classes to: " + Arrays.asList(systemPrefixes));
    }

    transient JavaClass jc;
    transient ConstantPoolGen cpg;    
    transient RTClass parent;
    transient RTMethod[] methods;


    /* Private/Public Variables */

    /**
     *  The default constructor
     */
    public RTClass (String name) {
        jc = Repository.lookupClass (name);

        Assert.assertion(jc != null, "Couldn't find class " + name);
    }

    public RTClass (JavaClass jc) {
            this.jc = jc;
    }

    public void setSuperClass (RTClass p) {
        parent = p;
    }

    /** Create an empty RTClass suitable for use with the parse 
     ** function 
     */
    public RTClass (File classfile) {
        try { 
            jc = new ClassParser (classfile.getCanonicalPath()).parse();
        } catch (java.io.IOException io) {
            System.out.println ("Problem reading : "+classfile);
        }
    }

    /** JavaClass Accessor */
    public RTClass getSuperClass () {
        return parent;
    }
    public boolean isInterface() {
        return jc.isInterface();
    }

    /** JavaClass Accessor */
    public JavaClass getJavaClass () {
        return jc;
    }
    /** JavaClass Accessor */
    public String getClassName () {
        return jc.getClassName();
    }
    /** JavaClass Accessor */
    public String[] getInterfaceNames () {
        return jc.getInterfaceNames();
    }
    /** JavaClass Accessor */
    public String getSuperclassName () {
        return jc.getSuperclassName();
    }


    /** JavaClass Accessor */
    public ConstantPool getConstantPool() {
        return jc.getConstantPool();
    }
    /** JavaClass Accessor */
    public ConstantPoolGen getConstantPoolGen() {
        if (cpg == null) 
            cpg = new ConstantPoolGen(jc.getConstantPool());
        return cpg;
    }

    public boolean isSystemClass () {
        for (int si=0; si < systemPrefixes.length; si++) 
            if (getClassName().startsWith (systemPrefixes[si])){
                return true;
            }
        return false;
    }
    public static boolean isSystemClass (String name) {
        for (int si=0; si < systemPrefixes.length; si++) 
            if (name.startsWith (systemPrefixes[si])){
                return true;
            }
        return false;
    }

    public RTMethod[] getMethods () {
        if (methods == null) {
            Method[] ms = jc.getMethods();
            methods = new RTMethod[ms.length];
            for (int mi = 0; mi < ms.length; mi++) {
                methods[mi] = new RTMethod(this, ms[mi]);
            }
        }
        return methods;
    }

    public void setMethods(RTMethod[] rtmethods) {
        Method[] jcmethods = jc.getMethods();
        this.methods = rtmethods;
        for (int mi = 0; mi < rtmethods.length; mi++) {
            jcmethods[mi] = rtmethods[mi].getMethod();
        }
    }

    /** Get a method by it's signature 
     *  @param signature soemthing like "main([Ljava/lang/String;)V"
     */
    public RTMethod getMethod (String signature) {
        //log.info ("Looking for " + signature + " in " + getClassName());

        RTMethod[] ms = getMethods();
        for (int mi=0; mi < ms.length; mi++) {
            //System.out.println ("  sig = " + ms[mi].getSignature());
            if (signature.equals(ms[mi].getSignature()))
                return ms[mi];
        }
        return null;
    }
    public int compareTo (Object o) {
        return this.getClassName().compareTo (((RTClass)o).getClassName());
    }

    /**
     *  @return a string representation of this <code>RTClass</code> 
     */
    public String toString () {
        return jc.getClassName();
    }
};
