/*
 * @(#)$Id: RMIWorkLoad.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 * Copyright (c)1999 Regents of the University of California
 *
 */

package mokan.test.rmiTest;

import java.net.InetAddress;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.registry.LocateRegistry;
import java.util.Vector;

/**
 * Set up the system to run tests.  The actual tests are in the
 * <code>RMIDriver</code> class.  This class just creates the distributed
 * objects and then launches one <code>RMIDriver</code> per node.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public final class RMIWorkLoad {

    /**
     * Type of call to make
     */
    public static final int NULL_ARG     = 0;
    public static final int INT_ARG      = 1;
    public static final int INT_ARGS     = 2;
    public static final int INT_ARRAY    = 3;
    public static final int NODE_ARG     = 4;
    public static final int TREE_ARG     = 5;
    public static final int NODE_ARG_INT = 6;
    public static final int TREE_ARG_INT = 7;
    public static final int KAN_ARGS     = 8;

    /**
     * The local RMI server port
     */
    private static final int SERVER_PORT = 5009;

    public static void main(String[] argv) throws Exception {
	// Parse the arguments
	int numObjects = Integer.parseInt(argv[0]);
	int callType   = Integer.parseInt(argv[1]);   // Type of method call
	int numCalls   = Integer.parseInt(argv[2]);

	// Install the security manager, if needed
        if (System.getSecurityManager() == null) { 
	    System.setSecurityManager(new RMISecurityManager()); 
	}

	// Get all the hosts involved in this test
	InetAddress local = InetAddress.getLocalHost();
	int localNumber = -1;
	int numNodes = argv.length - 3;
	InetAddress[] nodes = new InetAddress[numNodes];
	int j = 0;
	for (int i = 3; i < argv.length; i++, j++) {
	    nodes[j] = InetAddress.getByName(argv[i]);
	    if (local.equals(nodes[j]))
		localNumber = j;
	}

	// Create the registry
	LocateRegistry.createRegistry(SERVER_PORT);

	// Create the local objects
	int numLocals = numObjects / numNodes +
	    (localNumber < numObjects % numNodes ? 1 : 0);
	RMITestObject[] locals = new RMITestObject[numLocals];
	String hostAddr = local.getHostAddress();
	for (int i = 0; i < numLocals; i++)
	    locals[i] = new RMITestObject(numNodes, hostAddr);

	// We now pause for station identification
	Thread.sleep(2000);

	// Get refs to all other remote objects via remote lookup
	RMITestObjectInterface[] handles =
	    new RMITestObjectInterface[numObjects - numLocals];
	int i = 0;
	for (int node = 0; node < nodes.length; node++) {
	    if (node != localNumber) {
		int numObjs = numObjects / numNodes +
		    (node < numObjects % numNodes ? 1 : 0);
		String hostName = nodes[node].getHostName();
		for (j = 0; j < numObjs; j++, i++) {
		    handles[i] = getHandle(hostName, j);
		}
	    }
	}

        RMIDriver driver = new RMIDriver(handles);
	long[] results = driver.startSim(callType, numCalls);

	// Pause to let everyone else finish, then signal completion
	Thread.sleep(2000);
	for (i = 0; i < handles.length; i++) {
	    handles[i].done();
	}
	for (i = 0; i < numLocals; i++) {
	    locals[i].done();
	}
	for (i = 0; i < handles.length; i++) {
	    try {
		handles[i].waitUntilDone();
	    } catch (RemoteException remoteEx) {
		// That counts as done!
	    }
	}
	for (i = 0; i < numLocals; i++) {
	    locals[i].waitUntilDone();
	}

	// Compute the raw mean
	double average = 0.0;
	for (i = 0; i < numCalls; i++) {
	    average += results[i];
	}
	average /= numCalls;

	// Compute the standard deviation
	double stdDev = 0.0;
	for (i = 0; i < numCalls; i++) {
	    double diff = results[i] - average;
	    stdDev += diff * diff;
	}
	stdDev = Math.sqrt(stdDev / numCalls);

	// Compute the adjusted mean
	double adjMean = 0.0;
	int counted = 0;
	double comparator = stdDev * 2;
	for (i = 0; i < numCalls; i++) {
	    double num = Math.abs(results[i] - average);
	    if (num < comparator) {
		adjMean += results[i];
		counted++;
	    }
	}
	adjMean /= counted;
	final double MILL = 1000000.0;
	System.out.println("Mean          = " + (average / MILL));
	System.out.println("Standard Dev. = " + (stdDev  / MILL));
	System.out.println("Adjusted mean = " + (adjMean / MILL));
	System.exit(0);
    }

    private static RMITestObjectInterface getHandle(String hostName, int num)
	throws Exception {

	String name = "rmi://" + hostName + ":" + SERVER_PORT +
	    "/RMITestObject" + num;
	for (int i = 0; ; i++) {
	    try {
		return (RMITestObjectInterface)Naming.lookup(name);
	    } catch (Exception ex) {
		System.out.println("Failed to connect: " + i);
		if (i < 25)
		    Thread.currentThread().sleep(3000);
		else
		    throw ex;
	    }
	}
    }
}
