/*
 * @(#)$Id: Deserialization.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 * Copyright (c)1999 Regents of the University of California
 */

package mokan.test.speed;

import java.io.*;
import java.util.Date;
import java.util.Random;

/**
 * A test of the speed of Java deserialization.  We use a
 * <code>StringReader</code> from a buffer so that the reading itself consists
 * of just reads from memory.  That way, we avoid introducing extra costs into
 * the system.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public final class Deserialization {

    private final static Random rand = new Random();

    public static void main(String[] argv)
	throws ClassNotFoundException, FileNotFoundException, IOException {

	int loops = Integer.parseInt(argv[0]);
	deserializeInteger(loops);
	System.gc();	// Clean up as much memory as possible
	deserializeDate(loops);
	System.gc();	// Clean up as much memory as possible
	deserializeArrays(loops);
    }

    private static void deserializeInteger(int loops)
	throws ClassNotFoundException, IOException {

	// On x86 Solaris, an Integer takes 10 bytes to serialize, and there
	// are 71 bytes of overhead to serialize the class.  I think.
	int bufsize = loops * 10 + 71;
	byte[] buffer = new byte[bufsize];
	FileInputStream fil = new FileInputStream("serialInt");
	fil.read(buffer, 0, bufsize);
	fil.close();
	fil = null;  // Feed the garbage collector

	ByteArrayInputStream stream = new ByteArrayInputStream(buffer);
	ObjectInputStream in = new ObjectInputStream(stream);

	new Integer(42);	// Skip the class loading cost
	long startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    Integer myInt = (Integer)in.readObject();
	}
	long endTime = System.currentTimeMillis();
	long totalTime = endTime - startTime;

	startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    // Empty loop
	}
	endTime = System.currentTimeMillis();
	long adjustedTime = totalTime - endTime + startTime;
	System.out.println("Adjusted time: " + adjustedTime);
	System.out.println("Milliseconds per Integer read: " +
			   ((double)adjustedTime / loops));
	in.close();
	stream.close();
    }

    private static void deserializeDate(int loops)
	throws ClassNotFoundException, IOException {

	// On x86 Solaris, a Date takes 17 bytes to serialize, and there are
	// 29 bytes of overhead to serialize the class.  I think.
	int bufsize = loops * 17 + 29;
	byte[] buffer = new byte[bufsize];
	FileInputStream fil = new FileInputStream("serialDate");
	fil.read(buffer, 0, bufsize);
	fil.close();
	fil = null;  // Feed the garbage collector

	ByteArrayInputStream stream = new ByteArrayInputStream(buffer);
	ObjectInputStream in = new ObjectInputStream(stream);

	new Date();	// Skip the class loading cost
	long startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    Date myDate = (Date)in.readObject();
	}
	long endTime = System.currentTimeMillis();
	long totalTime = endTime - startTime;

	startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    // Empty loop
	}
	endTime = System.currentTimeMillis();
	long adjustedTime = totalTime - endTime + startTime;
	System.out.println("Adjusted time: " + adjustedTime);
	System.out.println("Milliseconds per Date read: " +
			   ((double)adjustedTime / loops));
	in.close();
	stream.close();
    }

    private static void deserializeArrays(int loops)
	throws ClassNotFoundException, IOException {

	// On x86 Solaris, an array of 300 bytes takes 620 bytes to serialize,
	// and there are 17 bytes of overhead to serialize the class.  I
	// think.
	int bufsize = loops * 620 + 17;
	byte[] buffer = new byte[bufsize];
	FileInputStream fil = new FileInputStream("serialArray");
	fil.read(buffer, 0, bufsize);
	fil.close();
	fil = null;  // Feed the garbage collector

	ByteArrayInputStream stream = new ByteArrayInputStream(buffer);
	ObjectInputStream in = new ObjectInputStream(stream);

	long startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    byte[] array = (byte[])in.readObject();
	}
	long endTime = System.currentTimeMillis();
	long totalTime = endTime - startTime;

	startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    // Empty loop
	}
	endTime = System.currentTimeMillis();
	long adjustedTime = totalTime - endTime + startTime;
	System.out.println("Adjusted time: " + adjustedTime);
	System.out.println("Milliseconds per array read: " +
			   ((double)adjustedTime / loops));
	in.close();
	stream.close();
    }
}
