/*
 * @(#)$Id: Serialization.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 * Copyright (c)1999 Regents of the University of California
 */

package mokan.test.speed;

import java.io.*;
import java.util.Date;
import java.util.Random;

/**
 * A test of the speed of Java serialization.  We use a
 * <code>StringWriter</code> into a suitably large buffer so that the writing
 * itself consists of just writes into memory.  That way, we avoid introducing 
 * extra costs into the system.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public final class Serialization {

    private final static Random rand = new Random();

    public static void main(String[] argv) throws IOException {
	int loops = Integer.parseInt(argv[0]);
	serializeInteger(loops);
	serializeDate(loops);
	serializeArrays(loops);
    }

    private static void serializeInteger(int loops) throws IOException {
	// On x86 Solaris, an Integer takes 10 bytes to serialize, and there
	// are 71 bytes of overhead to serialize the class.  I think.
	int bufsize = loops * 10 + 80;
	ByteArrayOutputStream buffer = new ByteArrayOutputStream(bufsize);
	ObjectOutputStream out = new ObjectOutputStream(buffer);

	new Integer(42);	// Skip the class loading cost
	long startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    Integer myInt = new Integer(rand.nextInt());
	    out.writeObject(myInt);
	}
	out.flush();
	long endTime = System.currentTimeMillis();
	long totalTime = endTime - startTime;

	System.gc();	// Clean up as much memory as possible
	startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    Integer myInt = new Integer(rand.nextInt());
	}
	endTime = System.currentTimeMillis();
	long adjustedTime = totalTime - endTime + startTime;
	System.out.println("Adjusted time: " + adjustedTime);
	System.out.println("Milliseconds per Integer write: " +
			   ((double)adjustedTime / loops));
	out.close();
	System.gc();	// Clean up as much memory as possible
    }

    private static void serializeDate(int loops) throws IOException {
	// On x86 Solaris, a Date takes 17 bytes to serialize, and there are
	// 29 bytes of overhead to serialize the class.  I think.
	int bufsize = loops * 17 + 30;
	ByteArrayOutputStream buffer = new ByteArrayOutputStream(bufsize);
	ObjectOutputStream out = new ObjectOutputStream(buffer);

	new Date();	// Skip the class loading cost
	long startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    Date d = new Date();
	    out.writeObject(d);
	}
	out.flush();
	long endTime = System.currentTimeMillis();
	long totalTime = endTime - startTime;
	System.out.println("Final size: " + buffer.size());
	System.out.println("Total time: " + totalTime);
	System.gc();	// Clean up as much memory as possible

	startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    Date d = new Date();
	}
	endTime = System.currentTimeMillis();
	long adjustedTime = totalTime - endTime + startTime;
	System.out.println("Adjusted time: " + adjustedTime);
	System.out.println("Milliseconds per object write: " +
			   ((double)adjustedTime / loops));
	out.close();
	System.gc();	// Clean up as much memory as possible
    }

    private static void serializeArrays(int loops) throws IOException {
	// On x86 Solaris, an array of 300 bytes takes 620 bytes to serialize,
	// and there are 17 bytes of overhead to serialize the class.  I
	// think.
	int bufsize = loops * 620 + 20;
	ByteArrayOutputStream buffer = new ByteArrayOutputStream(bufsize);
	ObjectOutputStream out = new ObjectOutputStream(buffer);

	long startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    byte[] b = new byte[300];
	    out.writeObject(b);
	}
	out.flush();
	long endTime = System.currentTimeMillis();
	long totalTime = endTime - startTime;
	System.out.println("Final size: " + buffer.size());
	System.out.println("Total time: " + totalTime);
	System.gc();	// Clean up as much memory as possible

	startTime = System.currentTimeMillis();
	for (int i = 0; i < loops; i++) {
	    byte[] b = new byte[300];
	}
	endTime = System.currentTimeMillis();
	long adjustedTime = totalTime - endTime + startTime;
	System.out.println("Adjusted time: " + adjustedTime);
	System.out.println("Milliseconds per object write: " +
			   ((double)adjustedTime / loops));
	out.close();
	System.gc();	// Clean up as much memory as possible
    }
}
