/*
 * @(#)$Id: Test.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 * Copyright (c)1999 Regents of the University of California
 *
 */

package mokan.test.trans;

import java.io.*;
import java.util.Random;
import mokan.KanSystem;
import mokan.app.Application;
import mokan.obj.ObjectID;
import mokan.util.Timer;

/**
 * Test the transaction capabilities of the Kan system
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public final class Test {

    // Type of transaction
    public static final int NONBLOCKING = 0;
    public static final int BLOCKING    = 1;
    public static final int NESTED      = 2;

    // Type of access
    public static final int READ        = 0;
    public static final int WRITE       = 1;

    static final Random rand = new Random();

    public static void main(final Application app, final String argv[]) {

	// Parse the arguments
	final int type  = Integer.parseInt(argv[0]);	// Type of transaction
	final int acc   = Integer.parseInt(argv[1]);	// Type of access
	final int depth = Integer.parseInt(argv[2]);	// Depth of trans.
	final int nobjs = Integer.parseInt(argv[3]);	// Objects per node
	final int loops = Integer.parseInt(argv[4]);	// Number of loops

	// Get the number of nodes involved in the test
	final int numNodes = app.numNodes();

	try {
	    // Create the Targets on all nodes
	    final ObjectID[] objs = new ObjectID[nobjs * numNodes];
	    for (int node = 0; node < numNodes; node++) {
		for (int i = 0; i < nobjs; i++) {
		    objs[node * nobjs + i] =
			KanSystem.newKanObject(Target.class, null, true, node);
		}
	    }

	    short meth;
	    Object[] args;
	    final long[] results = new long[loops];
	    switch(type) {
	    case NONBLOCKING:
		meth = (short)acc;
		args = new Object[acc];
		break;
	    case BLOCKING:
		meth = (short)(acc + 2);
		args = new Object[acc];
		break;
	    case NESTED:
		if (depth < 2) {
		    meth = (short)(acc + 6);
		    args = new Object[acc + 1];
		} else {
		    meth = (short)(acc + 4);
		    args = new Object[acc + 2];
		}
		break;
	    default:
		args = null;	// The stupid compiler insists on this
		meth = 0;	// And this
		System.out.println("Say what?");
		System.exit(1);
	    }

	    // Gentlemen, start your engines!
	    for (int i = 0; i < loops; i++) {
		final ObjectID nextObj = randomObj(objs);
		switch(type) {
		case NONBLOCKING:
		case BLOCKING:
		    if (acc == WRITE) {
			args[0] = new Integer(rand.nextInt());
			final long startTime = Timer.currentTimeNanos();
			KanSystem.invokeMethod(nextObj, meth, args).join();
			results[i] = Timer.currentTimeNanos() - startTime;
		    } else {
			final long startTime = Timer.currentTimeNanos();
			int result =
			    KanSystem.invokeMethod(nextObj, meth,
						   args).joinInt();
			results[i] = Timer.currentTimeNanos() - startTime;
		    }
		    break;
		case NESTED:
		    if (depth >= 2) {
			final ObjectID[] moreObjs = new ObjectID[nobjs - 1];
			for (int j = 0; j < nobjs - 1; j++)
			    moreObjs[j] = randomObj(objs);
			if (acc == WRITE) {
			    args[0] = new Integer(rand.nextInt());
			    args[1] = moreObjs;
			    args[2] = new Integer(0);
			    final long startTime = Timer.currentTimeNanos();
			    KanSystem.invokeMethod(nextObj, meth, args).join();
			    results[i] = Timer.currentTimeNanos() - startTime;
			} else {
			    args[0] = moreObjs;
			    args[1] = new Integer(0);
			    final long startTime = Timer.currentTimeNanos();
			    int result =
				KanSystem.invokeMethod(nextObj, meth,
						       args).joinInt();
			    results[i] = Timer.currentTimeNanos() - startTime;
			}
		    } else {
			final ObjectID[] moreObjs = new ObjectID[nobjs - 1];
			for (int j = 0; j < nobjs - 1; j++)
			    moreObjs[j] = randomObj(objs);
			if (acc == WRITE) {
			    args[0] = new Integer(rand.nextInt());
			    args[1] = moreObjs;
			    final long startTime = Timer.currentTimeNanos();
			    KanSystem.invokeMethod(nextObj, meth, args).join();
			    results[i] = Timer.currentTimeNanos() - startTime;
			} else {
			    args[0] = moreObjs;
			    final long startTime = Timer.currentTimeNanos();
			    int result =
				KanSystem.invokeMethod(nextObj, meth,
						       args).joinInt();
			    results[i] = Timer.currentTimeNanos() - startTime;
			}
		    }
		    break;
		}
	    }

	    // Now we're all done, so go ahead and compute intensively
	    // Compute the raw mean
	    double average = 0.0;
	    for (int i = 0; i < loops; i++) {
		average += results[i];
	    }
	    average /= loops;

	    // Compute the standard deviation
	    double stdDev = 0.0;
	    for (int i = 0; i < loops; i++) {
		double diff = results[i] - average;
		stdDev += diff * diff;
	    }
	    stdDev = Math.sqrt(stdDev / loops);

	    // Compute the adjusted mean
	    double mean = 0.0;
	    int counted = 0;
	    double comparator = stdDev * 2;
	    for (int i = 0; i < loops; i++) {
		double num = Math.abs(results[i] - average);
		if (num < comparator) {
		    mean += results[i];
		    counted++;
		}
	    }
	    mean /= counted;
	    final double MILL = 1000000.0;
	    System.out.println("Mean          = " + (average / MILL));
	    System.out.println("Standard Dev. = " + (stdDev / MILL));
	    System.out.println("Adjusted mean = " + (mean / MILL));
	} catch(Exception e) {
	    e.printStackTrace();
	}
    }

    private static ObjectID randomObj(final ObjectID[] objs) {
	final int num = objs.length;
	return (num > 0) ? objs[rand.nextInt(num)] : null;
    }
}
