/*
 * @(#)AsynchCall.java AsynchCall.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.thread;

import java.io.*;
import java.util.Iterator;
import java.util.LinkedList;

import mokan.Kan;
import mokan.comm.LogicalNode;
import mokan.debug.Event;
import mokan.obj.ObjectID;
import mokan.transact.TransactionID;
import mokan.transact.NestedTransactionID;
import mokan.transact.TransactionException;
import mokan.transact.TransactionAbortException;

import mokan.debug.Event;

/**
 * AsynchCall contains information on work to be done for the user
 * application.  It is a representation of an asynchronous method call.  When
 * a thread is assigned this object, it will make the method call and transmit
 * the results to a waiting <code>Future</code> object.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jing Wang
 * @author	Suk Y. Lee
 * @author	Jerry James
 * @seemokan.thread.ThreadScheduler
 */
public final class AsynchCall implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 1165921117549967679L;






    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;




    /**
     * The current transaction being executed by this asynchronous call
     *
     * @serial an identifier for the current transaction for this call
     */
    public TransactionID currTrans = null;

    /**
     * The driving (or executing) real thread for this asynchronous call
     */
    transient KanThread driver;

    /**
     * A unique identifier for this asynchronous call.  This is a pseudo-final
     * field.
     *
     * @serial the identifier for this asynchronous call
     */
    ThreadID threadID;

    /**
     * A token for mutually exclusive work stealing among inlined threads
     */
    transient boolean Token = true;

    /**
     * The object on which this call will operate.  This is a pseudo-final
     * field.
     *
     * @serial the object on which this call will operate
     */
    ObjectID objectID;

    /**
     * A boolean indicating whether we should try to inline threads or not.
     */
    static  boolean inlining = false;
//KKFIX 	Boolean.valueOf(Kan.systemProps.getProperty("thread.inlining"))
//KKFIX 	.booleanValue();

    /**
     * <code>true</code> if this is the root call to <code>main</code>.  This
     * is a pseudo-final field.
     *
     * @serial an indication of whether this is a call to the
     *         <code>main</code> method
     */
    private boolean root;

    /**
     * When <var>root</var> is <code>true</code>, the name of the class on
     * which to invoke <code>main</code>; otherwise, this is
     * <code>null</code>.  This is a pseudo-final field.
     *
     * @serial the name of the class on which to invoke <code>main</code>
     */
    private String rootClass;

    /**
     * The logical node in which this asynchronous call takes place.  This is
     * a pseudo-final field.
     *
     * @serial the logical node that will host this method call
     */
    private LogicalNode parentNode;

    /**
     * The list of nested transactions spawned by this thread, when it is
     * running as a subthread of a nested transaction.
     *
     * @serial the nested transactions spawned by this thread
     */
    private LinkedList nestedTrans = null;

    /**
     * A list of identifiers for children threads
     *
     * @serial a linked list of children thread identifiers
     */
    private LinkedList childTIDList = null;

    /**
     * A list of unprocessed AsynchCalls for sibling inlining
     *
     * @serial the list of unprocessed sibling inlined threads
     */
    private LinkedList tasks = null;

    /**
     * <code>True</code> only if a real thread has begun processing
     * (executing) this asynchronous call
     *
     * @serial an indication of whether an actual thread has begun processing
     *         this call
     */
    private boolean processed = false;
    // REVIEW: Shouldn't this be transient?

    /**
     * The method number this call will invoke on <var>object</var>.  Numbers
     * are emitted by the compiler.  This is a pseudo-final field.
     *
     * @serial the method number this call will invoke
     */
    private int methodNum;

    /**
     * The arguments to the method.  This is a pseudo-final field.
     *
     * @serial an array of arguments to the method
     */
    private Object[] argObjects;

    /**
     * The managing <code>ThreadScheduler</code> for this call
     */
    private transient ThreadScheduler myScheduler;

    /**
     * The <code>FutureID</code> to which this call will send its results.
     * This is a pseudo-final field.
     *
     * @serial an identifier for the future to which this call will send its
     *         results
     */
    private FutureID futID;

    /**
     * The no-args constructor, needed for deserialization
     */
    public AsynchCall() {
        super();
    }

    /**
     * Create a new <code>AsynchCall</code> to contain the information needed
     * to make an asynchronous method call
     *
     * @param parent the logical node that will host this method call
     * @param info a description of the method call to be made
     * @param obj the object on which this method call will be made
     * @param scheduler the managing thread scheduler for this call
     */
    public AsynchCall(final LogicalNode parent, final MethodInfo info,
                      final ObjectID objID, final ThreadScheduler scheduler) {
        root = false;
        rootClass = null;
        objectID = objID;
        methodNum = info.methodNum;
        argObjects = info.args;
        parentNode = parent;
        myScheduler = scheduler;
        threadID = (info == null || info.futID == null)
            ? new ThreadID(parent)		   // top-level thread
            : new ThreadID(parent, info.futID.getParentThread()); // sub-thread
        futID = info.futID;
        Event.DEBUG(Event.ASYNCH_CALL_START,  new Object[] { "AsynchCall:Starting ", this, info });
    }

    /**
     * Create a new <code>AsynchCall</code> to contain the information needed
     * to make an asynchronous method call on the static <code>main</code>
     * method of some class
     *
     * @param parent the logical node that will host this method call
     * @param className the name of the class containing <code>main</code>
     * @param args the arguments to pass to <code>main</code>
     * @param scheduler the managing thread scheduler for this call
     */
    public AsynchCall(final LogicalNode parent, final String className,
                      final Object[] args, final ThreadScheduler scheduler) {
        root = true;
        rootClass = className;
        objectID = null;
        argObjects = args;
        parentNode = parent;
        myScheduler = scheduler;
        threadID = new ThreadID(parent);
        futID = null;
    }

    /**
     * Retrieve the managing logical node for this asynchronous call
     *
     * @return the managing logical node for this asynchronous call
     */
    public LogicalNode getLogicalNode() {
        return parentNode;
    }

    /**
     * Retrieve the logical node that initiated this asynchronous call
     *
     * @return the logical node that initiated this asynchronous call
     */
    public LogicalNode getAccessor() {
//	return (threadID.nesting > 0) ? parentNode : futID.getNode();
        return  futID.getNode();
    }

    /**
     * Record the addition of a child thread
     *
     * @param fut a <code>Future</code> object which will hold the child
     *        thread's results
     */
    void addChild(final Future fut) {
        // Only one thread at a time will call this method
        // (see ThreadScheduler)
        if (childTIDList == null)
            childTIDList = new LinkedList();	// ... so this is safe
        synchronized (childTIDList) {
            childTIDList.add(fut);
        }
    }

    /**
     * Record the completion of a child thread
     *
     * @param fut a <code>Future</code> object which holds the child thread's
     *        results
     */
    void removeChild(final Future fut) {
        if (childTIDList != null) {
            synchronized (childTIDList) {
                childTIDList.remove(fut);
            }
        }
    }

    /**
     * Make the actual method call.  This method is called by the thread
     * assigned responsibility for this asynchronous method call.
     */
    void execute() {
        processed = true;
        Object retValue = null;

        /*
         * This method is the one that actually performs the work that this
         * AsynchCall represents.
         * New Algorithm :
         * ===============
         * There is no deadPool in the scheduler. The scheduler executes an
         * AsynchCall as soon as it gets to the scheduler. This implies, if an
         * AsynchCall receives a TransactionAbort (not TransactionUser)
         * exception, it should implicitly kill all it's childern (in worst
         * case, wait for them to report their results and then discard them),
         * and redo its work.
         */

        for (;;) {
            // Invoke the method
            try {
                myScheduler.begin(this);
                retValue = callMethod();
                myScheduler.done(this, threadID, futID, retValue, true,
                                 nestedTrans);
                break;
            } catch (TransactionException ex) {
                Event.DEBUG(Event.ASYNCH_CALL_DONE,
                            "AsyncCall.execute Transaction Exception in " + currTrans);
                ex.printStackTrace();

                /* Transaction was aborted due to some reason. */
                if (currTrans != null) {
                    // should kill all children transcations
                    currTrans.abort(ex);
                }

                // clear the child Threads and unprocessed tasks queue
                // kill all the children and try to execute again
                if (childTIDList != null) {
                    synchronized (childTIDList) {
                        childTIDList.clear();
                    }
                }
                if (tasks != null) {
                    synchronized (tasks) {
                        tasks.clear();
                    }
                }
                //KK test

                myScheduler.done(this, threadID, futID, ex, false,
                                 nestedTrans);
                break;
            } catch (Exception userEx) {
                Event.DEBUG(Event.ASYNCH_CALL_DONE,
                            "AsyncCall.execute Exception in" + currTrans);
                userEx.printStackTrace();
                myScheduler.done(this, threadID, futID, userEx, false,
                                 nestedTrans);
                break;
            }
        }

        // REVIEW: Can this test succeed?  If so, under what circumstances?
        if (driver == null)
            return;

        /* Release the driver.
         * Setting call to null will cause the KanThread running this
         * AsynchCall to go to sleep, and hence stop running.
         */
        final KanThread currDriver = driver;
        driver = null;
        threadID.driver = null;
        currDriver.call = null;

        AsynchCall call = null;
        try {
            ThreadID ptid = threadID.getParentTID();
            if (ptid != null) {
                call = ptid.findCall();
            } else {	// ptid == null : I am the top-level thread
                return;
            }
        } catch (NoSuchAsynchCallException e) {
            // No parent to report to or inline with; just exit
            return;
        }
        // parent local, and parent/child inlining WT with the token
        /* If there are no children(all have executed), check the siblings and
         * run any AsynchCalls if necessary.
         */
        if (currDriver.childThreadsListIsEmpty() && Token) {
            // switch driver, sibling inlining
            final AsynchCall newCall = call.getTask();
            if (newCall != null) {
                currDriver.call = newCall;
                newCall.driver = currDriver;
            }
        }
    }

    /**
     * Invoke the method corresponding to this <code>AsynchCall</code>
     *
     * @return the result of the method call.  Primitive types are wrapped.
     *         For <code>void</code> methods, <code>null</code> is returned.
     * @exception Exception if the called method throws an exception
     */
    private Object callMethod() throws Exception {
        if (!root) {
          // Added by Huahai
          // Warning: might be incorrect
          // If the object is invalid, the transaction should restart
          if (!objectID.isLocalValid()) {
            System.err.println("AsynchCall.callMethod: Get an object copy here. Might be incorrect.");
            objectID.getObjectCopy();
          }

            return objectID.getRealObj().kanCallMethod(methodNum,
                                                           argObjects);

        }

        // Root call
        try {
            final Class[] params = { mokan.app.Application.class, String[].class };
            final Class appClass = Class.forName(rootClass);
            return appClass.getMethod("main", params).invoke(null, argObjects);
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Remove an asynchronous call from the task list.  The task list is used
     * for thread inlining.
     *
     * @param call the asynchronous call to remove from the list
     * @return <code>true</code> if the call was in the list;
     *         <code>false</code> otherwise
     */
    private boolean removeTask(final AsynchCall call) {
        if (tasks == null)
            return false;
        synchronized (tasks) {
            return tasks.remove(call);
        }
    }

    /**
     * Add an asynchronous call to the task list.  The task list is used for
     * thread inlining.
     *
     * @param call the asynchronous call to add to the list
     */
    void addTask(final AsynchCall call) {
        // Only one thread at a time will call this method (see
        // ThreadScheduler)
        if (tasks == null)
            tasks = new LinkedList();		// ... so this is safe
        synchronized (tasks) {
            tasks.add(call);
        }
    }

    /**
     * Remove the first task from the task list and return it
     *
     * @return the first task in the task list, or <code>null</code> if there
     *         is not one
     */
    AsynchCall getTask() {
        if (tasks == null || tasks.isEmpty())
            return null;
        synchronized (tasks) {
            return (AsynchCall)tasks.removeFirst();
        }
    }

    /**
     * Handle the return of an inlined <code>AsynchCall</code>.  When a thread
     * is inlined, the parent thread pushes its <code>AsynchCall</code> onto a
     * stack and starts to execute the child <code>AsynchCall</code>. Hence,
     * when an <code>AsynchCall</code> returns, it needs to pop the
     * <code>AsynchCall</code> of the parent, and continue executing it.  The
     * <var>token</var> is to decide if I am the sibling (in case of multiple
     * child threads) who needs to inline other siblings.  If there are no
     * siblings, it recursively calls on its parent.
     */
    private void scheduleNext() {
        if (!inlining || threadID == null || !Token) {
            return;	// Not inlining, so leave
        }

        // Identify the parent of the AsynchCall
        final ThreadID parentID = threadID.getParentTID();
        AsynchCall parentCall;

        try {
            parentCall = (parentID != null) ? parentID.findCall() : null;
        } catch (NoSuchAsynchCallException e) {
            return;	// Non-local parent thread, so no inlining possible
        }

        if (parentCall != null) {
            // Get my sibling from parent
            AsynchCall newCall = parentCall.getTask();
            if (newCall != null) {
                // I give the chain to sibling, I won't be chain
                Token = false;
                myScheduler.bindDriver(newCall);
            } else {
                parentCall.scheduleNext(); // Pop to the parent
            }
        }
    }

    /**
     * Suspend operation until this asynchronous method call has returned
     *
     * @param fut the <code>Future</code> corresponding to this asynchronous
     *        method call
     * @exception TransactionException if the called method throws an
     *        exception
     */
    public void join(final Future fut) throws TransactionException {
        if (!fut.done) {
            if (inlining) {
                try {
                    if (!fut.done && fut.tid != null) {
                        final AsynchCall child =
                            fut.tid.findCall();
                        if (removeTask(child) ||
                            (!child.processed && child.driver == null)) {

                            // Parent/Child Inlining
                            driver.setChildKThread(child);
                            child.execute();
                            driver.restoreParentKThread();
                        }
                    }
                } catch (NoSuchAsynchCallException e) {
                    // child already finished!
                }
            }

            while (driver != null && !driver.isInterrupted() && !fut.done) {
                final AsynchCall childCall = getTask();

                if (childCall != null) {
                    myScheduler.createLocThread(childCall);
                } else {
                    scheduleNext();
                }

                synchronized (fut) {
                    if (!fut.done) {
                        try {
                            Event.DEBUG(Event.ASYNCH_CALL_JOIN,
                                        "AsynchCall.join:Blocking to wait for ", fut);
                            fut.wait(); //blocking for the child result
                        } catch (InterruptedException intEx) {
                            throw currTrans.killed;
                        } finally {
                            Event.DEBUG(Event.ASYNCH_CALL_JOIN,
                                        "AsynchCall.join: Joined with ", fut);
                        }
                    }
                }
            }
            if (driver != null && driver.interrupted())	// Aborted transaction?
                throw currTrans.killed;
        }
        if (fut.tid != null) {
            removeChild(fut);
            fut.tid = null;
        }
        Event.DEBUG(Event.ASYNCH_CALL_JOIN, "AsynchCall.join: Joined with ", fut);
    }

    /**
     * Wait for all child threads to complete
     */
    public void joinAll() {
        try {
            final Iterator iter = childTIDList.iterator();
            while (iter.hasNext()) {
                join((Future)iter.next());
            }
        } catch (Exception e) { // List is empty
            return;
        }
    }

    /**
     * Determine the thread that is executing this asynchronous method call
     *
     * @return the thread that is executing this asynchronous method call
     */
    public KanThread getDriver() {
        return driver;
    }

    /**
     *   Cancel this Asynchronous call , interrupting if necessary
     */
    public void cancel () {
        Event.DEBUG(Event.ASYNCH_CALL, "AsynchCall.cancel: "+  this + " " + driver + " " + currTrans);
        if (driver != null){
            if (currTrans != null) {
                currTrans.abort (new TransactionAbortException("Call cancelled"));
                return;
            }
            //driver.interrupt();
        }
    }

    /**
     * Add a nested transaction identifier to the list of nested transactions
     * descended from this method call
     *
     * @param trID the <code>ID</code> of a descendant nested transaction
     */
    public synchronized void addNestedTrans(final NestedTransactionID trID) {
        if (nestedTrans == null)
            nestedTrans = new LinkedList();
        nestedTrans.add(trID);
    }

    public String toString() {
        return threadID.toString();
    }

    /**
     * Save the state of the <code>AsynchCall</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The indication of whether this is the root call is written
     *             out (boolean), then the name of the root class (String),
     *             then the parent node (LogicalNode), then the child
     *             nested transactions of this call (LinkedList), then the
     *             identifiers of the child threads (LinkedList), then the
     *             inlined siblings of this call (LinkedList), then an
     *             indication of whether a real thread has begun executing
     *             this call (boolean), then the object on which the call will
     *             be made (GlobalObj), then the method number to invoke
     *             (int), then the arguments to the method (Object[]), then
     *             the identifier of this call (ThreadID), followed by an
     *             identifier for the transaction currently being executed in
     *             this call (TransactionID).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        out.writeBoolean(root);
        out.writeObject(rootClass);
        out.writeObject(parentNode);
        out.writeObject(nestedTrans);
        out.writeObject(childTIDList);
        out.writeObject(tasks);
        out.writeBoolean(processed);
        out.writeObject(objectID);
        out.writeInt(methodNum);
        out.writeObject(argObjects);
        out.writeObject(threadID);
        out.writeObject(currTrans);
    }

    public void readExternal(final ObjectInput in)
        throws ClassNotFoundException, IOException {

        root = in.readBoolean();
        rootClass = (String)in.readObject();
        parentNode = (LogicalNode)in.readObject();
        nestedTrans = (LinkedList)in.readObject();
        childTIDList = (LinkedList)in.readObject();
        tasks = (LinkedList)in.readObject();
        processed = in.readBoolean();
        objectID = (ObjectID)in.readObject();
        methodNum = in.readInt();
        argObjects = (Object[])in.readObject();
        threadID = (ThreadID)in.readObject();
        currTrans = (TransactionID)in.readObject();
    }
}
