/*
 * @(#)$Id: DoneMsg.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */
 
package mokan.thread;

import java.io.*;
import java.util.LinkedList;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.debug.Event;

/**
 * Report completion of a method, with return value or exceptional result
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Sean Brydon
 * @author	Jerry James
 */
final class DoneMsg extends AppMessage implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 7509268010686363607L;






    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;


    /**
     * The <code>ID</code> of the <code>Future</code> to which the return
     * value is being sent.  This is a pseudo-final field.
     *
     * @serial the <code>ID</code> of the <code>Future</code> to which the
     * return value is being sent
     */
    private FutureID futID;

    /**
     * The return value, for a normal return, or the exceptional value
     * otherwise.  This is a pseudo-final field.
     *
     * @serial the return value or exception produced by a method
     */
    private Object retValue;

    /**
     * <code>true</code> for a normal return; <code>false</code> for an
     * exceptional return.  This is a pseudo-final field.
     *
     * @serial an indication of whether this is a normal or exceptional return
     *         value
     */
    private boolean normal;

    /**
     * A set of child nested transactions to report to the parent.  This is a
     * pseudo-final field.
     *
     * @serial a set of child nested transactions to report to the parent
     */
    private LinkedList nestedTrans;

    /**
     * The no-args constructor, needed for deserialization
     */
    public DoneMsg() {
	super(false, false);
    }

    /**
     * Create a new <code>DoneMsg</code> to hold the results of a method call
     *
     * @param from the originating logical node for this message
     * @param futID the <code>ID</code> of the future to which results should
     *        be reported
     * @param retValue the normal or exceptional result of the method
     * @param normal <code>true</code> if the method returned normally;
     *        <code>false</code> if it was terminated by an exception
     * @param nestedTrans a set of nested transactions to report to the parent
     */
    DoneMsg(final LogicalNode from, final FutureID futID,
	    final Object retValue, final boolean normal,
	    final LinkedList nestedTrans) {
	super(from, retValue != null, false);
	this.futID       = futID;
	this.retValue    = retValue;
	this.normal      = normal;
	this.nestedTrans = nestedTrans;
	Event.DEBUG(Event.MSG_CREATE, "Created ", this);
    }

    /**
     * Take the receive action for this message, which is to notify a future
     * of the results of the method call
     */
    public void handle () {
	super.handle();
	final ThreadScheduler toScheduler = to.getLocalHandler().scheduler;
	toScheduler.localDone(futID, retValue, normal, nestedTrans);
    }

    /**
     * Convert the <code>DoneMsg</code> to a string
     *
     * @return a string representation of this <code>DoneMsg</code>
     */
    public String toString() {
	return "DoneMsg(" + futID + ',' + (normal ? "normal" : "exception") +
	    "): " + retValue;
    }

    /**
     * Save the state of the <code>DoneMsg</code> instance to a stream (i.e.,
     * serialize it).
     *
     * @serialData The superclass data is written out, then the ID of the
     *             future to which results should be reported (FutureID), then
     *             the results of the method (Object), then an indication of
     *             whether the method terminated normally (boolean), followed
     *             by a list of nested transactions completed by the child
     *             (LinkedList).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
	super.writeExternal(out);
	futID.writeExternal(out);	// Short cut!
	out.writeObject(retValue);
	out.writeBoolean(normal);
	out.writeObject(nestedTrans);
    }

    public void readExternal(final ObjectInput in)
	throws IOException, ClassNotFoundException {

	super.readExternal(in);
	futID = FutureID.readCompactly(in, false);
	retValue = in.readObject();
	normal = in.readBoolean();
	nestedTrans = (LinkedList)in.readObject();
    }
}
