/*
 * @(#)Future.java Future.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 *
 */

package mokan.thread;

import mokan.comm.LogicalNode;
import mokan.debug.Event;

/**
 * A representation of a value to be returned from an asynchronous method
 * call.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jing Wang
 * @author	Suk Y. Lee
 * @author	Jerry James
 * @author	Sandeep Bhatia
 */
public class Future {

    /**
     * An <code>ID</code> to uniquely identify this <code>Future</code>
     */
    FutureID futID;

    /**
     * <code>true</code> if the associated thread has finished
     */
    boolean done = false;

    /**
     * The <code>ID</code> of the associated thread, if it is local
     */
    ThreadID tid = null;

    /**
     * The <code>ID</code> of the parent thread
     */
    ThreadID parentTID;

    /**
     * The normal result of the associated thread (if any)
     */
    private Object result;

    /**
     * If the associated thread throws an exception, it gets stored here
     */
    private Exception ex;

    /**
     * Setup this <code>Future</code> by specifying its creating node and the
     * <code>ID</code> of the parent thread
     *
     * @param node the creating node for this future
     * @param parentThrdID the <code>ID</code> of the parent thread
     */
    void setup(final LogicalNode node, final ThreadID parentThrdID) {
        futID = new FutureID(node, this);
        parentTID = parentThrdID;
    }

    /**
     * Report the results of the called method to this future
     *
     * @param result the result object.  If the result was a primitive type,
     *        it was wrapped.  If the called method was <code>void</code>,
     *        this is <code>null</code>.
     * @param normal <code>true</code> to indicate a normal return value;
     *        <code>false</code> to indicate an exceptional return value
     */
    final void setResult(final Object result, final boolean normal) {
        if (normal)
            this.result = result;
        else
            ex = (Exception)result;
        futID.remove();
        synchronized (this) {
            done = true;
            notifyAll();	// Don't know how many threads are waiting
        }
    }

    /**
     * Determine whether the asynchronous method call associated with this
     * future has completed or not
     *
     * @return <code>true</code> if it has completed; <code>false</code>
     *         otherwise
     */
    public final boolean done() {
        return done;
    }

    /**
     * The root of all joining calls.  This call blocks until the associated
     * asychronous method call has reported completion.  The Kan compiler
     * emits calls to this method.
     */
    protected final void topJoin()  {
        final AsynchCall call = KanThread.currCall();
        if (!done) {
            synchronized (this) {
                while (!done) {
                    try {
                        Event.DEBUG(Event.ASYNCH_CALL_JOIN,
                                    "Future.topJoin:Blocking to wait for ", this);
                        wait(); //blocking for the child result
                    } catch (InterruptedException intEx) {
                        Event.DEBUG(Event.ASYNCH_CALL_JOIN, 
                                    "Future.topJoin: interrupException", this);
                        // There is no currTrans when an async call is cancelled
                        if (call.currTrans != null) 
                            throw call.currTrans.killed;
                        throw new AsynchCallInterruptedException ();
                    } finally {
                        Event.DEBUG(Event.ASYNCH_CALL_JOIN, 
                                    "Future.topJoin: Joined with ", this);
                    }
                }
            }
        }
        if (call != null) {
            call.removeChild(this);
            tid = null;
        }
        Event.DEBUG(Event.ASYNCH_CALL_JOIN, "Future.topJoin: Joined with ", this);
    }

    /**
     * Join and return the result as an object.  The Kan compiler emits calls
     * to this method.
     *
     * @return the object that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final Object joinObject() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return result;
    }

    /**
     * Join and return the result as a boolean.  The Kan compiler emits calls
     * to this method.
     *
     * @return the boolean value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final boolean joinBoolean() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Boolean)result).booleanValue();
    }

    /**
     * Join and return the result as a character.  The Kan compiler emits
     * calls to this method.
     *
     * @return the character value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final char joinCharacter() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Character)result).charValue();
    }

    /**
     * Join and return the result as a byte.  The Kan compiler emits calls to
     * this method.
     *
     * @return the byte value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final byte joinByte() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Byte)result).byteValue();
    }

    /**
     * Join and return the result as a short.  The Kan compiler emits calls to
     * this method.
     *
     * @return the short value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final short joinShort() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Short)result).shortValue();
    }

    /**
     * Join and return the result as an integer.  The Kan compiler emits calls
     * to this method.
     *
     * @return the integer value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final int joinInteger() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Integer)result).intValue();
    }

    /**
     * Join and return the result as a long.  The Kan compiler emits calls to
     * this method.
     *
     * @return the long value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final long joinLong() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Long)result).longValue();
    }

    /**
     * Join and return the result as a float.  The Kan compiler emits calls to
     * this method.
     *
     * @return the float value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final float joinFloat() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Float)result).floatValue();
    }

    /**
     * Join and return the result as a double.  The Kan compiler emits calls
     * to this method.
     *
     * @return the double value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final double joinDouble() throws Exception {
        topJoin();
        if (ex != null)
            throw ex;
        return ((Double)result).doubleValue();
    }

    /**
     * The root of all result fetching calls.  This call returns the result of
     * the called asychronous method without blocking for completion.  The Kan
     * compiler emits calls to this method.
     *
     * @exception Exception if the called method threw an exception
     */
    protected final void topResult() throws Exception {
        if (ex != null)
            throw ex;
    }

    /**
     * Return the result as an object.  The Kan compiler emits calls to this
     * method.
     *
     * @return the object that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final Object resultObject() throws Exception {
        if (ex != null)
            throw ex;
        return result;
    }

    /**
     * Return the result as a boolean.  The Kan compiler emits calls to this
     * method.
     *
     * @return the boolean value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final boolean resultBoolean() throws Exception {
        if (ex != null)
            throw ex;
        return ((Boolean)result).booleanValue();
    }

    /**
     * Return the result as a character.  The Kan compiler emits calls to this
     * method.
     *
     * @return the character value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final char resultCharacter() throws Exception {
        if (ex != null)
            throw ex;
        return ((Character)result).charValue();
    }

    /**
     * Return the result as a byte.  The Kan compiler emits calls to this
     * method.
     *
     * @return the byte value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final byte resultByte() throws Exception {
        if (ex != null)
            throw ex;
        return ((Byte)result).byteValue();
    }

    /**
     * Return the result as a short.  The Kan compiler emits calls to this
     * method.
     *
     * @return the short value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final short resultShort() throws Exception {
        if (ex != null)
            throw ex;
        return ((Short)result).shortValue();
    }

    /**
     * Return the result as an integer.  The Kan compiler emits calls to this
     * method.
     *
     * @return the integer value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final int resultInteger() throws Exception {
        if (ex != null)
            throw ex;
        return ((Integer)result).intValue();
    }

    /**
     * Return the result as a long.  The Kan compiler emits calls to this
     * method.
     *
     * @return the long value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final long resultLong() throws Exception {
        if (ex != null)
            throw ex;
        return ((Long)result).longValue();
    }

    /**
     * Return the result as a float.  The Kan compiler emits calls to this
     * method.
     *
     * @return the float value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final float resultFloat() throws Exception {
        if (ex != null)
            throw ex;
        return ((Float)result).floatValue();
    }

    /**
     * Return the result as a double.  The Kan compiler emits calls to this
     * method.
     *
     * @return the double value that was the result of the method call
     * @exception Exception if the called method threw an exception
     */
    protected final double resultDouble() throws Exception {
        if (ex != null)
            throw ex;
        return ((Double)result).doubleValue();
    }

    /**
     * Test two futures for equality
     *
     * @param fut the future to compare against this one
     * @return <code>true</code> if they are the same future;
     *        <code>false</code> otherwise
     */
    public final boolean equals(final Future fut) {
        return futID == fut.futID;
    }

    public final boolean equals(final Object obj) {
        return (obj instanceof Future) && equals((Future)obj);
    }

    public final String toString() {
        return "Future[" + futID + "]: " + tid;
    }
}
