/*
 * @(#)KanThread.java KanThread.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 *
 */

package mokan.thread;

import java.io.*;
import java.net.Socket;
import java.util.LinkedList;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.comm.Message;
import mokan.comm.PhysicalNode;
import mokan.comm.KanComm;

import mokan.debug.Event;

/**
 * A subclass of <code>java.lang.Thread</code> that runs Kan user code.
 * Basically, this is just a Java Thread that you can bind to some
 * <code>AsynchCall</code>.  To execute a call, the thread manager binds a
 * <code>KanThread</code> to an <code>AsynchCall</code> and wakes up the
 * waiting <code>KanThread</code>.  When the <code>AsynchCall</code>'s
 * <code>execute()</code> method returns, the <code>KanThread</code> either
 * re-joins the pool of threads in the Thread Manager or kills itself.  When a
 * <code>KanThread</code> is bound to an <code>AsynchCall</code>, the
 * <code>KanThread</code> is associated with the particular logical node that
 * is making the call.  This is because more than one logical node can be on
 * the same machine.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jing Wang
 * @author	Jerry James
 * @author	Sean Brydon
 * @seemokan.thread.AsynchCall
 */
public final class KanThread extends Thread {

    /**
     * A counter for thread names
     */
    private static int count = 0;

    /**
     * The next pointer in a linked list, used by the thread pool
     */
    KanThread next = null;

    /**
     * The logical node with which this thread is currently associated
     *
     * @seemokan.comm.LogicalNode
     */
    LogicalNode myLogicalNode = null;

    /**
     * The local threads spawned by this thread; used for thread inlining
     */
    private final LinkedList childThreadsList = new LinkedList();

    /**
     * The object input stream from which this handler reads
     */
    private PhysicalNode inputNode;

    /**
     * The physical node that will send to this request handler.  This field
     * is currently unused, but is set in anticipation of being required by
     * the fault handling code.
     */
    private PhysicalNode myPhysNode;

    /**
     * The <code>AsynchCall</code> this thread is currently executing
     */
    AsynchCall call = null;

    /**
     *  The thread pool that contains this thread.
     */

    ThreadPool pool;

    /**
     * This constructor is mainly used by <code>ThreadPool</code> to construct
     * lots of threads to be used later.  It can call
     * <code>setLogicalNode()</code> to bind the thread to the appropriate
     * logical node later.
     */
    KanThread(ThreadPool pool) {
        super(pool, "KT-" + count++);
        inputNode = null;
        myPhysNode = null;
        this.pool = pool;
        //setDaemon(true);
        start();
    }

    /**
     * Create a new request handler KanThread
     *
     * @param socket the socket on which requests will arrive
     * @param physNode the physical node that will send on <var>socket</var>
     */
//     public KanThread(final PhysicalNode ios, final PhysicalNode physNode) {
// 	super("KanThread-" + count++);
//         inputNode = ios;
//         myPhysNode = physNode;
//         start();
//     }

    /**
     * Copy constructor.  This is used to generate multiple threads waiting on
     * a socket.
     *
     * @param thrd the <code>KanThread</code> to copy
     */
    public KanThread(final KanThread thrd) {
        myLogicalNode = thrd.myLogicalNode;
        inputNode = thrd.inputNode;
        myPhysNode = thrd.myPhysNode;
        pool = thrd.pool;
        start();
    }

    /**
     * Set the asynchronous method call this thread will handle
     *
     * @param call the asynchronous method call this thread will handle
     */
    void setAsynchCall(final AsynchCall call) {
        this.call = call;
        inputNode = null;
        myPhysNode = null;
        synchronized (this) {
            notify();	// Wake up the thread so it will run
        }
    }

    /**
     * When processing a blocking message, the thread needs to know which
     * logical node is doing the processing.
     *
     * @param localNode the logical node on which this thread resides
     */
    public void setLocalNode(final LogicalNode localNode) {
        myLogicalNode = localNode;
    }

    /**
     * The method called when this thread starts.  It consists of a loop, in
     * which the thread sleeps waiting to be told to execute an asynchronous
     * method call or become a communication handler.
     */
    public void run() {
        for (;;) {
            while (call == null && inputNode == null) {
                Event.DEBUG(Event.THREAD_SLEEP, "Thread sleeping: ", this);
                synchronized (this) {
                    try {
                        wait();
                    } catch (InterruptedException e) {
                    }
                }
            }
            if (call != null) {
                Event.DEBUG(Event.THREAD_WAKEUP,
                            "Waking up to handle a call: ", this);
                // Handle an asynchronous method call
                myLogicalNode = call.getLogicalNode();
                LogicalNode.setMyNode(myLogicalNode);
                call.execute();		// THIS CAN CHANGE <var>call</var>!
            } else {
                // Become a communication handler
                Event.DEBUG(Event.THREAD_WAKEUP,
                            "Waking up to receive messages: ", this);
                setPriority(MAX_PRIORITY);
                handleMessages();
                setPriority(NORM_PRIORITY);
                inputNode = null;
            }
            if (call == null) {	// no new AsynchCall
// 		final ThreadScheduler sched =
// 		    myLogicalNode.getLocalHandler().scheduler;
// 		if (!sched.returnThreadToPool(this)) {
                if (!pool.returnThreadToPool (this)){
                    Event.DEBUG(Event.THREAD_DONE,
                                "Exiting due to full pool: ", this);
                    return; // Kill myself since the thread pool is full
                }
            } // else a new AsynchCall has been set already by inlining
        }
    }

    /**
     * Process incoming messages until one of them blocks.  When a message
     * blocks, choose another thread to process incoming messages before
     * processing the blocking message.
     */
    private void handleMessages() {
        Message newMessage = null;
        PhysicalNode input = inputNode;
        try {
            while (input != null) {
                //Event.DEBUG(Event.MSG_RECEIVE,
                //            "KanThread.handleMessage: Waiting to receive a message from: ", input);
                newMessage = KanComm.recv (input);
                //Event.DEBUG(Event.MSG_RECEIVE, "KanThread.handleMessages: Received ", newMessage);

                if (newMessage.block) {
                    if (newMessage instanceof AppMessage) {
                        LogicalNode toNode;
                        toNode = ((AppMessage)newMessage).to;
                        setLocalNode(toNode);
                    }

                    KanThread newHandler = pool.getThreadFromPool(true);
                    Event.DEBUG(Event.THREAD_OP,
                                "Switching handlers: to ", newHandler);
                    newHandler.setHandler(input, myPhysNode);
                    input = null;
                }
                newMessage.handle();
                Event.DEBUG(Event.MSG_HANDLE, "Done handling ", newMessage);
            }
        } catch (IOException eof) {
            // The physical node has lost a connection, stop handling
            Event.DEBUG(Event.MSG_HANDLE,
                        "KanThread.handleMessages recv died from ", input);
            return;
        } catch (Exception ex) {
            // Right now, do something stupid.  Later, we will have the fault
            // detector clean up after us.
            ex.printStackTrace();
            Event.ERROR("KanThread.handleMessages: "
                        +"Unhandled exception; NOT exiting but closing pipe!");
            Event.ERROR("KanThread.handleMessages: on msg " +
                        newMessage + " from " + input);

            // REVIEW: What to do with messages in waiting queue????
        }
    }

    /**
     * Make this thread a communication handler, to replace another thread
     * that is processing a message that might block
     *
     * @param objStream the object input stream from which to read
     * @param physNode the physical node on which this thread resides
     */
    public void setHandler(final PhysicalNode inpNode,
                           final PhysicalNode physNode) {
        call = null;
        inputNode = inpNode;
//	myPhysNode = physNode;
        synchronized (this) {
            notify();	// Wake up the thread so it will run
        }
    }

    /**
     * Determine the asynchronous method call being executed by the current
     * thread
     *
     * @return the asynchronous method call being executed by the current
     *         thread
     */
    public static AsynchCall currCall() {
        final Thread curr = Thread.currentThread();
        return (curr instanceof KanThread) ? ((KanThread)curr).call : null;
    }

    /**
     * Determine the currently running <code>KanThread</code>
     *
     * @return the currently running <code>KanThread</code>
     */
    public static KanThread currKThread() {
        return (KanThread)Thread.currentThread();
    }

    /**
     * Add a child thread to the child thread list
     *
     * @param call the asynchronous method call spawned by this thread
     */
    void setChildKThread(final AsynchCall call) {
        childThreadsList.add(this.call);
        this.call = call;
    }

    /**
     * Pop the top thread off of the child thread list, and return to
     * processing that asynchronous method call.  This is used by the thread
     * inlining code.
     */
    final void restoreParentKThread() {
        try {
            call = (AsynchCall)childThreadsList.removeLast();
        } catch (Exception e) {
            call = null;
        }
    }

    /**
     * Determine whether the child threads list is empty
     *
     * @return <code>true</code> if there are no child threads in the list;
     *         </code>false</code> if there are child threads in the list
     */
    boolean childThreadsListIsEmpty() {
        return childThreadsList.isEmpty();
    }
}
