/*
 * @(#)MethodInfo.java MethodInfo.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.thread;

import java.io.*;
import mokan.comm.LocalLogicalNode;
import mokan.comm.LogicalNode;
import mokan.obj.*;
import mokan.transact.NestedTransactionID;
import mokan.transact.TransactionAbortException;

/**
 * Information on a method to be invoked.  This is not made
 * <code>Externalizable</code> because it is sent in only one message
 * (<code>ExecMsg</code>), and it is cheaper for us to write the fields out,
 * so we skip writing the class descriptor.  We make it
 * <code>Serializable</code> anyway, in case the fault tolerance module needs
 * to write it out.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Suk Y. Lee
 * @author	Jing Wang
 * @author	Jerry James
 */
final class MethodInfo implements Serializable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 4603168075914725938L;






    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;


    /**
     * The <code>ID</code> of the future to which results will be returned
     *
     * @serial the <code>ID</code> of the future to which results will be
     *         returned
     */
    FutureID futID;

    /**
     * The parent transaction, if this is part of a nested transaction
     *
     * @serial the parent transaction
     */
    final NestedTransactionID parentTrans;

    /**
     * The object on which the method will be invoked
     *
     * @serial the object on which the method will be invoked
     */
    final ObjectID objID;

    /**
     * The number of the method to invoke (assigned by the Kan compiler)
     *
     * @seemokan.obj.GlobalObj
     * @serial the number of the method to invoke
     */
    final int methodNum;

    /**
     * The arguments to the method
     *
     * @serial an array of objects, containing the arguments to the method
     */
    final Object[] args;

    /**
     * Create a new <code>MethodInfo</code> object
     *
     * @param futureID the future to return results to
     * @param parentT the parent (or creating) transaction, if any
     * @param oid the object on which to invoke the method
     * @param num the number of the method to invoke
     * @param methArgs the arguments to the method
     */
    MethodInfo(final FutureID futureID, final NestedTransactionID parentT,
               final ObjectID oid, final int num, final Object[] methArgs) {
        futID = futureID;
        parentTrans = parentT;
        objID = oid;
        methodNum = num;
        args = methArgs;
    }

    /** 
     * Get the object to operate on, in preparation for calling this method on
     * it
     *
     * @param node the logical node creating the local thread
     * @return the object on which to operate
     * @exception NoSuchKanObjectException if the object on which to invoke
     *        the method cannot be found
     * @exception TransactionAbortException if the creating transaction was
     *        aborted
     */
    GlobalObj getObject(final LogicalNode node)
        throws NoSuchKanObjectException, TransactionAbortException {

        KanObject kObj = objID.getLocKanObject();
        if (kObj == null) { // I don't have a valid Kan object
            if (objID.isLocal(node)) {  // unrecoverable error
                System.err.println("MethodInfo.createLocThread: Can't find object on home node.");
                throw new NoSuchKanObjectException("Can't find " + kObj);
            }
            try {
                System.err.println("Had to fetch a copy for a local call!!");
                kObj = objID.getObjectCopy();
                if (kObj == null) {
                    System.err.println("MethodInfo.createLocThread: Could not fetch object." );
                    throw new NoSuchKanObjectException("Can't fetch " + kObj);
                }
            } catch (NoSuchKanObjectException e) {
                throw e;
            } catch (Exception e) {
                e.printStackTrace();
                throw new TransactionAbortException();
            }
        }
        return kObj.getObject();
    }

    /**
     * A shortcut to decrease the serialization costs for this object.  The
     * method writes the <code>MethodInfo</code> object on an output stream,
     * but skips certain object handles that are not needed.
     *
     * @param out the output stream on which to write
     * @exception IOException if there are any stream errors
     */
    void writeCompactly(final ObjectOutput out) throws IOException {
        futID.writeExternal(out);	// Short cut!
        out.writeObject(parentTrans);
        out.writeObject(objID);
        out.writeInt(methodNum);
        out.writeObject(args);
    }

    /**
     * A shortcut to decrease the serialization costs for
     * <code>MethodInfo</code> objects.  The method reads a
     * <code>MethodInfo</code> object from an input stream, that was written
     * by the <code>writeCompactly</code> method.
     *
     * @param in the input stream from which to read
     * @return the <code>MethodInfo</code> object read from the stream
     */
    static MethodInfo readCompactly(final ObjectInput in)
        throws ClassNotFoundException, IOException {

        final FutureID futID = FutureID.readCompactly(in, true);
        final NestedTransactionID parentTrans =
            (NestedTransactionID)in.readObject();
        final ObjectID objID = (ObjectID)in.readObject();
        final int methodNum = in.readInt();
        final Object[] args = (Object[])in.readObject();
        return new MethodInfo(futID, parentTrans, objID, methodNum, args);
    }

    /**
     * Create a string representation of this <code>MethodInfo</code>
     *
     * @return a string representation of this <code>MethodInfo</code>
     */
    public String toString() {
        return "Method " + methodNum + " on " + objID;
    }
}
