/**
 * @(#)ThreadID.java ThreadID.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.thread;

import java.io.*;
import java.util.Iterator;
import mokan.comm.LogicalNode;
import mokan.util.ID;
import mokan.util.CanonicalizingIDSet;
import mokan.transact.TransactionException;
import mokan.debug.Event;

/**
 * A unique identifier for a Kan Thread
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jing Wang
 * @author	Suk Y. Lee
 * @author	Jerry James
 */
public final class ThreadID extends ID implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -8368108570552255346L;






    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
        ObjectStreamClass.NO_FIELDS;


    /**
     * A container for canonicalizing <code>ThreadID</code>s
     */
    private static final CanonicalizingIDSet allThrd =
        new CanonicalizingIDSet();

    /**
     * A unique sequence number generator for each node.
     */
    private static long thrdCounter = 0;

    /**
     * The <code>ThreadID</code> of the parent thread.  This is a pseudo-final
     * field.
     *
     * @serial the <code>ThreadID</code> of the parent thread
     */
    private ThreadID parentThread;

    /**
     * Nonzero if this thread is currently executing inside an inlined or
     * directly called method
     */
    transient int nesting = 0;  //KK2001-03-01 Used in AsynchCall but never changed

    /**
     * The <code>ID</code> of the future to which this thread will return its
     * results.
     *
     * @serial a <code>FutureID</code> for the future to which this thread
     *         will return its results
     */
    FutureID futID;

    /**
     * The <code>KanThread</code> executing this thread, when it is local
     */
    transient KanThread driver;

    /**
     * The no-args constructor, needed for deserialization
     */
    public ThreadID() {
        super();
    }

    /** 
     * Construct a child thread ID from its parent's ID
     *
     * @param node the node on which the thread ID is created
     * @param ptid the thread ID of the parent thread
     */
    ThreadID(final LogicalNode node, final ThreadID ptid) {
        super(node, thrdCounter++);
        futID = null;
        parentThread = ptid;
        allThrd.add(this);
    }

    /** 
     * Construct the top level thread ID (i.e., the thread with no parent)
     *
     * @param node the node on which the top level thread ID is created
     */
    ThreadID(final LogicalNode node) {
        super(node, thrdCounter++);
        parentThread = null;
        allThrd.add(this);
        final Future f = new Future();
        f.setup(node, null);
        futID = f.futID;
        f.tid = this;
        // REVIEW: Should we make a special Future constructor so we don't
        // have to do all this?
    }

    /**
     * Find the future associated with this thread
     *
     * @return the future associated with this thread
     */
    public Future getFuture() {
        return futID.theFuture;
    }

    /**
     * Find the <code>ID</code> of the parent of this thread
     *
     * @return the <code>ID</code> of the parent of this thread
     */
    ThreadID getParentTID() {
        return parentThread;
    }

    /**
     * Find the <code>AsynchCall</code> corresponding to this
     * <code>ThreadID</code>
     *
     * @return the <code>AsynchCall</code> corresponding to this
     *        <code>ThreadID</code>
     * @exception NoSuchAsynchCallException if this <code>ThreadID</code> does
     *        not correspond to an active <code>AsynchCall</code>
     */
    AsynchCall findCall() throws NoSuchAsynchCallException {

        if (driver == null) {
            throw new NoSuchAsynchCallException("KanThread isn't active");
        }
        final AsynchCall call = driver.call;
        if (call == null) {
            throw new NoSuchAsynchCallException("No associated AsynchCall");
        }
        return call;
    }

    public static void abortTransactions (LogicalNode node) {
        Event.DEBUG(Event.APP_END, "ThreadID.abortTransactions", node);
        Iterator ti = allThrd.iterator();
        while (ti.hasNext()) {
            ThreadID tid = (ThreadID)ti.next();
            if (tid.driver != null && tid.driver.call != null){
                if (node != null && tid.futID.getNode() != node)
                    continue;
                tid.driver.call.cancel();
            }
        }
    }

    //KK 2001-02-23 Remove push/pop nesting level (not used)?
    //     /**
    //      * Increase the nesting level of this thread
    //      */
    //     public void push() {
    // 	nesting++;
    //     }

    //     /**
    //      * Decrease the nesting level of this thread
    //      */
    //     public void pop() {
    // 	nesting--;
    //     }

    /**
     * Save the state of the <code>ThreadID</code> instance to a stream (i.e.,
     * serialize it).
     *
     * @serialData The superclass data is written out, then the ID of the
     *             immediate parent of this thread (ThreadID), followed by the
     *             ID of the future to which this thread will report its
     *             results (FutureID).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeObject(parentThread);
        if (futID != null) {
            futID.writeExternal(out);	// Short cut!
        } else {
            out.writeLong(0L);
            out.writeObject(null);
        }
    }

    public void readExternal(final ObjectInput in)
        throws ClassNotFoundException, IOException {

        super.readExternal(in);
        parentThread = (ThreadID)in.readObject();
        futID = FutureID.readCompactly(in, false);
    }

    public Object readResolve() throws ObjectStreamException {
        return allThrd.canonicalize(this);
    }

    public String toString() {
        return "Thread" + super.toString();
    }
}
