/*
 * @(#)ThreadPool.java ThreadPool.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 *
 */

package mokan.thread;

import mokan.Kan;
//import mokan.comm.LogicalNode;
import mokan.debug.Event;

/**
 * A collection of Java threads waiting to be assigned work.  The purpose of
 * this class is to avoid paying thread creation cost at runtime.  Instead, we 
 * make some threads in advance, and hand them out when needed.  We try to
 * adapt to demand by creating more threads as needed.  We also try to
 * conserve memory by destroying threads that are not being used.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jing Wang
 * @author	Jerry James
 * @author	Sandeep Bhatia
 * @author	Sean Brydon
 */
public final class ThreadPool extends ThreadGroup{

    /*
     * PROGRAMMING NOTE: This class has been designed with two goals in mind:
     * 1. Keep synchronized blocks as short as possible to minimize contention 
     * 2. Use our own linked list rather than a general implementation (like
     *    the data structures in java.util), to avoid costly type casts and
     *    other overhead.  Our interface needs are simple, so don't pay for a
     *    fancy implementation.
     */

    /**
     * The number of pool operations before we check the pool size again
     */
    private static final int INTERVAL;

    static {
        final String str = Kan.systemProps.getProperty("thread.pool.interval");
        int interval = 40;
        try {
            interval = Integer.parseInt(str);
        } catch (NumberFormatException numEx) {
        }
        INTERVAL = interval;
    }

    /**
     * The threshold ratio of created threads versus all thread pool removing
     * operations.  If the actual ratio is above this threshold, increase the
     * size of the pool by the number of created threads.
     */
    private static final float GROW_RATIO;

    static {
        final String st = Kan.systemProps.getProperty("thread.pool.growRatio");
        float growRatio = 0.2F;
        try {
            growRatio = Float.parseFloat(st);
        } catch (NumberFormatException numEx) {
        }
        GROW_RATIO = growRatio;
    }

    /**
     * The threshold ratio of idle threads versus pool size.  If the actual
     * ratio is above this threshold, decrease the size of the pool by half
     * the average number of idle threads.
     */
    private static final float SHRINK_RATIO;

    static {
        final String ratioText =
            Kan.systemProps.getProperty("thread.pool.shrinkRatio");
        float shrinkRatio = 0.25F;
        try {
            shrinkRatio = Float.parseFloat(ratioText);
        } catch (NumberFormatException numEx) {
        }
        SHRINK_RATIO = shrinkRatio;
    }

    /**
     * The minimum size of the pool
     */
    private static final int POOLMIN;

    static {
        final String pool = Kan.systemProps.getProperty("thread.pool.minimum");
        int poolMin = 10;
        try {
            poolMin = Integer.parseInt(pool);
        } catch (NumberFormatException numEx) {
        }
        POOLMIN = poolMin;
    }

    /**
     * The number of threads to reserver for request handlers
     */
    private static final int REQMIN;

    static {
        final String req = Kan.systemProps.getProperty("thread.pool.request");
        int reqMin = 3;
        try {
            reqMin = Integer.parseInt(req);
        } catch (NumberFormatException numEx) {
        }
        REQMIN = reqMin;
    }

    /**
     * Number of free threads currently in the pool
     */
    private int idleThreadCount = POOLMIN;

    /**
     * The maximum number of free idle threads you can currently keep in the
     * thread pool.  This number can change adaptively.  It can <em>never</em>
     * be less than <code>POOLMIN</code>.
     */
    private int poolSize = POOLMIN;

    /**
     * <code>True</code> to use the adaptive thread pool; <code>false</code>
     * to use a fixed size pool
     */
    private final boolean adaptive;

    // The following are for the adaptive pool.

    /**
     * The number of pool operations since the last pool size adjustment
     */
    private int poolOps = 0;

    /**
     * Measure the # of idle threads over a certain time period.
     */
    private int totalIdleThreads = 0;

    /**
     * The head element of the pool, which consists of a linked list of
     * <code>KanThread</code>s.  This thread is never used; it is a dummy
     * value that lets us avoid race conditions that arise when we let
     * <var>head</var> and <var>tail</var> become <code>null</code>.
     */
    private final KanThread head;

    /**
     * The tail element of the pool, which consists of a linked list of
     * <code>KanThread</code>s
     */
    private KanThread tail;

    // These 2 fields keep track of how often you reuse threads from the pool
    // and how often you run out of threads in the pool, requiring the
    // creation of a new one

    /**
     * The number of times we had to create a thread (because the pool was
     * empty)
     */
    private int created = 0;

    /**
     * The number of times we found a thread in the pool when we needed one
     */
    private int found = 0;

    /**
     * Create a new <code>ThreadPool</code>
     *
     * @param adaptive <code>true</code> to make the thread pool adjust its
     *        size adaptively; <code>false</code> to use a fixed size pool
     */
    public ThreadPool(String name, final boolean adaptive) {
        super (name);
        this.adaptive = adaptive;
        head = new KanThread(this);

        // Create some idle threads and put them into the pool
        KanThread last = head;
        for (int i = 0; i < POOLMIN; i++) {
            tail = new KanThread(this);
            last.next = tail;
            last = tail;
        }
    }

    /**
     * Get a free idle <code>KanThread</code> from the pool or create one.
     *
     * @param request <code>true</code> if this is a request handler thread
     * @return a free <code>KanThread</code>
     */
    public KanThread getThreadFromPool(final boolean request) {
        KanThread kt = null;

        if (request || idleThreadCount > REQMIN) {
            synchronized (head) {
                kt = head.next;			 // Get the head element
                if (kt != null) {		 // Did we get something?
                    head.next = kt.next;	 // Swing the head pointer
                    if (head.next == null) {	 // It's null!  Change the tail
                        synchronized (this) {
                            if (kt.next != null) // Somebody changed tail
                                head.next = kt.next;
                            else		 // No, tail is still wrong
                                tail = head;
                        }
                    }
                }
            }
        }

        if (kt != null) {
            kt.next = null;
            // Synchronization is expensive, and we don't really care if these
            // numbers are slightly inaccurate, so just change them
            idleThreadCount--;
            found++;
            Event.DEBUG(Event.THREAD_POOL_GET,
                        "ThreadPool.get: in pool = " + idleThreadCount +
                        ": ", kt);
        } else {
            // The pool is empty.  Create a new thread.
            kt = new KanThread(this);
            created++;
            Event.DEBUG(Event.THREAD_POOL_EMPTY,
                        "ThreadPool.get: empty; creating a new thread: ", kt);
        }

        totalIdleThreads += idleThreadCount;
        if (adaptive && ++poolOps == INTERVAL)
            adapt();
        return kt;
    }

    /**
     * Try to put a thread back into the pool
     *
     * @param kt the <code>KanThread</code> to return to the pool
     * @return <code>true</code> if we succeeded; <code>false</code> if the
     *         thread must be destroyed due to a full pool
     */
    public boolean returnThreadToPool(final KanThread kt) {
        // Disassociate the KanThread from a particular logical node
        kt.myLogicalNode = null;

        // Is the pool full?  (Don't need to synchronize; can be inaccurate)
        if (idleThreadCount >= poolSize) {
            Event.DEBUG(Event.THREAD_POOL_FULL,
                        "ThreadPool.put: full discarding ", kt);
            return false;
        }

        // Try to put the thread back on the pool
        synchronized (this) {
            tail.next = kt;
            tail = kt;
        }

        // Synchronization is expensive, and we don't really care if this
        // number is slightly inaccurate, so just change it
        totalIdleThreads += ++idleThreadCount;
        Event.DEBUG(Event.THREAD_POOL_PUT, "ThreadPool.put: in pool = " +
                    idleThreadCount + ": ", kt);
        if (adaptive && ++poolOps == INTERVAL)
            adapt();
        return true;
    }

    /**
     * Check if we should grow or shrink the pool of idle threads.  We use a
     * sort of notion of time, since we tick <var>poolOps</var> at each call.
     * If we have been creating a lot of new threads in a certain interval of
     * time, because the idle threads in the pool are always being used, then
     * we want to increase the size of the thread pool.  If the number of
     * reused threads is much bigger than the number of new threads that had
     * to be created, then the pool is probably too big, so try to shrink it
     * and save memory.  Note that when the pool shrinks, idle threads in the
     * pool are not immediately destroyed.  Instead we lazily shrink the
     * number of threads by not putting them back onto the list.
     */
    // Note that, for absolute accuracy, this method should be synchronized.
    // However, we can live with a little bit of inaccuracy, as that affects
    // efficiency only, not correctness.  Even then, not making this
    // synchronized will probably have a bigger positive impact on efficiency
    // than the negative impact from inaccuracy herein.
    private void adapt() {
        Event.DEBUG(Event.THREAD_POOL, "Adapting thread pool size");
        if (created != 0) {
            // We created some.  Is the pool big enough?
            float ratio = created / (float)(created + found);
            if (ratio > GROW_RATIO) {
                poolSize += created;
                Event.DEBUG(Event.THREAD_POOL_GROW,
                            "Pool size is now " + poolSize);
            }
        } else if (poolOps > 0) {  // If part necessary for concurrent calls
            // We didn't create any.  Is the pool too big?
            float avgThreads = totalIdleThreads / (float)poolOps;
            if (avgThreads / poolSize > SHRINK_RATIO) {
                poolSize -= avgThreads / 2;
                if (poolSize < POOLMIN)
                    poolSize = POOLMIN;
                Event.DEBUG(Event.THREAD_POOL_SHRINK,
                            "Pool size is now " + poolSize);
            }
        }

        // Reset the variables used for the adaptive heuristic
        created = 0;
        found = 0;
        totalIdleThreads = 0;
        poolOps = 0;
    }
}
