/*
 * @(#)ThreadScheduler.java ThreadScheduler.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 */

package mokan.thread;

import java.util.LinkedList;
import mokan.Kan;
import mokan.comm.LogicalNode;
import mokan.debug.Event;
import mokan.obj.GlobalObj;
import mokan.obj.KanObject;
import mokan.obj.NoSuchKanObjectException;
import mokan.obj.ObjectID;
import mokan.transact.*;
import mokan.util.ID;
import mokan.nameservice.*;

/**
 * The Thread scheduler is responsible for tracking <code>KanThread</code>s in
 * a distributed system.  Its primary purpose is to provide a way of finding
 * and invoking <code>AsynchCall</code>s.  It also schedules threads in the
 * following way:
 * <ul>
 *   <li>It initializes a <code>ThreadPool</code>, which contains <var>n</var>
 *     threads.</li>
 *   <li>When an application wants an <code>AsynchCall</code>, it finds a free
 *     thread and assigns the work to that thread.</li>
 *   <li>If all threads are busy, it invokes a temporary thread.</li>
 * </ul>
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Suk Y. Lee
 * @author      Jing Wang
 * @author	Jerry James
 */
public final class ThreadScheduler {

    /**
     * The logical node associated with this thread scheduler
     */
    private final LogicalNode myLogicalNode;

    /**
     * A pool of idle <code>KanThread</code>s
     */
    private final ThreadPool idleThreads;

    /**
     * <code>True</code> to make the idle thread pool resize adaptively;
     * <code>false</code> to use a fixed size pool
     */
    static final boolean adaptive =
	Boolean.valueOf(Kan.systemProps.getProperty("thread.pool.adaptive"))
        .booleanValue();

    /**
     * Create a new <code>ThreadScheduler</code>
     *
     * @param node the <code>LogicalNode</code> for which this scheduler
     *        schedules
     */
    public ThreadScheduler(final LogicalNode node) {
        myLogicalNode = node;
        idleThreads = new ThreadPool("node", adaptive);
    }

    public void destroy () {
        idleThreads.interrupt();
    }

    /**
     * Get a <code>KanThread</code> from the thread pool
     *
     * @param request <code>true</code> if this is a request handler thread
     * @return a <code>KanThread</code>
     */
    public KanThread getThreadFromPool(final boolean request) {
        KanThread kt = idleThreads.getThreadFromPool(request);

        // Associate the KanThread with the logical node using it.
        kt.myLogicalNode = myLogicalNode;
        return kt;
    }

    /**
     * Try to put a thread back into the thread pool
     *
     * @param thread the <code>KanThread</code> to return to the pool
     * @return <code>true</code> if the thread went into the pool;
     *         <code>false</code> if the thread must be destroyed due to a
     *         full pool
     */
    public boolean returnThreadToPool(final KanThread thread) {
        return idleThreads.returnThreadToPool(thread);
    }

    /**
     * Invoke an asycnronous method on some <code>KanObject</code>
     *
     * @param objID the <code>ObjectID</code> for the object on which to
     *        invoke the method
     * @param fut a <code>Future</code> to hold the results of the method call
     * @param num the method number to invoke
     * @param args an array of objects holding the method arguments
     * @exception TransactionException if the containing transaction (if any)
     *        is aborted
     * @seemokan.obj.GlobalObj
     */
    public void callAsyncMethod(final ObjectID objID, final Future fut,
                                final int num, final Object[] args)
        throws NoSuchKanObjectException, TransactionException {

        final AsynchCall currCall = KanThread.currCall();
        final ThreadID currTID = (currCall != null) ? currCall.threadID : null;
        final TransactionID currTrans =
            (currCall != null) ? currCall.currTrans : null;

        // Has this thread been aborted?
        if (Thread.interrupted())
            throw currTrans.killed;

        // compose the MethodInfo to hold method invocation information
        final NestedTransactionID parent =
            (currTrans instanceof NestedTransactionID)
            ? (NestedTransactionID)currTrans
            : null;
        fut.setup(myLogicalNode, currTID);
        if (currCall != null)
            currCall.addChild(fut);
        final MethodInfo info =
            new MethodInfo(fut.futID, parent, objID, num, args);

        if (objID.isLocalValid()) {
            final GlobalObj obj = info.getObject(myLogicalNode);
            final AsynchCall call =
                new AsynchCall(myLogicalNode, info, objID, this);
            call.threadID.futID = info.futID;
            if (currCall != null) {
                call.currTrans = currCall.currTrans;
            }

            // If inlining, add the AsynchCall to the parent's call list.
            // Otherwise, bind a new thread from the thread pool.
            if (AsynchCall.inlining && currCall != null) {
                currCall.addTask(call);
            } else {
                bindDriver(call);
            }
            fut.tid = call.threadID;
        } else {
            //Modified by Huahai
            //new ExecMsg(myLogicalNode, info).send(objID.getNode());
            LogicalNode home = Globals.nameService.discoverHome(objID);
            new ExecMsg(myLogicalNode, info).send(home);
        }
    }

    /**
     * Create a thread on the local node to make a static method call.  Create
     * an <code>AsynchCall</code> (which is the unit of work) and bind it to a
     * free {@linkmokan.thread.KanThread KanThread} extracted from the
     * {@linkmokan.thread.ThreadPool ThreadPool}.
     *
     * @param staticMethodCall the method call to make
     * @return the <code>ID</code> of the newly creatd thread
     */
    public ThreadID createLocThread(final AsynchCall staticMethodCall) {
        return bindDriver(staticMethodCall).call.threadID;
    }

    /**
     * Make the current {@linkmokan.thread.KanThread KanThread} execute a new
     * {@linkmokan.thread.AsynchCall AsynchCall}.  This is for the handling of
     * {@linkmokan.thread.ExecMsg ExecMsgs}, which are already running in a
     * {@linkmokan.thread.KanThread KanThread} (the one handling requests).
     *
     * @param kObj the object on which to execute
     * @param info a description of the method call to make on <var>obj</var>
     * @exception TransactionAbortException if anything else goes wrong
     */
    void makeLocalCall(final ObjectID objID, final MethodInfo info)
        throws TransactionAbortException {

        final GlobalObj obj = objID.getRealObj();
        final AsynchCall call=new AsynchCall(myLogicalNode, info, objID, this);
        final ThreadID thrdID = call.threadID;	// Avoid multiple dereferences
        final KanThread theThread = KanThread.currKThread();

        thrdID.futID = info.futID;
        thrdID.driver = theThread;
        call.driver = theThread;
        call.currTrans = info.parentTrans;
        theThread.setAsynchCall(call);
    }

    /**
     * An <code>AsynchCall</code> binds a driver thread and runs
     *
     */
    KanThread bindDriver(final AsynchCall call) {
        final KanThread freeThread =
            idleThreads.getThreadFromPool(false);

        if (freeThread.call != null) {
            System.err.println
                ("ThreadScheduler.bindDriver(): call == " + freeThread.call);
            Thread.dumpStack();
            System.exit(0);
            // FIXME: We don't really want to exit the whole program.
            // Maybe just kill this app or restart.
        }
        freeThread.myLogicalNode = myLogicalNode;
        call.driver = freeThread;  // FIXME: use an accessor method
        call.threadID.driver = freeThread;
        freeThread.setAsynchCall(call);
        return freeThread;
    }

    final void begin(AsynchCall acall){
        final KanThread theThread = KanThread.currKThread();
        final ThreadID thrdID = acall.threadID;	// Avoid multiple dereferences

        acall.driver = theThread;
        thrdID.driver = theThread;

        if (acall.objectID != null)
            acall.objectID.startCall();
    }

    /**
     * Report completion of an asynchronous method call.  This results in
     * transmission of the result to the appropriate <code>Future</code>.
     *
     * @param me the completed asynchronous method call
     * @param tid the <code>ThreadID</code> of <var>me</var>
     * @param futID the <code>ID</code> of the future to report to
     * @param retValue the return value of the method
     * @param normal <code>true</code> for normal completion;
     *        <code>false</code> for exceptional completion
     * @param nestedTrans the nested transactions that completed as part of
     *        this method call completion
     */
    final void done(final AsynchCall me, final ThreadID tid, FutureID futID,
                    final Object retValue, final boolean normal,
                    final LinkedList nestedTrans) {


        if (me.objectID != null)
            me.objectID.endCall();
        // If not static the let the object know that the call
// 	try {
// 	    me.objectID.endCall();
// 	} catch (NullPointerException nullEx) {
// 	    // Static method call; do nothing
// 	}

        if (futID == null) {	// The top-level thread; special actions
            futID = tid.futID;
        }

        Event.DEBUG(Event.THREAD_DONE, "Thread done: ", tid);

        if (futID.isLocal(myLogicalNode)) {  // Local parent
            localDone(futID, retValue, normal, nestedTrans);

            // process the unprocessed tasks.
            AsynchCall newAC = me.getTask();
            while (newAC != null) {
                bindDriver(newAC);
                newAC = me.getTask();
            }
            tid.driver = null;
        } else {			// Remote parent
            new DoneMsg(myLogicalNode, futID, retValue, normal,
                        nestedTrans).send(futID.getNode());
        }
    }

    /**
     * Report completion of an asynchronous method call to a local future
     *
     * @param futID the <code>ID</code> of the future to which return values
     *        should be transmitted
     * @param retValue the return value to transmit
     * @param normal <code>true</code> for normal completion;
     *        <code>false</code> for exceptional completion
     * @param nestedTrans the nested transactions that completed as part of
     *        this method call completion
     */
    void localDone(final FutureID futID, final Object retValue,
                   final boolean normal, final LinkedList nestedTrans) {
        final Future f = futID.theFuture;
        final ThreadID parentThrdID = f.parentTID;
        if (nestedTrans != null && !nestedTrans.isEmpty()) {
            try {
                final AsynchCall parCall = parentThrdID.findCall();
                final NestedTransaction parTrans =
                    (NestedTransaction)(parCall.currTrans.getTransaction());
                parTrans.addDescTrans(nestedTrans);
            } catch (Exception noCallEx) {
                // NoSuchAsynchCallException: parent has aborted
                // ClassCastException: parent isn't running nested transaction
                // NullPointerException: parent isn't running a transaction
                // In any case, abort the children
                final LogicalNode node =
                    ((ID)nestedTrans.getFirst()).getNode();
                new KillChildrenMsg(myLogicalNode, nestedTrans).send(node);
            }
        }
        Event.DEBUG(Event.ASYNCH_CALL_DONE, "Ready to join with ", f);
        f.setResult(retValue, normal);
    }
}
