/*
 * @(#)BlockingTransaction.java BlockingTransaction.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 *
 */

package mokan.transact;

import mokan.comm.LogicalNode;
import mokan.obj.*;
import mokan.thread.*;

/**
 * A representation of an atomic transaction that might block or abort.  To
 * guard against this eventuality, we make a copy of the object before writing
 * to it.  In the event of an abort, we restore the copy.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public class BlockingTransaction extends Transaction {

    /**
     * A copy of the object, to restore in case of an abort
     */
    protected transient GlobalObj old;

    /**
     * The <em>normal</em> constructor, used for single-level transactions
     *
     * @param node the <code>LogicalNode</code> on which this transaction is
     *        initiated
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param objID the object on which this transaction will operate
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param call the creating thread for this transaction
     */
    protected BlockingTransaction(final LogicalNode node,
                                  final GuardCondition gc,
                                  final ObjectID objID,
                                  final boolean write,
                                  final AsynchCall call) {
        super(node, gc, objID, write, call);
    }

    /**
     * Subclasses of <code>BlockingTransaction</code> use this constructor to
     * install a specialized <code>TransactionID</code>
     *
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param objID the object on which this transaction will operate
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param trID the <code>TransactionID</code> to assign to this
     *        transaction
     * @param call the creating thread for this transaction
     * @seemokan.transact.TransactionID
     */
    protected BlockingTransaction(final GuardCondition gc,
                                  final ObjectID objID, final boolean write,
                                  final AsynchCall call,
                                  final TransactionID trID) {
        super(gc, objID, write, call, trID);
    }

    /**
     * Begin the transaction by locking the object in a state satisfying the
     * guard, and storing a copy of its original state
     *
     * @exception TransactionException if this transaction is aborted (e.g.,
     *            due to the wound-wait algorithm)
     */
    public final void begin() throws TransactionException {
        // Update access counts and lock the object
        // First lock in preemptible mode, then upgrade to commit
        super.begin(true);
        // copy object on a write
        old = write ? kObj.copyGlobalObj() : null;
    }

    /**
     * Install the object copy as the final version and unlock the object
     *
     * @exception TransactionException if this transaction was aborted or
     *            preempted
     */
    public synchronized void commit() throws TransactionException {
        if (transID.killed == null)
            super.commit();
        else
            super.end(false, old);
    }

    /**
     * Convert the <code>BlockingTransaction</code> to a string
     *
     * @return a string representation of the <code>BlockingTransaction</code>
     */
    public String toString() {
        return "BlockTrans" + transID;
    }
}
