/*
 * @(#)CanonicalizingLRSet.java CanonicalizingLRSet.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.transact;

import java.io.*;
import mokan.transact.NestedTransactionID;
import mokan.transact.TransactionID;

/**
 * A set that allows lookups of canonical versions of
 * {@linkmokan.obj.LockRequest LockRequests}.  The underlying data structure is
 * a hash set.  This implementation is <strong>unsynchronized</strong>.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 * @see	mokan.obj.LockRequest
 */
final class CanonicalizingLRSet implements Cloneable {

    /**
     * The elements in the set
     */
    private LockRequest[] table;

    /**
     * The total number of elements in the set
     */
    private int count;

    /**
     * The set is rehashed when its size exceeds this threshold.  (The value
     * of this field is (int)(capacity * loadFactor))
     */
    private int threshold;

    /**
     * The load factor for the hashtable
     *
     * @serial the load factor
     */
    private final float loadFactor;

    /**
     * Constructs a new, empty set with the specified initial capacity and the
     * specified load factor
     *
     * @param initialCapacity the initial capacity of the set
     * @param loadFactor the load factor of the set
     * @exception IllegalArgumentException if the initial capacity is less
     *        than zero, or if the load factor is nonpositive
     */
    CanonicalizingLRSet(int initialCapacity, final float loadFactor) {
	if (initialCapacity < 0)
	    throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        if (loadFactor <= 0.0f)
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
        if (initialCapacity == 0)
            initialCapacity = 1;
	this.loadFactor = loadFactor;
	table = new LockRequest[initialCapacity];
	threshold = (int)(initialCapacity * loadFactor);
    }

    /**
     * Constructs a new, empty set with the specified initial capacity and
     * default load factor, which is <tt>0.75</tt>.
     *
     * @param initialCapacity the initial capacity of the set
     * @exception IllegalArgumentException if the initial capacity is less
     *        than zero
     */
    CanonicalizingLRSet(final int initialCapacity) {
	this(initialCapacity, 0.75f);
    }

    /**
     * Constructs a new, empty set with a default capacity, which is
     * <tt>101</tt>, and a default load factor, which is <tt>0.75</tt>.
     */
    CanonicalizingLRSet() {
	this(101, 0.75f);
    }

    /**
     * Returns the number of entries in this set
     *
     * @return the number of entries in this set
     */
    int size() {
	return count;
    }

    /**
     * Returns <tt>true</tt> if this set contains no <code>LockRequest</code>s
     *
     * @return <tt>true</tt> if this set contains no <code>LockRequest</code>s
     */
    boolean isEmpty() {
	return count == 0;
    }

    /**
     * Rehashes the contents of this set into a new
     * <code>CanonicalizingLRSet</code> instance with a larger capacity.  This
     * method is called automatically when the number of elements in this set
     * exceeds its capacity and load factor.
     */
    private void rehash() {
	final int oldCapacity = table.length;
	final LockRequest oldSet[] = table;

	final int newCapacity = oldCapacity * 2 + 1;
	final LockRequest newSet[] = new LockRequest[newCapacity];

	threshold = (int)(newCapacity * loadFactor);
	table = newSet;

	for (int i = oldCapacity ; i-- > 0 ;) {
	    for (LockRequest old = oldSet[i] ; old != null ; ) {
		final LockRequest lockReq = old;
		old = old.next;

		final int index =
		    (lockReq.hashCode() & 0x7FFFFFFF) % newCapacity;
		lockReq.next = newSet[index];
		newSet[index] = lockReq;
	    }
	}
    }

    /**
     * Adds the specified {@linkmokan.obj.LockRequest LockRequest} to the set.
     * This assumes that it is not already present in the set.
     *
     * @param lockReq the {@linkmokan.obj.LockRequest LockRequest} to add to
     *        the set
     */
    void add(final LockRequest lockReq) {
	final int key = lockReq.hashCode();
	LockRequest[] tab = table;
	int index = (key & 0x7FFFFFFF) % tab.length;

	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();

	    tab = table;
	    index = (key & 0x7FFFFFFF) % tab.length;
	}

	// Creates the new entry.
	lockReq.next = tab[index];
	tab[index] = lockReq;
	count++;
    }

    /**
     * Find (and possibly set) the canonical version of a
     * {@linkmokan.obj.LockRequest LockRequest}
     *
     * @param lockReq the {@linkmokan.obj.LockRequest LockRequest} to
     *        canonicalize
     * @return the canonical version of <var>lockReq</var> if it is in the
     *        set, or <var>lockReq</var> itself if not.  In the latter case,
     *        <var>lockReq</var> is added to the set.
     */
    LockRequest canonicalize(final LockRequest lockReq) {
	final int key = lockReq.hashCode();
	LockRequest[] tab = table;
	int index = (key & 0x7FFFFFFF) % tab.length;

	// See if it is in there already
	for (LockRequest lr = tab[index]; lr != null; lr = lr.next) {
	    if (key == lr.hashCode() && lockReq.equals(lr)) {
		return lr;
	    }
	}

	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();

	    tab = table;
	    index = (key & 0x7FFFFFFF) % tab.length;
	}

	// Creates the new entry.
	lockReq.next = tab[index];
	tab[index] = lockReq;
	count++;
	return lockReq;
    }

    /**
     * Determine whether a particular {@linkmokan.obj.LockRequest LockRequest}
     * is in the set
     *
     * @param lockReq the {@linkmokan.obj.LockRequest LockRequest} to check
     * @return <code>true</code> iff <var>lockReq</var> is in the set
     */
    boolean contains(final LockRequest lockReq) {
	final LockRequest[] tab = table;
	final int key = lockReq.hashCode();
	final int index = (key & 0x7FFFFFFF) % tab.length;

	for (LockRequest lr = tab[index]; lr != null; lr = lr.next) {
	    if (key == lr.hashCode() && lr.equals(lockReq)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Find the canonical version of a {@linkmokan.obj.LockRequest LockRequest}
     * by its {@linkmokan.transact.TransactionID TransactionID}
     *
     * @param transID the transaction executing the
     *        {@linkmokan.obj.LockRequest LockRequest} to find
     * @return the canonical version of the
     *        {@linkmokan.obj.LockRequest LockRequest} to find
     */
    LockRequest find(final TransactionID transID) {
	final LockRequest[] tab = table;
	final int key = transID.hashCode();
	final int index = (key & 0x7FFFFFFF) % tab.length;

	for (LockRequest lockReq = tab[index]; lockReq != null;
	     lockReq = lockReq.next) {

	    if (key == lockReq.hashCode() && lockReq.equals(transID)) {
		return lockReq;
	    }
	}
	return null;
    }

    /**
     * Removes this {@linkmokan.obj.LockRequest LockRequest} from the set if
     * present
     *
     * @param lockReq the {@linkmokan.obj.LockRequest LockRequest} to be
     *        removed from the set
     * @return <code>true</code> if <var>lockReq</var> was in the set
     */
    boolean remove(final LockRequest lockReq) {
	final LockRequest[] tab = table;
	final int key = lockReq.hashCode();
	final int index = (key & 0x7FFFFFFF) % tab.length;

	for (LockRequest lr = tab[index], prev = null; lr != null;
	     prev = lr, lr = lr.next) {
	    if (key == lr.hashCode() && lockReq.equals(lr)) {
		if (prev != null)
		    prev.next = lr.next;
		else
		    tab[index] = lr.next;

		count--;
		lr.next = null;
		return true;
	    }
	}
	return false;
    }

    /**
     * Removes all {@linkmokan.obj.LockRequest LockRequests} from this set
     */
    void clear() {
	final LockRequest[] tab = table;
	for (int index = tab.length; --index >= 0; )
	    tab[index] = null;
	count = 0;
    }

    /**
     * Find all of the {@linkmokan.obj.LockRequest LockRequests} that are:
     * <ol>
     *    <li>initiated by a
     *      {@linkmokan.transact.NestedTransaction NestedTransaction}; and</li>
     *    <li>are <em>younger</em> than the argument</li>
     * </ol>
     *
     * @param transID the
     *        {@linkmokan.transact.NestedTransactionID NestedTransactionID} to
     *        compare against
     * @return an array containing all of the <code>ID</code>s found
     */
    NestedTransactionID[] findYounger(final NestedTransactionID transID) {
	final LockRequest[] tab = table;
	int capacity = 10;	// We expect very few of these
	int size = 0;
	NestedTransactionID[] ids = new NestedTransactionID[capacity];

	for (int i = 0; i < tab.length; i++) {
	    for (LockRequest lr = tab[i]; lr != null; lr = lr.next) {
		TransactionID id = lr.locker;
		if (id instanceof NestedTransactionID) {
		    NestedTransactionID nestedID = (NestedTransactionID)id;
		    if (transID.compareTo(nestedID) < 0) {
			if (size == capacity) {
			    capacity += 10;
			    final NestedTransactionID[] newIDs =
				new NestedTransactionID[capacity];
			    System.arraycopy(ids, 0, newIDs, 0, size);
			    ids = newIDs;
			}
			ids[size++] = nestedID;
		    }
		}
	    }
	}
	if (size != capacity) {		// Do we need a trim?
	    final NestedTransactionID[] newIDs = new NestedTransactionID[size];
	    System.arraycopy(ids, 0, newIDs, 0, size);
	    ids = newIDs;
	}
	return ids;
    }

    /**
     * Find a lock-holding ancestor of this transaction (if any)
     *
     * @param transID the
     *        {@linkmokan.transact.NestedTransactionID NestedTransactionID} to
     *        compare against
     * @return a lock-holding ancestor, or <code>null</code> if none
     */
    NestedTransactionID findAncestor(final NestedTransactionID transID) {
	final LockRequest[] tab = table;

	for (int i = 0; i < tab.length; i++) {
	    for (LockRequest lr = tab[i]; lr != null; lr = lr.next) {
		final TransactionID id = lr.locker;
		if (id instanceof NestedTransactionID) {
		    final NestedTransactionID nestedID =
			(NestedTransactionID)id;
		    if (nestedID.isAncestor(transID))
			return nestedID;
		}
	    }
	}
	return null;
    }

    /**
     * Returns a shallow copy of this <code>CanonicalizingLRSet</code>
     * instance: the values themselves are not cloned
     *
     * @return a shallow copy of this set
     */
    public Object clone() {
	final CanonicalizingLRSet cl =
	    new CanonicalizingLRSet(table.length, loadFactor);
	for (int i = table.length ; i-- > 0 ; ) {
	    cl.table[i] = table[i];
	}
	cl.count = count;
	return cl;
    }

    /**
     * Return the size of the internal array used to implement this set
     *
     * @return the size of the internal array
     */
    int capacity() {
        return table.length;
    }

    /**
     * Return the load factor of this set
     *
     * @return the load factor of this set
     */
    float loadFactor() {
        return loadFactor;
    }

    /**
     * Create a string representation of this <code>CanonicalizingLRSet</code>
     *
     * @return a string representation of this <code>CanonicalizingLRSet</code>
     */
    public String toString() {
	final LockRequest[] tab = table;
	final StringBuffer buf = new StringBuffer("CanonicalizingLRSet:\n");
	for (int i = 0; i < tab.length; i++) {
	    for (LockRequest req = tab[i]; req != null; req = req.next) {
		buf.append("    ");
		buf.append(req.toString());
	    }
	}
	return buf.toString();
    }
}
