/*
 * @(#)$Id: ChildDoneMsg.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.transact;

import java.io.*;
import java.util.Iterator;
import java.util.TreeSet;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.debug.Event;

/**
 * Inform the parent transaction that this transaction has completed
 * successfully and pass the locks acquired by all children
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 */
final class ChildDoneMsg extends AppMessage implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -1358499125759226095L;







    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;



    /**
     * The <code>ID</code> of the parent in the nested transaction tree
     *
     * @serial the <code>ID</code> of the parent in the nested transaction tree
     */
    private NestedTransactionID parentID;

    /**
     * The cumulative list of all children <code>ID</code>s in this branch of
     * the nested transaction tree
     *
     * @serial a set of children <code>ID</code>s in this branch of the nested
     *         transaction tree
     */
    private TreeSet childIDs;

    /**
     * The no-args constructor, needed for deserialization
     */
    public ChildDoneMsg() {
	super(false, false);
    }

    /**
     * Create a new <code>ChildDoneMsg</code>
     *
     * @param node the originating node for this message
     * @param parentID the <code>ID</code> of the parent transaction
     * @param childIDs the <code>ID</code>s of all children of the committing
     *        transaction, plus the committing transaction's <code>ID</code>
     */
    ChildDoneMsg(final LogicalNode node, final NestedTransactionID parentID,
		 final TreeSet childIDs) {
	super(node, false, false);
	this.parentID = parentID;
	this.childIDs = childIDs;
	Event.DEBUG(Event.MSG_CREATE, "Created ", this);
    }

    /**
     * Handle this message by notifying the parent transaction that the child
     * has finished
     */
    public void handle() {
	super.handle();
	final NestedTransaction parent = (NestedTransaction)parentID.theTrans;
	if (parent != null) {
	    parent.childDone(childIDs);
	} else {
	    // the parent is already dead -- kill the children
	    new KillChildrenMsg(to, childIDs).send(from);
	}
    }

    /**
     * Convert the <code>ChildDoneMsg</code> to a string
     *
     * @return a string representation of the <code>ChildDoneMsg</code>
     */
    public String toString() {
	return "ChildDoneMsg(" + parentID + "): " + childIDs;
    }

    /**
     * Save the state of the <code>ChildDoneMsg</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The superclass data is written out, then the identifier for
     *             the parent transaction (TransactionID), then the list of
     *             identifiers for all child transactions (TreeSet).  The
     *             child transaction identifiers are written out in their
     *             natural order.
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
	super.writeExternal(out);
	out.writeObject(parentID);
	out.writeInt(childIDs.size());
	final Iterator iter = childIDs.iterator();
	while (iter.hasNext())
	    out.writeObject(iter.next());
    }

    public void readExternal(final ObjectInput in)
	throws IOException, ClassNotFoundException {

	super.readExternal(in);
	parentID = (NestedTransactionID)in.readObject();
	childIDs = new TreeSet(NestedTransaction.comp);
	for (int i = in.readInt(); i > 0; i--)
	    childIDs.add(in.readObject());
    }
}
