/*
 * @(#)$Id: KillChildrenMsg.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.transact;

import java.io.*;
import java.util.Collection;
import java.util.Iterator;
import mokan.comm.LogicalNode;
import mokan.comm.AppMessage;
import mokan.debug.Event;

/**
 * Notify a set of remote transactions that they have been aborted.  This is
 * used in the nested transaction code to abort all children that have
 * completed the first phase of 2-phase commit.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author      Jerry James
 */
public final class KillChildrenMsg extends AppMessage
    implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 5664384287127456103L;







    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;



    /**
     * The <code>TransactionID</code>s of the transactions to abort.  This is
     * a pseudo-final field.
     *
     * @serial a collection of <code>TransactionID</code>s of the transactions
     *         to abort
     */
    private Collection children;

    /**
     * The no-args constructor, needed for deserialization
     */
    public KillChildrenMsg() {
	super(false, false);
    }

    /**
     * Make a new <code>KillChildrenMsg</code> from a particular logical node
     *
     * @param node the originating <code>LogicalNode</code> of this message
     * @param desc the <code>ID</code>s of the transactions to abort
     */
    public KillChildrenMsg(final LogicalNode node, final Collection desc) {
	super(node, false, false);
	children = desc;
	Event.DEBUG(Event.MSG_CREATE, "Created ", this);
    }

    /**
     * Handle this message by aborting the target transactions
     */
    public void handle() {
	super.handle();
	final Iterator iter = children.iterator();
	while (iter.hasNext()) {
	    final NestedTransactionID transID =
		(NestedTransactionID)iter.next();
	    transID.abort(new TransactionAbortException("Parent killed"));
	    final Transaction trans = transID.theTrans;
	    // If trans is null, then the transaction is already dead
	    if (trans != null) {
		try {
		    trans.commit();	// Kill it now
		} catch (TransactionException transEx) {
		    // We don't care; we just want it dead!
		}
	    }
	}
    }

    /**
     * Convert the <code>KillChildrenMsg</code> to a string
     *
     * @return a string representation of the <code>KillChildrenMsg</code>
     */
    public String toString() {
	return "KillChildrenMsg: " + children;
    }

    /**
     * Save the state of the <code>KillChildrenMsg</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The superclass information is written out, followed by the
     *             transaction IDs to abort (Collection).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
	super.writeExternal(out);
	out.writeObject(children);
    }

    public void readExternal(final ObjectInput in)
	throws IOException, ClassNotFoundException {

	super.readExternal(in);
	children = (Collection)in.readObject();
    }
}
