/*
 * @(#)LockRequest.java LockRequest.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 */

package mokan.transact;

import java.io.*;

import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.obj.GlobalObj;
import mokan.comm.LogicalNode;
import mokan.thread.AsynchCall;

import mokan.nameservice.*;

/**
 * A request for a lock on a <code>KanObject</code>, for reading, writing, or
 * copying the object, only when some predicate (guard) is satisfied.
 *
 * @seemokan.obj.Lock
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 */
public final class LockRequest
    implements IsolationRequest, Comparable, Externalizable {


    private final static int LF_RW = 0x01;
    private final static int LF_COPY = 0x02;
    private final static int LF_PREEMPT = 0x04;

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 1331628327060237480L;


    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /**
     * A next pointer, for use in constructing chains of lock requests
     *
     * @seemokan.obj.CanonicalizingLRSet
     */
    transient LockRequest next;

    /**
     * The <code>ID</code> of the transaction waiting for this lock request to
     * be satisfied.
     *
     * @serial the <code>ID</code> of the transaction waiting for this lock
     *         request to be satisfied
     */
    public TransactionID locker;

    /**
     * The <code>LogicalNode</code> that issued this lock request (may be
     * different from the node that originally issued the lock request)
     *
     * @serial the <code>LogicalNode</code> of the immediate request issuer
     */
    public LogicalNode requester;

    /**
     * The guard condition that must be satisfied for this lock request to be
     * granted.  A <code>null</code> condition indicates a trivial guard.
     * This is a pseudo-final field.
     *
     * @serial the guard condition that must be satisfied for this lock
     *         request to be granted
     * @seemokan.obj.GuardCondition
     */
    public GuardCondition guardCond;

    /**
     * The object <code>ID</code> for the object to be locked.  This is a
     * pseudo-final field.
     *
     * @serial the object <code>ID</code> for the object to be locked
     * @seemokan.obj.ObjectID
     */
    ObjectID objID;

    /**
     * The version of object when the request is successful
     */
    public transient long      version =-1;

//     public boolean read;
//     boolean copy;
//     public boolean preemptible;


    /**
     * <code>True</code> if this is a read lock request; <code>false</code> if
     * it is a write lock request.
     */

    /**
     * <code>True</code> if this is a copy request; <code>false</code> for a
     * normal lock request.  This is a pseudo-final field.
     *
     * @serial an indication of whether this is a copy or normal lock request
     */
    /**
     * <code>True</code> if this is a request for a preemptible lock;
     * <code>false</code> for a nonpreemptible lock.  This is a pseudo-final
     * field.
     *
     * @serial an indication of whether this is a request for a preemptible or
     *         nonpreemptible lock
     */
    private final static int LR_READONLY   = 0x0001;
    private final static int LR_PREEMPTIBLE= 0x0002;
    private final static int LR_COPY       = 0x0004;
    private final static int LR_UPGRADE    = 0x0008;

    /**
     * Various flags for the request.
     * field.
     *
     */

    private int flags;

    /**
     * <code>True</code> if the guard is satisfied for this request;
     * <code>false</code> if it is not satisfied
     */
    transient boolean guardSat = false;

    /**
     * Create a new <code>LockRequest</code> for copying an object
     */
    public LockRequest() {
        locker	    = null;
        requester   = null;
        guardCond   = null;
        objID	    = null;
//         read	    = true;
//         copy	    = true;
//         preemptible = true;
        flags = (LR_READONLY|LR_COPY|LR_PREEMPTIBLE);
    }

    /**
     * Create a new <code>LockRequest</code> on behalf of a transaction
     *
     * @param locker the <code>ID</code> of the <code>Transaction</code>
     *        requesting the lock
     * @param requester the <code>LogicalNode</code> that is the immediate
     *        issuer of this request
     * @param guardCond the condition that must be satisfied when the lock is
     *        granted
     * @param objID the <code>ID</code> of the object to lock
     * @param read <code>true</code> for a read lock; <code>false</code> for a
     *        write lock
     * @param preemptible <code>true</code> for a preemptible lock;
     *        <code>false</code> for a nonpreemptible lock
     */
    public LockRequest(final TransactionID locker, final LogicalNode requester,
                       final GuardCondition guardCond, final ObjectID objID,
                       final boolean read, final boolean preemptible) {
        this.locker	 = locker;
        this.requester	 = requester;
        this.guardCond	 = guardCond;
        this.objID	 = objID;
//         this.read	 = read;
//         this.copy	 = false;
//         this.preemptible = preemptible;
        setFlag(LR_READONLY, read);
        setFlag(LR_PREEMPTIBLE, preemptible);
    }

    /**
     * Copy a <code>LockRequest</code>, but change the locker.  This is used
     * to transfer a lock from a child transaction to its parent.
     *
     * @param request the lock request to copy
     * @param newLocker the <code>ID</code> of the <code>Transaction</code> to
     *        transfer the lock to
     * @seemokan.transact.TransactionID
     */
    public LockRequest(final LockRequest request,
                       final TransactionID newLocker) {
        locker	    = newLocker;
        requester   = newLocker.getNode();
        guardCond   = request.guardCond;
        objID	    = request.objID;
        version     = request.version;
//         read	    = request.read;
//         copy	    = request.copy;
//         preemptible = request.preemptible;
        flags = request.flags;
    }

    /**
     * Copy a <code>LockRequest</code>, but change its preemptible status.
     * This is used to upgrade locks during 2-phase commit.
     *
     * @param request the lock request to copy
     * @param preemptible <code>true</code> for a preemptible lock;
     *                    <code>false</code> for a nonpreemptible lock
     * @seemokan.transact.Transaction
     */
    public LockRequest(final LockRequest request, final boolean preempt) {
        locker		 = request.locker;
        requester	 = request.requester;
        guardCond	 = request.guardCond;
        objID		 = request.objID;
        guardSat     = request.guardSat;
        version      = request.version;

//         read		 = request.read;
//         copy		 = request.copy;
//         preemptible  = preempt;
        flags        = request.flags;
        setFlag(LR_PREEMPTIBLE, preempt);
        addFlag(LR_UPGRADE);
    }

    public final  TransactionID getTransactionID(){
        return locker;
    }
    public final  LogicalNode   getRequestingNode(){
        return requester;
    }
    public final  GuardCondition getGuard(){
        return guardCond;
    }

    public final boolean guardSatisfied (GlobalObj obj) {
        return ((guardCond == null)
                || guardSat
                || (guardSat = guardCond.isSatisfied(obj)));
    }

    public final  ObjectID   getObjectID(){
        return objID;
    }


    private final void addFlag(int f){
        flags |= f;
    }
    private final void remFlag(int f){
        flags &= ~f;
    }
    private final void setFlag(int f, boolean v){
        if (v)
            flags |= f;
        else
            flags &= ~f;
    }
    private final boolean checkFlag(int f){
        return (flags & f) != 0;
    }



    public final boolean isReadOnly(){
        return checkFlag(LR_READONLY);
    }
    public final boolean isPreemptible() {
        return checkFlag(LR_PREEMPTIBLE);
    }
    public final boolean isCopy() {
        return checkFlag(LR_COPY);
    }
    public final boolean isUpgrade() {
        return checkFlag(LR_UPGRADE);
    }

    public final void setTimestamp(long version){
        this.version = version;
    }
    public final long getTimestamp(){
        return version;
    }


    /**
     * Determine whether two lock requests can be satisfied concurrently
     *
     * @param lr lock request to test for compatibility
     * @return <code>true</code> if the lock request is compatible with this
     *        one
     * @seemokan.obj.Lock
     */
    boolean isCompatible(final LockRequest lr) {
        return ((isReadOnly() || lr.isReadOnly()) && (isPreemptible() || lr.isPreemptible()));
    }

    public int compareTo(final LockRequest req) {
        return locker.compareTo(req.locker);
    }

    public int compareTo(final Object obj) throws ClassCastException {
        return compareTo((LockRequest)obj);
    }

    public boolean equals(final LockRequest req) {
        return (this == req || locker.equals(req.locker));
    }

    public boolean equals(final TransactionID transID) {
        return locker.equals(transID);
    }

    public boolean equals(final Object obj) {
        return ((obj instanceof LockRequest) && equals((LockRequest)obj) ||
                (obj instanceof TransactionID) && equals((TransactionID)obj));
    }

    // NOTE: CanonicalizingLRSet depends on the hash code being computed in
    // this way.
    public int hashCode() {
        return locker.hashCode();
    }

    public String toString() {
        return "LockReq[" + locker + ',' + guardCond + ',' + objID + ',' +
            (isCopy()        ? 'C' : (isReadOnly()        ? 'R' : 'W')) +
            (isPreemptible() ? 'P' : 'N') + ',' + version +
            ']';
    }

    /**
     * Save the state of the <code>LockRequest</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The ID of the transaction requesting the lock is written
     *             out (TransactionID), then the issuing node (LogicalNode),
     *             then information on the guard (GuardCondition), then the ID
     *             of the object to lock (ObjectID), then an indication of
     *             whether a read or a write lock is requested (boolean), then
     *             an indication of whether this is a copy or a normal lock
     *             request (boolean), followed by an indication of whether
     *             this is a request for a preemptible or nonpreemtible lock
     *             (boolean).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        out.writeObject(locker);
        out.writeObject(requester);
        out.writeObject(guardCond);
        out.writeObject(objID);
//         out.writeBoolean(read);
//         out.writeBoolean(copy);
//         out.writeBoolean(preemptible);
        out.writeLong(version);
        out.writeInt(flags);
    }

    public void readExternal(final ObjectInput in)
        throws ClassNotFoundException, IOException {

        locker = (TransactionID)in.readObject();
        requester = (LogicalNode)in.readObject();
        guardCond = (GuardCondition)in.readObject();
        objID = (ObjectID)in.readObject();

//         read = in.readBoolean();
//         copy = in.readBoolean();
//         preemptible = in.readBoolean();
        version = in.readLong();
        flags = in.readInt();
    }
}
