/*
 * @(#)NestedTransaction.java NestedTransaction.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 *
 */

package mokan.transact;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.TreeSet;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.obj.GlobalObj;
import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.thread.AsynchCall;
import mokan.util.NodeComparator;

/**
 * A representation of an atomic transaction that is part of a larger
 * transaction
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public final class NestedTransaction extends BlockingTransaction {

    /**
     * An object for grouping <code>LockRequest</code>s from the same node
     * together
     *
     * @seemokan.obj.LockRequest
     */
    static final NodeComparator comp = new NodeComparator();

    /**
     * All the descendant transactions of this transaction
     */
    private final transient LinkedList descTrans = new LinkedList();

    /**
     * All the locks acquired by children of this transaction
     *
     * @seemokan.obj.LockRequest
     */
    private final transient TreeSet ids = new TreeSet(comp);

    /*
     * Not yet used; the set of objects to lock in the future
     */
    // final Set future = null;

    /**
     * The toplevel transaction constructor
     *
     * @param node the <code>LogicalNode</code> on which this transaction is
     *        initiated
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param objID the object on which this transaction will operate
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param call the creating thread for this transaction
     */
    protected NestedTransaction(final LogicalNode node,
                                final GuardCondition gc, final ObjectID objID,
                                final boolean write,
                                final AsynchCall call) {
        super(gc, objID, write, call, new NestedTransactionID(node));
    }

    /**
     * The subtransaction constructor
     *
     * @param node the <code>LogicalNode</code> on which this transaction is
     *        initiated
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param objID the object on which this transaction will operate
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param call the creating thread for this transaction
     * @param parent the <code>ID</code> of the parent transaction
     */
    protected NestedTransaction(final LogicalNode node,
                                final GuardCondition gc, final ObjectID objID,
                                final boolean write, final AsynchCall call,
                                final NestedTransactionID parent) {
        super(gc, objID, write, call, new NestedTransactionID(node, parent));
    }

    /**
     * End the transaction by releasing the object lock.  If the transaction
     * has aborted, then restore the original version of the object and kill
     * all children transactions.
     *
     * @param install <code>true</code> if the new version of the object
     *        should be installed permanently; <code>false</code> otherwise
     * @param oldObj the old version of the object (ignored)
     * @exception TransactionException if this transaction has been aborted
     */
    protected void end(final boolean install, final GlobalObj oldObj)
        throws TransactionException {

        if (install) {
            // Add to set of transactions for this AsynchCall if not toplevel
            if (!isTopLevel()) {
                call.addNestedTrans((NestedTransactionID)transID);
            }
        } else {
            // If we are not installing, then this is an abort.  Make sure all
            // the children are dead.
            Iterator iter = descTrans.iterator();
            LogicalNode from = transID.getNode();
            while (iter.hasNext()) {
                NestedTransactionID id = (NestedTransactionID)iter.next();
                id.abort(new TransactionAbortException("Parent killed"));
            }
        }
    }

    /**
     * Peform a preliminary commit to the parent transaction for
     * subtransactions, or the final 2-phase commit for a toplevel transaction
     *
     * @exception TransactionException if this transaction was aborted or
     *            preempted
     */
    public void commit() throws TransactionException {
        if (transID.killed != null) {
            // This transaction has been aborted; just exit
            super.commit();
        }

        // Make sure all child transactions are done
        while (!descTrans.isEmpty()) {
            try {
                synchronized (descTrans) {
                    while (!descTrans.isEmpty())
                        descTrans.wait();
                }
            } catch (InterruptedException intEx) {
                if (transID.killed != null)
                    super.commit();
            }
        }

        // Try to get a nonpreemptible lock now
        // This will throw an exception if it is not possible
        try {
            lockReq = kObj.upgrade(lockReq);
        } catch (TransactionAbortException abortEx) {
            transID.abort(abortEx);
            super.commit();
        }

        // Do a toplevel commit or child transaction commit as appropriate
        final NestedTransactionID trID = (NestedTransactionID)transID;
        if (trID.parent == null)
            toplevelCommit();
        else
            subCommit(trID);
    }

    /**
     * Perform a preliminary commit of a subtransaction to its parent
     * transaction
     *
     * @param trID the <code>ID</code> of this subtransaction
     */
    private void subCommit(final NestedTransactionID trID) {
        final NestedTransactionID parentID = trID.parent;

        synchronized (ids) {
            ids.add(trID);
        }
        final LogicalNode myNode = trID.getNode();
        if (parentID.isLocal(myNode)) {
            final NestedTransaction parent =
                (NestedTransaction)parentID.theTrans;
            parent.childDone(ids);
        } else {
            new ChildDoneMsg(myNode, parentID, ids).send(parentID.getNode());
        }
    }

    /**
     * Peform the final 2-phase commit for a toplevel transaction
     */
    private void toplevelCommit() {
        // If we reach this point, then all child transactions are done, we
        // have received their IDs, and the toplevel transaction got its
        // nonpreemptible lock; in other words, we have completed phase 1 of
        // 2-phase commit.  Now we do the second phase.

        if (ids.size() > 0) {	// Some nested transactions aren't nested
            final Iterator iter = ids.iterator();
            final long[] SSNs = new long[ids.size()];
            final LinkedList idsPerNode = new LinkedList();
            LogicalNode node = ((NestedTransactionID)ids.first()).getNode();
            int index = 0;
            while (iter.hasNext()) {
                NestedTransactionID trans = (NestedTransactionID)iter.next();
                LogicalNode destNode = trans.getNode();
                if (destNode != node) {  // switching to a new node
                    long SSN = commitAll(node, idsPerNode);
                    if (SSN != -1)
                        SSNs[index++] = SSN;
                    idsPerNode.clear();
                    node = destNode;
                }
                idsPerNode.add(trans);
            }

            // Do the last set
            final long SSN = commitAll(node, idsPerNode);
            if (SSN != -1)
                SSNs[index++] = SSN;

            // Get replies
            for (int i = 0; i < index; i++) {
                try {
                    AppMessage.await(SSNs[i]);
                } catch (TransactionAbortException abortEx) {
                    // Can't happen, but print it just in case!
                    abortEx.printStackTrace();
                }
            }
        }
        commitFinal();
    }

    /**
     * Tell a set of subtransactions on a common logical node to commit
     *
     * @param dest the logical node on which the subtransactions executed
     * @param setOfID a set of subtransactions on a common logical node
     */
    static long commitAll(final LogicalNode dest, final LinkedList setOfID) {
        final LogicalNode myNode = LogicalNode.getMyNode();
        if (dest == myNode) {
            final Iterator iter = setOfID.iterator();
            while (iter.hasNext()) {
                final NestedTransactionID trID =
                    (NestedTransactionID)iter.next();
                ((NestedTransaction)trID.theTrans).commitFinal();
            }
        } else {
            return new CommitMsg(myNode, setOfID).register(dest);
        }
        return -1L;
    }

    /**
     * Make a subtransaction do its final commit
     */
    private void commitFinal() {
        try {
            super.end(true, null);
        } catch (TransactionException transEx) {
            // Can't happen, but print it just in case!
            transEx.printStackTrace();
        }
    }

    /**
     * Add a set of subtransactions to my list of children transactions
     *
     * @param newDescTrans a list of subtransactions
     */
    public void addDescTrans(final LinkedList newDescTrans) {
        descTrans.addAll(newDescTrans);
    }

    /**
     * Notify a parent transaction of completed subtransactions
     *
     * @param childIDs the <code>ID</code>s of the completed subtransactions
     */
    void childDone(final TreeSet childIDs) {
        synchronized (descTrans) {
            ids.addAll(childIDs);
            descTrans.removeAll(childIDs);
            descTrans.notify();
        }
    }

    /**
     * Determine whether the current transaction is toplevel or a
     * subtransaction
     *
     * @return <code>true</code> if the current transaction is a toplevel
     *         transaction; <code>false</code> if it is a subtransaction
     */
    private boolean isTopLevel() {
        NestedTransactionID trID = (NestedTransactionID)transID;
        return (trID.parent == null);
    }

    /**
     * Convert the <code>NestedTransaction</code> to a string
     *
     * @return a string representation of the <code>NestedTransaction</code>
     */
    public String toString() {
        return transID.toString();
    }
}
