/*
 * @(#)NestedTransactionID.java NestedTransactionID.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.transact;

import java.io.*;
import mokan.comm.LogicalNode;

/**
 * A unique identifier for a nested transaction.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 */
public final class NestedTransactionID extends TransactionID
    implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -5197334887385344799L;

    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when 
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;



    /**
     * The immediate parent <code>ID</code> of this transaction.  This is a
     * pseudo-final field.
     *
     * @serial an <code>ID</code> for the immediate parent of this transaction
     */
    NestedTransactionID parent;

    /**
     * The top-level parent <code>ID</code> of this transaction.  This is a
     * pseudo-final field.
     *
     * @serial an <code>ID</code> for the top level parent of this transaction
     */
    private NestedTransactionID topLevel;

    /**
     * The level in the hierarchy of this transaction (0-based).  This is a
     * pseudo-final field.
     *
     * @serial an integer describing the (0-based) level in the hierarchy of
     *         this transaction 
     */
    private int level;

    /**
     * The no-args constructor, needed for deserialization
     */
    public NestedTransactionID() {
        super();
    }

    /**
     * Create a <code>NestedTransactionID</code> for a top-level transaction
     *
     * @param node the <code>LogicalNode</code> on which to create this
     *        <code>NestedTransactionID</code>
     */
    NestedTransactionID(final LogicalNode node) {
        super(node);
        parent = null;
        topLevel = this;
        level = 0;
    }

    /**
     * Create a <code>NestedTransactionID</code> with an immediate parent
     *
     * @param node the <code>LogicalNode</code> on which to create this
     *        <code>NestedTransactionID</code>
     * @param parent the <code>ID</code> of the parent transaction to this
     *        transaction
     */
    NestedTransactionID(final LogicalNode node,
                        final NestedTransactionID parent) {
        super(node);
        this.parent = parent;
        topLevel = parent.topLevel;
        level = parent.level + 1;
    }

    /**
     * Create a <code>NestedTransactionID</code> with a particular sequence
     * number
     *
     * @param node the <code>LogicalNode</code> on which to create this
     *        <code>TransactionID</code>
     * @param parent the <code>ID</code> of the parent transaction to this
     *        transaction
     * @param seq a unique sequence number for this node
     */
    NestedTransactionID(final LogicalNode node,
                        final NestedTransactionID parent,
                        final long seq) {
        super(node, seq);
        this.parent = parent;
        topLevel = parent.topLevel;
        level = parent.level + 1;
    }

    /**
     * Determine whether this is a top-level transaction
     *
     * @return <code>true</code> if this is a top-level transaction;
     *         <code>false</code> otherwise
     */
    public boolean isTopLevel() {
        return level == 0;
    }

    /**
     * Find out if this transaction is an ancestor of another
     *
     * @param transID the transaction to check for descendancy from this
     *        transaction
     * @return <code>true</code> if this transaction is an ancestor of the
     *        argument transaction
     */
    public boolean isAncestor(final NestedTransactionID transID) {
        return topLevel == transID.topLevel;
    }

    /**
     * Compare two <code>NestedTransactionID</code>s hierarchically
     *
     * @param transID the <code>NestedTransactionID</code> to be compared
     * @return the value 0 if the argument <code>ID</code> is equal to this
     *         <code>ID</code>; a value less than 0 if this <code>ID</code> is
     *         less than the <code>ID</code> argument; and a value greater
     *         than 0 if this <code>ID</code> is greater than the
     *         <code>ID</code> argument (hierarchical comparison).
     */
    public int compareTo(final NestedTransactionID transID) {
        // Toplevels compare like ordinary transaction IDs
        if (parent == null && transID.parent == null)
            return super.compareTo(transID);

        // Toplevels are different; compare them against each other
        if (topLevel != transID.topLevel)
            return topLevel.compareTo(transID.topLevel);

        // The toplevel is the same; we're in the same hierarchy
        // TODO: Make this iterative, rather than recursive
        if (level < transID.level) {
            int c = compareTo(transID.parent);
            return (c == 0) ? -1 : c;
        } else if (level > transID.level) {
            int c = parent.compareTo(transID);
            return (c == 0) ? 1 : c;
        } else {	// level == transID.level
            if (this == transID)
                return 0;
            int c = parent.compareTo(transID.parent);
            return (c == 0) ? super.compareTo(transID) : c;
        }
    }

    /**
     * Compare this <code>NestedTransactionID</code> to another
     * <code>TransactionID</code>.  If the <code>TransactionID</code> is a
     * <code>NestedTransactionID</code>, this function behaves like
     * <code>compareTo(NestedTransactionID)</code>.
     *
     * @param transID the <code>TransactionID</code> to be compared
     * @return the value 0 if the argument <code>ID</code> is equal to this
     *         <code>ID</code>; a value less than 0 if the <code>ID</code> of
     *         the toplevel transaction of this transaction is less than the
     *         <code>ID</code> argument; and a value greater than 0 if the
     *         <code>ID</code> of the toplevel transaction of this transaction
     *         is greater than the <code>ID</code> argument.
     */
    public int compareTo(final TransactionID transID) {
        if (transID instanceof NestedTransactionID)
            return compareTo((NestedTransactionID)transID);
        if (parent == null)
            return super.compareTo(transID);
        return topLevel.compareTo(transID);
    }

    /**
     * Compare this <code>NestedTransactionID</code> to another
     * <code>Object</code>.  If the <code>Object</code> is a
     * <code>TransactionID</code>, this function behaves like
     * <code>compareTo(TransactionID)</code>.  Otherwise, it throws a
     * <code>ClassCastException</code>.
     *
     * @param obj the <code>Object</code> to be compared
     * @return the value 0 if the argument is an <code>ID</code> equal to this
     *         <code>ID</code>; a value less than 0 if the <code>ID</code> of
     *         the toplevel transaction of this transaction is less than the
     *         <code>ID</code> argument; and a value greater than 0 if the
     *         <code>ID</code> of the toplevel transaction of this transaction
     *         is greater than the <code>ID</code> argument.
     * @exception ClassCastException if the argument is not a
     *            <code>TransactionID</code>
     */
    public int compareTo(final Object obj) throws ClassCastException {
        return compareTo((TransactionID)obj);
    }

    public String toString() {
        return "Nested" + super.toString();
    }

    /**
     * Save the state of the <code>NestedTransactionID</code> instance to a
     * stream (i.e., serialize it).
     *
     * @serialData The superclass data is written out, then the ID of the
     *             immediate parent of this transaction (NestedTransactionID),
     *             then ID of the top level transaction in the nested
     *             transaction tree (NestedTransactionID), followed by the
     *             level of this transaction in that tree (int).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeObject(parent);
        out.writeObject(topLevel);
        out.writeInt(level);
    }

    public void readExternal(final ObjectInput in)
        throws ClassNotFoundException, IOException {

        super.readExternal(in);
        parent = (NestedTransactionID)in.readObject();
        topLevel = (NestedTransactionID)in.readObject();
        level = in.readInt();
    }
}
