/**
 * OptimisticLock
 * Copyright kkvilekval@acm.org.   See COPYRIGHT for details.
 */
package mokan.transact;



import java.util.TreeSet;
import java.util.LinkedList;
import java.util.Iterator;

import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.obj.GlobalObj;
import mokan.obj.ReplicaDropMsg;
import mokan.obj.KanObject;
import mokan.obj.GlobalObjCopy;
import mokan.obj.UnsatisfiedGuardException;

import mokan.comm.LogicalNode;
import mokan.thread.AsynchCall;
import mokan.thread.KanThread;

import mokan.debug.Event;


/**
 *  Provide implementation of an optimistic lock.   The locking
 *  of an optimistic lock always succeds, however an exception
 *  may be thrown on commit if any other
 *  <ul>
 *  <li>Forward Validation:
 *  <li>Backward Validation:
 *  </ul>
 *  @version $Revision$, $Date$
 *  @author Kristian Kvilekval (kkvilekval@acm.org)
 */
public class OptimisticLock implements IsolationPrimitive {
    /**
     * The set of <code>LockRequest</code>s corresponding to lock holders
     */
    private final CanonicalizingLRSet lockHolders = new CanonicalizingLRSet();

    /**
     * The set of <code>LockRequest</code>s waiting to get the lock
     */
    private final LinkedList waiters = new LinkedList();


    /** The version of the object being used by this lock */
    private long version;
    private KanObject kObj;

    boolean precommit = false;

    /**
     * Added by Huahai
     * The number of waiters to be aborted
     */
    private int abort_flag = 0;

    public OptimisticLock (KanObject ko){
        kObj=ko;
    }

    public synchronized void lock(IsolationRequest ir)
        throws TransactionException {
        final LockRequest lr = (LockRequest)ir;
        ObjectID oid = kObj.getID();
        if (!oid.isLocalValid()) {
            // Force a valid copy to be local when we first
            // attempt a lock.   
            oid.getObjectCopy();
        }
        final TransactionID locker = ir.getTransactionID();
        final GlobalObj obj = kObj.getObject();	// The object to operate on
        final boolean guardSat = ir.guardSatisfied(obj);

        Event.DEBUG(Event.LOCK, "The guard is" + (guardSat ? "" : " not") +
                    " satisfied for ", ir);

        if (guardSat) {

            if (isCompatible(lr)){
                grantLock(lr);
                return;
            }

            Event.DEBUG(Event.LOCK,
                        "OptimisticLock.lock: precommit="+precommit
                        + "      waiters:\n" + getWaiters());

        }

        waiters.add (ir);
        //  An unlock will remove this request when the guard becomes
        //  valid
        try {
            while (waiters.contains(ir)) {
                Event.DEBUG(Event.LOCK, "Waiting on ", lr);
                wait();
                if (!kObj.valid) {
                    Event.DEBUG(Event.LOCK, "Awoke to invalid object", lr);
                    locker.kill (
                        new OptimisticVersionException(kObj.version,
                                                       lr.version));
                }

                //Added by Huahai
                if(abort_flag > 0) {
                  abort_flag--;
                  throw new TransactionAbortException("Abort waiters");
                }

            }
        } catch (InterruptedException intEx) {
            if (locker != null && locker.killed != null)
                throw (TransactionException)locker.killed;
        }
    }

    public void unlock(IsolationRequest ir, boolean install){
        unlock((LockRequest)ir, install);
    }
    public void unlock(IsolationRequest ir, boolean install, GlobalObj gob){

        long version = ir.isReadOnly()?kObj.version:kObj.version+1;
        unlock((LockRequest)ir, install, gob, version);
    }
    /**
     *  Unlock and install a version (used from remote)
     */
    public void unlock(IsolationRequest ir, boolean install,
                       GlobalObj gob, long version){
        unlock((LockRequest)ir, install, gob, version);
    }

    /** Attempt to commit */
    /**
     * Release a lock, possibly change the object version, and wake up any
     * potential lockers
     *
     * @param lockReq a previously granted lock request
     * @param install whether to install a new version
     * @param newObj the version of the object to install; if this is
     *        <code>null</code>, the version number is bumped only
     * @exception TransactionException if this transaction was killed
     */
    private synchronized void unlock(LockRequest ir, boolean install,
                                     GlobalObj gob, long version){

        Event.DEBUG(Event.LOCK,"OptimisticLock.unlock " + kObj + " for " +ir);

        if (ir.locker != null)  // Null when remote transaction
            lockHolders.remove (ir);

        if (install) {
            // Install the new object if nobody else has installed
            // a newer version of the object.

            // The value has not been changed install our new version.
            if (!ir.isReadOnly()) {
                kObj.version = version;
                GlobalObj realob = kObj.getObject();
                if (gob != null) {
                    realob.kanSetObject (gob);
                }
                Event.DEBUG(Event.LOCK,"OptimisticLock.unlock installed " + kObj);
                reevaluate (realob);
            }
        }
        releaseLock(ir, true);
    }


    /**
     * Break the lock.  This is used for a lock on a replica
     * that is stuck waiting on a guard condition.
     */
    public synchronized void nudge () {
        Event.DEBUG(Event.LOCK,"OptimisticLock.nudge");
        notifyAll();
    }


    public boolean idle(){
        return true;
    }

    /**
     * Get a optimistic copy of this object.  This involves waiting for any
     * nonpreemptible writers to finish, since they may be altering the main
     * copy of the object.
     *
     * @return a consistent copy of the object
     * @exception InterruptedException if the copier is aborted during the
     *            copying process
     */
//     public GlobalObjCopy copy()  throws InterruptedException{
//         if (waiters.isEmpty() && isCompatible(lockReq)) {
//             grantLock(lockReq);
//         } else {
//             copiers.add(lockReq);
//             while (copiers.contains(lockReq)) {
//                 wait();
//             }
//         }
//         final GlobalObjCopy copy =
//             new GlobalObjCopy(kObj.getObject().kanClone(),
//                               kObj.version,
//                               Protocol.OPTIMISTIC);
//         try {
//             unlock(lockReq, false, null);
//         } catch (TransactionException transEx) {
//             // This can't happen
//             // REVIEW: Can it?
//         }
//         Event.DEBUG(Event.OBJECT_COPY, "Copying ", kObj);
//         return copy;
//     }


    /**
     * Get a specialized copy lock
     */
    public synchronized IsolationRequest lockForCopy() {
        final LockRequest lr = new LockRequest();

        if (isCompatible(lr)) {
            grantLock(lr);
        } else {
            try {
                waiters.add(lr);
                Event.DEBUG(Event.LOCK, "Waiting on ", lr);
                while (waiters.contains(lr)) {
                    wait();
                    if (!kObj.valid) {
                        Event.DEBUG(Event.LOCK, "Awoke to invalid object", lr);
                        lr.locker.kill(
                            new OptimisticVersionException(kObj.version,
                                                             lr.version));
                    }
                }
            } catch (InterruptedException ex) {
                if (lr.locker != null && lr.locker.killed != null)
                    throw (TransactionException)lr.locker.killed;
            }
        }
        return lr;
    }


    /**
     * Unlock copy lock
     */
    public synchronized void unlockForCopy(IsolationRequest lockReq) {
        try {
            unlock(lockReq, false, null);
        } catch (TransactionException transEx) {
            // This can't happen
            // REVIEW: Can it?
            transEx.printStackTrace();
        }
    }


    /**
     *   Perform a validation and pre-commit of the lock
     *
     */
    public synchronized LockRequest upgrade(LockRequest lockReq){

        Event.DEBUG(Event.UPGRADE, "Upgrading lock for ", lockReq);
        if (kObj.version == lockReq.version) {
            final LockRequest newReq = new LockRequest(lockReq, false);
            lock(newReq);
            return newReq;
        }

        Event.DEBUG(Event.UPGRADE,"failed upgrade for "+ lockReq + " != version" + kObj.version );

        lockReq.locker.kill(
            new OptimisticVersionException(kObj.version,
                                             lockReq.version));
        return null;            // abort will throw exception.
    }

    public LockRequest lookup (TransactionID tid){
        Event.DEBUG(Event.OBJECT_LOCK, "Looking up lock  ", tid);
        final LockRequest lockReq = lockHolders.find(tid);
        if (lockReq == null && tid.killed != null)
            throw tid.killed;
        return lockReq;
    }



    /**
     * An optimistic lock request is valid unless
     * the lock is precommit mode
     * synchronized:
     */
    private boolean isCompatible (LockRequest lr) {
        return !precommit;
    }
    /**
     *  This is called when we may capture the lock
     * synchronized:
     *
     */
    private void grantLock (LockRequest lr) {
        // We are upgrading
        if (lr.locker != null) { // Copies are not part of a transaction
            LockRequest oldReq = lockHolders.find (lr.locker);
            if (oldReq != null)
                releaseLock (oldReq, false);

            lockHolders.add(lr);
        }

        if (!lr.isPreemptible())
            precommit = true;

        lr.setTimestamp (kObj.version);

        Event.DEBUG(Event.LOCK, "OptimisticLock for " + kObj + " granted to ", lr);

        notifyAll();
    }

    /**
     *  Release the lock.
     *  Wake any calls waiting on precommit or guards
     * synchronized:
     */
    private void releaseLock (LockRequest lr, boolean check) {
        Event.DEBUG(Event.UNLOCK, "OptimisticLock.releaseLock: ", lr);

        if (!lr.isPreemptible())
            precommit = false;

        if (lr.locker != null)
            lockHolders.remove(lr);

        if (check) {
            final Iterator iter = waiters.iterator();
            boolean satWaiter = false;
            while (iter.hasNext()) {
                final LockRequest req = (LockRequest)iter.next();
                if (req.guardSat) {
                    if (isCompatible(req)) {
                        iter.remove(); // removes from waiters/copiers
                        grantLock(req);
                    } else {
                        satWaiter = true;
                    }
                }
            }
            //satisfiedWaiter = satWaiter;
        }
    }

    /**
     * Reevaluate guards of waiting requests
     *
     * @param newObj the new state of the object the requesters are trying to
     *        lock
     */
    private void reevaluate(final GlobalObj obj) {
        final Iterator iter = waiters.iterator();
        while (iter.hasNext()) {
            final LockRequest req = (LockRequest)iter.next();
            final GuardCondition guard = req.getGuard();
            final boolean guardSat = (guard == null)||guard.isSatisfied(obj);
            req.guardSat = guardSat;

            Event.DEBUG(Event.GUARD_EVAL, "Guard now " + guardSat +
                        " for ", req);
            if (!req.guardSat && req.locker instanceof NestedTransactionID)
                req.locker.abort(new UnsatisfiedGuardException());
        }
    }

    private String getWaiters() {
        String r="";

        Iterator i =  waiters.iterator();
        while (i.hasNext()) {
            r = r+"     "+((LockRequest)i.next()) + "\n";
        }
        return r;
    }

    //Added by Huahai
    public synchronized void abortWaiters() {
      //Cause all threads blocked in the lock() to abort
      if(!waiters.isEmpty()) {
        abort_flag = waiters.size();
        waiters.clear();
        notifyAll();
      }

    }
}
