/**
 * OptimisticScheduler
 * Copyright kkvilekval@acm.org.   See COPYRIGHT for details.
 */
package mokan.transact;

import mokan.KanSystem;
import mokan.comm.LogicalNode;
import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.obj.GlobalObj;
import mokan.obj.KanObject;
import mokan.obj.Protocol;
import mokan.obj.NoSuchKanObjectException;

import mokan.thread.AsynchCall;
import mokan.thread.Future;
import mokan.thread.KanThread;
import mokan.thread.ThreadScheduler;

import mokan.debug.Event;


/**
 *  The <code>OptimisticScheduler</code> description.
 *  @version $Revision$, $Date$
 *  @author Kristian Kvilekval (kkvilekval@acm.org)
 */
public final class OptimisticScheduler implements ConcurrencyScheduler{
    /* Private/Public Variables */


    /** Create a Transaction type for this concurrency scheduler */
    public Transaction makeTransaction (final GuardCondition gc, 
                                        final ObjectID objID, 
                                        final boolean write,
                                        final AsynchCall call, 
                                        final int type){

        final NestedTransactionID parent = (NestedTransactionID)call.currTrans;
        final LogicalNode node = call.getLogicalNode();
        if (parent instanceof NestedTransactionID)
            return new NestedOptimisticTransaction(node, gc, objID, write, call, parent);

        return new NestedOptimisticTransaction(node, gc, objID, write, call);
    }

    /** Initiate the transaction.  Optimistic Transactions are
     *  preemptible.
     *  @param t  The transation 
     */
    public void beginTransaction (Transaction trans)
        throws TransactionException {

        Event.DEBUG(Event.TRANSACTION, "Scheduler.beginTrans for " + 
                    trans + " on " + this);

        trans.begin();
    }

    /** 
     * End the transtaction
     * @param t the transaction to finish
     */
    public void endTransaction(Transaction trans, boolean install){
        trans.commit();
        Event.DEBUG(Event.TRANSACTION, "Scheduler.endTrans "+
            trans + " on " + this );
    }

    public void abortTransaction(TransactionID transID, 
                                 TransactionException tex){
        transID.abort(tex);
    }


    public IsolationPrimitive makePrimitive (KanObject kobj, int flags){
        IsolationPrimitive p = new OptimisticLock (kobj);
        return p;
    }


    /** 
     * Capture a strict Lock 
     * @param guardCond   The guard of this lock
     * @param objID       The Object (contains the lock)
     * @param flags       Type of request.
     */
    public IsolationRequest beginIsolation(final IsolationPrimitive lock,
                                           final GuardCondition guardCond, 
                                           final ObjectID objID,
                                           int   flags){

        final TransactionID creator = KanThread.currCall().currTrans;
        final IsolationRequest lr =
            new LockRequest(creator, creator.getNode(), guardCond, objID,
                            false, false);





        beginIsolation(lr, lock);
        return lr;

    }

    /** 
     * Capture a strict Lock 
     * @param ir  A previously created <code>IsolationRequest</code>
     */
    public void beginIsolation(IsolationRequest lockReq, 
                               IsolationPrimitive lock){
        lock.lock(lockReq);
    }

    /** 
     *  Release a strict lock 
     */
    public void endIsolation(final IsolationRequest lockReq,
                             final IsolationPrimitive lock,
                             final GlobalObj oldObj,
                             final boolean install){

        lock.unlock(lockReq, install, oldObj);
    }

    public int protocol (int type){
        if (type == KanSystem.IMMOBILE)
            return Protocol.IMMOBILE;
        return Protocol.OPTIMISTIC_HOME;
    }



    /**
     * Create a thread to invoke a method on some <code>KanObject</code>
     *
     * @param objID the {@link mokan.obj.ObjectID ObjectID} for the object on
     *        which to invoke the method 
     * @param fut a {@linkmokan.thread.Future Future} to hold the results of
     *        the method call
     * @param num the method number to invoke
     * @param args an array of objects holding the method arguments
     * @exception TransactionException if the containing transaction (if any)
     *        is aborted
     */
    public void invokeMethod(final ObjectID objID, final Future fut,
                                    final int num, final Object[] args)
        throws TransactionException {

        try {

              if (!objID.isLocalValid()) {
                  Event.DEBUG(Event.OPTIMISTIC_OBJECT,
                              "Request local for method "+ num +" of "+ objID);
                  objID.getObjectCopy();
              }

            final LogicalNode myNode = LogicalNode.getMyNode();
            final ThreadScheduler myScheduler =
                myNode.getLocalHandler().scheduler;
            myScheduler.callAsyncMethod(objID, fut, num, args);
        } catch (NoSuchKanObjectException noObjEx) {
            // TODO: What should we do here?
            // This shouldn't be possible, given the compiler, but...
            noObjEx.printStackTrace();
        }
    }
}
