/**
 * OptimisticTransaction
 * Copyright kkvilekval@acm.org.   See COPYRIGHT for details.
 */
package mokan.transact;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.TreeSet;
import mokan.comm.AppMessage;
import mokan.comm.LogicalNode;
import mokan.obj.GlobalObj;
import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.thread.AsynchCall;
import mokan.util.NodeComparator;

import mokan.debug.Event;

/**
 *  The <code>OptimisticTransaction</code> description.
 *  @version $Revision$, $Date$
 *  @author Kristian Kvilekval (kkvilekval@acm.org)
 */
public class OptimisticTransaction extends Transaction {
  /* Private/Public Variables */

  /**
   * A copy of the object, to restore in case of an abort
   */
  protected transient GlobalObj old;

  /**
   * The toplevel transaction constructor
   *
   * @param node the <code>LogicalNode</code> on which this transaction is
   *        initiated
   * @param gc the guard that has to be satisfied for this transaction to
   *        begin
   * @param objID the object on which this transaction will operate
   * @param write <code>true</code> if this transaction might change the
   *        state of the object; <code>false</code> otherwise
   * @param call the creating thread for this transaction
   */
  protected OptimisticTransaction(final LogicalNode node,
                                  final GuardCondition gc, final ObjectID objID,
                                  final boolean write,
                                  final AsynchCall call) {
    super(node, gc, objID, write, call);
  }

  /**
   * The subtransaction constructor
   *
   * @param node the <code>LogicalNode</code> on which this transaction is
   *        initiated
   * @param gc the guard that has to be satisfied for this transaction to
   *        begin
   * @param objID the object on which this transaction will operate
   * @param write <code>true</code> if this transaction might change the
   *        state of the object; <code>false</code> otherwise
   * @param call the creating thread for this transaction
   * @param parent the <code>ID</code> of the parent transaction
   */
  protected OptimisticTransaction(final GuardCondition gc,
                                  final ObjectID objID,
                                  final boolean write, final AsynchCall call,
                                  final NestedTransactionID parent) {
    super(gc, objID, write, call, parent);
  }

  /**
   * Begin the transaction by locking the object in a state satisfying the
   * guard, and storing a copy of its original state
   *
   * @exception TransactionException if this transaction is aborted (e.g.,
   *            due to the wound-wait algorithm)
   */
  public final void begin() throws TransactionException {
    // Added by Huahai
    if (!objID.isLocalValid()) {
      Event.DEBUG(Event.OPTIMISTIC_OBJECT,
                  "Request local for method of " + objID);
      objID.getObjectCopy();
      kObj = objID.getLocKanObject();
    }

    // Update access counts and lock the object
    // First lock in preemptible mode, then upgrade to commit
    super.begin(true);

    old = write ? kObj.copyGlobalObj() : null;
  }

  /**
   * Install the object copy as the final version and unlock the object
   *
   * @exception TransactionException if this transaction was aborted or
   *            preempted
   */
  public synchronized void commit() throws TransactionException {
    if (transID.killed == null) {
      Event.DEBUG(Event.TRANSACTION,
                  "OptimisticTransaction commiting " + transID);
      super.commit();
    } else {
      Event.DEBUG(Event.TRANSACTION,
                  "OptimisticTransaction aborted unlock begin" + transID);
      super.end(false, old);
      Event.DEBUG(Event.TRANSACTION,
                  "OptimisticTransaction aborted unlock end" + transID);
    }
  }

  public synchronized void fail() {
    super.end(false, old);
  }

  /**
   * Convert the <code>BlockingTransaction</code> to a string
   *
   * @return a string representation of the <code>BlockingTransaction</code>
   */
  public String toString() {
    return "OptTrans" + transID;
  }

}
