/*
 * @(#)StrictLock.java StrictLock.java,v 1.1.1.1 2002/11/06 22:10:27 kris Exp $
 *
 */

package mokan.transact;

//import java.util.*;
import java.util.TreeSet;
import java.util.LinkedList;
import java.util.Iterator;

import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.obj.GlobalObj;
import mokan.obj.ReplicaDropMsg;
import mokan.obj.KanObject;
import mokan.obj.GlobalObjCopy;
import mokan.obj.UnsatisfiedGuardException;

import mokan.comm.LogicalNode;
import mokan.thread.AsynchCall;
import mokan.thread.KanThread;

import mokan.debug.Event;


/**
 * Locks for distributed objects.  Each lock request is for some combination
 * of the following two binary attributes:
 * <ul>
 *   <li>read or write</li>
 *   <li>preemptible or nonpreemptible</li>
 * </ul>
 * <p>The locking rules are as follows:</p>
 * <ul>
 *   <li>Writers conflict with everything</li>
 *   <li>Readers do not conflict with each other</li>
 *   <li>Non-preemptible requests are allowed to kill preemptible requests</li>
 *   <li>Non-preemptible requests cannot be killed</li>
 * </ul>
 * <p>Note that preemptible lockers may get a WoundWaitException when
 * unlocking if an older transaction locked in an incompatible mode in the
 * meantime.  Also, a TransactionAbortException may be thrown if the
 * transaction was aborted while waiting for the lock (due, for example, to a
 * parent transaction being killed).</p>
 * <p>The locking algorithm used is a solution to the readers-writers problem
 * with the addition of preemptible readers and writers, and
 * conditionalization of lockers (due to guards).</p>
 * <p>As an optimization, we give a different interface for the system to use
 * when making a copy of an object (for replication purposes).  This amounts
 * to a read operation that is known to have a short period.  We treat copy
 * requests like nonpreemptible read requests, except that we allow all
 * pending copy requests to proceed after a writer releases the lock, even if
 * there are other writers waiting ahead of the copiers.</p>
 * <p>The algorithm we use is given next, in pseudocode.  The notation
 * <code>with lock</code> means that that section of code is executed while
 * holding a lock, and the lock is released by both normal and abnormal
 * exits.  There are locks for each individual lock request and a lock
 * (<code>waiters</code>) for all waiting requests.</p>
 * <pre><code>
 * lock(transaction, guard, read/write):
 *   with lock (waiters)
 *     if the guard is satisfied then
 *       if there are lockHolders that transaction can preempt then
 *         kill (wound-wait) all such lockHolders
 *       if waiters is empty AND transaction is compatible with all
 *           lockHolders then
 *         grant the lock
 *         return
 *     record version # and whether the guard is satisfied for this version
 *     add transaction to waiters
 *     sleep on waiters
 *
 * unlock(transaction [, new version]):
 *   with lock (waiters)
 *     if new version
 *       install new version
 *       reevaluate and mark guards of waiters
 *     remove transaction from lockHolders
 *     if holders is empty
 *       if copiers is nonempty
 *         wake up all copiers
 *       else if waiters is nonempty
 *         head = first enabled waiter
 *         wake up head
 *         wake up all waiters compatible with head
 * [where wake up = remove from waiters/copiers, add to holders, notify]
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:27 $
 * @author	Jerry James
 * @seemokan.obj.LockRequest
 */
final public  class StrictLock implements IsolationPrimitive{
    /**
     * The set of <code>LockRequest</code>s corresponding to lock holders
     */
    private final CanonicalizingLRSet lockHolders = new CanonicalizingLRSet();

    /**
     * The set of <code>LockRequest</code>s waiting to get the lock
     */
    private final TreeSet waiters = new TreeSet();

    /**
     * The set of <code>LockRequest</code>s waiting to copy the object
     */
    private final LinkedList copiers = new LinkedList();

    /**
     * The <code>KanObject</code> guarded by this lock.
     */
    private final KanObject kObj; // TODO: Generalize this to a set

    /**
     * The number of preemptible readers holding the lock
     */
    private int pReaders = 0;

    /**
     * The number of nonpreemptible readers holding the lock
     */
    private int nReaders = 0;

    /**
     * The number of preemptible writers holding the lock
     */
    private int pWriters = 0;

    /**
     * The number of nonpreemptible writers holding the lock
     */
    private int nWriters = 0;

    /**
     * <code>true</code> if some waiting thread has a satisfied guard
     */
    private boolean satisfiedWaiter = false;

    /**
     * Added by Huahai
     * The number of waiters to be aborted
     */
    private int abort_flag = 0;

    /**
     * Create a <code>StrictLock</code> in the unlocked state for a particular
     * <code>KanObject</code>
     *
     * @param obj the object associated with this lock
     */
    public StrictLock(final KanObject obj) {
        kObj = obj;
        Event.DEBUG(Event.OBJECT_LOCK, "Created a lock for ", obj);
    }

    /**
     * Look up a lock by its associated transaction ID
     *
     * @param transID a transaction holding the lock
     * @exception TransactionException if this transaction was killed
     */
    public LockRequest lookup(final TransactionID transID)
        throws TransactionException {

        Event.DEBUG(Event.OBJECT_LOCK, "Looking up lock  ", transID);
        final LockRequest lockReq = lockHolders.find(transID);
        if (lockReq == null && transID.killed != null)
            throw transID.killed;
        return lockReq;
    }

    /**
     * Get a consistent copy of this object.  This involves waiting for any
     * nonpreemptible writers to finish, since they may be altering the main
     * copy of the object.
     *
     * @return a consistent copy of the object
     * @exception InterruptedException if the copier is aborted during the
     *            copying process
     */
//     public synchronized GlobalObjCopy copy() throws InterruptedException {
//         final LockRequest lockReq = new LockRequest();
//         if (waiters.isEmpty() && isCompatible(lockReq)) {
//             grantLock(lockReq);
//         } else {
//             copiers.add(lockReq);
//             while (copiers.contains(lockReq)) {
//                 wait();
//             }
//         }
//         final GlobalObjCopy copy =
//             new GlobalObjCopy(kObj.getObject().kanClone(),
//                               kObj.version,
//                               Protocol.REPLICA);
//         try {
//             unlock(lockReq, false, null);
//         } catch (TransactionException transEx) {
//             // This can't happen
//             // REVIEW: Can it?
//         }
//         Event.DEBUG(Event.OBJECT_COPY, "Copying ", kObj);
//         return copy;
//     }


    public synchronized IsolationRequest lockForCopy(){
        final LockRequest lockReq = new LockRequest();
        if (waiters.isEmpty() && isCompatible(lockReq)) {
            grantLock(lockReq);
        } else {
            try {
                copiers.add(lockReq);
                while (copiers.contains(lockReq)) {
                    wait();
                }
            } catch (InterruptedException ex) {
                if (lockReq.locker != null && lockReq.locker.killed != null)
                    throw (TransactionException)lockReq.locker.killed;
            }
        }
        return lockReq;
    }

    public synchronized void unlockForCopy(IsolationRequest lockReq) {
        try {
            unlock(lockReq, false, null);
        } catch (TransactionException transEx) {
            // This can't happen
            // REVIEW: Can it?
        }
    }


    /**
     * Test whether this lock is in use.  This method is useless unless the
     * caller has synchronized on the lock.
     *
     * @return <code>true</code> if there are no current or waiting lockers;
     *         <code>false</code> otherwise
     */
    public boolean idle() {
        return lockHolders.isEmpty() && waiters.isEmpty();
    }

    public void lock(IsolationRequest ir){
        lock ((LockRequest)ir);
    }

    /**
     * Acquire a lock, ensuring that a guard condition is satisfied first.
     * The lock can be acquired in read/write and preemptible/nonpreemptible
     * modes.
     *
     * @param lockReq a lock request, in a form suitable for queueing
     * @exception TransactionAbortException if the transaction is aborted
     *            while waiting for the lock (due to wound-wait or an
     *            unsatisfied guard)
     */
    public synchronized void lock(final LockRequest lockReq)
        throws TransactionException {

        final TransactionID locker = lockReq.locker;
        final GlobalObj obj = kObj.getObject();	// The object to operate on
        final boolean guardSat = (lockReq.guardCond == null)
            || lockReq.guardCond.isSatisfied(obj);
        Event.DEBUG(Event.LOCK, "The guard is" + (guardSat ? "" : " not") +
                    " satisfied for ", lockReq);
        if (guardSat) {
            if (locker != null && locker instanceof NestedTransactionID) {
                final NestedTransactionID rqst = (NestedTransactionID)locker;
                // are we part of the same nested transaction?
                if (lockHolders.findAncestor(rqst) != null) {
                    grantLock(lockReq);
                    return;
                }
                // Otherwise kill younger nested transactions (wound-wait)
                final NestedTransactionID[] younger =
                    lockHolders.findYounger(rqst);
                for (int i = 0; i < younger.length; i++) {
                    Event.DEBUG(Event.WOUND, "Wounding ", younger[i]);
                    younger[i].abort(new WoundWaitException(locker));
                }
            }
            boolean compatibleReq = isCompatible(lockReq);
            if (!satisfiedWaiter && compatibleReq) {
                grantLock(lockReq);
                return;
            }
            Event.DEBUG(Event.LOCK,
                        "Lock.lock: satisfied="+satisfiedWaiter
                        + " compatible="+compatibleReq
                        + " pw="+pWriters+" pr="+pReaders
                        + " nw="+nWriters+" nr="+nReaders
                        + "      waiters:\n" + getWaiters());
        } else if (locker instanceof NestedTransactionID &&
                   !((NestedTransactionID)locker).isTopLevel()) {
            throw new UnsatisfiedGuardException();
        }
        //  Could not grant immediately, put lock on waitque.
        lockReq.guardSat = guardSat;
        waiters.add(lockReq);
        try {
            while (waiters.contains(lockReq)) {
                Event.DEBUG(Event.LOCK, "Waiting on ", lockReq);
                wait();
            }

            //Added by Huahai
            if(abort_flag > 0) {
              abort_flag--;
              throw new TransactionAbortException("Abort waiters");
            }

        } catch (InterruptedException intEx) {
            if (locker != null && locker.killed != null)
                throw (TransactionException)locker.killed;
        }
    }

    //Added by Huahai
    /**
     * Abort all lock waiters.
     */
    public synchronized void abortWaiters() {
      //Cause all threads blocked in the lock() to abort
      if(!waiters.isEmpty()) {
        abort_flag = waiters.size();
        waiters.clear();
        notifyAll();
      }
    }
    /**
     * Release a lock, possibly change the object version, and wake up any
     * potential lockers
     *
     * @param lockReq a previously granted lock request
     * @param install whether to install a new version
     * @param newObj the version of the object to install; if this is
     *        <code>null</code>, the version number is bumped only
     * @exception TransactionException if this transaction was killed
     */
    public void unlock(final IsolationRequest ir, final boolean install,
                       final GlobalObj newObj) throws TransactionException {

        //lockReq.setGlobalObject(newObj, kObj.version+1);
        unlock(ir, install, newObj, ir.isReadOnly()? kObj.version:kObj.version+1);
    }


    public void unlock(IsolationRequest ir, boolean install){
        long timestamp = ir.isReadOnly()? kObj.version:kObj.version+1;
        unlock((LockRequest)ir, install, null, timestamp);
    }

    public void unlock(final IsolationRequest ir, final boolean install, GlobalObj gob, long version){
        unlock((LockRequest)ir, install, gob, version);
    }


    /**
     * Release a lock, change the object version and version number, and wake
     * up any potential lockers.  This method is for remote object changes to
     * be installed locally.
     *
     * @param lockReq a previously granted lock request
     * @param install whether to install a new version
     * @exception TransactionException if this transaction was killed
     */
    public synchronized void unlock(final LockRequest lockReq, final boolean install, GlobalObj gob, long version)
        throws TransactionException {

        if (install) {
            if (version != kObj.version)
                Event.DEBUG(Event.OBJECT_NEW_VERSION, "New version: ", kObj);
            kObj.version = version;
            final GlobalObj masterObj = kObj.getObject();
            if (gob != null){		// Null for NonblockingTransaction
                Event.DEBUG(Event.OBJECT_NEW_VERSION, "Copying new global object: ");
                masterObj.kanSetObject(gob);
            }
            reevaluate(masterObj);
        }
        releaseLock(lockReq, true);
        final TransactionID locker = lockReq.locker;
        if (locker != null && locker.killed != null)
            throw locker.killed;
    }

    /**
     * Upgrade a lock from preemptible to nonpreemptible.  This requires going
     * through the same algorithm as the original lock, only we are careful to
     * avoid having the <code>LockRequest</code> conflict with itself.
     *
     * @param lockReq a previously granted preemptible lock request
     * @return the upgraded <code>LockRequest</code>
     * @exception TransactionAbortException if the locker was preempted before
     *            making the upgrade
     */
    public LockRequest upgrade(final LockRequest lockReq)
        throws TransactionAbortException {

        Event.DEBUG(Event.UPGRADE, "Upgrading lock for ", lockReq);
        final LockRequest newReq = new LockRequest(lockReq, false);
        lock(newReq);
        return newReq;
    }

    /**
     *  Nudge the lock .. ask it to reevalute any guarded actions
     *  that have been blocked.  No-op for StrictLock
     */
    public void nudge () { }



    /**
     * Give this lock requester the lock.  Note that calls to this method
     * should be synchronized.
     *
     * @param lockReq a request for the lock
     */
    private void grantLock(final LockRequest lockReq) {
        // First check whether it is in there already.  If so, we are
        // upgrading.
        final LockRequest oldReq = lockHolders.find(lockReq.locker);
        if (oldReq != null)
            releaseLock(oldReq, false);

        // Now add the new version
        lockHolders.add(lockReq);
        if (lockReq.isReadOnly()) {
            if (lockReq.isPreemptible()) {
                pReaders++;
            } else {
                nReaders++;
            }
        } else {
            if (lockReq.isPreemptible()) {
                pWriters++;
            } else {
                nWriters++;
            }
        }
        Event.DEBUG(Event.LOCK, "Lock for " + kObj + " granted to ", lockReq);
        notifyAll();
    }

    /**
     * Check whether any incompatible requests are holding the lock
     *
     * @param lockReq the <code>LockRequest</code> to check for compatibility
     * @return <code>true</code> if this request is compatible with all lock
     *         holders; <code>false</code> otherwise
     */
    private boolean isCompatible(final LockRequest lockReq) {
        int pr = pReaders;
        int nr = nReaders;
        int pw = pWriters;
        int nw = nWriters;

        // First remove it in case we are upgrading, so it does not conflict
        // with itself
        if (lockHolders.contains(lockReq)) {
            if (lockReq.isReadOnly()) {
                pr--;
            } else {
                pw--;
            }
        }

        // Succeed when:
        // no writers (both preemptible and non-preemptible)
        //  and (it is a read request
        //       or  it is a write request and there are no readers).
        return (nw == 0) && (pw == 0) &&
            (lockReq.isReadOnly() || (!lockReq.isReadOnly() && pr == 0 && nr == 0));
    }

    /**
     * Release the lock from some requester.  Note that calls to this method
     * should be synchronized.
     *
     * @param lockReq a request for the lock
     * @param allowCopies <code>true</code> if copiers can be allowed to go
     *        ahead.  This is <code>false</code> when upgrading locks.
     */
    private void releaseLock(final LockRequest lockReq,
                             final boolean allowCopies) {
        Event.DEBUG(Event.UNLOCK, "Lock.releaseLock: ", lockReq);
        if (!lockHolders.contains(lockReq))
            return;
        if (lockReq.isReadOnly()) {
            if (lockReq.isPreemptible()) {
                pReaders--;
            } else {
                nReaders--;
            }
        } else {
            if (lockReq.isPreemptible()) {
                pWriters--;
            } else {
                nWriters--;
            }
        }
        lockHolders.remove(lockReq);
        //
        if (allowCopies && lockHolders.isEmpty()) {
            final boolean emptyCopiers = copiers.isEmpty();
            if (!kObj.valid && emptyCopiers && waiters.isEmpty())
                new ReplicaDropMsg(kObj.getID()).send(kObj.getID().getNode());

            final Iterator iter = (emptyCopiers)
                ? waiters.iterator()
                : copiers.iterator();
            boolean satWaiter = false;
            while (iter.hasNext()) {
                final LockRequest req = (LockRequest)iter.next();
                if (req.guardSat) {
                    if (isCompatible(req)) {
                        iter.remove(); // removes from waiters/copiers
                        grantLock(req);
                    } else {
                        satWaiter = true;
                    }
                }
            }
            if (emptyCopiers)
                satisfiedWaiter = satWaiter;
        }
    }

    /**
     * Reevaluate guards of waiting requests
     *
     * @param newObj the new state of the object the requesters are trying to
     *        lock
     */
    private void reevaluate(final GlobalObj newObj) {
        final Iterator iter = waiters.iterator();
        while (iter.hasNext()) {
            final LockRequest req = (LockRequest)iter.next();
            req.guardSat = (req.guardCond == null)
                || req.guardCond.isSatisfied(newObj);
            Event.DEBUG(Event.GUARD_EVAL, "Guard now " + req.guardSat +
                        " for ", req);
            if (!req.guardSat && req.locker instanceof NestedTransactionID)
                req.locker.abort(new UnsatisfiedGuardException());
        }
    }

    private String getWaiters() {
        String r="";

        Iterator i =  waiters.iterator();
        while (i.hasNext()) {
            r = r+"     "+((LockRequest)i.next()) + "\n";
        }
        return r;
    }
}
