/*
 * @(#)Transaction.java Transaction.java,v 1.1.1.1 2002/11/06 22:10:28 kris Exp $
 *
 *
 */

package mokan.transact;

import mokan.comm.LogicalNode;
import mokan.thread.AsynchCall;
import mokan.obj.GuardCondition;
import mokan.obj.ObjectID;
import mokan.obj.GlobalObj;
import mokan.obj.KanObject;

import mokan.debug.Event;

/**
 * A representation of an atomic transaction.  This class is not meant to be
 * instantiated directly.  It merely contains code and provides the interface
 * that is common to all transaction types.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:28 $
 * @author	Jerry James
 */
public class Transaction implements Comparable {

    // NOTE: For future expansion purposes, DO NOT USE THE NUMBER ZERO unless
    // you also change the logic in the compiler (see GuardAST.transType).
    // Briefly, we use the positive/negative sign on these numbers to
    // represent reading/writing transactions of those types, and Java just
    // doesn't have a negative zero.

    /**
     * A constant indicating a nonblocking transaction
     */
    public static final int NONBLOCKING = 1;

    /**
     * A constant indicating a blocking transaction
     */
    public static final int BLOCKING = 2;

    /**
     * A constant indicating a nested transaction
     */
    public static final int NESTED = 3;

    /**
     * The <code>LockRequest</code> for this transaction, which is needed for
     * unlocking
     */
    public LockRequest lockReq;

    /**
     * The object <code>ID</code> for the object that this transaction is
     * accessing
     */
    public final ObjectID objID;

    /**
     * The <code>ID</code> of this transaction
     */
    public final TransactionID transID;

    /**
     * The guard condition that must be satisfied for this transaction to
     * begin.  <code>Null</code> indicates the trivial (always satisfied)
     * guard.
     */
    protected final GuardCondition gc;

    /**
     * <code>False</code> if this is a read-only transaction,
     * <code>true</code> otherwise
     */
    protected final boolean write;

    /**
     * The creating thread for this transaction
     */
    protected final Thread creatingThread;

    /**
     * The <code>KanObject</code> for the object that this transaction is
     * accessing
     */
    protected KanObject kObj;

    /**
     * The creating thread for this transaction
     */
    protected final AsynchCall call;

    /**
     * The <em>normal</em> constructor, used for single-level transactions
     *
     * @param node the <code>LogicalNode</code> on which this transaction is
     *        initiated
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param objID the object on which this transaction will operate
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param call the creating thread for this transaction
     */
    protected Transaction(final LogicalNode node, final GuardCondition gc,
                          final ObjectID objID, final boolean write,
                          final AsynchCall call) {
        this(gc, objID, write, call, new TransactionID(node));
    }

    /**
     * Subclasses of <code>Transaction</code> use this constructor to install
     * a specialized <code>TransactionID</code>
     *
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param objID the object on which this transaction will operate
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param call the creating thread for this transaction
     * @param trID the <code>TransactionID</code> to assign to this
     *        transaction
     * @seemokan.transact.TransactionID
     */
    protected Transaction(final GuardCondition gc, final ObjectID objID,
                          final boolean write, final AsynchCall call,
                          final TransactionID trID) {
        this.gc = gc;
        this.objID = objID;
        this.write = write;
        this.call = call;
        kObj = objID.getLocKanObject();
        creatingThread = call.getDriver();
        call.currTrans = transID = trID;
        trID.theTrans = this;
    }

    /**
     * A factory for creating new <code>Transaction</code>s of an appropriate
     * type
     *
     * @param objID the object on which this transaction will operate
     * @param gc the guard that has to be satisfied for this transaction to
     *        begin
     * @param write <code>true</code> if this transaction might change the
     *        state of the object; <code>false</code> otherwise
     * @param call the creating thread for this transaction
     * @param type whether the transaction should be nonblocking, blocking, or
     *        nested
     * @return a new <code>Transaction</code> object of the appropriate type
     * @seemokan.obj.GuardCondition
     */
    public static final Transaction makeTransaction
	(final GuardCondition gc, final ObjectID objID, final boolean write,
	 final AsynchCall call, final int type) {

        final NestedTransactionID parent = (NestedTransactionID)call.currTrans;
        final LogicalNode node = call.getLogicalNode();
        if (parent instanceof NestedTransactionID)
            return new NestedTransaction(node, gc, objID, write, call, parent);
        switch (type) {
        case NONBLOCKING:
            return new NonblockingTransaction(node, gc, objID, write, call);
        case BLOCKING:
            return new BlockingTransaction(node, gc, objID, write, call);
        default:
            return new NestedTransaction(node, gc, objID, write, call);
        }
    }


    public IsolationRequest getLock(){
        return null;
    }
    public ObjectID         getObject(){
        return null;
    }
    public TransactionID    getID(){
        return null;
    }
    public GuardCondition   getGuard(){
        return null;
    }



    /**
     * Find a transaction from its <code>TransactionID</code>
     *
     * @param transID the <code>ID</code> of the transaction to look up
     * @return the <code>Transaction</code> corresponding to
     *         <var>transID</var>, or <code>null</code> if no such transaction
     *         exists
     */
    public static final Transaction lookup(final TransactionID transID) {
        return transID.theTrans;
    }

    /**
     * Begin the transaction by locking the object in nonpreemtible mode, in a
     * state satisfying the guard, and storing a copy of its original state
     *
     * @exception TransactionException if this transaction is aborted (e.g.,
     *            due to the wound-wait algorithm)
     */
     public void begin() throws TransactionException {
         begin(false);
     }

    /**
     * Begin the transaction by locking the object in a state satisfying the
     * guard, and storing a copy of its original state
     *
     * @param preemptible <code>true</code> if this transaction can be
     *        preempted; <code>false</code> if it cannot
     * @exception TransactionException if this transaction is aborted (e.g.,
     *            due to the wound-wait algorithm)
     */
    public void begin(final boolean preemptible)
        throws TransactionException {

        Event.DEBUG(Event.TRANSACTION_BEGIN, "Transaction.begin ", this);

        // lock the object and check the guard
        lockReq = new LockRequest(transID, transID.getNode(), gc, objID,
                                  !write, preemptible);


        kObj.lock(lockReq);
    }

    /**
     * End the transaction by releasing the object lock.  If the transaction
     * has aborted, then restore the original version of the object and kill
     * all children transactions.
     *
     * @param install <code>true</code> if the new version of the object
     *        should be installed permanently; <code>false</code> otherwise
     * @param oldObj the old version of the object, to restore in case of an
     *        abort
     * @exception TransactionException if this transaction has been aborted
     */
    protected void end(final boolean install, final GlobalObj oldObj)
        throws TransactionException {

        transID.theTrans = null;
        transID.remove();
        kObj.unlock(lockReq, install, oldObj);
        Event.DEBUG(Event.TRANSACTION_COMMIT, "Transaction.end ", this);
    }

    /**
     * Install the object copy as the final version and unlock the object
     *
     * @exception TransactionException if this transaction was aborted or
     *            preempted
     */
    public void commit() throws TransactionException {
        Event.DEBUG(Event.TRANSACTION_COMMIT, "Committing ", this);
        end(true, null);
    }

    /**
     * Abort the transaction.  The abort is signalled with an interrupt.
     */
    public void abort(TransactionException ex) {
        Event.DEBUG(Event.TRANSACTION_ABORT,
                    "Thread " + Thread.currentThread()+ " Aborting " + this + " this " + creatingThread);

        if (Thread.currentThread() != creatingThread)
            creatingThread.interrupt();	// Signal the abort
        else
            throw ex;
    }


    public final boolean isAborted() {
        return transID.killed != null;
    }
    public final void  rethrowAbort () throws TransactionException {
        throw transID.killed;
    }

    /**
     * Compare two <code>Transaction</code>s by <code>ID</code>
     *
     * @param trans the <code>Transaction</code> to be compared
     * @return the value 0 if the argument <code>Transaction</code> is equal
     *         to this one; a value less than 0 if this
     *         <code>Transaction</code> is less than the
     *         <code>Transaction</code> argument; and a value greater than 0
     *         if this <code>Transaction</code> is greater than the
     *         <code>Transaction</code> argument.
     */
    public int compareTo(final Transaction trans) {
        return transID.compareTo(trans.transID);
    }

    /**
     * Compare this <code>TransactionID</code> of this
     * <code>Transaction</code> to another <code>TransactionID</code>.
     *
     * @param transID the <code>TransactionID</code> to be compared
     * @return the value 0 if the argument <code>ID</code> is equal to this
     *         transaction's <code>ID</code>; a value less than 0 if the
     *         <code>ID</code> of this transaction is less than the
     *         <code>TransactionID</code> argument; and a value greater than 0
     *         if the <code>ID</code> of this transaction is greater than the
     *         <code>TransactionID</code> argument.
     */
    public int compareTo(final TransactionID transID) {
        return transID.compareTo(transID);
    }

    /**
     * Compare this <code>Transaction</code> to another <code>Object</code>.
     * If the <code>Object</code> is a <code>Transaction</code>, this
     * function behaves like <code>compareTo(Transaction)</code>.  If the
     * <code>Object</code> is a <code>TransactionID</code>, this function
     * behaves like <code>compareTo(TransactionID)</code>.  Otherwise, it
     * throws a <code>ClassCastException</code>.
     *
     * @param obj the <code>Object</code> to be compared
     * @return the value 0 if the argument is a <code>Transaction</code> or
     *         <code>TransactionID</code> equal to this
     *         <code>Transaction</code>; a value less than 0 if this
     *         transaction is less than the <code>Transaction</code> or
     *         <code>TransactionID</code> argument; and a value greater than 0
     *         if this transaction is greater than the
     *         <code>Transaction</code> or <code>TransactionID</code>
     *         argument.
     */
    public int compareTo(final Object obj) throws ClassCastException {
        return (obj instanceof TransactionID)
            ? transID.compareTo((TransactionID)obj)
            : transID.compareTo(((Transaction)obj).transID);
    }

    /**
     * Returns a hashcode value for the <code>Transaction</code>
     *
     * @return a hashcode value for this <code>Transaction</code>
     * @see java.lang.Object
     */
    public int hashCode() {
        return objID.hashCode();
    }

    /**
     * Indicates whether the argument transaction is the same as this one
     *
     * @param trans the <code>Transaction</code> with which to compare
     * @return <code>true</code> if this transaction is the same as the
     *         argument transaction; <code>false</code> otherwise
     */
    public boolean equals(final Transaction trans) {
        return transID.equals(trans.transID);
    }

    /**
     * Indicates whether this transaction has the argument
     * <code>TransactionID</code> as its <code>ID</code>
     *
     * @param trID the <code>TransactionID</code> with which to compare
     * @return <code>true</code> if this transaction's <code>ID</code> is the
     *         same as the argument <code>TransactionID</code>;
     *         <code>false</code> otherwise
     */
    public boolean equals(final TransactionID trID) {
        return transID.equals(trID);
    }

    /**
     * Indicates whether this transaction is operating on a particular object
     *
     * @param oID the <code>ObjectID</code> with which to compare
     * @return <code>true</code> if this transaction is operating on the
     *         object identified by the argument <code>ObjectID</code>;
     *         <code>false</code> otherwise
     */
    public boolean equals(final ObjectID oID) {
        return objID.equals(oID);
    }

    /**
     * Indicates whether this transaction is equal to some other object
     *
     * @param obj the <code>Object</code> with which to compare
     * @return <code>true</code> if this transaction is equal to the argument
     *         object; <code>false</code> otherwise
     */
    public boolean equals(final Object obj) {
        return ((obj instanceof Transaction && equals((Transaction)obj)) ||
                (obj instanceof TransactionID && equals((TransactionID)obj)) ||
                (obj instanceof ObjectID && objID.equals((ObjectID)obj)));
    }

    /**
     * Convert the <code>Transaction</code> to a string
     *
     * @return a string representation of the <code>Transaction</code>
     */
    public String toString() {
        return transID.toString();
    }
}
