/*
 * @(#)TransactionID.java TransactionID.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.transact;

import java.io.*;
import java.util.Iterator;
import mokan.comm.LogicalNode;
//import mokan.obj.LockRequest;
import mokan.thread.KanThread;
import mokan.util.ID;
import mokan.util.CanonicalizingIDSet;
import mokan.debug.Event;


/**
 * A unique identifier for a transaction.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 */
public class TransactionID extends ID implements Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 9118389336110150361L;

    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /**
     * A unique sequence number generator for each node.
     */
    private static long transCounter = 0;

    /**
     * A container for canonicalizing <code>TransactionID</code>s
     */
    protected static final CanonicalizingIDSet allTrans =
	new CanonicalizingIDSet();

    /**
     * The logical node responsible for this transaction.  This is the node
     * that issued the method call containing this transaction.  This is a
     * pseudo-final field.
     *
     * @serial the logical node that created this transaction
     */
    protected LogicalNode creator;

    /**
     * The <code>Transaction</code> identified by this <code>ID</code>, if it
     * is local
     */
    transient Transaction theTrans;

    /**
     * An exception that killed this transaction.  This is used when some
     * external agent kills the transaction.
     *
     * @see mokan.obj.Lock
     * @serial an exception that killed this transaction
     */
    public TransactionException killed;

    /**
     * The no-args constructor, needed for deserialization
     */
    public TransactionID() {
        super();
    }

    /**
     * Create a <code>TransactionID</code>
     *
     * @param node the <code>LogicalNode</code> on which to create this
     *             <code>TransactionID</code>
     */
    public TransactionID(final LogicalNode node) {
        this(node, transCounter++);
    }

    /**
     * Create a <code>TransactionID</code> with a particular sequence number
     *
     * @param seq a unique sequence number for this node
     * @param node the <code>LogicalNode</code> on which to create this
     *             <code>TransactionID</code>
     */
    TransactionID(final LogicalNode node, final long seq) {
        super(node, seq);
        allTrans.add(this);
        creator = KanThread.currCall().getAccessor();
    }

    public Transaction getTransaction() {
        return theTrans;
    }

    /**
     * Abort the transaction.  The abort is signalled with an interrupt if the
     * transaction is local, or with an <code>AbortMsg</code> otherwise.
     *
     * @param killed the <code>TransactionException</code> that killed this
     *        transaction
     */
    public void abort(final TransactionException killed) {
        this.killed = killed;
        if (theTrans != null) {
            Event.DEBUG(Event.TRANSACTION, this + " aborting");
            theTrans.abort(killed);
        } else if (node != LogicalNode.getMyNode()) {
            new AbortMsg(null, this).send(node);
        } // Otherwise the transaction <em>was</em> local, but is already dead
    }


    /**
     * Kill the transaction.  The abort is signalled with an interrupt
     * or exception when transaction is local
     *
     * @param killed the <code>TransactionException</code> that killed this
     *        transaction
     */
    public void kill(final TransactionException killed) {
        this.killed = killed;
        if (theTrans != null) {
            theTrans.abort(killed);
        } else
            throw killed;
    }

    /**
     *  Abort all transtaction for a node.
     */
    public static void abortTransactions (LogicalNode node, Exception e) {
        Event.DEBUG(Event.TRANSACTION, "TransactionID.abortTransactions: ", node);
        CanonicalizingIDSet tmp = (CanonicalizingIDSet) allTrans.clone();
        Iterator ti = tmp.iterator();

        while (ti.hasNext()) {
            TransactionID tid = (TransactionID)ti.next();
            //XXX What about the creator
            if (node== null || tid.getNode () == node)
                tid.abort (new TransactionUserException (e));
        }
    }

    /**
     * Remove this transaction ID from the list of all transaction IDs
     */
    void remove() {
        allTrans.remove(this);
    }

    /**
     * Return the logical node responsible for this transaction.  This is the
     * node that issued the method call containing this transaction.
     */
    public LogicalNode getAccessor() {
        return creator;
    }

    public boolean equals(final LockRequest lockReq) {
        return lockReq.equals(this);
    }

    public boolean equals(final TransactionID transID) {
        return super.equals(transID);
    }

    public boolean equals(final Object obj) {
        return ((obj instanceof LockRequest) && equals((LockRequest)obj) ||
                super.equals(obj));
    }

    public Object readResolve() throws ObjectStreamException {
        final TransactionID canonicalID =
            (TransactionID)allTrans.canonicalize(this);
        canonicalID.killed = killed;
        return canonicalID;
    }

    /**
     * Save the state of the <code>TransactionID</code> instance to a stream
     * (i.e., serialize it).
     *
     * @serialData The superclass data is written out, then the logical node
     *             that created this transaction (LogicalNode), followed by an
     *             exception that killed this transaction (Exception).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeObject(creator);
        out.writeObject(killed);
    }

    public void readExternal(final ObjectInput in)
        throws ClassNotFoundException, IOException {

        super.readExternal(in);
        creator = (LogicalNode)in.readObject();
        killed = (TransactionException)in.readObject();
    }

    public String toString() {
        return "Trans" + super.toString();
    }
}
