/*
 * @(#)$Id
 *
 *
 */

package mokan.ui;

import java.awt.*;
import java.awt.event.*;
import java.net.UnknownHostException;
import java.util.Vector;
import javax.swing.*;
import mokan.Kan;
import mokan.app.AppInfo;
import mokan.comm.KanComm;
import mokan.comm.MachinePort;
import mokan.debug.Event;

/**
 * Choose an application to run on Kan
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:29 $
 * @author	Sandeep Bhatia
 * @author	Jerry James
 */
final class AppLauncherGUI extends JFrame implements ActionListener {

    /**
     * The first part of the message to give the user for an invalid number of
     * machines
     */
    private static final String machineMsg1 =
	"Invalid choice for # of machines.\nPlease check input and try again.\n\nMachines available = ";

    /**
     * The second part of the message to give the user for an invalid number
     * of machines
     */
    private static final String machineMsg2 = "\nYou chose: ";

    /**
     * The message to give the user when no application was selected
     */
    private static final String appMsg = "No application selected!";

    /**
     * The window listener, which closes the window when the user clicks the
     * appropriate button
     *
     * @serial the window listener for this frame
     */
    private final WindowAdapter winListener;

    /**
     * The visual list of applications that can be executed on Kan
     *
     * @serial a GUI list of applications that Kan knows about
     */
    private final JList lstAvailApps;

    /**
     * The list of applications that have been launched on Kan
     *
     * @serial the list of applications that have been launched on Kan
     */
    private final DefaultListModel launchedAppList = new DefaultListModel();

    /**
     * The number of machines on which to run the application
     *
     * @serial a text field containing the number of machines on which to run
     */
    private final JTextField txtNoMachines;

    /**
     * The arguments to pass to the application
     *
     * @serial a text field containing the arguments to the application
     */
    private final JTextField txtArguments;

    /**
     * The debug events to print
     *
     * @serial a GUI element containing the debug events to print
     */
    //private final DebugList listEvents = new DebugList();

    /**
     * The list of applications that are available to be chosen
     *
     * @serial an array of application names
     */
    private final String[] availableApps;

    /**
     * The list of applications that are running
     *
     * @serial a vector of running applications
     */
    private final Vector runningApps;

    /**
     * The connected machines and launch status of applications on them
     *
     * @serial the connected machines and launch status of applications on
     *         those machine
     */
    private final KanStatus currKVMStatus;

    /**
     * The connected machines
     *
     * @serial an array of strings containing names of connected machines
     */
    private final String[] KVMMachines;

    /**
     * The command to start Kan on a remote node
     *
     * @serial a string containing a command to start Java remotely
     */
    private final String kanCommand;

    /**
     * The button for launching an application
     *
     * @serial a button for launching applications
     */
    private final JButton btnOk = new JButton("Ok");

    /**
     * The button for not launching an application
     *
     * @serial a button for not launching applications
     */
    private final JButton btnCancel = new JButton("Cancel");

    /**
     * Construct a new <code>AppLauncherGUI</code> and let the user choose
     * applications to launch
     *
     * @param availApps the applications that can be launched
     * @param runApps the applications that are already running
     * @param status the connected machines and application status
     * @param kanStart the command to start Kan on a remote node
     */
    AppLauncherGUI(final String[] availApps, final Vector runApps,
		   final KanStatus status, final String kanStart) {
	// Set up the JFrame
	super("Kan Application Launcher");

	// Set up the data holding fields
	availableApps = availApps;
	runningApps = runApps;
	currKVMStatus = status;
	kanCommand = kanStart;
	KVMMachines = status.getMachines();

	// Set up the application lists
	for (int i = 0; i < runApps.size(); i++) {
	    final AppChoice curApp = (AppChoice)runApps.elementAt(i);
	    launchedAppList.addElement(curApp.getAppName());
	}

	// Initialize the GUI elements
	lstAvailApps = new JList(availApps);
	final JList lstLaunchedApps = new JList(launchedAppList);
	txtNoMachines = new JTextField(String.valueOf(KVMMachines.length));
	txtArguments = new JTextField();

	final JLabel lblAvailApps = new JLabel("Available Applications");
	lblAvailApps.setLabelFor(lstAvailApps);

	final JLabel lblLaunchedApps = new JLabel("Launched Applications");
	lblLaunchedApps.setLabelFor(lstLaunchedApps);

	final JLabel lblNoMachines = new JLabel("No. of Machines:");
	lblNoMachines.setLabelFor(txtNoMachines);
	lblNoMachines.setDisplayedMnemonic('N');

	final JLabel lblArguments = new JLabel("App Arguments:");
	lblArguments.setLabelFor(txtArguments);
	lblArguments.setDisplayedMnemonic('A');

	// Set up the left pane
	final GridBagLayout bag = new GridBagLayout();
	final JPanel leftPane = new JPanel(bag, true);

	// Add the GUI elements
	final GridBagConstraints cons = new GridBagConstraints();
	final Insets deepInsets = new Insets(0, 4, 4, 4);
	final Insets standardInsets = new Insets(0, 4, 2, 4);

	// First label
	cons.gridwidth = GridBagConstraints.RELATIVE;
	cons.insets = standardInsets;
	cons.weightx = 1.0;
	bag.setConstraints(lblAvailApps, cons);
	leftPane.add(lblAvailApps);

	// Second label
	cons.gridwidth = GridBagConstraints.REMAINDER;
	bag.setConstraints(lblLaunchedApps, cons);
	leftPane.add(lblLaunchedApps);

	// First list
	cons.fill = GridBagConstraints.BOTH;
	cons.gridwidth = GridBagConstraints.RELATIVE;
	cons.insets = deepInsets;
	cons.weighty = 1.0;
	bag.setConstraints(lstAvailApps, cons);
	leftPane.add(lstAvailApps);

	// Second list
	cons.gridwidth = GridBagConstraints.REMAINDER;
	bag.setConstraints(lstLaunchedApps, cons);
	leftPane.add(lstLaunchedApps);

	// Third label
	cons.anchor = GridBagConstraints.EAST;
	cons.fill = GridBagConstraints.NONE;
	cons.gridwidth = GridBagConstraints.RELATIVE;
	cons.weightx = 0.0;
	cons.weighty = 0.0;
	bag.setConstraints(lblNoMachines, cons);
	leftPane.add(lblNoMachines);

	// First text field
	cons.anchor = GridBagConstraints.WEST;
	cons.fill = GridBagConstraints.BOTH;
	cons.gridwidth = GridBagConstraints.REMAINDER;
	cons.weightx = 1.0;
	bag.setConstraints(txtNoMachines, cons);
	leftPane.add(txtNoMachines);

	// Fourth label
	cons.anchor = GridBagConstraints.EAST;
	cons.fill = GridBagConstraints.NONE;
	cons.gridwidth = GridBagConstraints.RELATIVE;
	cons.weightx = 0.0;
	cons.weighty = 0.0;
	bag.setConstraints(lblArguments, cons);
	leftPane.add(lblArguments);

	// Second text field
	cons.anchor = GridBagConstraints.WEST;
	cons.fill = GridBagConstraints.BOTH;
	cons.gridwidth = GridBagConstraints.REMAINDER;
	cons.weightx = 1.0;
	bag.setConstraints(txtArguments, cons);
	leftPane.add(txtArguments);

	// First button
	cons.anchor = GridBagConstraints.CENTER;
	cons.gridwidth = 1;
	cons.insets = standardInsets;
	bag.setConstraints(btnOk, cons);
	leftPane.add(btnOk);

	// Second button
	bag.setConstraints(btnCancel, cons);
	leftPane.add(btnCancel);

	// Set up the right pane
	final JPanel rightPane = new JPanel(new BorderLayout());
	rightPane.setBorder(BorderFactory.createEmptyBorder(0, 2, 2, 2));
	final JLabel lblEvents = new JLabel("Debug Events:");
	//lblEvents.setLabelFor(listEvents);
	lblEvents.setDisplayedMnemonic('D');
	rightPane.add(lblEvents, BorderLayout.NORTH);
	//rightPane.add(listEvents, BorderLayout.CENTER);

	// Set up the content pane
	final JSplitPane mainPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
						   leftPane, rightPane);
	setContentPane(mainPane);

	// Add the listeners
	btnOk.addActionListener(this);
	btnCancel.addActionListener(this);
	txtNoMachines.addActionListener(this);
	txtArguments.addActionListener(this);
	winListener = new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {
		AppLauncherGUI.this.dispose();
		System.exit(0);
	    }
	};
	addWindowListener(winListener);

	// Show the window
	pack();
	setVisible(true);
    }

    /**
     * Respond to a user action
     *
     * @param ae the user action to process
     */
    public void actionPerformed(final ActionEvent ae) {
	if (ae.getActionCommand().equals("Ok")) {
	    int machinesSelected;
	    try {
		machinesSelected = Integer.parseInt(txtNoMachines.getText());
	    } catch (NumberFormatException badNumEx) {
		machinesSelected = 0;
	    }
	    if (machinesSelected > KVMMachines.length || machinesSelected<1) {
		JOptionPane.showMessageDialog(this, machineMsg1 +
					      KVMMachines.length + machineMsg2
					      + machinesSelected,
					      "Choose machines",
					      JOptionPane.ERROR_MESSAGE);
	    } else if (lstAvailApps.isSelectionEmpty()) {
		JOptionPane.showMessageDialog(this, appMsg,
					      "Choose application",
					      JOptionPane.ERROR_MESSAGE);
	    } else {
		final String[] realArgs = parseArgs();
		final int[] events = null;//listEvents.getList();
	mokan.debug.Event.setEventMask(events);
		try {
		    final AppInfo newApplication =
			new AppInfo((String)lstAvailApps.getSelectedValue(),
				    realArgs, false);

		    // have to use a round robin assignment for the machines
		    final int start = currKVMStatus.lastMachMarker;
		    final int end = start + machinesSelected;
		    final int newMachMarker = end % KVMMachines.length;
		    final String portName =
			Kan.systemProps.getProperty("tcp.port");
		    final MachinePort[] launchMachines =
			new MachinePort[machinesSelected];
		    for (int i = start; i < end; i++) {
			launchMachines[i - start] =
			    new MachinePort(KVMMachines[i%KVMMachines.length]);
			System.out.println("AppLauncherGUI::Adding " +
					   KVMMachines[i % KVMMachines.length]
					   + " to the list" );
		    }
		    currKVMStatus.lastMachMarker = newMachMarker;
		    System.out.print("AppLauncherGUI::Launching the application (");
		    System.out.println(lstAvailApps.getSelectedValue() +
				       ") with " + machinesSelected +
				       " machines");
		    KanComm.startApp(newApplication, launchMachines);
		    final String val = (String)lstAvailApps.getSelectedValue();
		    runningApps.addElement(new AppChoice(val, kanCommand));
		    launchedAppList.addElement(val);
		    // lstAvailApps.validate();
		} catch (UnknownHostException uhe) {
		    // FIXME: Do something reasonable here
		    uhe.printStackTrace();
		}
	    }
	} else if (ae.getActionCommand().equals("Cancel")) {
	    // Remove the listeners
	    btnOk.removeActionListener(this);
	    btnCancel.removeActionListener(this);
	    txtNoMachines.removeActionListener(this);
	    txtArguments.removeActionListener(this);
	    removeWindowListener(winListener);

	    // Clean up
	    dispose();
	}
    }

    /**
     * Break the application arguments up into an array of strings
     *
     * @return an array of strings holding the application arguments
     */
    private String[] parseArgs() {
	final Vector arguments = new Vector();
	final String argString = txtArguments.getText().trim();

	boolean quote = false;
	int index = 0;
	for (int i = 0; i < argString.length(); i++) {
	    switch (argString.charAt(i)) {
	    case ' ':
		if (!quote) {
		    arguments.add(argString.substring(index, i));
		    // Don't need to check argString.length here, because we
		    // trimmed argString (it can't end in a space)
		    while (argString.charAt(i + 1) == ' ') i++;
		    index = i + 1;
		}
		break;
	    case '"':
		if (i == 0 || argString.charAt(i - 1) != '\\') {
		    quote = !quote;
		    if (!quote) {
			arguments.add(argString.substring(index + 1, i));
			while (i + 1 < argString.length() &&
			       argString.charAt(i + 1) == ' ')
			    i++;
			index = i + 1;
		    }
		}
		// FIXME: Replace escaped quotes with actual quote characters
		break;
	    default:
		break;
	    }
	}

	final int numArgs = arguments.size();
	final String[] realArgs = new String[numArgs];
	arguments.copyInto(realArgs);
	return realArgs;
    }
}
