/*
 * @(#)$Id: DebugEntry.java,v 1.1.1.1 2002/11/06 22:10:29 kris Exp $
 *
 *
 */

package mokan.ui;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import javax.swing.tree.TreeNode;

/**
 * One type of debug event that the user can select to view
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:29 $
 * @author	Jerry James
 */
final class DebugEntry implements TreeNode {

    /**
     * The type of this entry, from a canonical list of types
     */
    private final int value;

    /**
     * The message to print when this debug event occurs
     */
    private final String msg;

    /**
     * For generalized debug entry types, the list of their more specific
     * subtypes
     */
    private final DebugEntry[] children;

    /**
     * The next more general debug entry type, or <code>null</code> if none
     */
    private DebugEntry parent;

    /**
     * Create a new debug entry object
     *
     * @param value the type of this entry, from a canonical list of types
     * @param msg the message to print when this event occurs
     * @param children the subtypes of this type
     */
    DebugEntry(final int value, final String msg, final DebugEntry[] children)
    {
	this.value    = value;
	this.msg      = msg;
	this.children = (children != null) ? children : new DebugEntry[0];
    }

    /**
     * Get the type of this entry
     *
     * @return an integer representing the type of this entry
     */
    int getValue() {
	return value;
    }

    /**
     * Set the parent (next more general type) of this entry
     *
     * @param parent the parent of this entry
     */
    void setParent(final DebugEntry parent) {
	this.parent = parent;
    }

    /**
     * Get a specific child of this entry
     *
     * @param childIndex the index of the child to retrieve
     * @return a <code>TreeNode</code> containing the selected child
     */
    public TreeNode getChildAt(final int childIndex) {
	return children[childIndex];
    }

    /**
     * Determine how many children this entry has
     *
     * @return the number of children this entry has
     */
    public int getChildCount() {
	return children.length;
    }

    /**
     * Retrieve the parent of this entry
     *
     * @return the parent of this entry
     */
    public TreeNode getParent() {
	return parent;
    }

    /**
     * Determine the index of a child of this entry
     *
     * @param node the child whose index is to be determined
     * @return the index of <var>node</var>, or -1 if it is not a child of
     *        this entry
     */
    public int getIndex(final TreeNode node) {
	for (int i = 0; i < children.length; i++)
	    if (children[i] == node)
		return i;
	return -1;
    }

    /**
     * Determine whether this entry can have any children at all
     *
     * @return <code>true</code> if this entry can have children;
     *         <code>false</code> otherwise
     */
    public boolean getAllowsChildren() {
	return (children.length != 0);
    }

    /**
     * Determine whether this entry does have any children
     *
     * @return <code>true</code> if this entry has no children (i.e., it is a
     *         leaf node); <code>false</code> otherwise
     */
    public boolean isLeaf() {
	return (children.length == 0);
    }

    /**
     * Get an enumeration of the children of this entry
     *
     * @return an enumeration of the children
     */
    public Enumeration children() {
	class DebugEntryEnumeration implements Enumeration {
	    private int count = 0;

	    public boolean hasMoreElements() {
		return (count < children.length);
	    }

	    public Object nextElement() throws NoSuchElementException {
		if (count >= children.length)
		    throw new NoSuchElementException();
		return children[count++];
	    }
	}

	return new DebugEntryEnumeration();
    }

    /**
     * Convert this <code>DebugEntry</code> into a string
     *
     * @return a string representation of this <code>DebugEntry</code>
     */
    public String toString() {
	return msg;
    }
}
