/*
 * @(#)$Id
 *
 *
 */

package mokan.ui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * Ask the user to input a valid username and password
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:29 $
 * @author	Jerry James
 */
final class LoginGUI implements ActionListener, FocusListener {

    private static final String WINDOWNAME = "Kan Login";
    private static final String USERNAME = "user";
    private static final String PASSWORD = "password";
    private static final String OKAY = "okay";
    private static final String CANCEL = "cancel";
    private static final int CLICKLENGTH = 200;

    /**
     * A <code>JTextField</code> for typing in a valid username.  This will be 
     * used to login to remote systems.
     */
    private final JTextField usernameField = new JTextField();

    /**
     * A <code>JPasswordField</code> for typing in a valid password.  This
     * will be used to login to remote systems.
     */
    private final JPasswordField passwordField = new JPasswordField();

    /**
     * A <code>JButton</code> for telling the UI to go ahead
     */
    private final JButton okayButton = new JButton("OK");

    /**
     * A <code>JButton</code> for telling the UI to quit
     */
    private final JButton cancelButton = new JButton("Cancel");

    /**
     * The main dialog for the GUI
     */
    private final JDialog dialog;

    /**
     * The main interactive pane of the dialog
     */
    private final JOptionPane pane;

    /**
     * A class of JLabel's that never accept the focus.  We have to do this to 
     * avoid having <var>usernameLabel</var> take the focus right away.
     */
    private static final class UnfocusableJLabel extends JLabel {

	/**
	 * Creates an <code>UnfocusableJLabel</code> instance with the
	 * specified text and horizontal alignment.  The label is centered
	 * vertically in its display area.
	 *
	 * @param text  The text to be displayed by the label.
	 * @param horizontalAlignment  One of the following constants
	 *           defined in <code>SwingConstants</code>:
	 *           <code>LEFT</code>,
	 *           <code>CENTER</code>, or
	 *           <code>RIGHT</code>.
	 */
	UnfocusableJLabel(final String text, final int horizontalAlignment) {
	    super(text, horizontalAlignment);
	}

	/**
	 * This prevents an <code>UnfocusableJLabel</code> from ever accepting 
	 * the focus
	 */
	public boolean isFocusTraversable() {
	    return false;
	}
    }

    /**
     * Create a new GUI to accept login information from the user
     */
    LoginGUI() {

	// Create the labels
	final UnfocusableJLabel usernameLabel =
	    new UnfocusableJLabel("User name: ", JLabel.LEFT);
	final UnfocusableJLabel passwordLabel =
	    new UnfocusableJLabel("Password: ", JLabel.LEFT);

	// Set up the labels
	usernameLabel.setLabelFor(usernameField);
	usernameLabel.setDisplayedMnemonic('U');
	passwordLabel.setLabelFor(passwordField);
	passwordLabel.setDisplayedMnemonic('P');

	// Set the actions to generate
	usernameField.setActionCommand(USERNAME);
	passwordField.setActionCommand(PASSWORD);
	okayButton.setActionCommand(OKAY);
	cancelButton.setActionCommand(CANCEL);

	// Set the listeners to this object
	usernameField.addActionListener(this);
	passwordField.addActionListener(this);
	okayButton.addActionListener(this);
	cancelButton.addActionListener(this);

	// Set tool tips (balloon help)
	usernameLabel.setToolTipText("Username for remote systems");
	passwordLabel.setToolTipText("Password for remote systems");
	okayButton.setToolTipText("Use this username and password");
	cancelButton.setToolTipText("Quit Kan");

	// Set the focus order
	usernameField.setNextFocusableComponent(passwordField);
	passwordField.setNextFocusableComponent(okayButton);
	okayButton.setNextFocusableComponent(cancelButton);
	cancelButton.setNextFocusableComponent(usernameField);

	// Create the panel holding the labels and text fields
	final JPanel textPanel = new JPanel(new GridLayout(2, 2), false);
	textPanel.add(usernameLabel);
	textPanel.add(usernameField);
	textPanel.add(passwordLabel);
	textPanel.add(passwordField);

	// Create the panel that will hold the entire UI
	final JPanel mainPanel = new JPanel(false);

	// Set the user's options
	final Object[] options = { okayButton, cancelButton };

	// Create the pane containing the buttons
	pane = new JOptionPane(textPanel, JOptionPane.PLAIN_MESSAGE,
			       JOptionPane.DEFAULT_OPTION, null, options,
			       usernameField);

	// Create the main dialog, set the default button, and make the dialog 
	// give the focus to usernameField if it gets it (only initially)
	dialog = pane.createDialog(mainPanel, WINDOWNAME);
	dialog.getRootPane().setDefaultButton(okayButton);
	dialog.addFocusListener(this);

	// Show the dialog and get the user's input
	dialog.show();
	if (pane.getValue() != okayButton)
	    System.exit(0);
	final String password = new String(passwordField.getPassword());
	Commander.setLogin(usernameField.getText(), password);

	// Remove the listeners; otherwise they are never GCed!
	dialog.removeFocusListener(this);
	usernameField.removeActionListener(this);
	passwordField.removeActionListener(this);
	okayButton.removeActionListener(this);
	cancelButton.removeActionListener(this);
    }

    /**
     * Process an action taken by the user (button press)
     *
     * @param ev the event spawned by the user action
     */
    public void actionPerformed(final ActionEvent ev) {
	final String command = ev.getActionCommand();
	if (command.equals(USERNAME)) {
	    passwordField.requestFocus();
	} else if (command.equals(PASSWORD)) {
	    okayButton.doClick(CLICKLENGTH);
	    pane.setValue(okayButton);
	    dialog.dispose();
	} else if (command.equals(OKAY)) {
	    pane.setValue(okayButton);
	    dialog.dispose();
	} else if (command.equals(CANCEL)) {
	    pane.setValue(cancelButton);
	    dialog.dispose();
	}
    }

    /**
     * When the dialog gets the focus, make it give it to
     * <var>usernameField</var>
     *
     * @param ev the event created by the change in focus
     */
    public void focusGained(final FocusEvent ev) {
	usernameField.requestFocus();
    }

    /**
     * Take no action when the focus goes away
     *
     * @param ev the event created by the change in focus
     */
    public void focusLost(final FocusEvent ev) { }
}
