/*
 * @(#)$Id: RemoteTelnet.java,v 1.1.1.1 2002/11/06 22:10:29 kris Exp $
 *
 *
 */

package mokan.ui;

import java.net.InetAddress;

/** 
 * This class is a wrapper around a Telnet object
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:29 $
 * @author	Sean Brydon
 * @author	Jing Wang
 * @author	Sandeep Bhatia
 */
final class RemoteTelnet implements Runnable {

    /**
     * The wrapped <code>Telnet</code> object
     */
    private Telnet connection;

    /**
     * The user name for logging into a remote system
     */
    private final String user;

    /**
     * The password for logging into a remote system
     */
    private final String password;

    /**
     * Accessed by run--will cause race condition if each telnet client does
     * not have its own <code>RemoteTelnet</code> object
     */
    private InetAddress dest;

    /**
     * This is a string composed of "set classpath", "change to proper
     * directory", "javamokan.Kan args" where <var>args</var> are the
     * commandline switches and list of machines in your world
     */
    private final String command;  

    /**
     * The frame around the telnet client
     */
    private MonitorClient monitor;

    /**
     * Used to kill this thread
     */
    private boolean isRunning = false;

    /**
     * Show monitor telnet window
     */
    private boolean isMonitorOn = false;

    /**
     * Show output on monitor telnet window
     */
    private boolean isPrintOn = true;

    /**
     * Create a new <code>RemoteTelnet</code> object
     *
     * @param user the user name for logging into a remote system
     * @param password the password for logging into a remote system
     * @param command the command to issue to start Kan on a remote system
     */
    RemoteTelnet(final String user, final String password,
		 final String command) {
        this.user = user;
        this.password = password;
        this.command = command;
    }

    /**
     * Start the telnet session
     *
     * @param dest the remote computer to telnet to
     * @return <code>true</code> if successful, <code>false</code> if not
     */
    boolean doTelnet(final InetAddress dest) {
        this.dest = dest; //used to pass into run method also

	if (connection == null) {
	    try {
		connection = new Telnet(dest, user, password);
	    } catch (Exception e) {
		System.out.println("Can not telnet to " + dest.getHostName());
		return false;
	    }
	}
	connection.sendComm(command);
	new Thread(this).start();
	return true;
    }

    /**
     * A newly spawned thread executes here to watch the telnet connection
     */
    public void run() {
	if (connection == null) {
	    System.err.println("Telnet error at RemoteTelnet");
	    return;
	}

        isRunning = true;
	isMonitorOn = false; //initially open window-
	isPrintOn = true;   //initially print to windows
	if (isMonitorOn == true) {
	    monitor = new MonitorClient(dest);   //initially create a gui frame
	}

	while (isRunning) {
	    if (isMonitorOn) {
		final char c = connection.readch();
		if ((int)c == -1) {
		    monitor.disconnect();
		    break;
		}
		if (isPrintOn) {
		    monitor.displaych(c);
		}
	    }
	    Thread.yield();
	}
	System.out.println("RemoteTelnet:run is EXITING Thread *************");
    }

    /**
     * Kill a remote telnet session
     *
     * @param mName the name of the remote machine
     */
    void killRemoteTelnet(final String mName) {
	sendRemoteCommand("killall", mName ); 
	System.out.println("RemoteTelnet.killRemoteTelnet::" + mName +
			   " is now Dead");
    }

    /**
     * Send a command to a remote telnet session
     *
     * @param command the command to send
     * @param machineName the name of the remote machine
     */
    void sendRemoteCommand(final String command, final String machineName) {
	System.out.println("doTelnet.sendRemoteCommand executing "  + command
			   + " at" +  machineName);
	connection.sendComm(command);
    }

    /**
     * Determine whether the monitor is on
     *
     * @return <code>true</code> if the monitor is on; <code>false</code>
     *         otherwise
     */
    boolean getIsMonitorOn() {
        return isMonitorOn;
    }

    /**
     * Toggle the monitor's on/off state
     */
    void toggleMonitor() {
        if (isMonitorOn) {
	    turnMonitornOff();
	} else { 
	    turnMonitornOn();
	}
    }

    // REVIEW: Can't we just create one GUI frame and then hide/show it?
    /**
     * Turn the monitor on, by creating a GUI frame for it
     */
    void turnMonitornOn() {
        if (!isMonitorOn) {
	    monitor = new MonitorClient(dest);   //create a gui frame
	    isMonitorOn = true;
	}
    }

    /**
     * Turn the monitor off, by killing its GUI frame
     */
    void turnMonitornOff() {
        if (isMonitorOn) {
	    monitor.killMe();     //destroy window
	    isMonitorOn = false;
	}
    }

    /**
     * Toggle printing of captured output to the monitor window
     */
    void togglePrint() {
	isPrintOn = !isPrintOn;
    }
}
