/*
 * @(#)$Id: CanonicalizingIDSet.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 */

package mokan.util;

import java.io.Serializable;
import mokan.comm.LogicalNode;
import java.util.*;

/**
 * A set that allows lookups of canonical versions of {@linkmokan.util.ID IDs}.
 * The underlying data structure is a hash set.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 * @see	mokan.util.ID
 */
public final class CanonicalizingIDSet implements Cloneable, Serializable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = -4517533184094935136L;






    /**
     * The elements in the set
     *
     * @serial an array containing the elements of the set
     */
    private ID[] table;

    /**
     * The total number of elements in the set
     *
     * @serial an integer containing the number of elements in the set
     */
    private int count;

    /**
     * The set is rehashed when its size exceeds this threshold.  (The value
     * of this field is <code>(int)(capacity * loadFactor))</code>
     *
     * @serial an integer containing the threshold size for the set
     */
    private int threshold;

    /**
     * The load factor for the hashtable
     *
     * @serial a floating point number containing the load factor for the set
     */
    private float loadFactor;
    /**
     * The number of times this Set has been structurally modified
     * Structural modifications are those that change the number of mappings in
     * the Set or otherwise modify its internal structure (e.g.,
     * rehash).  This field is used to make iterators on Collection-views of
     * the Set fail-fast.  (See ConcurrentModificationException).
     */
    private transient int modCount = 0;


    /**
     * Constructs a new, empty set with the specified initial capacity and the
     * specified load factor
     *
     * @param initialCapacity the initial capacity of the set
     * @param loadFactor the load factor of the set
     * @exception IllegalArgumentException if the initial capacity is less
     *        than zero, or if the load factor is nonpositive
     */
    public CanonicalizingIDSet(int initialCapacity, final float loadFactor) {
	if (initialCapacity < 0)
	    throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        if (loadFactor <= 0.0f)
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
        if (initialCapacity == 0)
            initialCapacity = 1;
	this.loadFactor = loadFactor;
	table = new ID[initialCapacity];
	threshold = (int)(initialCapacity * loadFactor);
    }

    /**
     * Constructs a new, empty set with the specified initial capacity and
     * default load factor, which is <code>0.75</code>.
     *
     * @param initialCapacity the initial capacity of the set
     * @exception IllegalArgumentException if the initial capacity is less
     *        than zero
     */
    public CanonicalizingIDSet(final int initialCapacity) {
	this(initialCapacity, 0.75f);
    }

    /**
     * Constructs a new, empty set with a default capacity, which is
     * <code>101</code>, and a default load factor, which is <code>0.75</code>.
     */
    public CanonicalizingIDSet() {
	this(101, 0.75f);
    }

    /**
     * Returns the number of entries in this set
     *
     * @return the number of entries in this set
     */
    public int size() {
	return count;
    }

    /**
     * Returns <code>true</code> if this set contains no
     * {@linkmokan.util.ID IDs}
     *
     * @return <code>true</code> if this set contains no
     *         {@linkmokan.util.ID IDs}
     */
    public boolean isEmpty() {
	return count == 0;
    }

    /**
     * Rehashes the contents of this set into a new
     * <code>CanonicalizingIDSet</code> instance with a larger capacity.  This
     * method is called automatically when the number of elements in this set
     * exceeds its capacity and load factor.
     */
    private void rehash() {
	final int oldCapacity = table.length;
	final ID[] oldSet = table;

	final int newCapacity = oldCapacity * 2 + 1;
	final ID[] newSet = new ID[newCapacity];

        modCount++;
	threshold = (int)(newCapacity * loadFactor);
	table = newSet;

	for (int i = oldCapacity ; i-- > 0 ;) {
	    for (ID old = oldSet[i] ; old != null ; ) {
		final ID id = old;
		old = old.next;

		final int index = (id.myHashCode & 0x7FFFFFFF) % newCapacity;
		id.next = newSet[index];
		newSet[index] = id;
	    }
	}
    }

    /**
     * Adds the specified {@linkmokan.util.ID ID} to the set.  This method
     * assumes that the {@linkmokan.util.ID ID} is not already present in the
     * set.  <strong>If this assumption is violated, the result will be a
     * circular data structure that will cause infinite looping later in the
     * program!</strong>
     *
     * @param id the {@linkmokan.util.ID ID} to add to the set
     */
    public synchronized void add(final ID id) {
	final int key = id.myHashCode;
	ID[] tab = table;
	int index = (key & 0x7FFFFFFF) % tab.length;

	modCount++;
	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();

	    tab = table;
	    index = (key & 0x7FFFFFFF) % tab.length;
	}

	// Create the new entry.
	count++;
	id.next = tab[index];
	tab[index] = id;
    }

    /**
     * Find (and possibly set) the canonical version of an
     * {@linkmokan.util.ID ID}
     *
     * @param id the {@linkmokan.util.ID ID} to canonicalize
     * @return the canonical version of <var>id</var> if it is in the set, or
     *        <var>id</var> itself if not.  In the latter case, <var>id</var>
     *        is added to the set.
     */
    public synchronized ID canonicalize(final ID id) {
	final int key = id.myHashCode;
	ID[] tab = table;
	int index = (key & 0x7FFFFFFF) % tab.length;

	// See if it is in there already
	for (ID theID = tab[index]; theID != null; theID = theID.next) {
	    if (key == theID.myHashCode && id.equals(theID)) {
		return theID;
	    }
	}

	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();

	    tab = table;
	    index = (key & 0x7FFFFFFF) % tab.length;
	}

	// Create the new entry.
	count++;
	id.next = tab[index];
	tab[index] = id;
	return id;
    }

    /**
     * Find the canonical version of an {@linkmokan.util.ID ID}
     *
     * @param node the logical node of the {@linkmokan.util.ID ID} to find
     * @param count the counter of the {@linkmokan.util.ID ID} to find
     * @return the canonical version of the {@linkmokan.util.ID ID} to find
     */
    public synchronized ID find(final LogicalNode node, final long count) {
	final int key = (int)(3 * count + 191 * node.hashCode());
	final ID[] tab = table;
	final int index = (key & 0x7FFFFFFF) % tab.length;

	for (ID id = tab[index]; id != null; id = id.next) {
	    if (key == id.myHashCode && id.equals(node, count)) {
		return id;
	    }
	}
	return null;
    }

    /**
     * Remove this {@linkmokan.util.ID ID} from the set if present
     *
     * @param id the {@linkmokan.util.ID ID} to be removed from the set
     * @return <code>true</code> if <var>id</var> was in the set
     */
    public synchronized boolean remove(ID id) {
	final int key = id.myHashCode;
	final ID[] tab = table;

	final int index = (key & 0x7FFFFFFF) % tab.length;
	for (ID theID = tab[index], prev = null; theID != null;
	     prev = theID, theID = theID.next) {
	    if (key == theID.myHashCode && id.equals(theID)) {
                modCount++;
		if (prev != null)
		    prev.next = theID.next;
		else
		    tab[index] = theID.next;

		count--;
		theID.next = null;
		return true;
	    }
	}
	return false;
    }

    /**
     * Remove all {@linkmokan.util.ID IDs} from this set
     */
    public synchronized void clear() {
	final ID[] tab = table;
	for (int index = tab.length; --index >= 0; )
	    tab[index] = null;
	count = 0;
	modCount++;
    }

    /**
     * Returns a shallow copy of this <code>CanonicalizingIDSet</code>
     * instance: the values themselves are not cloned
     *
     * @return a shallow copy of this set
     */
    public synchronized Object clone() {
	final CanonicalizingIDSet cl =
	    new CanonicalizingIDSet(table.length, loadFactor);
	for (int i = table.length ; i-- > 0 ; ) {
	    cl.table[i] = table[i];
	}
	cl.count = count;
	cl.modCount = 0;

	return cl;
    }

    /**
     * Return the size of the internal array comprising the set
     *
     * @return the length of the internal array
     */
    int capacity() {
        return table.length;
    }

    /**
     * Return the load factor for this set
     *
     * @return the load factor for this set
     */
    float loadFactor() {
        return loadFactor;
    }

    public Iterator iterator() {
	return new IDIterator();
    }


    private class IDIterator implements Iterator {
	final ID[] table = CanonicalizingIDSet.this.table;
	int index = table.length;
	ID entry = null;
	ID lastReturned = null;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	private int expectedModCount = modCount;

	IDIterator() { }

	public boolean hasNext() {
	    while (entry == null && index > 0)
		entry = table[--index];
	    return entry != null;
	}

	public Object next() {
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    while (entry == null && index > 0)
		entry = table[--index];

	    if (entry != null) {
		final ID e = lastReturned = entry;
		entry = e.next;
		return e;
	    }
	    throw new NoSuchElementException();
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    final ID[] tab = CanonicalizingIDSet.this.table;
	    final int index = (lastReturned.myHashCode & 0x7FFFFFFF) % tab.length;

	    for (ID e = tab[index], prev = null; e != null;
		 prev = e, e = e.next) {
		if (e == lastReturned) {
		    modCount++;
		    expectedModCount++;
		    if (prev == null)
			tab[index] = e.next;
		    else
			prev.next = e.next;
		    count--;
		    lastReturned = null;
		    return;
		}
	    }
	    throw new ConcurrentModificationException();
	}
    }



}
