/*
 * @(#)ID.java ID.java,v 1.2 2002/11/22 23:26:07 kris Exp $
 *
 *
 */

package mokan.util;

import java.io.*;
import mokan.app.Application;
import mokan.comm.LogicalNode;

/**
 * An abstract representation of an identifier.  This is merely a node (where
 * something is created), and a sequence number for that node.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Jerry James
 */
public abstract class ID implements Cloneable, Comparable, Externalizable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 4718922518402729349L;





    /**
     * The list of persistent fields for this class.  This is empty, because
     * that is the fastest option for <code>Externalizable</code> classes when
     * we are not worried about versioning.
     */
    private static final ObjectStreamField[] serialPersistentFields =
	ObjectStreamClass.NO_FIELDS;

    /**
     * A sequence number for the creating node.  This should generally be
     * unique for the node.  We use a 64-bit quantity to make the possibility
     * of overflow very remote.  (Even incrementing the counter at the rate of
     * once per nanosecond, it would still take close to 300 years for
     * overflow to occur if we started at zero.  Maybe we should start at the
     * most negative long, so that overflow would take nigh on 600 years!)
     * <p>This is a pseudo-final field.</p>
     *
     * @serial a 64-bit value containing a sequence number
     */
    protected long count;

    /**
     * The node where this <code>ID</code> was created.  This is a
     * pseudo-final field.
     *
     * @serial the creating logical node
     */
    protected LogicalNode node;

    /**
     * An integer holding the hashcode of this <code>ID</code> We compute this
     * once when the <code>ID</code> is created to cut down on costs arising
     * from excessive computation of the hashcode.  This is a pseudo-final
     * field.
     */
    protected transient int myHashCode;

    /**
     * A next pointer, for use in constructing chains of IDs
     *
     * @seemokan.util.CanonicalizingIDSet
     */
    transient ID next;

    /**
     * The no-args constructor, needed for deserialization
     */
    public ID() { }

    /**
     * Create an <code>ID</code>
     *
     * @param node the <code>LogicalNode</code> creating this <code>ID</code>
     * @param count the sequence number for this <code>ID</code>
     * @seemokan.comm.LogicalNode
     */
    public ID(final LogicalNode node, final long count) {
        this.node = (node == null) ? LogicalNode.getMyNode() : node;
        this.count = count;
        myHashCode = (int)(3*count + 191*node.hashCode() );
    }

    /**
     * Copy constructor
     *
     * @param id the <code>ID</code> to copy
     */
    public ID(final ID id) {
        this(id.node, id.count);
    }

    /**
     * Retrieve the node portion of this <code>ID</code>
     *
     * @return the creating node for this <code>ID</code>
     * @seemokan.comm.LogicalNode
     */
    public LogicalNode getNode() {
        return node;
    }

    /**
     * Retrieve the sequence number for this <code>ID</code>
     *
     * @return the sequence number for this <code>ID</code>
     */
    public final long getSeq() {
        return count;
    }

    /**
     * Retrieve the application that created this <code>ID</code>
     *
     * @return the application that created this <code>ID</code>
     */
    public final Application getApp() {
        return node.getApplication();
    }

    /**
     * Determine whether this <code>ID</code> was created locally
     *
     * @param node the <code>LogicalNode</code> that is considered
     *        <em>local</em>
     * @return <code>true</code> if this <code>ID</code> was created locally,
     *         else <code>false</code>
     * @seemokan.comm.LogicalNode
     */
    public final boolean isLocal(final LogicalNode node) {
        return this.node == node;
    }

    /**
     * Compare two <code>ID</code>s, where the <var>count</var> is most
     * significant, and the <var>node</var> is the least significant
     *
     * @param id the <code>ID</code> to be compared
     * @return the value 0 if the argument <code>ID</code> is equal to this
     *         <code>ID</code>; a value less than 0 if this <code>ID</code>
     *         has a lower sequence number or equal sequence number and lesser
     *         node than the <code>ID</code> argument; and a value greater
     *         than 0 if this <code>ID</code> has a greater sequence number of
     *         than the equal sequence number and greater node than the
     *         <code>ID</code> argument.
     */
    public final int compareTo(final ID id) {
        // Pseudo-final handling
        final long myCount = count;
        final long herCount = id.count;

        return (myCount == herCount)
            ? node.compareTo(id.node)
            : ((myCount > herCount) ? 1 : -1);
        // It is tempting to return the difference between count and herCount,
        // but that is a long, and casting a long to an int is not guaranteed
        // to preserve sign.
    }

    /**
     * Compare this <code>ID</code> to another object.  If the object is an
     * <code>ID</code>, this function behaves like
     * {@link #compareTo(ID) compareTo(ID)}.  Otherwise, it throws a
     * {@link java.lang.ClassCastException ClassCastException}.
     *
     * @param obj the object to be compared
     * @return the value 0 if the argument is an <code>ID</code> equal to this
     *         <code>ID</code>; a value less than 0 if this <code>ID</code> is
     *         less than the argument, which is an <code>ID</code>; and a
     *         value greater than 0 if this <code>ID</code> is greater than
     *         the argument, which is an <code>ID</code>.
     * @exception ClassCastException if the argument is not an <code>ID</code>
     */
    public int compareTo(final Object obj) throws ClassCastException {
        return compareTo((ID)obj);
    }

    /**
     * Returns a hashcode value for the <code>ID</code>
     *
     * @return a hashcode value for this <code>ID</code>
     * @see java.lang.Object
     */
    public final int hashCode() {
        return myHashCode;
    }

    /**
     * Indicates whether the argument <code>ID</code> is the same as the one
     * represented by the node and count arguments
     *
     * @param node the logical node to compare against
     * @param count the count to compare against
     * @return <code>true</code> if this <code>ID</code> is the one
     *        represented by <var>node</var> and <var>count</var>;
     *        <code>false</code> otherwise
     */
    public final boolean equals(final LogicalNode node, final long count) {
        return (this.count == count) && this.node.equals(node);
    }

    /**
     * Indicates whether the argument <code>ID</code> is the same as this one
     *
     * @param id the <code>ID</code> with which to compare
     * @return <code>true</code> if this <code>ID</code> is the same as the
     *         argument <code>ID</code>; <code>false</code> otherwise
     */
    public final boolean equals(final ID id) {
        return this == id || ((count == id.count) && node.equals(id.node));
    }

    /**
     * Indicates whether this <code>ID</code> is equal to some other object
     *
     * @param obj the object with which to compare
     * @return <code>true</code> if this <code>ID</code> is equal to the
     *         argument object; <code>false</code> otherwise
     */
    public boolean equals(final Object obj) {
        return this == obj || ((obj instanceof ID) && equals((ID)obj));
    }

    /**
     * Convert the <code>ID</code> to a string
     *
     * @return a string representation of this <code>ID</code>
     */
    public String toString() {
        return "ID[" + node + ',' + count + ']';
    }

    /**
     * Save the state of the <code>ID</code> instance to a stream (i.e.,
     * serialize it).
     *
     * @serialData The unique counter is written out (long), followed by the
     *             logical node component (LogicalNode).
     */
    public void writeExternal(final ObjectOutput out) throws IOException {
        out.writeLong(count);
        out.writeObject(node);
    }

    public void readExternal(final ObjectInput in)
        throws IOException, ClassNotFoundException {

        count = in.readLong();
        node = (LogicalNode)in.readObject();
        myHashCode = (int)(3*count + 191*node.hashCode());
    }

    public void setNode(LogicalNode ln) {
      this.node = ln;
    }
}
