/*
 * @(#)$Id
 *
 * derived from code that is Copyright 1997, 1998 by Sun Microsystems, Inc.
 *
 * In particular, this class was derived from the implementation of
 * java.util.HashMap in Sun's distribution of JDK 1.2.1.
 */

package mokan.util;

import java.io.*;
import java.util.*;

/**
 * A hashtable with <code>int</code> keys.  This class makes no guarantees as
 * to the order of the map; in particular, it does not guarantee that the
 * order will remain constant over time.<p>
 *
 * This implementation provides constant-time performance for the basic
 * operations (<tt>get</tt> and <tt>put</tt>), assuming the hash function
 * disperses the elements properly among the buckets.  Iteration over
 * collection views requires time proportional to the "capacity" of the
 * <tt>IntHashMap</tt> instance (the number of buckets) plus its size (the
 * number of key-value mappings).  Thus, it's very important not to set the
 * intial capacity too high (or the load factor too low) if iteration
 * performance is important.<p>
 *
 * An instance of <tt>IntHashMap</tt> has two parameters that affect its
 * performance: <i>initial capacity</i> and <i>load factor</i>.  The
 * <i>capacity</i> is the number of buckets in the hash table, and the initial
 * capacity is simply the capacity at the time the hash table is created.  The
 * <i>load factor</i> is a measure of how full the hash table is allowed to
 * get before its capacity is automatically increased.  When the number of
 * entries in the hash table exceeds the product of the load factor and the
 * current capacity, the capacity is roughly doubled by calling the
 * <tt>rehash</tt> method.<p>
 *
 * As a general rule, the default load factor (.75) offers a good tradeoff
 * between time and space costs.  Higher values decrease the space overhead
 * but increase the lookup cost (reflected in most of the operations of the
 * <tt>IntHashMap</tt> class, including <tt>get</tt> and <tt>put</tt>).  The
 * expected number of entries in the map and its load factor should be taken
 * into account when setting its initial capacity, so as to minimize the
 * number of <tt>rehash</tt> operations.  If the initial capacity is greater
 * than the maximum number of entries divided by the load factor, no
 * <tt>rehash</tt> operations will ever occur.<p>
 *
 * If many mappings are to be stored in a <tt>IntHashMap</tt> instance,
 * creating it with a sufficiently large capacity will allow the mappings to
 * be stored more efficiently than letting it perform automatic rehashing as
 * needed to grow the table.<p>
 *
 * <b>Note that this implementation is not synchronized.</b> If multiple
 * threads access this map concurrently, and at least one of the threads
 * modifies the map structurally, it <i>must</i> be synchronized externally.
 * (A structural modification is any operation that adds or deletes one or
 * more mappings; merely changing the value associated with a key that an
 * instance already contains is not a structural modification.)  This is
 * typically accomplished by synchronizing on some object that naturally
 * encapsulates the map.<p>
 *
 * The iterators returned by all of this class's "collection view methods" are
 * <i>fail-fast</i>: if the map is structurally modified at any time after the
 * iterator is created, in any way except through the iterator's own
 * <tt>remove</tt> or <tt>add</tt> methods, the iterator will throw a
 * <tt>ConcurrentModificationException</tt>.  Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the
 * future.
 *
 * @version	$Revision: 1.2 $, $Date: 2002/11/22 23:26:07 $
 * @author	Josh Bloch
 * @author	Arthur van Hoff
 * @author	Jerry James 
 * @see java.util.HashMap
 */
public final class IntHashMap implements Cloneable, Serializable {

    /**
     * The <var>serialVersionUID</var> for this class.  If you change the
     * signature of this class (field names or types, method names, parameter
     * names or types, adding or removing fields or methods, etc.), you
     * <strong>must</strong> change this value.  The recommended way is to use
     * the <code>serialver</code> tool.
     */
    static final long serialVersionUID = 2957434530168623020L;






    /**
     * The hash table data.
     */
    private transient Entry[] table;

    /**
     * The total number of mappings in the hash table.
     */
    private transient int count;

    /**
     * The table is rehashed when its size exceeds this threshold.  (The
     * value of this field is (int)(capacity * loadFactor).)
     *
     * @serial an integer containing the threshold value for this map
     */
    private int threshold;

    /**
     * The load factor for the hashtable.
     *
     * @serial a floating point number containing the load factor for this map
     */
    private float loadFactor;

    /**
     * The number of times this IntHashMap has been structurally modified
     * Structural modifications are those that change the number of mappings
     * in the IntHashMap or otherwise modify its internal structure (e.g.,
     * rehash).  This field is used to make iterators on Collection-views of
     * the IntHashMap fail-fast.  (See ConcurrentModificationException).
     */
    private transient int modCount = 0;

    /**
     * Constructs a new, empty hash table with the specified initial 
     * capacity and the specified load factor. 
     *
     * @param     initialCapacity   the initial capacity of the IntHashMap.
     * @param     loadFactor        the load factor of the IntHashMap
     * @exception IllegalArgumentException  if the initial capacity is less
     *              than zero, or if the load factor is nonpositive.
     */
    public IntHashMap(int initialCapacity, final float loadFactor) {
	if (initialCapacity < 0)
	    throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        if (loadFactor <= 0)
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
        if (initialCapacity == 0)
            initialCapacity = 1;
	this.loadFactor = loadFactor;
	table = new Entry[initialCapacity];
	threshold = (int)(initialCapacity * loadFactor);
    }

    /**
     * Constructs a new, empty hash table with the specified initial capacity
     * and default load factor, which is <tt>0.75</tt>.
     *
     * @param   initialCapacity   the initial capacity of the IntHashMap.
     * @exception IllegalArgumentException if the initial capacity is less
     *              than zero.
     */
    public IntHashMap(final int initialCapacity) {
	this(initialCapacity, 0.75f);
    }

    /**
     * Constructs a new, empty hash table with a default capacity, which is
     * <tt>101</tt>, and a default load factor, which is <tt>0.75</tt>.
     */
    public IntHashMap() {
	this(101, 0.75f);
    }

    /**
     * Returns the number of entries in this hash table.
     *
     * @return the number of entries in this hash table.
     */
    public int size() {
	return count;
    }

    /**
     * Returns <tt>true</tt> if this hash table contains no entries.
     *
     * @return <tt>true</tt> if this hash table contains no entries.
     */
    public boolean isEmpty() {
	return count == 0;
    }

    /**
     * Returns <tt>true</tt> if this hash table maps one or more keys to the
     * specified value.
     *
     * @param value value whose presence in this hash table is to be tested.
     * @return <tt>true</tt> if this hash table maps one or more keys to the
     *         specified value.
     */
    public boolean containsValue(final Object value) {
	final Entry[] tab = table;

	for (int i = tab.length ; i-- > 0 ;)
	    for (Entry e = tab[i]; e != null; e = e.next)
		if (value.equals(e.value))
		    return true;
	return false;
    }

    /**
     * Returns <tt>true</tt> if this hash table contains a mapping for the
     * specified key.
     * 
     * @return <tt>true</tt> if this hash table contains a mapping for the
     * specified key.
     * @param key key whose presence in this hash table is to be tested.
     */
    public boolean containsKey(final int key) {
	final Entry[] tab = table;
	final int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index]; e != null; e = e.next)
	    if (key == e.key)
		return true;
	return false;
    }

    /**
     * Returns the value to which this hash table maps the specified key.
     * Returns <tt>null</tt> if the hash table contains no mapping for this
     * key.
     *
     * @return the value to which this hash table maps the specified key.
     * @param key key whose associated value is to be returned.
     */
    public Object get(final int key) {
	final Entry[] tab = table;
	final int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index]; e != null; e = e.next){
	    if (key == e.key)
		return e.value;
        }
	return null;
    }

    /**
     * Rehashes the contents of this hash table into a new
     * <tt>IntHashMap</tt> instance with a larger capacity.  This method is
     * called automatically when the number of keys in this map exceeds its
     * capacity and load factor.
     */
    private void rehash() {
	final int oldCapacity = table.length;
	final Entry[] oldMap = table;

	final int newCapacity = oldCapacity * 2 + 1;
	final Entry[] newMap = new Entry[newCapacity];

	modCount++;
	threshold = (int)(newCapacity * loadFactor);
	table = newMap;

	for (int i = oldCapacity ; i-- > 0 ;) {
	    for (Entry old = oldMap[i] ; old != null ; ) {
		final Entry e = old;
		old = old.next;

		final int index = (e.key & 0x7FFFFFFF) % newCapacity;
		e.next = newMap[index];
		newMap[index] = e;
	    }
	}
    }

    /**
     * Associates the specified value with the specified key in this hash
     * table.  If the hash table previously contained a mapping for this key,
     * the old value is replaced.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.
     */
    public Object put(final int key, final Object value) {
	// Makes sure the key is not already in the IntHashMap.
	Entry[] tab = table;
        
	int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index]; e != null; e = e.next) {
	    if (key == e.key) {
		final Object old = e.value;
		e.value = value;
		return old;
	    }
	}

	modCount++;
	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();

            tab = table;
            index = (key & 0x7FFFFFFF) % tab.length;
	}

	// Creates the new entry.
	final Entry e = new Entry(key, value, tab[index]);
	tab[index] = e;
	count++;
	return null;
    }

    /**
     * Removes the mapping for this key from this hash table if present.
     *
     * @param key key whose mapping is to be removed from the map.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.
     */
    public Object remove(final int key) {
	final Entry[] tab = table;

	final int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index], prev = null; e != null;
	     prev = e, e = e.next) {
	    if (key == e.key) {
		modCount++;
		if (prev != null)
		    prev.next = e.next;
		else
		    tab[index] = e.next;

		count--;
		final Object oldValue = e.value;
		e.value = null;
		return oldValue;
	    }
	}
	return null;
    }

    /**
     * Removes all mappings from this map.
     */
    public void clear() {
	final Entry[] tab = table;
	modCount++;
	for (int index = tab.length; --index >= 0; )
	    tab[index] = null;
	count = 0;
    }

    /**
     * Returns a shallow copy of this <tt>IntHashMap</tt> instance: the
     * values themselves are not cloned.
     *
     * @return a shallow copy of this map.
     */
    public Object clone() {
	final IntHashMap t = new IntHashMap(table.length, loadFactor);
	for (int i = table.length ; i-- > 0 ; ) {
	    t.table[i] = (table[i] != null) ? (Entry)table[i].clone() : null;
	}
	t.values = null;
	t.modCount = 0;
	return t;
    }

    // Views -- only one, sorry!
    private transient Collection values = null;

    /**
     * Returns a collection view of the values contained in this hash table.
     * The collection is backed by the hash table, so changes to the hash
     * table are reflected in the collection, and vice-versa.  The collection
     * supports element removal, which removes the corresponding mapping from
     * this hash table, via the <tt>Iterator.remove</tt>,
     * <tt>Collection.remove</tt>, <tt>removeAll</tt>, <tt>retainAll</tt>, and
     * <tt>clear</tt> operations.  It does not support the <tt>add</tt> or
     * <tt>addAll</tt> operations.
     *
     * @return a collection view of the values contained in this map.
     */
    public Collection values() {
	if (values == null) {
	    values = new AbstractCollection() {
                public Iterator iterator() {
                    return new EntryIterator();
                }
                public int size() {
                    return count;
                }
                public boolean contains(final Object o) {
                    return containsValue(o);
                }
                public void clear() {
                    IntHashMap.this.clear();
                }
            };
        }
	return values;
    }

    public Iterator entryIterator() {
	return new EntryIterator();
    }

    public Iterator valueIterator() {
	return new ValueIterator();
    }

    /**
     * Compares the specified object with this hash table for equality.
     * Returns <tt>true</tt> if the given object is also a hash table and the
     * two hash tables represent the same mappings.  More formally, two hash
     * tables <tt>t1</tt> and <tt>t2</tt> represent the same mappings if the
     * contain the same keys, and for every key <tt>k</tt> in <tt>t1</tt>,
     * <tt>(t1.get(k)==null ? t2.get(k)==null : t1.get(k).equals(t2.get(k)))
     * </tt>.<p>
     *
     * This implementation first checks if the specified object is this hash
     * table; if so it returns <tt>true</tt>.  Then, it checks if the
     * specified object is a hash table whose size is identical to the size of
     * this set; if not, it it returns <tt>false</tt>.  If so, it iterates
     * over this map, and checks that the specified map contains each mapping
     * that this map contains.  If the specified map fails to contain such a
     * mapping, <tt>false</tt> is returned.  If the iteration completes,
     * <tt>true</tt> is returned.
     *
     * @param o object to be compared for equality with this map.
     * @return <tt>true</tt> if the specified object is equal to this map.
     */
    public boolean equals(final Object o) {
	if (o == this)
	    return true;

	if (!(o instanceof IntHashMap))
	    return false;
	final IntHashMap t = (IntHashMap)o;
	if (t.size() != size())
	    return false;

	final Iterator i = entryIterator();
	while (i.hasNext()) {
	    final Entry e = (Entry)i.next();
	    final int key = e.getKey();
	    final Object value = e.getValue();
	    if (!value.equals(t.get(key)))
		return false;
	}
	return true;
    }


    public String toString() {
        String r = "{";
        Iterator ei = entryIterator ();
        while (ei.hasNext()) {
            Entry e = (Entry)ei.next();
            r = r + e.key + ":" + e.value + ",";
        }
        
        return r + "}";
    }

    /**
     * Returns the hash code value for this hash table.  The hash code of a
     * hash table is defined to be the sum of the hash codes of each entry in
     * the hash table's <tt>entrySet()</tt> view.  This ensures that
     * <tt>t1.equals(t2)</tt> implies that <tt>t1.hashCode()==
     * t2.hashCode()</tt> for any two maps <tt>t1</tt> and <tt>t2</tt>, as
     * required by the general contract of Object.hashCode.<p>
     *
     * This implementation iterates over <tt>entrySet()</tt>, calling
     * <tt>hashCode</tt> on each element (entry) in the Collection, and adding
     * up the results.
     *
     * @return the hash code value for this map.
     * @see Map.Entry#hashCode()
     * @see Object#hashCode()
     * @see Object#equals(Object)
     * @see Set#equals(Object)
     */
    public int hashCode() {
	int h = 0;
	final Iterator i = entryIterator();
	while (i.hasNext())
	    h += i.next().hashCode();
	return h;
    }

    /**
     * IntHashMap collision list entry.
     */
    public static class Entry {
	final int key;
	Object value;
	Entry next;

	Entry(final int key, final Object value, final Entry next) {
	    this.key = key;
	    this.value = value;
	    this.next = next;
	}

	protected Object clone() {
	    return new Entry(key, value,
			     (next==null ? null : (Entry)next.clone()));
	}

	// Map.Entry Ops 

	public int getKey() {
	    return key;
	}

	public Object getValue() {
	    return value;
	}

	public Object setValue(final Object value) {
	    final Object oldValue = this.value;
	    this.value = value;
	    return oldValue;
	}

	public boolean equals(final Object o) {
	    if (!(o instanceof Entry))
		return false;
	    final Entry e = (Entry)o;

	    return key == e.getKey() &&
	       (value==null ? e.getValue()==null : value.equals(e.getValue()));
	}

	public int hashCode() {
	    return key ^ (value == null ? 0 : value.hashCode());
	}

	public String toString() {
	    return String.valueOf(key) + "=" + value.toString();
	}
    }

    /***** Iterators *****/

    private final class EntryIterator implements Iterator {
	final Entry[] table = IntHashMap.this.table;
	int index = table.length;
	Entry entry = null;
	Entry lastReturned = null;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	private int expectedModCount = modCount;

	EntryIterator() { }

	public boolean hasNext() {
	    while (entry == null && index > 0)
		entry = table[--index];
	    return entry != null;
	}

	public Object next() {
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    while (entry == null && index > 0)
		entry = table[--index];

	    if (entry != null) {
		final Entry e = lastReturned = entry;
		entry = e.next;
		return e;
	    }
	    throw new NoSuchElementException();
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    final Entry[] tab = IntHashMap.this.table;
	    final int index = (lastReturned.key & 0x7FFFFFFF) % tab.length;

	    for (Entry e = tab[index], prev = null; e != null;
		 prev = e, e = e.next) {
		if (e == lastReturned) {
		    modCount++;
		    expectedModCount++;
		    if (prev == null)
			tab[index] = e.next;
		    else
			prev.next = e.next;
		    count--;
		    lastReturned = null;
		    return;
		}
	    }
	    throw new ConcurrentModificationException();
	}
    }

    private final class ValueIterator implements Iterator {
	final Entry[] table = IntHashMap.this.table;
	int index = table.length;
	Entry entry = null;
	Entry lastReturned = null;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	private int expectedModCount = modCount;

	ValueIterator() { }

	public boolean hasNext() {
	    while (entry == null && index > 0)
		entry = table[--index];
	    return entry != null;
	}

	public Object next() {
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    while (entry == null && index > 0)
		entry = table[--index];

	    if (entry != null) {
		final Entry e = lastReturned = entry;
		entry = e.next;
		return e.value;
	    }
	    throw new NoSuchElementException();
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    final Entry[] tab = IntHashMap.this.table;
	    final int index = (lastReturned.key & 0x7FFFFFFF) % tab.length;

	    for (Entry e = tab[index], prev = null; e != null;
		 prev = e, e = e.next) {
		if (e == lastReturned) {
		    modCount++;
		    expectedModCount++;
		    if (prev == null)
			tab[index] = e.next;
		    else
			prev.next = e.next;
		    count--;
		    lastReturned = null;
		    return;
		}
	    }
	    throw new ConcurrentModificationException();
	}
    }

    /**
     * Save the state of the <tt>IntHashMap</tt> instance to a stream (i.e.,
     * serialize it).
     *
     * @serialData The <i>capacity</i> of the IntHashMap (the length of the
     *		   bucket array) is emitted (int), followed  by the
     *		   <i>size</i> of the IntHashMap (the number of key-value
     *		   mappings), followed by the key (int) and value (Object)
     *		   for each key-value mapping represented by the IntHashMap
     * The key-value mappings are emitted in no particular order.
     */
    private void writeObject(final ObjectOutputStream s) throws IOException {
	// Write out the threshold, loadfactor, and any hidden stuff
	s.defaultWriteObject();

	// Write out number of buckets
	s.writeInt(table.length);

	// Write out size (number of Mappings)
	s.writeInt(count);

        // Write out keys and values (alternating)
	for (int index = table.length-1; index >= 0; index--) {
	    for (Entry entry = table[index]; entry!=null; entry = entry.next) {
		s.writeInt(entry.key);
		s.writeObject(entry.value);
	    }
	}
    }

    /**
     * Reconstitute the <tt>IntHashMap</tt> instance from a stream (i.e.,
     * deserialize it).
     */
    private void readObject(final ObjectInputStream s)
         throws IOException, ClassNotFoundException
    {
	// Read in the threshold, loadfactor, and any hidden stuff
	s.defaultReadObject();

	// Read in number of buckets and allocate the bucket array;
	final int numBuckets = s.readInt();
	table = new Entry[numBuckets];

	// Read in size (number of Mappings)
	final int size = s.readInt();

	// Read the keys and values, and put the mappings in the IntHashMap
	for (int i = 0; i < size; i++) {
	    final int key = s.readInt();
	    final Object value = s.readObject();
	    put(key, value);
	}
    }

    int capacity() {
        return table.length;
    }

    float loadFactor() {
        return loadFactor;
    }

}
