/*
 * @(#)$Id
 *
 *  Specalize Mapping from int to int
 */

package mokan.util;

import java.io.*;
import java.util.*;




public class IntIntMap {


    /**
     * The hash table data.
     */
    private transient Entry[] table;

    /**
     * The total number of mappings in the hash table.
     */
    private transient int count;

    /**
     * The table is rehashed when its size exceeds this threshold.  (The
     * value of this field is (int)(capacity * loadFactor).)
     *
     * @serial an integer containing the threshold value for this map
     */
    private int threshold;

    /**
     * The load factor for the hashtable.
     *
     * @serial a floating point number containing the load factor for this map
     */
    private float loadFactor;

    /**
     * The number of times this IntHashMap has been structurally modified
     * Structural modifications are those that change the number of mappings
     * in the IntHashMap or otherwise modify its internal structure (e.g.,
     * rehash).  This field is used to make iterators on Collection-views of
     * the IntHashMap fail-fast.  (See ConcurrentModificationException).
     */
    private transient int modCount = 0;


    /**
     * Rehashes the contents of this hash table into a new
     * <tt>IntHashMap</tt> instance with a larger capacity.  This method is
     * called automatically when the number of keys in this map exceeds its
     * capacity and load factor.
     */
    private void rehash() {
	final int oldCapacity = table.length;
	final Entry[] oldMap = table;

	final int newCapacity = oldCapacity * 2 + 1;
	final Entry[] newMap = new Entry[newCapacity];

	modCount++;
	threshold = (int)(newCapacity * loadFactor);
	table = newMap;

	for (int i = oldCapacity ; i-- > 0 ;) {
	    for (Entry old = oldMap[i] ; old != null ; ) {
		final Entry e = old;
		old = old.next;

		final int index = (e.key & 0x7FFFFFFF) % newCapacity;
		e.next = newMap[index];
		newMap[index] = e;
	    }
	}
    }

    public IntIntMap(int initialCapacity, final float loadFactor) {
	if (initialCapacity < 0)
	    throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        if (loadFactor <= 0)
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
        if (initialCapacity == 0)
            initialCapacity = 1;
	this.loadFactor = loadFactor;
	table = new Entry[initialCapacity];
	threshold = (int)(initialCapacity * loadFactor);
    }
    public IntIntMap () {
        this (37, 0.75f);
    }

    public boolean containsKey(final int key) {
	final Entry[] tab = table;
	final int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index]; e != null; e = e.next)
	    if (key == e.key)
		return true;
	return false;
    }

    /**
     * Removes all mappings from this map.
     */
    public void clear() {
	final Entry[] tab = table;
	modCount++;
	for (int index = tab.length; --index >= 0; )
	    tab[index] = null;
	count = 0;
    }

    public int get (int key) {
	final Entry[] tab = table;
	final int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index]; e != null; e = e.next){
	    if (key == e.key)
		return e.value;
        }
	return 0;
    }

    public int put (int key, int value){
	// Makes sure the key is not already in the IntHashMap.
	Entry[] tab = table;
        
	int index = (key & 0x7FFFFFFF) % tab.length;
	for (Entry e = tab[index]; e != null; e = e.next) {
	    if (key == e.key) {
		final int old = e.value;
		e.value = value;
		return old;
	    }
	}

	modCount++;
	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();

            tab = table;
            index = (key & 0x7FFFFFFF) % tab.length;
	}

	// Creates the new entry.
	final Entry e = new Entry(key, value, tab[index]);
	tab[index] = e;
	count++;
	return value;
    }

    public Iterator entryIterator() {
	return new EntryIterator();
    }

    public Iterator valueIterator() {
	return new ValueIterator();
    }

    public static class Entry {
        public int key;
	public int value;
        Entry next;

	Entry(final int key, final int value, final Entry next) {
	    this.key = key;
	    this.value = value;
	    this.next = next;
	}
    }

    /***** Iterators *****/

    private final class EntryIterator implements Iterator {
	final Entry[] table = IntIntMap.this.table;
	int index = table.length;
	Entry entry = null;
	Entry lastReturned = null;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	private int expectedModCount = modCount;

	EntryIterator() { }

	public boolean hasNext() {
	    while (entry == null && index > 0)
		entry = table[--index];
	    return entry != null;
	}

	public Object next() {
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    while (entry == null && index > 0)
		entry = table[--index];

	    if (entry != null) {
		final Entry e = lastReturned = entry;
		entry = e.next;
		return e;
	    }
	    throw new NoSuchElementException();
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    final Entry[] tab = IntIntMap.this.table;
	    final int index = (lastReturned.key & 0x7FFFFFFF) % tab.length;

	    for (Entry e = tab[index], prev = null; e != null;
		 prev = e, e = e.next) {
		if (e == lastReturned) {
		    modCount++;
		    expectedModCount++;
		    if (prev == null)
			tab[index] = e.next;
		    else
			prev.next = e.next;
		    count--;
		    lastReturned = null;
		    return;
		}
	    }
	    throw new ConcurrentModificationException();
	}
    }

    private final class ValueIterator implements Iterator {
	final Entry[] table = IntIntMap.this.table;
	int index = table.length;
	Entry entry = null;
	Entry lastReturned = null;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	private int expectedModCount = modCount;

	ValueIterator() { }

	public boolean hasNext() {
	    while (entry == null && index > 0)
		entry = table[--index];
	    return entry != null;
	}

	public Object next() {
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    while (entry == null && index > 0)
		entry = table[--index];

	    if (entry != null) {
		final Entry e = lastReturned = entry;
		entry = e.next;
		return new Integer(e.value);
	    }
	    throw new NoSuchElementException();
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    if (modCount != expectedModCount)
		throw new ConcurrentModificationException();

	    final Entry[] tab = IntIntMap.this.table;
	    final int index = (lastReturned.key & 0x7FFFFFFF) % tab.length;

	    for (Entry e = tab[index], prev = null; e != null;
		 prev = e, e = e.next) {
		if (e == lastReturned) {
		    modCount++;
		    expectedModCount++;
		    if (prev == null)
			tab[index] = e.next;
		    else
			prev.next = e.next;
		    count--;
		    lastReturned = null;
		    return;
		}
	    }
	    throw new ConcurrentModificationException();
	}
    }

}
