/*
 * @(#)$Id: NodeComparator.java,v 1.1.1.1 2002/11/06 22:10:29 kris Exp $
 *
 *
 */

package mokan.util;

import java.util.Comparator;
import mokan.comm.LogicalNode;

/**
 * A class for sorting <code>ID</code>s by node.  It considers node to be most
 * significant, and sequence number to be least significant.
 *
 * @version	$Revision: 1.1.1.1 $, $Date: 2002/11/06 22:10:29 $
 * @author	Jerry James
 */
public final class NodeComparator implements Comparator {

    /**
     * Compares two {@linkmokan.util.ID IDs} for order.  Returns a negative
     * integer, zero, or a positive integer as the first {@linkmokan.util.ID ID}
     * is less than, equal to, or greater than the second.
     *
     * @param id1 the first comparator
     * @param id2 the second comparator
     * @return a negative integer, zero, or a positive integer as the first
     *        {@linkmokan.util.ID ID} is less than, equal to, or greater than
     *        the second.
     */
    public int compare(final ID id1, final ID id2) {
	final LogicalNode node1 = id1.getNode();
	final LogicalNode node2 = id2.getNode();

	// Don't use equals, because we canonicalize LogicalNodes
	if (node1 == node2) {
	    final long seq1 = id1.getSeq();
	    final long seq2 = id2.getSeq();
	    return (seq1 == seq2) ? 0 : ((seq1 < seq2) ? -1 : 1);
	    // Don't subtract sequence numbers, because they are longs and we
	    // need to return an int.  Casting a long to an int can change the
	    // sign, which is disastrous for a Comparator.
	}
	return node1.compareTo(node2);
    }

    /**
     * Compares two objects for order.  If the objects are
     * {@linkmokan.util.ID IDs}, this function behaves like
     * {@link #compare(ID, ID) compare(ID,ID)}.  Otherwise, its behavior is
     * arbitrary.
     *
     * @param obj1 the first comparator
     * @param obj2 the second comparator
     * @return a negative integer, zero, or a positive integer as the first
     *        {@linkmokan.util.ID ID} is less than, equal to, or greater than
     *        the second.
     */
    public int compare(final Object obj1, final Object obj2) {
	return compare((ID)obj1, (ID)obj2);
    }

    /**
     * Indicates that all <code>NodeComparator</code>s impose the same order.
     *
     * @param nodeComp the <code>NodeComparator</code> to compare to this one
     *       (ignored)
     * @return <code>true</code>, since all <code>NodeComparator</code>s
     *         impose the same order
     */
    public boolean equals(final NodeComparator nodeComp) {
	return true;
    }

    /**
     * Indicates whether the argument object is a <code>NodeComparator</code>
     *
     * @param obj the object to compare to this <code>NodeComparator</code>
     * @return <code>true</code> if <var>obj</var> is a
     *         <code>NodeComparator</code>
     */
    public boolean equals(final Object obj) {
	return (obj instanceof NodeComparator);
    }
}
